extern blitter_func blitdofast_0;
extern blitter_func blitdofast_desc_0;
extern blitter_func blitdofast_a;
extern blitter_func blitdofast_desc_a;
extern blitter_func blitdofast_2a;
extern blitter_func blitdofast_desc_2a;
extern blitter_func blitdofast_30;
extern blitter_func blitdofast_desc_30;
extern blitter_func blitdofast_3a;
extern blitter_func blitdofast_desc_3a;
extern blitter_func blitdofast_3c;
extern blitter_func blitdofast_desc_3c;
extern blitter_func blitdofast_4a;
extern blitter_func blitdofast_desc_4a;
extern blitter_func blitdofast_6a;
extern blitter_func blitdofast_desc_6a;
extern blitter_func blitdofast_8a;
extern blitter_func blitdofast_desc_8a;
extern blitter_func blitdofast_8c;
extern blitter_func blitdofast_desc_8c;
extern blitter_func blitdofast_9a;
extern blitter_func blitdofast_desc_9a;
extern blitter_func blitdofast_a8;
extern blitter_func blitdofast_desc_a8;
extern blitter_func blitdofast_aa;
extern blitter_func blitdofast_desc_aa;
extern blitter_func blitdofast_b1;
extern blitter_func blitdofast_desc_b1;
extern blitter_func blitdofast_ca;
extern blitter_func blitdofast_desc_ca;
extern blitter_func blitdofast_cc;
extern blitter_func blitdofast_desc_cc;
extern blitter_func blitdofast_d8;
extern blitter_func blitdofast_desc_d8;
extern blitter_func blitdofast_e2;
extern blitter_func blitdofast_desc_e2;
extern blitter_func blitdofast_ea;
extern blitter_func blitdofast_desc_ea;
extern blitter_func blitdofast_f0;
extern blitter_func blitdofast_desc_f0;
extern blitter_func blitdofast_fa;
extern blitter_func blitdofast_desc_fa;
extern blitter_func blitdofast_fc;
extern blitter_func blitdofast_desc_fc;
