sudo apt-get install libsdl1.2-dev 
sudo apt-get install libguichan-dev
sudo apt-get install libsdl-ttf2.0-dev
sudo apt-get install libsdl-gfx1.2-dev
sudo apt-get install libxml2-dev
sudo apt-get install libflac-dev
sudo apt-get install libmpg123-dev
sudo apt-get install libmpeg2-4-dev
sudo apt-get install autoconf
