extern cpuop_func op_0000_0_nf;
extern cpuop_func op_0000_0_ff;
extern cpuop_func op_0010_0_nf;
extern cpuop_func op_0010_0_ff;
extern cpuop_func op_0018_0_nf;
extern cpuop_func op_0018_0_ff;
extern cpuop_func op_0020_0_nf;
extern cpuop_func op_0020_0_ff;
extern cpuop_func op_0028_0_nf;
extern cpuop_func op_0028_0_ff;
extern cpuop_func op_0030_0_nf;
extern cpuop_func op_0030_0_ff;
extern cpuop_func op_0038_0_nf;
extern cpuop_func op_0038_0_ff;
extern cpuop_func op_0039_0_nf;
extern cpuop_func op_0039_0_ff;
extern cpuop_func op_003c_0_nf;
extern cpuop_func op_003c_0_ff;
extern cpuop_func op_0040_0_nf;
extern cpuop_func op_0040_0_ff;
extern cpuop_func op_0050_0_nf;
extern cpuop_func op_0050_0_ff;
extern cpuop_func op_0058_0_nf;
extern cpuop_func op_0058_0_ff;
extern cpuop_func op_0060_0_nf;
extern cpuop_func op_0060_0_ff;
extern cpuop_func op_0068_0_nf;
extern cpuop_func op_0068_0_ff;
extern cpuop_func op_0070_0_nf;
extern cpuop_func op_0070_0_ff;
extern cpuop_func op_0078_0_nf;
extern cpuop_func op_0078_0_ff;
extern cpuop_func op_0079_0_nf;
extern cpuop_func op_0079_0_ff;
extern cpuop_func op_007c_0_nf;
extern cpuop_func op_007c_0_ff;
extern cpuop_func op_0080_0_nf;
extern cpuop_func op_0080_0_ff;
extern cpuop_func op_0090_0_nf;
extern cpuop_func op_0090_0_ff;
extern cpuop_func op_0098_0_nf;
extern cpuop_func op_0098_0_ff;
extern cpuop_func op_00a0_0_nf;
extern cpuop_func op_00a0_0_ff;
extern cpuop_func op_00a8_0_nf;
extern cpuop_func op_00a8_0_ff;
extern cpuop_func op_00b0_0_nf;
extern cpuop_func op_00b0_0_ff;
extern cpuop_func op_00b8_0_nf;
extern cpuop_func op_00b8_0_ff;
extern cpuop_func op_00b9_0_nf;
extern cpuop_func op_00b9_0_ff;
extern cpuop_func op_00d0_0_nf;
extern cpuop_func op_00d0_0_ff;
extern cpuop_func op_00e8_0_nf;
extern cpuop_func op_00e8_0_ff;
extern cpuop_func op_00f0_0_nf;
extern cpuop_func op_00f0_0_ff;
extern cpuop_func op_00f8_0_nf;
extern cpuop_func op_00f8_0_ff;
extern cpuop_func op_00f9_0_nf;
extern cpuop_func op_00f9_0_ff;
extern cpuop_func op_00fa_0_nf;
extern cpuop_func op_00fa_0_ff;
extern cpuop_func op_00fb_0_nf;
extern cpuop_func op_00fb_0_ff;
extern cpuop_func op_0100_0_nf;
extern cpuop_func op_0100_0_ff;
extern cpuop_func op_0108_0_nf;
extern cpuop_func op_0108_0_ff;
extern cpuop_func op_0110_0_nf;
extern cpuop_func op_0110_0_ff;
extern cpuop_func op_0118_0_nf;
extern cpuop_func op_0118_0_ff;
extern cpuop_func op_0120_0_nf;
extern cpuop_func op_0120_0_ff;
extern cpuop_func op_0128_0_nf;
extern cpuop_func op_0128_0_ff;
extern cpuop_func op_0130_0_nf;
extern cpuop_func op_0130_0_ff;
extern cpuop_func op_0138_0_nf;
extern cpuop_func op_0138_0_ff;
extern cpuop_func op_0139_0_nf;
extern cpuop_func op_0139_0_ff;
extern cpuop_func op_013a_0_nf;
extern cpuop_func op_013a_0_ff;
extern cpuop_func op_013b_0_nf;
extern cpuop_func op_013b_0_ff;
extern cpuop_func op_013c_0_nf;
extern cpuop_func op_013c_0_ff;
extern cpuop_func op_0140_0_nf;
extern cpuop_func op_0140_0_ff;
extern cpuop_func op_0148_0_nf;
extern cpuop_func op_0148_0_ff;
extern cpuop_func op_0150_0_nf;
extern cpuop_func op_0150_0_ff;
extern cpuop_func op_0158_0_nf;
extern cpuop_func op_0158_0_ff;
extern cpuop_func op_0160_0_nf;
extern cpuop_func op_0160_0_ff;
extern cpuop_func op_0168_0_nf;
extern cpuop_func op_0168_0_ff;
extern cpuop_func op_0170_0_nf;
extern cpuop_func op_0170_0_ff;
extern cpuop_func op_0178_0_nf;
extern cpuop_func op_0178_0_ff;
extern cpuop_func op_0179_0_nf;
extern cpuop_func op_0179_0_ff;
extern cpuop_func op_0180_0_nf;
extern cpuop_func op_0180_0_ff;
extern cpuop_func op_0188_0_nf;
extern cpuop_func op_0188_0_ff;
extern cpuop_func op_0190_0_nf;
extern cpuop_func op_0190_0_ff;
extern cpuop_func op_0198_0_nf;
extern cpuop_func op_0198_0_ff;
extern cpuop_func op_01a0_0_nf;
extern cpuop_func op_01a0_0_ff;
extern cpuop_func op_01a8_0_nf;
extern cpuop_func op_01a8_0_ff;
extern cpuop_func op_01b0_0_nf;
extern cpuop_func op_01b0_0_ff;
extern cpuop_func op_01b8_0_nf;
extern cpuop_func op_01b8_0_ff;
extern cpuop_func op_01b9_0_nf;
extern cpuop_func op_01b9_0_ff;
extern cpuop_func op_01c0_0_nf;
extern cpuop_func op_01c0_0_ff;
extern cpuop_func op_01c8_0_nf;
extern cpuop_func op_01c8_0_ff;
extern cpuop_func op_01d0_0_nf;
extern cpuop_func op_01d0_0_ff;
extern cpuop_func op_01d8_0_nf;
extern cpuop_func op_01d8_0_ff;
extern cpuop_func op_01e0_0_nf;
extern cpuop_func op_01e0_0_ff;
extern cpuop_func op_01e8_0_nf;
extern cpuop_func op_01e8_0_ff;
extern cpuop_func op_01f0_0_nf;
extern cpuop_func op_01f0_0_ff;
extern cpuop_func op_01f8_0_nf;
extern cpuop_func op_01f8_0_ff;
extern cpuop_func op_01f9_0_nf;
extern cpuop_func op_01f9_0_ff;
extern cpuop_func op_0200_0_nf;
extern cpuop_func op_0200_0_ff;
extern cpuop_func op_0210_0_nf;
extern cpuop_func op_0210_0_ff;
extern cpuop_func op_0218_0_nf;
extern cpuop_func op_0218_0_ff;
extern cpuop_func op_0220_0_nf;
extern cpuop_func op_0220_0_ff;
extern cpuop_func op_0228_0_nf;
extern cpuop_func op_0228_0_ff;
extern cpuop_func op_0230_0_nf;
extern cpuop_func op_0230_0_ff;
extern cpuop_func op_0238_0_nf;
extern cpuop_func op_0238_0_ff;
extern cpuop_func op_0239_0_nf;
extern cpuop_func op_0239_0_ff;
extern cpuop_func op_023c_0_nf;
extern cpuop_func op_023c_0_ff;
extern cpuop_func op_0240_0_nf;
extern cpuop_func op_0240_0_ff;
extern cpuop_func op_0250_0_nf;
extern cpuop_func op_0250_0_ff;
extern cpuop_func op_0258_0_nf;
extern cpuop_func op_0258_0_ff;
extern cpuop_func op_0260_0_nf;
extern cpuop_func op_0260_0_ff;
extern cpuop_func op_0268_0_nf;
extern cpuop_func op_0268_0_ff;
extern cpuop_func op_0270_0_nf;
extern cpuop_func op_0270_0_ff;
extern cpuop_func op_0278_0_nf;
extern cpuop_func op_0278_0_ff;
extern cpuop_func op_0279_0_nf;
extern cpuop_func op_0279_0_ff;
extern cpuop_func op_027c_0_nf;
extern cpuop_func op_027c_0_ff;
extern cpuop_func op_0280_0_nf;
extern cpuop_func op_0280_0_ff;
extern cpuop_func op_0290_0_nf;
extern cpuop_func op_0290_0_ff;
extern cpuop_func op_0298_0_nf;
extern cpuop_func op_0298_0_ff;
extern cpuop_func op_02a0_0_nf;
extern cpuop_func op_02a0_0_ff;
extern cpuop_func op_02a8_0_nf;
extern cpuop_func op_02a8_0_ff;
extern cpuop_func op_02b0_0_nf;
extern cpuop_func op_02b0_0_ff;
extern cpuop_func op_02b8_0_nf;
extern cpuop_func op_02b8_0_ff;
extern cpuop_func op_02b9_0_nf;
extern cpuop_func op_02b9_0_ff;
extern cpuop_func op_02d0_0_nf;
extern cpuop_func op_02d0_0_ff;
extern cpuop_func op_02e8_0_nf;
extern cpuop_func op_02e8_0_ff;
extern cpuop_func op_02f0_0_nf;
extern cpuop_func op_02f0_0_ff;
extern cpuop_func op_02f8_0_nf;
extern cpuop_func op_02f8_0_ff;
extern cpuop_func op_02f9_0_nf;
extern cpuop_func op_02f9_0_ff;
extern cpuop_func op_02fa_0_nf;
extern cpuop_func op_02fa_0_ff;
extern cpuop_func op_02fb_0_nf;
extern cpuop_func op_02fb_0_ff;
extern cpuop_func op_0400_0_nf;
extern cpuop_func op_0400_0_ff;
extern cpuop_func op_0410_0_nf;
extern cpuop_func op_0410_0_ff;
extern cpuop_func op_0418_0_nf;
extern cpuop_func op_0418_0_ff;
extern cpuop_func op_0420_0_nf;
extern cpuop_func op_0420_0_ff;
extern cpuop_func op_0428_0_nf;
extern cpuop_func op_0428_0_ff;
extern cpuop_func op_0430_0_nf;
extern cpuop_func op_0430_0_ff;
extern cpuop_func op_0438_0_nf;
extern cpuop_func op_0438_0_ff;
extern cpuop_func op_0439_0_nf;
extern cpuop_func op_0439_0_ff;
extern cpuop_func op_0440_0_nf;
extern cpuop_func op_0440_0_ff;
extern cpuop_func op_0450_0_nf;
extern cpuop_func op_0450_0_ff;
extern cpuop_func op_0458_0_nf;
extern cpuop_func op_0458_0_ff;
extern cpuop_func op_0460_0_nf;
extern cpuop_func op_0460_0_ff;
extern cpuop_func op_0468_0_nf;
extern cpuop_func op_0468_0_ff;
extern cpuop_func op_0470_0_nf;
extern cpuop_func op_0470_0_ff;
extern cpuop_func op_0478_0_nf;
extern cpuop_func op_0478_0_ff;
extern cpuop_func op_0479_0_nf;
extern cpuop_func op_0479_0_ff;
extern cpuop_func op_0480_0_nf;
extern cpuop_func op_0480_0_ff;
extern cpuop_func op_0490_0_nf;
extern cpuop_func op_0490_0_ff;
extern cpuop_func op_0498_0_nf;
extern cpuop_func op_0498_0_ff;
extern cpuop_func op_04a0_0_nf;
extern cpuop_func op_04a0_0_ff;
extern cpuop_func op_04a8_0_nf;
extern cpuop_func op_04a8_0_ff;
extern cpuop_func op_04b0_0_nf;
extern cpuop_func op_04b0_0_ff;
extern cpuop_func op_04b8_0_nf;
extern cpuop_func op_04b8_0_ff;
extern cpuop_func op_04b9_0_nf;
extern cpuop_func op_04b9_0_ff;
extern cpuop_func op_04d0_0_nf;
extern cpuop_func op_04d0_0_ff;
extern cpuop_func op_04e8_0_nf;
extern cpuop_func op_04e8_0_ff;
extern cpuop_func op_04f0_0_nf;
extern cpuop_func op_04f0_0_ff;
extern cpuop_func op_04f8_0_nf;
extern cpuop_func op_04f8_0_ff;
extern cpuop_func op_04f9_0_nf;
extern cpuop_func op_04f9_0_ff;
extern cpuop_func op_04fa_0_nf;
extern cpuop_func op_04fa_0_ff;
extern cpuop_func op_04fb_0_nf;
extern cpuop_func op_04fb_0_ff;
extern cpuop_func op_0600_0_nf;
extern cpuop_func op_0600_0_ff;
extern cpuop_func op_0610_0_nf;
extern cpuop_func op_0610_0_ff;
extern cpuop_func op_0618_0_nf;
extern cpuop_func op_0618_0_ff;
extern cpuop_func op_0620_0_nf;
extern cpuop_func op_0620_0_ff;
extern cpuop_func op_0628_0_nf;
extern cpuop_func op_0628_0_ff;
extern cpuop_func op_0630_0_nf;
extern cpuop_func op_0630_0_ff;
extern cpuop_func op_0638_0_nf;
extern cpuop_func op_0638_0_ff;
extern cpuop_func op_0639_0_nf;
extern cpuop_func op_0639_0_ff;
extern cpuop_func op_0640_0_nf;
extern cpuop_func op_0640_0_ff;
extern cpuop_func op_0650_0_nf;
extern cpuop_func op_0650_0_ff;
extern cpuop_func op_0658_0_nf;
extern cpuop_func op_0658_0_ff;
extern cpuop_func op_0660_0_nf;
extern cpuop_func op_0660_0_ff;
extern cpuop_func op_0668_0_nf;
extern cpuop_func op_0668_0_ff;
extern cpuop_func op_0670_0_nf;
extern cpuop_func op_0670_0_ff;
extern cpuop_func op_0678_0_nf;
extern cpuop_func op_0678_0_ff;
extern cpuop_func op_0679_0_nf;
extern cpuop_func op_0679_0_ff;
extern cpuop_func op_0680_0_nf;
extern cpuop_func op_0680_0_ff;
extern cpuop_func op_0690_0_nf;
extern cpuop_func op_0690_0_ff;
extern cpuop_func op_0698_0_nf;
extern cpuop_func op_0698_0_ff;
extern cpuop_func op_06a0_0_nf;
extern cpuop_func op_06a0_0_ff;
extern cpuop_func op_06a8_0_nf;
extern cpuop_func op_06a8_0_ff;
extern cpuop_func op_06b0_0_nf;
extern cpuop_func op_06b0_0_ff;
extern cpuop_func op_06b8_0_nf;
extern cpuop_func op_06b8_0_ff;
extern cpuop_func op_06b9_0_nf;
extern cpuop_func op_06b9_0_ff;
extern cpuop_func op_06c0_0_nf;
extern cpuop_func op_06c0_0_ff;
extern cpuop_func op_06c8_0_nf;
extern cpuop_func op_06c8_0_ff;
extern cpuop_func op_06d0_0_nf;
extern cpuop_func op_06d0_0_ff;
extern cpuop_func op_06e8_0_nf;
extern cpuop_func op_06e8_0_ff;
extern cpuop_func op_06f0_0_nf;
extern cpuop_func op_06f0_0_ff;
extern cpuop_func op_06f8_0_nf;
extern cpuop_func op_06f8_0_ff;
extern cpuop_func op_06f9_0_nf;
extern cpuop_func op_06f9_0_ff;
extern cpuop_func op_06fa_0_nf;
extern cpuop_func op_06fa_0_ff;
extern cpuop_func op_06fb_0_nf;
extern cpuop_func op_06fb_0_ff;
extern cpuop_func op_0800_0_nf;
extern cpuop_func op_0800_0_ff;
extern cpuop_func op_0810_0_nf;
extern cpuop_func op_0810_0_ff;
extern cpuop_func op_0818_0_nf;
extern cpuop_func op_0818_0_ff;
extern cpuop_func op_0820_0_nf;
extern cpuop_func op_0820_0_ff;
extern cpuop_func op_0828_0_nf;
extern cpuop_func op_0828_0_ff;
extern cpuop_func op_0830_0_nf;
extern cpuop_func op_0830_0_ff;
extern cpuop_func op_0838_0_nf;
extern cpuop_func op_0838_0_ff;
extern cpuop_func op_0839_0_nf;
extern cpuop_func op_0839_0_ff;
extern cpuop_func op_083a_0_nf;
extern cpuop_func op_083a_0_ff;
extern cpuop_func op_083b_0_nf;
extern cpuop_func op_083b_0_ff;
extern cpuop_func op_0840_0_nf;
extern cpuop_func op_0840_0_ff;
extern cpuop_func op_0850_0_nf;
extern cpuop_func op_0850_0_ff;
extern cpuop_func op_0858_0_nf;
extern cpuop_func op_0858_0_ff;
extern cpuop_func op_0860_0_nf;
extern cpuop_func op_0860_0_ff;
extern cpuop_func op_0868_0_nf;
extern cpuop_func op_0868_0_ff;
extern cpuop_func op_0870_0_nf;
extern cpuop_func op_0870_0_ff;
extern cpuop_func op_0878_0_nf;
extern cpuop_func op_0878_0_ff;
extern cpuop_func op_0879_0_nf;
extern cpuop_func op_0879_0_ff;
extern cpuop_func op_0880_0_nf;
extern cpuop_func op_0880_0_ff;
extern cpuop_func op_0890_0_nf;
extern cpuop_func op_0890_0_ff;
extern cpuop_func op_0898_0_nf;
extern cpuop_func op_0898_0_ff;
extern cpuop_func op_08a0_0_nf;
extern cpuop_func op_08a0_0_ff;
extern cpuop_func op_08a8_0_nf;
extern cpuop_func op_08a8_0_ff;
extern cpuop_func op_08b0_0_nf;
extern cpuop_func op_08b0_0_ff;
extern cpuop_func op_08b8_0_nf;
extern cpuop_func op_08b8_0_ff;
extern cpuop_func op_08b9_0_nf;
extern cpuop_func op_08b9_0_ff;
extern cpuop_func op_08c0_0_nf;
extern cpuop_func op_08c0_0_ff;
extern cpuop_func op_08d0_0_nf;
extern cpuop_func op_08d0_0_ff;
extern cpuop_func op_08d8_0_nf;
extern cpuop_func op_08d8_0_ff;
extern cpuop_func op_08e0_0_nf;
extern cpuop_func op_08e0_0_ff;
extern cpuop_func op_08e8_0_nf;
extern cpuop_func op_08e8_0_ff;
extern cpuop_func op_08f0_0_nf;
extern cpuop_func op_08f0_0_ff;
extern cpuop_func op_08f8_0_nf;
extern cpuop_func op_08f8_0_ff;
extern cpuop_func op_08f9_0_nf;
extern cpuop_func op_08f9_0_ff;
extern cpuop_func op_0a00_0_nf;
extern cpuop_func op_0a00_0_ff;
extern cpuop_func op_0a10_0_nf;
extern cpuop_func op_0a10_0_ff;
extern cpuop_func op_0a18_0_nf;
extern cpuop_func op_0a18_0_ff;
extern cpuop_func op_0a20_0_nf;
extern cpuop_func op_0a20_0_ff;
extern cpuop_func op_0a28_0_nf;
extern cpuop_func op_0a28_0_ff;
extern cpuop_func op_0a30_0_nf;
extern cpuop_func op_0a30_0_ff;
extern cpuop_func op_0a38_0_nf;
extern cpuop_func op_0a38_0_ff;
extern cpuop_func op_0a39_0_nf;
extern cpuop_func op_0a39_0_ff;
extern cpuop_func op_0a3c_0_nf;
extern cpuop_func op_0a3c_0_ff;
extern cpuop_func op_0a40_0_nf;
extern cpuop_func op_0a40_0_ff;
extern cpuop_func op_0a50_0_nf;
extern cpuop_func op_0a50_0_ff;
extern cpuop_func op_0a58_0_nf;
extern cpuop_func op_0a58_0_ff;
extern cpuop_func op_0a60_0_nf;
extern cpuop_func op_0a60_0_ff;
extern cpuop_func op_0a68_0_nf;
extern cpuop_func op_0a68_0_ff;
extern cpuop_func op_0a70_0_nf;
extern cpuop_func op_0a70_0_ff;
extern cpuop_func op_0a78_0_nf;
extern cpuop_func op_0a78_0_ff;
extern cpuop_func op_0a79_0_nf;
extern cpuop_func op_0a79_0_ff;
extern cpuop_func op_0a7c_0_nf;
extern cpuop_func op_0a7c_0_ff;
extern cpuop_func op_0a80_0_nf;
extern cpuop_func op_0a80_0_ff;
extern cpuop_func op_0a90_0_nf;
extern cpuop_func op_0a90_0_ff;
extern cpuop_func op_0a98_0_nf;
extern cpuop_func op_0a98_0_ff;
extern cpuop_func op_0aa0_0_nf;
extern cpuop_func op_0aa0_0_ff;
extern cpuop_func op_0aa8_0_nf;
extern cpuop_func op_0aa8_0_ff;
extern cpuop_func op_0ab0_0_nf;
extern cpuop_func op_0ab0_0_ff;
extern cpuop_func op_0ab8_0_nf;
extern cpuop_func op_0ab8_0_ff;
extern cpuop_func op_0ab9_0_nf;
extern cpuop_func op_0ab9_0_ff;
extern cpuop_func op_0ad0_0_nf;
extern cpuop_func op_0ad0_0_ff;
extern cpuop_func op_0ad8_0_nf;
extern cpuop_func op_0ad8_0_ff;
extern cpuop_func op_0ae0_0_nf;
extern cpuop_func op_0ae0_0_ff;
extern cpuop_func op_0ae8_0_nf;
extern cpuop_func op_0ae8_0_ff;
extern cpuop_func op_0af0_0_nf;
extern cpuop_func op_0af0_0_ff;
extern cpuop_func op_0af8_0_nf;
extern cpuop_func op_0af8_0_ff;
extern cpuop_func op_0af9_0_nf;
extern cpuop_func op_0af9_0_ff;
extern cpuop_func op_0c00_0_nf;
extern cpuop_func op_0c00_0_ff;
extern cpuop_func op_0c10_0_nf;
extern cpuop_func op_0c10_0_ff;
extern cpuop_func op_0c18_0_nf;
extern cpuop_func op_0c18_0_ff;
extern cpuop_func op_0c20_0_nf;
extern cpuop_func op_0c20_0_ff;
extern cpuop_func op_0c28_0_nf;
extern cpuop_func op_0c28_0_ff;
extern cpuop_func op_0c30_0_nf;
extern cpuop_func op_0c30_0_ff;
extern cpuop_func op_0c38_0_nf;
extern cpuop_func op_0c38_0_ff;
extern cpuop_func op_0c39_0_nf;
extern cpuop_func op_0c39_0_ff;
extern cpuop_func op_0c3a_0_nf;
extern cpuop_func op_0c3a_0_ff;
extern cpuop_func op_0c3b_0_nf;
extern cpuop_func op_0c3b_0_ff;
extern cpuop_func op_0c40_0_nf;
extern cpuop_func op_0c40_0_ff;
extern cpuop_func op_0c50_0_nf;
extern cpuop_func op_0c50_0_ff;
extern cpuop_func op_0c58_0_nf;
extern cpuop_func op_0c58_0_ff;
extern cpuop_func op_0c60_0_nf;
extern cpuop_func op_0c60_0_ff;
extern cpuop_func op_0c68_0_nf;
extern cpuop_func op_0c68_0_ff;
extern cpuop_func op_0c70_0_nf;
extern cpuop_func op_0c70_0_ff;
extern cpuop_func op_0c78_0_nf;
extern cpuop_func op_0c78_0_ff;
extern cpuop_func op_0c79_0_nf;
extern cpuop_func op_0c79_0_ff;
extern cpuop_func op_0c7a_0_nf;
extern cpuop_func op_0c7a_0_ff;
extern cpuop_func op_0c7b_0_nf;
extern cpuop_func op_0c7b_0_ff;
extern cpuop_func op_0c80_0_nf;
extern cpuop_func op_0c80_0_ff;
extern cpuop_func op_0c90_0_nf;
extern cpuop_func op_0c90_0_ff;
extern cpuop_func op_0c98_0_nf;
extern cpuop_func op_0c98_0_ff;
extern cpuop_func op_0ca0_0_nf;
extern cpuop_func op_0ca0_0_ff;
extern cpuop_func op_0ca8_0_nf;
extern cpuop_func op_0ca8_0_ff;
extern cpuop_func op_0cb0_0_nf;
extern cpuop_func op_0cb0_0_ff;
extern cpuop_func op_0cb8_0_nf;
extern cpuop_func op_0cb8_0_ff;
extern cpuop_func op_0cb9_0_nf;
extern cpuop_func op_0cb9_0_ff;
extern cpuop_func op_0cba_0_nf;
extern cpuop_func op_0cba_0_ff;
extern cpuop_func op_0cbb_0_nf;
extern cpuop_func op_0cbb_0_ff;
extern cpuop_func op_0cd0_0_nf;
extern cpuop_func op_0cd0_0_ff;
extern cpuop_func op_0cd8_0_nf;
extern cpuop_func op_0cd8_0_ff;
extern cpuop_func op_0ce0_0_nf;
extern cpuop_func op_0ce0_0_ff;
extern cpuop_func op_0ce8_0_nf;
extern cpuop_func op_0ce8_0_ff;
extern cpuop_func op_0cf0_0_nf;
extern cpuop_func op_0cf0_0_ff;
extern cpuop_func op_0cf8_0_nf;
extern cpuop_func op_0cf8_0_ff;
extern cpuop_func op_0cf9_0_nf;
extern cpuop_func op_0cf9_0_ff;
extern cpuop_func op_0cfc_0_nf;
extern cpuop_func op_0cfc_0_ff;
extern cpuop_func op_0e10_0_nf;
extern cpuop_func op_0e10_0_ff;
extern cpuop_func op_0e18_0_nf;
extern cpuop_func op_0e18_0_ff;
extern cpuop_func op_0e20_0_nf;
extern cpuop_func op_0e20_0_ff;
extern cpuop_func op_0e28_0_nf;
extern cpuop_func op_0e28_0_ff;
extern cpuop_func op_0e30_0_nf;
extern cpuop_func op_0e30_0_ff;
extern cpuop_func op_0e38_0_nf;
extern cpuop_func op_0e38_0_ff;
extern cpuop_func op_0e39_0_nf;
extern cpuop_func op_0e39_0_ff;
extern cpuop_func op_0e50_0_nf;
extern cpuop_func op_0e50_0_ff;
extern cpuop_func op_0e58_0_nf;
extern cpuop_func op_0e58_0_ff;
extern cpuop_func op_0e60_0_nf;
extern cpuop_func op_0e60_0_ff;
extern cpuop_func op_0e68_0_nf;
extern cpuop_func op_0e68_0_ff;
extern cpuop_func op_0e70_0_nf;
extern cpuop_func op_0e70_0_ff;
extern cpuop_func op_0e78_0_nf;
extern cpuop_func op_0e78_0_ff;
extern cpuop_func op_0e79_0_nf;
extern cpuop_func op_0e79_0_ff;
extern cpuop_func op_0e90_0_nf;
extern cpuop_func op_0e90_0_ff;
extern cpuop_func op_0e98_0_nf;
extern cpuop_func op_0e98_0_ff;
extern cpuop_func op_0ea0_0_nf;
extern cpuop_func op_0ea0_0_ff;
extern cpuop_func op_0ea8_0_nf;
extern cpuop_func op_0ea8_0_ff;
extern cpuop_func op_0eb0_0_nf;
extern cpuop_func op_0eb0_0_ff;
extern cpuop_func op_0eb8_0_nf;
extern cpuop_func op_0eb8_0_ff;
extern cpuop_func op_0eb9_0_nf;
extern cpuop_func op_0eb9_0_ff;
extern cpuop_func op_0ed0_0_nf;
extern cpuop_func op_0ed0_0_ff;
extern cpuop_func op_0ed8_0_nf;
extern cpuop_func op_0ed8_0_ff;
extern cpuop_func op_0ee0_0_nf;
extern cpuop_func op_0ee0_0_ff;
extern cpuop_func op_0ee8_0_nf;
extern cpuop_func op_0ee8_0_ff;
extern cpuop_func op_0ef0_0_nf;
extern cpuop_func op_0ef0_0_ff;
extern cpuop_func op_0ef8_0_nf;
extern cpuop_func op_0ef8_0_ff;
extern cpuop_func op_0ef9_0_nf;
extern cpuop_func op_0ef9_0_ff;
extern cpuop_func op_0efc_0_nf;
extern cpuop_func op_0efc_0_ff;
extern cpuop_func op_1000_0_nf;
extern cpuop_func op_1000_0_ff;
extern cpuop_func op_1010_0_nf;
extern cpuop_func op_1010_0_ff;
extern cpuop_func op_1018_0_nf;
extern cpuop_func op_1018_0_ff;
extern cpuop_func op_1020_0_nf;
extern cpuop_func op_1020_0_ff;
extern cpuop_func op_1028_0_nf;
extern cpuop_func op_1028_0_ff;
extern cpuop_func op_1030_0_nf;
extern cpuop_func op_1030_0_ff;
extern cpuop_func op_1038_0_nf;
extern cpuop_func op_1038_0_ff;
extern cpuop_func op_1039_0_nf;
extern cpuop_func op_1039_0_ff;
extern cpuop_func op_103a_0_nf;
extern cpuop_func op_103a_0_ff;
extern cpuop_func op_103b_0_nf;
extern cpuop_func op_103b_0_ff;
extern cpuop_func op_103c_0_nf;
extern cpuop_func op_103c_0_ff;
extern cpuop_func op_1080_0_nf;
extern cpuop_func op_1080_0_ff;
extern cpuop_func op_1090_0_nf;
extern cpuop_func op_1090_0_ff;
extern cpuop_func op_1098_0_nf;
extern cpuop_func op_1098_0_ff;
extern cpuop_func op_10a0_0_nf;
extern cpuop_func op_10a0_0_ff;
extern cpuop_func op_10a8_0_nf;
extern cpuop_func op_10a8_0_ff;
extern cpuop_func op_10b0_0_nf;
extern cpuop_func op_10b0_0_ff;
extern cpuop_func op_10b8_0_nf;
extern cpuop_func op_10b8_0_ff;
extern cpuop_func op_10b9_0_nf;
extern cpuop_func op_10b9_0_ff;
extern cpuop_func op_10ba_0_nf;
extern cpuop_func op_10ba_0_ff;
extern cpuop_func op_10bb_0_nf;
extern cpuop_func op_10bb_0_ff;
extern cpuop_func op_10bc_0_nf;
extern cpuop_func op_10bc_0_ff;
extern cpuop_func op_10c0_0_nf;
extern cpuop_func op_10c0_0_ff;
extern cpuop_func op_10d0_0_nf;
extern cpuop_func op_10d0_0_ff;
extern cpuop_func op_10d8_0_nf;
extern cpuop_func op_10d8_0_ff;
extern cpuop_func op_10e0_0_nf;
extern cpuop_func op_10e0_0_ff;
extern cpuop_func op_10e8_0_nf;
extern cpuop_func op_10e8_0_ff;
extern cpuop_func op_10f0_0_nf;
extern cpuop_func op_10f0_0_ff;
extern cpuop_func op_10f8_0_nf;
extern cpuop_func op_10f8_0_ff;
extern cpuop_func op_10f9_0_nf;
extern cpuop_func op_10f9_0_ff;
extern cpuop_func op_10fa_0_nf;
extern cpuop_func op_10fa_0_ff;
extern cpuop_func op_10fb_0_nf;
extern cpuop_func op_10fb_0_ff;
extern cpuop_func op_10fc_0_nf;
extern cpuop_func op_10fc_0_ff;
extern cpuop_func op_1100_0_nf;
extern cpuop_func op_1100_0_ff;
extern cpuop_func op_1110_0_nf;
extern cpuop_func op_1110_0_ff;
extern cpuop_func op_1118_0_nf;
extern cpuop_func op_1118_0_ff;
extern cpuop_func op_1120_0_nf;
extern cpuop_func op_1120_0_ff;
extern cpuop_func op_1128_0_nf;
extern cpuop_func op_1128_0_ff;
extern cpuop_func op_1130_0_nf;
extern cpuop_func op_1130_0_ff;
extern cpuop_func op_1138_0_nf;
extern cpuop_func op_1138_0_ff;
extern cpuop_func op_1139_0_nf;
extern cpuop_func op_1139_0_ff;
extern cpuop_func op_113a_0_nf;
extern cpuop_func op_113a_0_ff;
extern cpuop_func op_113b_0_nf;
extern cpuop_func op_113b_0_ff;
extern cpuop_func op_113c_0_nf;
extern cpuop_func op_113c_0_ff;
extern cpuop_func op_1140_0_nf;
extern cpuop_func op_1140_0_ff;
extern cpuop_func op_1150_0_nf;
extern cpuop_func op_1150_0_ff;
extern cpuop_func op_1158_0_nf;
extern cpuop_func op_1158_0_ff;
extern cpuop_func op_1160_0_nf;
extern cpuop_func op_1160_0_ff;
extern cpuop_func op_1168_0_nf;
extern cpuop_func op_1168_0_ff;
extern cpuop_func op_1170_0_nf;
extern cpuop_func op_1170_0_ff;
extern cpuop_func op_1178_0_nf;
extern cpuop_func op_1178_0_ff;
extern cpuop_func op_1179_0_nf;
extern cpuop_func op_1179_0_ff;
extern cpuop_func op_117a_0_nf;
extern cpuop_func op_117a_0_ff;
extern cpuop_func op_117b_0_nf;
extern cpuop_func op_117b_0_ff;
extern cpuop_func op_117c_0_nf;
extern cpuop_func op_117c_0_ff;
extern cpuop_func op_1180_0_nf;
extern cpuop_func op_1180_0_ff;
extern cpuop_func op_1190_0_nf;
extern cpuop_func op_1190_0_ff;
extern cpuop_func op_1198_0_nf;
extern cpuop_func op_1198_0_ff;
extern cpuop_func op_11a0_0_nf;
extern cpuop_func op_11a0_0_ff;
extern cpuop_func op_11a8_0_nf;
extern cpuop_func op_11a8_0_ff;
extern cpuop_func op_11b0_0_nf;
extern cpuop_func op_11b0_0_ff;
extern cpuop_func op_11b8_0_nf;
extern cpuop_func op_11b8_0_ff;
extern cpuop_func op_11b9_0_nf;
extern cpuop_func op_11b9_0_ff;
extern cpuop_func op_11ba_0_nf;
extern cpuop_func op_11ba_0_ff;
extern cpuop_func op_11bb_0_nf;
extern cpuop_func op_11bb_0_ff;
extern cpuop_func op_11bc_0_nf;
extern cpuop_func op_11bc_0_ff;
extern cpuop_func op_11c0_0_nf;
extern cpuop_func op_11c0_0_ff;
extern cpuop_func op_11d0_0_nf;
extern cpuop_func op_11d0_0_ff;
extern cpuop_func op_11d8_0_nf;
extern cpuop_func op_11d8_0_ff;
extern cpuop_func op_11e0_0_nf;
extern cpuop_func op_11e0_0_ff;
extern cpuop_func op_11e8_0_nf;
extern cpuop_func op_11e8_0_ff;
extern cpuop_func op_11f0_0_nf;
extern cpuop_func op_11f0_0_ff;
extern cpuop_func op_11f8_0_nf;
extern cpuop_func op_11f8_0_ff;
extern cpuop_func op_11f9_0_nf;
extern cpuop_func op_11f9_0_ff;
extern cpuop_func op_11fa_0_nf;
extern cpuop_func op_11fa_0_ff;
extern cpuop_func op_11fb_0_nf;
extern cpuop_func op_11fb_0_ff;
extern cpuop_func op_11fc_0_nf;
extern cpuop_func op_11fc_0_ff;
extern cpuop_func op_13c0_0_nf;
extern cpuop_func op_13c0_0_ff;
extern cpuop_func op_13d0_0_nf;
extern cpuop_func op_13d0_0_ff;
extern cpuop_func op_13d8_0_nf;
extern cpuop_func op_13d8_0_ff;
extern cpuop_func op_13e0_0_nf;
extern cpuop_func op_13e0_0_ff;
extern cpuop_func op_13e8_0_nf;
extern cpuop_func op_13e8_0_ff;
extern cpuop_func op_13f0_0_nf;
extern cpuop_func op_13f0_0_ff;
extern cpuop_func op_13f8_0_nf;
extern cpuop_func op_13f8_0_ff;
extern cpuop_func op_13f9_0_nf;
extern cpuop_func op_13f9_0_ff;
extern cpuop_func op_13fa_0_nf;
extern cpuop_func op_13fa_0_ff;
extern cpuop_func op_13fb_0_nf;
extern cpuop_func op_13fb_0_ff;
extern cpuop_func op_13fc_0_nf;
extern cpuop_func op_13fc_0_ff;
extern cpuop_func op_2000_0_nf;
extern cpuop_func op_2000_0_ff;
extern cpuop_func op_2008_0_nf;
extern cpuop_func op_2008_0_ff;
extern cpuop_func op_2010_0_nf;
extern cpuop_func op_2010_0_ff;
extern cpuop_func op_2018_0_nf;
extern cpuop_func op_2018_0_ff;
extern cpuop_func op_2020_0_nf;
extern cpuop_func op_2020_0_ff;
extern cpuop_func op_2028_0_nf;
extern cpuop_func op_2028_0_ff;
extern cpuop_func op_2030_0_nf;
extern cpuop_func op_2030_0_ff;
extern cpuop_func op_2038_0_nf;
extern cpuop_func op_2038_0_ff;
extern cpuop_func op_2039_0_nf;
extern cpuop_func op_2039_0_ff;
extern cpuop_func op_203a_0_nf;
extern cpuop_func op_203a_0_ff;
extern cpuop_func op_203b_0_nf;
extern cpuop_func op_203b_0_ff;
extern cpuop_func op_203c_0_nf;
extern cpuop_func op_203c_0_ff;
extern cpuop_func op_2040_0_nf;
extern cpuop_func op_2040_0_ff;
extern cpuop_func op_2048_0_nf;
extern cpuop_func op_2048_0_ff;
extern cpuop_func op_2050_0_nf;
extern cpuop_func op_2050_0_ff;
extern cpuop_func op_2058_0_nf;
extern cpuop_func op_2058_0_ff;
extern cpuop_func op_2060_0_nf;
extern cpuop_func op_2060_0_ff;
extern cpuop_func op_2068_0_nf;
extern cpuop_func op_2068_0_ff;
extern cpuop_func op_2070_0_nf;
extern cpuop_func op_2070_0_ff;
extern cpuop_func op_2078_0_nf;
extern cpuop_func op_2078_0_ff;
extern cpuop_func op_2079_0_nf;
extern cpuop_func op_2079_0_ff;
extern cpuop_func op_207a_0_nf;
extern cpuop_func op_207a_0_ff;
extern cpuop_func op_207b_0_nf;
extern cpuop_func op_207b_0_ff;
extern cpuop_func op_207c_0_nf;
extern cpuop_func op_207c_0_ff;
extern cpuop_func op_2080_0_nf;
extern cpuop_func op_2080_0_ff;
extern cpuop_func op_2088_0_nf;
extern cpuop_func op_2088_0_ff;
extern cpuop_func op_2090_0_nf;
extern cpuop_func op_2090_0_ff;
extern cpuop_func op_2098_0_nf;
extern cpuop_func op_2098_0_ff;
extern cpuop_func op_20a0_0_nf;
extern cpuop_func op_20a0_0_ff;
extern cpuop_func op_20a8_0_nf;
extern cpuop_func op_20a8_0_ff;
extern cpuop_func op_20b0_0_nf;
extern cpuop_func op_20b0_0_ff;
extern cpuop_func op_20b8_0_nf;
extern cpuop_func op_20b8_0_ff;
extern cpuop_func op_20b9_0_nf;
extern cpuop_func op_20b9_0_ff;
extern cpuop_func op_20ba_0_nf;
extern cpuop_func op_20ba_0_ff;
extern cpuop_func op_20bb_0_nf;
extern cpuop_func op_20bb_0_ff;
extern cpuop_func op_20bc_0_nf;
extern cpuop_func op_20bc_0_ff;
extern cpuop_func op_20c0_0_nf;
extern cpuop_func op_20c0_0_ff;
extern cpuop_func op_20c8_0_nf;
extern cpuop_func op_20c8_0_ff;
extern cpuop_func op_20d0_0_nf;
extern cpuop_func op_20d0_0_ff;
extern cpuop_func op_20d8_0_nf;
extern cpuop_func op_20d8_0_ff;
extern cpuop_func op_20e0_0_nf;
extern cpuop_func op_20e0_0_ff;
extern cpuop_func op_20e8_0_nf;
extern cpuop_func op_20e8_0_ff;
extern cpuop_func op_20f0_0_nf;
extern cpuop_func op_20f0_0_ff;
extern cpuop_func op_20f8_0_nf;
extern cpuop_func op_20f8_0_ff;
extern cpuop_func op_20f9_0_nf;
extern cpuop_func op_20f9_0_ff;
extern cpuop_func op_20fa_0_nf;
extern cpuop_func op_20fa_0_ff;
extern cpuop_func op_20fb_0_nf;
extern cpuop_func op_20fb_0_ff;
extern cpuop_func op_20fc_0_nf;
extern cpuop_func op_20fc_0_ff;
extern cpuop_func op_2100_0_nf;
extern cpuop_func op_2100_0_ff;
extern cpuop_func op_2108_0_nf;
extern cpuop_func op_2108_0_ff;
extern cpuop_func op_2110_0_nf;
extern cpuop_func op_2110_0_ff;
extern cpuop_func op_2118_0_nf;
extern cpuop_func op_2118_0_ff;
extern cpuop_func op_2120_0_nf;
extern cpuop_func op_2120_0_ff;
extern cpuop_func op_2128_0_nf;
extern cpuop_func op_2128_0_ff;
extern cpuop_func op_2130_0_nf;
extern cpuop_func op_2130_0_ff;
extern cpuop_func op_2138_0_nf;
extern cpuop_func op_2138_0_ff;
extern cpuop_func op_2139_0_nf;
extern cpuop_func op_2139_0_ff;
extern cpuop_func op_213a_0_nf;
extern cpuop_func op_213a_0_ff;
extern cpuop_func op_213b_0_nf;
extern cpuop_func op_213b_0_ff;
extern cpuop_func op_213c_0_nf;
extern cpuop_func op_213c_0_ff;
extern cpuop_func op_2140_0_nf;
extern cpuop_func op_2140_0_ff;
extern cpuop_func op_2148_0_nf;
extern cpuop_func op_2148_0_ff;
extern cpuop_func op_2150_0_nf;
extern cpuop_func op_2150_0_ff;
extern cpuop_func op_2158_0_nf;
extern cpuop_func op_2158_0_ff;
extern cpuop_func op_2160_0_nf;
extern cpuop_func op_2160_0_ff;
extern cpuop_func op_2168_0_nf;
extern cpuop_func op_2168_0_ff;
extern cpuop_func op_2170_0_nf;
extern cpuop_func op_2170_0_ff;
extern cpuop_func op_2178_0_nf;
extern cpuop_func op_2178_0_ff;
extern cpuop_func op_2179_0_nf;
extern cpuop_func op_2179_0_ff;
extern cpuop_func op_217a_0_nf;
extern cpuop_func op_217a_0_ff;
extern cpuop_func op_217b_0_nf;
extern cpuop_func op_217b_0_ff;
extern cpuop_func op_217c_0_nf;
extern cpuop_func op_217c_0_ff;
extern cpuop_func op_2180_0_nf;
extern cpuop_func op_2180_0_ff;
extern cpuop_func op_2188_0_nf;
extern cpuop_func op_2188_0_ff;
extern cpuop_func op_2190_0_nf;
extern cpuop_func op_2190_0_ff;
extern cpuop_func op_2198_0_nf;
extern cpuop_func op_2198_0_ff;
extern cpuop_func op_21a0_0_nf;
extern cpuop_func op_21a0_0_ff;
extern cpuop_func op_21a8_0_nf;
extern cpuop_func op_21a8_0_ff;
extern cpuop_func op_21b0_0_nf;
extern cpuop_func op_21b0_0_ff;
extern cpuop_func op_21b8_0_nf;
extern cpuop_func op_21b8_0_ff;
extern cpuop_func op_21b9_0_nf;
extern cpuop_func op_21b9_0_ff;
extern cpuop_func op_21ba_0_nf;
extern cpuop_func op_21ba_0_ff;
extern cpuop_func op_21bb_0_nf;
extern cpuop_func op_21bb_0_ff;
extern cpuop_func op_21bc_0_nf;
extern cpuop_func op_21bc_0_ff;
extern cpuop_func op_21c0_0_nf;
extern cpuop_func op_21c0_0_ff;
extern cpuop_func op_21c8_0_nf;
extern cpuop_func op_21c8_0_ff;
extern cpuop_func op_21d0_0_nf;
extern cpuop_func op_21d0_0_ff;
extern cpuop_func op_21d8_0_nf;
extern cpuop_func op_21d8_0_ff;
extern cpuop_func op_21e0_0_nf;
extern cpuop_func op_21e0_0_ff;
extern cpuop_func op_21e8_0_nf;
extern cpuop_func op_21e8_0_ff;
extern cpuop_func op_21f0_0_nf;
extern cpuop_func op_21f0_0_ff;
extern cpuop_func op_21f8_0_nf;
extern cpuop_func op_21f8_0_ff;
extern cpuop_func op_21f9_0_nf;
extern cpuop_func op_21f9_0_ff;
extern cpuop_func op_21fa_0_nf;
extern cpuop_func op_21fa_0_ff;
extern cpuop_func op_21fb_0_nf;
extern cpuop_func op_21fb_0_ff;
extern cpuop_func op_21fc_0_nf;
extern cpuop_func op_21fc_0_ff;
extern cpuop_func op_23c0_0_nf;
extern cpuop_func op_23c0_0_ff;
extern cpuop_func op_23c8_0_nf;
extern cpuop_func op_23c8_0_ff;
extern cpuop_func op_23d0_0_nf;
extern cpuop_func op_23d0_0_ff;
extern cpuop_func op_23d8_0_nf;
extern cpuop_func op_23d8_0_ff;
extern cpuop_func op_23e0_0_nf;
extern cpuop_func op_23e0_0_ff;
extern cpuop_func op_23e8_0_nf;
extern cpuop_func op_23e8_0_ff;
extern cpuop_func op_23f0_0_nf;
extern cpuop_func op_23f0_0_ff;
extern cpuop_func op_23f8_0_nf;
extern cpuop_func op_23f8_0_ff;
extern cpuop_func op_23f9_0_nf;
extern cpuop_func op_23f9_0_ff;
extern cpuop_func op_23fa_0_nf;
extern cpuop_func op_23fa_0_ff;
extern cpuop_func op_23fb_0_nf;
extern cpuop_func op_23fb_0_ff;
extern cpuop_func op_23fc_0_nf;
extern cpuop_func op_23fc_0_ff;
extern cpuop_func op_3000_0_nf;
extern cpuop_func op_3000_0_ff;
extern cpuop_func op_3008_0_nf;
extern cpuop_func op_3008_0_ff;
extern cpuop_func op_3010_0_nf;
extern cpuop_func op_3010_0_ff;
extern cpuop_func op_3018_0_nf;
extern cpuop_func op_3018_0_ff;
extern cpuop_func op_3020_0_nf;
extern cpuop_func op_3020_0_ff;
extern cpuop_func op_3028_0_nf;
extern cpuop_func op_3028_0_ff;
extern cpuop_func op_3030_0_nf;
extern cpuop_func op_3030_0_ff;
extern cpuop_func op_3038_0_nf;
extern cpuop_func op_3038_0_ff;
extern cpuop_func op_3039_0_nf;
extern cpuop_func op_3039_0_ff;
extern cpuop_func op_303a_0_nf;
extern cpuop_func op_303a_0_ff;
extern cpuop_func op_303b_0_nf;
extern cpuop_func op_303b_0_ff;
extern cpuop_func op_303c_0_nf;
extern cpuop_func op_303c_0_ff;
extern cpuop_func op_3040_0_nf;
extern cpuop_func op_3040_0_ff;
extern cpuop_func op_3048_0_nf;
extern cpuop_func op_3048_0_ff;
extern cpuop_func op_3050_0_nf;
extern cpuop_func op_3050_0_ff;
extern cpuop_func op_3058_0_nf;
extern cpuop_func op_3058_0_ff;
extern cpuop_func op_3060_0_nf;
extern cpuop_func op_3060_0_ff;
extern cpuop_func op_3068_0_nf;
extern cpuop_func op_3068_0_ff;
extern cpuop_func op_3070_0_nf;
extern cpuop_func op_3070_0_ff;
extern cpuop_func op_3078_0_nf;
extern cpuop_func op_3078_0_ff;
extern cpuop_func op_3079_0_nf;
extern cpuop_func op_3079_0_ff;
extern cpuop_func op_307a_0_nf;
extern cpuop_func op_307a_0_ff;
extern cpuop_func op_307b_0_nf;
extern cpuop_func op_307b_0_ff;
extern cpuop_func op_307c_0_nf;
extern cpuop_func op_307c_0_ff;
extern cpuop_func op_3080_0_nf;
extern cpuop_func op_3080_0_ff;
extern cpuop_func op_3088_0_nf;
extern cpuop_func op_3088_0_ff;
extern cpuop_func op_3090_0_nf;
extern cpuop_func op_3090_0_ff;
extern cpuop_func op_3098_0_nf;
extern cpuop_func op_3098_0_ff;
extern cpuop_func op_30a0_0_nf;
extern cpuop_func op_30a0_0_ff;
extern cpuop_func op_30a8_0_nf;
extern cpuop_func op_30a8_0_ff;
extern cpuop_func op_30b0_0_nf;
extern cpuop_func op_30b0_0_ff;
extern cpuop_func op_30b8_0_nf;
extern cpuop_func op_30b8_0_ff;
extern cpuop_func op_30b9_0_nf;
extern cpuop_func op_30b9_0_ff;
extern cpuop_func op_30ba_0_nf;
extern cpuop_func op_30ba_0_ff;
extern cpuop_func op_30bb_0_nf;
extern cpuop_func op_30bb_0_ff;
extern cpuop_func op_30bc_0_nf;
extern cpuop_func op_30bc_0_ff;
extern cpuop_func op_30c0_0_nf;
extern cpuop_func op_30c0_0_ff;
extern cpuop_func op_30c8_0_nf;
extern cpuop_func op_30c8_0_ff;
extern cpuop_func op_30d0_0_nf;
extern cpuop_func op_30d0_0_ff;
extern cpuop_func op_30d8_0_nf;
extern cpuop_func op_30d8_0_ff;
extern cpuop_func op_30e0_0_nf;
extern cpuop_func op_30e0_0_ff;
extern cpuop_func op_30e8_0_nf;
extern cpuop_func op_30e8_0_ff;
extern cpuop_func op_30f0_0_nf;
extern cpuop_func op_30f0_0_ff;
extern cpuop_func op_30f8_0_nf;
extern cpuop_func op_30f8_0_ff;
extern cpuop_func op_30f9_0_nf;
extern cpuop_func op_30f9_0_ff;
extern cpuop_func op_30fa_0_nf;
extern cpuop_func op_30fa_0_ff;
extern cpuop_func op_30fb_0_nf;
extern cpuop_func op_30fb_0_ff;
extern cpuop_func op_30fc_0_nf;
extern cpuop_func op_30fc_0_ff;
extern cpuop_func op_3100_0_nf;
extern cpuop_func op_3100_0_ff;
extern cpuop_func op_3108_0_nf;
extern cpuop_func op_3108_0_ff;
extern cpuop_func op_3110_0_nf;
extern cpuop_func op_3110_0_ff;
extern cpuop_func op_3118_0_nf;
extern cpuop_func op_3118_0_ff;
extern cpuop_func op_3120_0_nf;
extern cpuop_func op_3120_0_ff;
extern cpuop_func op_3128_0_nf;
extern cpuop_func op_3128_0_ff;
extern cpuop_func op_3130_0_nf;
extern cpuop_func op_3130_0_ff;
extern cpuop_func op_3138_0_nf;
extern cpuop_func op_3138_0_ff;
extern cpuop_func op_3139_0_nf;
extern cpuop_func op_3139_0_ff;
extern cpuop_func op_313a_0_nf;
extern cpuop_func op_313a_0_ff;
extern cpuop_func op_313b_0_nf;
extern cpuop_func op_313b_0_ff;
extern cpuop_func op_313c_0_nf;
extern cpuop_func op_313c_0_ff;
extern cpuop_func op_3140_0_nf;
extern cpuop_func op_3140_0_ff;
extern cpuop_func op_3148_0_nf;
extern cpuop_func op_3148_0_ff;
extern cpuop_func op_3150_0_nf;
extern cpuop_func op_3150_0_ff;
extern cpuop_func op_3158_0_nf;
extern cpuop_func op_3158_0_ff;
extern cpuop_func op_3160_0_nf;
extern cpuop_func op_3160_0_ff;
extern cpuop_func op_3168_0_nf;
extern cpuop_func op_3168_0_ff;
extern cpuop_func op_3170_0_nf;
extern cpuop_func op_3170_0_ff;
extern cpuop_func op_3178_0_nf;
extern cpuop_func op_3178_0_ff;
extern cpuop_func op_3179_0_nf;
extern cpuop_func op_3179_0_ff;
extern cpuop_func op_317a_0_nf;
extern cpuop_func op_317a_0_ff;
extern cpuop_func op_317b_0_nf;
extern cpuop_func op_317b_0_ff;
extern cpuop_func op_317c_0_nf;
extern cpuop_func op_317c_0_ff;
extern cpuop_func op_3180_0_nf;
extern cpuop_func op_3180_0_ff;
extern cpuop_func op_3188_0_nf;
extern cpuop_func op_3188_0_ff;
extern cpuop_func op_3190_0_nf;
extern cpuop_func op_3190_0_ff;
extern cpuop_func op_3198_0_nf;
extern cpuop_func op_3198_0_ff;
extern cpuop_func op_31a0_0_nf;
extern cpuop_func op_31a0_0_ff;
extern cpuop_func op_31a8_0_nf;
extern cpuop_func op_31a8_0_ff;
extern cpuop_func op_31b0_0_nf;
extern cpuop_func op_31b0_0_ff;
extern cpuop_func op_31b8_0_nf;
extern cpuop_func op_31b8_0_ff;
extern cpuop_func op_31b9_0_nf;
extern cpuop_func op_31b9_0_ff;
extern cpuop_func op_31ba_0_nf;
extern cpuop_func op_31ba_0_ff;
extern cpuop_func op_31bb_0_nf;
extern cpuop_func op_31bb_0_ff;
extern cpuop_func op_31bc_0_nf;
extern cpuop_func op_31bc_0_ff;
extern cpuop_func op_31c0_0_nf;
extern cpuop_func op_31c0_0_ff;
extern cpuop_func op_31c8_0_nf;
extern cpuop_func op_31c8_0_ff;
extern cpuop_func op_31d0_0_nf;
extern cpuop_func op_31d0_0_ff;
extern cpuop_func op_31d8_0_nf;
extern cpuop_func op_31d8_0_ff;
extern cpuop_func op_31e0_0_nf;
extern cpuop_func op_31e0_0_ff;
extern cpuop_func op_31e8_0_nf;
extern cpuop_func op_31e8_0_ff;
extern cpuop_func op_31f0_0_nf;
extern cpuop_func op_31f0_0_ff;
extern cpuop_func op_31f8_0_nf;
extern cpuop_func op_31f8_0_ff;
extern cpuop_func op_31f9_0_nf;
extern cpuop_func op_31f9_0_ff;
extern cpuop_func op_31fa_0_nf;
extern cpuop_func op_31fa_0_ff;
extern cpuop_func op_31fb_0_nf;
extern cpuop_func op_31fb_0_ff;
extern cpuop_func op_31fc_0_nf;
extern cpuop_func op_31fc_0_ff;
extern cpuop_func op_33c0_0_nf;
extern cpuop_func op_33c0_0_ff;
extern cpuop_func op_33c8_0_nf;
extern cpuop_func op_33c8_0_ff;
extern cpuop_func op_33d0_0_nf;
extern cpuop_func op_33d0_0_ff;
extern cpuop_func op_33d8_0_nf;
extern cpuop_func op_33d8_0_ff;
extern cpuop_func op_33e0_0_nf;
extern cpuop_func op_33e0_0_ff;
extern cpuop_func op_33e8_0_nf;
extern cpuop_func op_33e8_0_ff;
extern cpuop_func op_33f0_0_nf;
extern cpuop_func op_33f0_0_ff;
extern cpuop_func op_33f8_0_nf;
extern cpuop_func op_33f8_0_ff;
extern cpuop_func op_33f9_0_nf;
extern cpuop_func op_33f9_0_ff;
extern cpuop_func op_33fa_0_nf;
extern cpuop_func op_33fa_0_ff;
extern cpuop_func op_33fb_0_nf;
extern cpuop_func op_33fb_0_ff;
extern cpuop_func op_33fc_0_nf;
extern cpuop_func op_33fc_0_ff;
extern cpuop_func op_4000_0_nf;
extern cpuop_func op_4000_0_ff;
extern cpuop_func op_4010_0_nf;
extern cpuop_func op_4010_0_ff;
extern cpuop_func op_4018_0_nf;
extern cpuop_func op_4018_0_ff;
extern cpuop_func op_4020_0_nf;
extern cpuop_func op_4020_0_ff;
extern cpuop_func op_4028_0_nf;
extern cpuop_func op_4028_0_ff;
extern cpuop_func op_4030_0_nf;
extern cpuop_func op_4030_0_ff;
extern cpuop_func op_4038_0_nf;
extern cpuop_func op_4038_0_ff;
extern cpuop_func op_4039_0_nf;
extern cpuop_func op_4039_0_ff;
extern cpuop_func op_4040_0_nf;
extern cpuop_func op_4040_0_ff;
extern cpuop_func op_4050_0_nf;
extern cpuop_func op_4050_0_ff;
extern cpuop_func op_4058_0_nf;
extern cpuop_func op_4058_0_ff;
extern cpuop_func op_4060_0_nf;
extern cpuop_func op_4060_0_ff;
extern cpuop_func op_4068_0_nf;
extern cpuop_func op_4068_0_ff;
extern cpuop_func op_4070_0_nf;
extern cpuop_func op_4070_0_ff;
extern cpuop_func op_4078_0_nf;
extern cpuop_func op_4078_0_ff;
extern cpuop_func op_4079_0_nf;
extern cpuop_func op_4079_0_ff;
extern cpuop_func op_4080_0_nf;
extern cpuop_func op_4080_0_ff;
extern cpuop_func op_4090_0_nf;
extern cpuop_func op_4090_0_ff;
extern cpuop_func op_4098_0_nf;
extern cpuop_func op_4098_0_ff;
extern cpuop_func op_40a0_0_nf;
extern cpuop_func op_40a0_0_ff;
extern cpuop_func op_40a8_0_nf;
extern cpuop_func op_40a8_0_ff;
extern cpuop_func op_40b0_0_nf;
extern cpuop_func op_40b0_0_ff;
extern cpuop_func op_40b8_0_nf;
extern cpuop_func op_40b8_0_ff;
extern cpuop_func op_40b9_0_nf;
extern cpuop_func op_40b9_0_ff;
extern cpuop_func op_40c0_0_nf;
extern cpuop_func op_40c0_0_ff;
extern cpuop_func op_40d0_0_nf;
extern cpuop_func op_40d0_0_ff;
extern cpuop_func op_40d8_0_nf;
extern cpuop_func op_40d8_0_ff;
extern cpuop_func op_40e0_0_nf;
extern cpuop_func op_40e0_0_ff;
extern cpuop_func op_40e8_0_nf;
extern cpuop_func op_40e8_0_ff;
extern cpuop_func op_40f0_0_nf;
extern cpuop_func op_40f0_0_ff;
extern cpuop_func op_40f8_0_nf;
extern cpuop_func op_40f8_0_ff;
extern cpuop_func op_40f9_0_nf;
extern cpuop_func op_40f9_0_ff;
extern cpuop_func op_4100_0_nf;
extern cpuop_func op_4100_0_ff;
extern cpuop_func op_4110_0_nf;
extern cpuop_func op_4110_0_ff;
extern cpuop_func op_4118_0_nf;
extern cpuop_func op_4118_0_ff;
extern cpuop_func op_4120_0_nf;
extern cpuop_func op_4120_0_ff;
extern cpuop_func op_4128_0_nf;
extern cpuop_func op_4128_0_ff;
extern cpuop_func op_4130_0_nf;
extern cpuop_func op_4130_0_ff;
extern cpuop_func op_4138_0_nf;
extern cpuop_func op_4138_0_ff;
extern cpuop_func op_4139_0_nf;
extern cpuop_func op_4139_0_ff;
extern cpuop_func op_413a_0_nf;
extern cpuop_func op_413a_0_ff;
extern cpuop_func op_413b_0_nf;
extern cpuop_func op_413b_0_ff;
extern cpuop_func op_413c_0_nf;
extern cpuop_func op_413c_0_ff;
extern cpuop_func op_4180_0_nf;
extern cpuop_func op_4180_0_ff;
extern cpuop_func op_4190_0_nf;
extern cpuop_func op_4190_0_ff;
extern cpuop_func op_4198_0_nf;
extern cpuop_func op_4198_0_ff;
extern cpuop_func op_41a0_0_nf;
extern cpuop_func op_41a0_0_ff;
extern cpuop_func op_41a8_0_nf;
extern cpuop_func op_41a8_0_ff;
extern cpuop_func op_41b0_0_nf;
extern cpuop_func op_41b0_0_ff;
extern cpuop_func op_41b8_0_nf;
extern cpuop_func op_41b8_0_ff;
extern cpuop_func op_41b9_0_nf;
extern cpuop_func op_41b9_0_ff;
extern cpuop_func op_41ba_0_nf;
extern cpuop_func op_41ba_0_ff;
extern cpuop_func op_41bb_0_nf;
extern cpuop_func op_41bb_0_ff;
extern cpuop_func op_41bc_0_nf;
extern cpuop_func op_41bc_0_ff;
extern cpuop_func op_41d0_0_nf;
extern cpuop_func op_41d0_0_ff;
extern cpuop_func op_41e8_0_nf;
extern cpuop_func op_41e8_0_ff;
extern cpuop_func op_41f0_0_nf;
extern cpuop_func op_41f0_0_ff;
extern cpuop_func op_41f8_0_nf;
extern cpuop_func op_41f8_0_ff;
extern cpuop_func op_41f9_0_nf;
extern cpuop_func op_41f9_0_ff;
extern cpuop_func op_41fa_0_nf;
extern cpuop_func op_41fa_0_ff;
extern cpuop_func op_41fb_0_nf;
extern cpuop_func op_41fb_0_ff;
extern cpuop_func op_4200_0_nf;
extern cpuop_func op_4200_0_ff;
extern cpuop_func op_4210_0_nf;
extern cpuop_func op_4210_0_ff;
extern cpuop_func op_4218_0_nf;
extern cpuop_func op_4218_0_ff;
extern cpuop_func op_4220_0_nf;
extern cpuop_func op_4220_0_ff;
extern cpuop_func op_4228_0_nf;
extern cpuop_func op_4228_0_ff;
extern cpuop_func op_4230_0_nf;
extern cpuop_func op_4230_0_ff;
extern cpuop_func op_4238_0_nf;
extern cpuop_func op_4238_0_ff;
extern cpuop_func op_4239_0_nf;
extern cpuop_func op_4239_0_ff;
extern cpuop_func op_4240_0_nf;
extern cpuop_func op_4240_0_ff;
extern cpuop_func op_4250_0_nf;
extern cpuop_func op_4250_0_ff;
extern cpuop_func op_4258_0_nf;
extern cpuop_func op_4258_0_ff;
extern cpuop_func op_4260_0_nf;
extern cpuop_func op_4260_0_ff;
extern cpuop_func op_4268_0_nf;
extern cpuop_func op_4268_0_ff;
extern cpuop_func op_4270_0_nf;
extern cpuop_func op_4270_0_ff;
extern cpuop_func op_4278_0_nf;
extern cpuop_func op_4278_0_ff;
extern cpuop_func op_4279_0_nf;
extern cpuop_func op_4279_0_ff;
extern cpuop_func op_4280_0_nf;
extern cpuop_func op_4280_0_ff;
extern cpuop_func op_4290_0_nf;
extern cpuop_func op_4290_0_ff;
extern cpuop_func op_4298_0_nf;
extern cpuop_func op_4298_0_ff;
extern cpuop_func op_42a0_0_nf;
extern cpuop_func op_42a0_0_ff;
extern cpuop_func op_42a8_0_nf;
extern cpuop_func op_42a8_0_ff;
extern cpuop_func op_42b0_0_nf;
extern cpuop_func op_42b0_0_ff;
extern cpuop_func op_42b8_0_nf;
extern cpuop_func op_42b8_0_ff;
extern cpuop_func op_42b9_0_nf;
extern cpuop_func op_42b9_0_ff;
extern cpuop_func op_42c0_0_nf;
extern cpuop_func op_42c0_0_ff;
extern cpuop_func op_42d0_0_nf;
extern cpuop_func op_42d0_0_ff;
extern cpuop_func op_42d8_0_nf;
extern cpuop_func op_42d8_0_ff;
extern cpuop_func op_42e0_0_nf;
extern cpuop_func op_42e0_0_ff;
extern cpuop_func op_42e8_0_nf;
extern cpuop_func op_42e8_0_ff;
extern cpuop_func op_42f0_0_nf;
extern cpuop_func op_42f0_0_ff;
extern cpuop_func op_42f8_0_nf;
extern cpuop_func op_42f8_0_ff;
extern cpuop_func op_42f9_0_nf;
extern cpuop_func op_42f9_0_ff;
extern cpuop_func op_4400_0_nf;
extern cpuop_func op_4400_0_ff;
extern cpuop_func op_4410_0_nf;
extern cpuop_func op_4410_0_ff;
extern cpuop_func op_4418_0_nf;
extern cpuop_func op_4418_0_ff;
extern cpuop_func op_4420_0_nf;
extern cpuop_func op_4420_0_ff;
extern cpuop_func op_4428_0_nf;
extern cpuop_func op_4428_0_ff;
extern cpuop_func op_4430_0_nf;
extern cpuop_func op_4430_0_ff;
extern cpuop_func op_4438_0_nf;
extern cpuop_func op_4438_0_ff;
extern cpuop_func op_4439_0_nf;
extern cpuop_func op_4439_0_ff;
extern cpuop_func op_4440_0_nf;
extern cpuop_func op_4440_0_ff;
extern cpuop_func op_4450_0_nf;
extern cpuop_func op_4450_0_ff;
extern cpuop_func op_4458_0_nf;
extern cpuop_func op_4458_0_ff;
extern cpuop_func op_4460_0_nf;
extern cpuop_func op_4460_0_ff;
extern cpuop_func op_4468_0_nf;
extern cpuop_func op_4468_0_ff;
extern cpuop_func op_4470_0_nf;
extern cpuop_func op_4470_0_ff;
extern cpuop_func op_4478_0_nf;
extern cpuop_func op_4478_0_ff;
extern cpuop_func op_4479_0_nf;
extern cpuop_func op_4479_0_ff;
extern cpuop_func op_4480_0_nf;
extern cpuop_func op_4480_0_ff;
extern cpuop_func op_4490_0_nf;
extern cpuop_func op_4490_0_ff;
extern cpuop_func op_4498_0_nf;
extern cpuop_func op_4498_0_ff;
extern cpuop_func op_44a0_0_nf;
extern cpuop_func op_44a0_0_ff;
extern cpuop_func op_44a8_0_nf;
extern cpuop_func op_44a8_0_ff;
extern cpuop_func op_44b0_0_nf;
extern cpuop_func op_44b0_0_ff;
extern cpuop_func op_44b8_0_nf;
extern cpuop_func op_44b8_0_ff;
extern cpuop_func op_44b9_0_nf;
extern cpuop_func op_44b9_0_ff;
extern cpuop_func op_44c0_0_nf;
extern cpuop_func op_44c0_0_ff;
extern cpuop_func op_44d0_0_nf;
extern cpuop_func op_44d0_0_ff;
extern cpuop_func op_44d8_0_nf;
extern cpuop_func op_44d8_0_ff;
extern cpuop_func op_44e0_0_nf;
extern cpuop_func op_44e0_0_ff;
extern cpuop_func op_44e8_0_nf;
extern cpuop_func op_44e8_0_ff;
extern cpuop_func op_44f0_0_nf;
extern cpuop_func op_44f0_0_ff;
extern cpuop_func op_44f8_0_nf;
extern cpuop_func op_44f8_0_ff;
extern cpuop_func op_44f9_0_nf;
extern cpuop_func op_44f9_0_ff;
extern cpuop_func op_44fa_0_nf;
extern cpuop_func op_44fa_0_ff;
extern cpuop_func op_44fb_0_nf;
extern cpuop_func op_44fb_0_ff;
extern cpuop_func op_44fc_0_nf;
extern cpuop_func op_44fc_0_ff;
extern cpuop_func op_4600_0_nf;
extern cpuop_func op_4600_0_ff;
extern cpuop_func op_4610_0_nf;
extern cpuop_func op_4610_0_ff;
extern cpuop_func op_4618_0_nf;
extern cpuop_func op_4618_0_ff;
extern cpuop_func op_4620_0_nf;
extern cpuop_func op_4620_0_ff;
extern cpuop_func op_4628_0_nf;
extern cpuop_func op_4628_0_ff;
extern cpuop_func op_4630_0_nf;
extern cpuop_func op_4630_0_ff;
extern cpuop_func op_4638_0_nf;
extern cpuop_func op_4638_0_ff;
extern cpuop_func op_4639_0_nf;
extern cpuop_func op_4639_0_ff;
extern cpuop_func op_4640_0_nf;
extern cpuop_func op_4640_0_ff;
extern cpuop_func op_4650_0_nf;
extern cpuop_func op_4650_0_ff;
extern cpuop_func op_4658_0_nf;
extern cpuop_func op_4658_0_ff;
extern cpuop_func op_4660_0_nf;
extern cpuop_func op_4660_0_ff;
extern cpuop_func op_4668_0_nf;
extern cpuop_func op_4668_0_ff;
extern cpuop_func op_4670_0_nf;
extern cpuop_func op_4670_0_ff;
extern cpuop_func op_4678_0_nf;
extern cpuop_func op_4678_0_ff;
extern cpuop_func op_4679_0_nf;
extern cpuop_func op_4679_0_ff;
extern cpuop_func op_4680_0_nf;
extern cpuop_func op_4680_0_ff;
extern cpuop_func op_4690_0_nf;
extern cpuop_func op_4690_0_ff;
extern cpuop_func op_4698_0_nf;
extern cpuop_func op_4698_0_ff;
extern cpuop_func op_46a0_0_nf;
extern cpuop_func op_46a0_0_ff;
extern cpuop_func op_46a8_0_nf;
extern cpuop_func op_46a8_0_ff;
extern cpuop_func op_46b0_0_nf;
extern cpuop_func op_46b0_0_ff;
extern cpuop_func op_46b8_0_nf;
extern cpuop_func op_46b8_0_ff;
extern cpuop_func op_46b9_0_nf;
extern cpuop_func op_46b9_0_ff;
extern cpuop_func op_46c0_0_nf;
extern cpuop_func op_46c0_0_ff;
extern cpuop_func op_46d0_0_nf;
extern cpuop_func op_46d0_0_ff;
extern cpuop_func op_46d8_0_nf;
extern cpuop_func op_46d8_0_ff;
extern cpuop_func op_46e0_0_nf;
extern cpuop_func op_46e0_0_ff;
extern cpuop_func op_46e8_0_nf;
extern cpuop_func op_46e8_0_ff;
extern cpuop_func op_46f0_0_nf;
extern cpuop_func op_46f0_0_ff;
extern cpuop_func op_46f8_0_nf;
extern cpuop_func op_46f8_0_ff;
extern cpuop_func op_46f9_0_nf;
extern cpuop_func op_46f9_0_ff;
extern cpuop_func op_46fa_0_nf;
extern cpuop_func op_46fa_0_ff;
extern cpuop_func op_46fb_0_nf;
extern cpuop_func op_46fb_0_ff;
extern cpuop_func op_46fc_0_nf;
extern cpuop_func op_46fc_0_ff;
extern cpuop_func op_4800_0_nf;
extern cpuop_func op_4800_0_ff;
extern cpuop_func op_4808_0_nf;
extern cpuop_func op_4808_0_ff;
extern cpuop_func op_4810_0_nf;
extern cpuop_func op_4810_0_ff;
extern cpuop_func op_4818_0_nf;
extern cpuop_func op_4818_0_ff;
extern cpuop_func op_4820_0_nf;
extern cpuop_func op_4820_0_ff;
extern cpuop_func op_4828_0_nf;
extern cpuop_func op_4828_0_ff;
extern cpuop_func op_4830_0_nf;
extern cpuop_func op_4830_0_ff;
extern cpuop_func op_4838_0_nf;
extern cpuop_func op_4838_0_ff;
extern cpuop_func op_4839_0_nf;
extern cpuop_func op_4839_0_ff;
extern cpuop_func op_4840_0_nf;
extern cpuop_func op_4840_0_ff;
extern cpuop_func op_4848_0_nf;
extern cpuop_func op_4848_0_ff;
extern cpuop_func op_4850_0_nf;
extern cpuop_func op_4850_0_ff;
extern cpuop_func op_4868_0_nf;
extern cpuop_func op_4868_0_ff;
extern cpuop_func op_4870_0_nf;
extern cpuop_func op_4870_0_ff;
extern cpuop_func op_4878_0_nf;
extern cpuop_func op_4878_0_ff;
extern cpuop_func op_4879_0_nf;
extern cpuop_func op_4879_0_ff;
extern cpuop_func op_487a_0_nf;
extern cpuop_func op_487a_0_ff;
extern cpuop_func op_487b_0_nf;
extern cpuop_func op_487b_0_ff;
extern cpuop_func op_4880_0_nf;
extern cpuop_func op_4880_0_ff;
extern cpuop_func op_4890_0_nf;
extern cpuop_func op_4890_0_ff;
extern cpuop_func op_48a0_0_nf;
extern cpuop_func op_48a0_0_ff;
extern cpuop_func op_48a8_0_nf;
extern cpuop_func op_48a8_0_ff;
extern cpuop_func op_48b0_0_nf;
extern cpuop_func op_48b0_0_ff;
extern cpuop_func op_48b8_0_nf;
extern cpuop_func op_48b8_0_ff;
extern cpuop_func op_48b9_0_nf;
extern cpuop_func op_48b9_0_ff;
extern cpuop_func op_48c0_0_nf;
extern cpuop_func op_48c0_0_ff;
extern cpuop_func op_48d0_0_nf;
extern cpuop_func op_48d0_0_ff;
extern cpuop_func op_48e0_0_nf;
extern cpuop_func op_48e0_0_ff;
extern cpuop_func op_48e8_0_nf;
extern cpuop_func op_48e8_0_ff;
extern cpuop_func op_48f0_0_nf;
extern cpuop_func op_48f0_0_ff;
extern cpuop_func op_48f8_0_nf;
extern cpuop_func op_48f8_0_ff;
extern cpuop_func op_48f9_0_nf;
extern cpuop_func op_48f9_0_ff;
extern cpuop_func op_49c0_0_nf;
extern cpuop_func op_49c0_0_ff;
extern cpuop_func op_4a00_0_nf;
extern cpuop_func op_4a00_0_ff;
extern cpuop_func op_4a10_0_nf;
extern cpuop_func op_4a10_0_ff;
extern cpuop_func op_4a18_0_nf;
extern cpuop_func op_4a18_0_ff;
extern cpuop_func op_4a20_0_nf;
extern cpuop_func op_4a20_0_ff;
extern cpuop_func op_4a28_0_nf;
extern cpuop_func op_4a28_0_ff;
extern cpuop_func op_4a30_0_nf;
extern cpuop_func op_4a30_0_ff;
extern cpuop_func op_4a38_0_nf;
extern cpuop_func op_4a38_0_ff;
extern cpuop_func op_4a39_0_nf;
extern cpuop_func op_4a39_0_ff;
extern cpuop_func op_4a3a_0_nf;
extern cpuop_func op_4a3a_0_ff;
extern cpuop_func op_4a3b_0_nf;
extern cpuop_func op_4a3b_0_ff;
extern cpuop_func op_4a40_0_nf;
extern cpuop_func op_4a40_0_ff;
extern cpuop_func op_4a48_0_nf;
extern cpuop_func op_4a48_0_ff;
extern cpuop_func op_4a50_0_nf;
extern cpuop_func op_4a50_0_ff;
extern cpuop_func op_4a58_0_nf;
extern cpuop_func op_4a58_0_ff;
extern cpuop_func op_4a60_0_nf;
extern cpuop_func op_4a60_0_ff;
extern cpuop_func op_4a68_0_nf;
extern cpuop_func op_4a68_0_ff;
extern cpuop_func op_4a70_0_nf;
extern cpuop_func op_4a70_0_ff;
extern cpuop_func op_4a78_0_nf;
extern cpuop_func op_4a78_0_ff;
extern cpuop_func op_4a79_0_nf;
extern cpuop_func op_4a79_0_ff;
extern cpuop_func op_4a7a_0_nf;
extern cpuop_func op_4a7a_0_ff;
extern cpuop_func op_4a7b_0_nf;
extern cpuop_func op_4a7b_0_ff;
extern cpuop_func op_4a80_0_nf;
extern cpuop_func op_4a80_0_ff;
extern cpuop_func op_4a88_0_nf;
extern cpuop_func op_4a88_0_ff;
extern cpuop_func op_4a90_0_nf;
extern cpuop_func op_4a90_0_ff;
extern cpuop_func op_4a98_0_nf;
extern cpuop_func op_4a98_0_ff;
extern cpuop_func op_4aa0_0_nf;
extern cpuop_func op_4aa0_0_ff;
extern cpuop_func op_4aa8_0_nf;
extern cpuop_func op_4aa8_0_ff;
extern cpuop_func op_4ab0_0_nf;
extern cpuop_func op_4ab0_0_ff;
extern cpuop_func op_4ab8_0_nf;
extern cpuop_func op_4ab8_0_ff;
extern cpuop_func op_4ab9_0_nf;
extern cpuop_func op_4ab9_0_ff;
extern cpuop_func op_4aba_0_nf;
extern cpuop_func op_4aba_0_ff;
extern cpuop_func op_4abb_0_nf;
extern cpuop_func op_4abb_0_ff;
extern cpuop_func op_4ac0_0_nf;
extern cpuop_func op_4ac0_0_ff;
extern cpuop_func op_4ad0_0_nf;
extern cpuop_func op_4ad0_0_ff;
extern cpuop_func op_4ad8_0_nf;
extern cpuop_func op_4ad8_0_ff;
extern cpuop_func op_4ae0_0_nf;
extern cpuop_func op_4ae0_0_ff;
extern cpuop_func op_4ae8_0_nf;
extern cpuop_func op_4ae8_0_ff;
extern cpuop_func op_4af0_0_nf;
extern cpuop_func op_4af0_0_ff;
extern cpuop_func op_4af8_0_nf;
extern cpuop_func op_4af8_0_ff;
extern cpuop_func op_4af9_0_nf;
extern cpuop_func op_4af9_0_ff;
extern cpuop_func op_4c00_0_nf;
extern cpuop_func op_4c00_0_ff;
extern cpuop_func op_4c10_0_nf;
extern cpuop_func op_4c10_0_ff;
extern cpuop_func op_4c18_0_nf;
extern cpuop_func op_4c18_0_ff;
extern cpuop_func op_4c20_0_nf;
extern cpuop_func op_4c20_0_ff;
extern cpuop_func op_4c28_0_nf;
extern cpuop_func op_4c28_0_ff;
extern cpuop_func op_4c30_0_nf;
extern cpuop_func op_4c30_0_ff;
extern cpuop_func op_4c38_0_nf;
extern cpuop_func op_4c38_0_ff;
extern cpuop_func op_4c39_0_nf;
extern cpuop_func op_4c39_0_ff;
extern cpuop_func op_4c3a_0_nf;
extern cpuop_func op_4c3a_0_ff;
extern cpuop_func op_4c3b_0_nf;
extern cpuop_func op_4c3b_0_ff;
extern cpuop_func op_4c3c_0_nf;
extern cpuop_func op_4c3c_0_ff;
extern cpuop_func op_4c40_0_nf;
extern cpuop_func op_4c40_0_ff;
extern cpuop_func op_4c50_0_nf;
extern cpuop_func op_4c50_0_ff;
extern cpuop_func op_4c58_0_nf;
extern cpuop_func op_4c58_0_ff;
extern cpuop_func op_4c60_0_nf;
extern cpuop_func op_4c60_0_ff;
extern cpuop_func op_4c68_0_nf;
extern cpuop_func op_4c68_0_ff;
extern cpuop_func op_4c70_0_nf;
extern cpuop_func op_4c70_0_ff;
extern cpuop_func op_4c78_0_nf;
extern cpuop_func op_4c78_0_ff;
extern cpuop_func op_4c79_0_nf;
extern cpuop_func op_4c79_0_ff;
extern cpuop_func op_4c7a_0_nf;
extern cpuop_func op_4c7a_0_ff;
extern cpuop_func op_4c7b_0_nf;
extern cpuop_func op_4c7b_0_ff;
extern cpuop_func op_4c7c_0_nf;
extern cpuop_func op_4c7c_0_ff;
extern cpuop_func op_4c90_0_nf;
extern cpuop_func op_4c90_0_ff;
extern cpuop_func op_4c98_0_nf;
extern cpuop_func op_4c98_0_ff;
extern cpuop_func op_4ca8_0_nf;
extern cpuop_func op_4ca8_0_ff;
extern cpuop_func op_4cb0_0_nf;
extern cpuop_func op_4cb0_0_ff;
extern cpuop_func op_4cb8_0_nf;
extern cpuop_func op_4cb8_0_ff;
extern cpuop_func op_4cb9_0_nf;
extern cpuop_func op_4cb9_0_ff;
extern cpuop_func op_4cba_0_nf;
extern cpuop_func op_4cba_0_ff;
extern cpuop_func op_4cbb_0_nf;
extern cpuop_func op_4cbb_0_ff;
extern cpuop_func op_4cd0_0_nf;
extern cpuop_func op_4cd0_0_ff;
extern cpuop_func op_4cd8_0_nf;
extern cpuop_func op_4cd8_0_ff;
extern cpuop_func op_4ce8_0_nf;
extern cpuop_func op_4ce8_0_ff;
extern cpuop_func op_4cf0_0_nf;
extern cpuop_func op_4cf0_0_ff;
extern cpuop_func op_4cf8_0_nf;
extern cpuop_func op_4cf8_0_ff;
extern cpuop_func op_4cf9_0_nf;
extern cpuop_func op_4cf9_0_ff;
extern cpuop_func op_4cfa_0_nf;
extern cpuop_func op_4cfa_0_ff;
extern cpuop_func op_4cfb_0_nf;
extern cpuop_func op_4cfb_0_ff;
extern cpuop_func op_4e40_0_nf;
extern cpuop_func op_4e40_0_ff;
extern cpuop_func op_4e50_0_nf;
extern cpuop_func op_4e50_0_ff;
extern cpuop_func op_4e58_0_nf;
extern cpuop_func op_4e58_0_ff;
extern cpuop_func op_4e60_0_nf;
extern cpuop_func op_4e60_0_ff;
extern cpuop_func op_4e68_0_nf;
extern cpuop_func op_4e68_0_ff;
extern cpuop_func op_4e70_0_nf;
extern cpuop_func op_4e70_0_ff;
extern cpuop_func op_4e71_0_nf;
extern cpuop_func op_4e71_0_ff;
extern cpuop_func op_4e72_0_nf;
extern cpuop_func op_4e72_0_ff;
extern cpuop_func op_4e73_0_nf;
extern cpuop_func op_4e73_0_ff;
extern cpuop_func op_4e74_0_nf;
extern cpuop_func op_4e74_0_ff;
extern cpuop_func op_4e75_0_nf;
extern cpuop_func op_4e75_0_ff;
extern cpuop_func op_4e76_0_nf;
extern cpuop_func op_4e76_0_ff;
extern cpuop_func op_4e77_0_nf;
extern cpuop_func op_4e77_0_ff;
extern cpuop_func op_4e7a_0_nf;
extern cpuop_func op_4e7a_0_ff;
extern cpuop_func op_4e7b_0_nf;
extern cpuop_func op_4e7b_0_ff;
extern cpuop_func op_4e90_0_nf;
extern cpuop_func op_4e90_0_ff;
extern cpuop_func op_4ea8_0_nf;
extern cpuop_func op_4ea8_0_ff;
extern cpuop_func op_4eb0_0_nf;
extern cpuop_func op_4eb0_0_ff;
extern cpuop_func op_4eb8_0_nf;
extern cpuop_func op_4eb8_0_ff;
extern cpuop_func op_4eb9_0_nf;
extern cpuop_func op_4eb9_0_ff;
extern cpuop_func op_4eba_0_nf;
extern cpuop_func op_4eba_0_ff;
extern cpuop_func op_4ebb_0_nf;
extern cpuop_func op_4ebb_0_ff;
extern cpuop_func op_4ed0_0_nf;
extern cpuop_func op_4ed0_0_ff;
extern cpuop_func op_4ee8_0_nf;
extern cpuop_func op_4ee8_0_ff;
extern cpuop_func op_4ef0_0_nf;
extern cpuop_func op_4ef0_0_ff;
extern cpuop_func op_4ef8_0_nf;
extern cpuop_func op_4ef8_0_ff;
extern cpuop_func op_4ef9_0_nf;
extern cpuop_func op_4ef9_0_ff;
extern cpuop_func op_4efa_0_nf;
extern cpuop_func op_4efa_0_ff;
extern cpuop_func op_4efb_0_nf;
extern cpuop_func op_4efb_0_ff;
extern cpuop_func op_5000_0_nf;
extern cpuop_func op_5000_0_ff;
extern cpuop_func op_5010_0_nf;
extern cpuop_func op_5010_0_ff;
extern cpuop_func op_5018_0_nf;
extern cpuop_func op_5018_0_ff;
extern cpuop_func op_5020_0_nf;
extern cpuop_func op_5020_0_ff;
extern cpuop_func op_5028_0_nf;
extern cpuop_func op_5028_0_ff;
extern cpuop_func op_5030_0_nf;
extern cpuop_func op_5030_0_ff;
extern cpuop_func op_5038_0_nf;
extern cpuop_func op_5038_0_ff;
extern cpuop_func op_5039_0_nf;
extern cpuop_func op_5039_0_ff;
extern cpuop_func op_5040_0_nf;
extern cpuop_func op_5040_0_ff;
extern cpuop_func op_5048_0_nf;
extern cpuop_func op_5048_0_ff;
extern cpuop_func op_5050_0_nf;
extern cpuop_func op_5050_0_ff;
extern cpuop_func op_5058_0_nf;
extern cpuop_func op_5058_0_ff;
extern cpuop_func op_5060_0_nf;
extern cpuop_func op_5060_0_ff;
extern cpuop_func op_5068_0_nf;
extern cpuop_func op_5068_0_ff;
extern cpuop_func op_5070_0_nf;
extern cpuop_func op_5070_0_ff;
extern cpuop_func op_5078_0_nf;
extern cpuop_func op_5078_0_ff;
extern cpuop_func op_5079_0_nf;
extern cpuop_func op_5079_0_ff;
extern cpuop_func op_5080_0_nf;
extern cpuop_func op_5080_0_ff;
extern cpuop_func op_5088_0_nf;
extern cpuop_func op_5088_0_ff;
extern cpuop_func op_5090_0_nf;
extern cpuop_func op_5090_0_ff;
extern cpuop_func op_5098_0_nf;
extern cpuop_func op_5098_0_ff;
extern cpuop_func op_50a0_0_nf;
extern cpuop_func op_50a0_0_ff;
extern cpuop_func op_50a8_0_nf;
extern cpuop_func op_50a8_0_ff;
extern cpuop_func op_50b0_0_nf;
extern cpuop_func op_50b0_0_ff;
extern cpuop_func op_50b8_0_nf;
extern cpuop_func op_50b8_0_ff;
extern cpuop_func op_50b9_0_nf;
extern cpuop_func op_50b9_0_ff;
extern cpuop_func op_50c0_0_nf;
extern cpuop_func op_50c0_0_ff;
extern cpuop_func op_50c8_0_nf;
extern cpuop_func op_50c8_0_ff;
extern cpuop_func op_50d0_0_nf;
extern cpuop_func op_50d0_0_ff;
extern cpuop_func op_50d8_0_nf;
extern cpuop_func op_50d8_0_ff;
extern cpuop_func op_50e0_0_nf;
extern cpuop_func op_50e0_0_ff;
extern cpuop_func op_50e8_0_nf;
extern cpuop_func op_50e8_0_ff;
extern cpuop_func op_50f0_0_nf;
extern cpuop_func op_50f0_0_ff;
extern cpuop_func op_50f8_0_nf;
extern cpuop_func op_50f8_0_ff;
extern cpuop_func op_50f9_0_nf;
extern cpuop_func op_50f9_0_ff;
extern cpuop_func op_50fa_0_nf;
extern cpuop_func op_50fa_0_ff;
extern cpuop_func op_50fb_0_nf;
extern cpuop_func op_50fb_0_ff;
extern cpuop_func op_50fc_0_nf;
extern cpuop_func op_50fc_0_ff;
extern cpuop_func op_5100_0_nf;
extern cpuop_func op_5100_0_ff;
extern cpuop_func op_5110_0_nf;
extern cpuop_func op_5110_0_ff;
extern cpuop_func op_5118_0_nf;
extern cpuop_func op_5118_0_ff;
extern cpuop_func op_5120_0_nf;
extern cpuop_func op_5120_0_ff;
extern cpuop_func op_5128_0_nf;
extern cpuop_func op_5128_0_ff;
extern cpuop_func op_5130_0_nf;
extern cpuop_func op_5130_0_ff;
extern cpuop_func op_5138_0_nf;
extern cpuop_func op_5138_0_ff;
extern cpuop_func op_5139_0_nf;
extern cpuop_func op_5139_0_ff;
extern cpuop_func op_5140_0_nf;
extern cpuop_func op_5140_0_ff;
extern cpuop_func op_5148_0_nf;
extern cpuop_func op_5148_0_ff;
extern cpuop_func op_5150_0_nf;
extern cpuop_func op_5150_0_ff;
extern cpuop_func op_5158_0_nf;
extern cpuop_func op_5158_0_ff;
extern cpuop_func op_5160_0_nf;
extern cpuop_func op_5160_0_ff;
extern cpuop_func op_5168_0_nf;
extern cpuop_func op_5168_0_ff;
extern cpuop_func op_5170_0_nf;
extern cpuop_func op_5170_0_ff;
extern cpuop_func op_5178_0_nf;
extern cpuop_func op_5178_0_ff;
extern cpuop_func op_5179_0_nf;
extern cpuop_func op_5179_0_ff;
extern cpuop_func op_5180_0_nf;
extern cpuop_func op_5180_0_ff;
extern cpuop_func op_5188_0_nf;
extern cpuop_func op_5188_0_ff;
extern cpuop_func op_5190_0_nf;
extern cpuop_func op_5190_0_ff;
extern cpuop_func op_5198_0_nf;
extern cpuop_func op_5198_0_ff;
extern cpuop_func op_51a0_0_nf;
extern cpuop_func op_51a0_0_ff;
extern cpuop_func op_51a8_0_nf;
extern cpuop_func op_51a8_0_ff;
extern cpuop_func op_51b0_0_nf;
extern cpuop_func op_51b0_0_ff;
extern cpuop_func op_51b8_0_nf;
extern cpuop_func op_51b8_0_ff;
extern cpuop_func op_51b9_0_nf;
extern cpuop_func op_51b9_0_ff;
extern cpuop_func op_51c0_0_nf;
extern cpuop_func op_51c0_0_ff;
extern cpuop_func op_51c8_0_nf;
extern cpuop_func op_51c8_0_ff;
extern cpuop_func op_51d0_0_nf;
extern cpuop_func op_51d0_0_ff;
extern cpuop_func op_51d8_0_nf;
extern cpuop_func op_51d8_0_ff;
extern cpuop_func op_51e0_0_nf;
extern cpuop_func op_51e0_0_ff;
extern cpuop_func op_51e8_0_nf;
extern cpuop_func op_51e8_0_ff;
extern cpuop_func op_51f0_0_nf;
extern cpuop_func op_51f0_0_ff;
extern cpuop_func op_51f8_0_nf;
extern cpuop_func op_51f8_0_ff;
extern cpuop_func op_51f9_0_nf;
extern cpuop_func op_51f9_0_ff;
extern cpuop_func op_51fa_0_nf;
extern cpuop_func op_51fa_0_ff;
extern cpuop_func op_51fb_0_nf;
extern cpuop_func op_51fb_0_ff;
extern cpuop_func op_51fc_0_nf;
extern cpuop_func op_51fc_0_ff;
extern cpuop_func op_52c0_0_nf;
extern cpuop_func op_52c0_0_ff;
extern cpuop_func op_52c8_0_nf;
extern cpuop_func op_52c8_0_ff;
extern cpuop_func op_52d0_0_nf;
extern cpuop_func op_52d0_0_ff;
extern cpuop_func op_52d8_0_nf;
extern cpuop_func op_52d8_0_ff;
extern cpuop_func op_52e0_0_nf;
extern cpuop_func op_52e0_0_ff;
extern cpuop_func op_52e8_0_nf;
extern cpuop_func op_52e8_0_ff;
extern cpuop_func op_52f0_0_nf;
extern cpuop_func op_52f0_0_ff;
extern cpuop_func op_52f8_0_nf;
extern cpuop_func op_52f8_0_ff;
extern cpuop_func op_52f9_0_nf;
extern cpuop_func op_52f9_0_ff;
extern cpuop_func op_52fa_0_nf;
extern cpuop_func op_52fa_0_ff;
extern cpuop_func op_52fb_0_nf;
extern cpuop_func op_52fb_0_ff;
extern cpuop_func op_52fc_0_nf;
extern cpuop_func op_52fc_0_ff;
extern cpuop_func op_53c0_0_nf;
extern cpuop_func op_53c0_0_ff;
extern cpuop_func op_53c8_0_nf;
extern cpuop_func op_53c8_0_ff;
extern cpuop_func op_53d0_0_nf;
extern cpuop_func op_53d0_0_ff;
extern cpuop_func op_53d8_0_nf;
extern cpuop_func op_53d8_0_ff;
extern cpuop_func op_53e0_0_nf;
extern cpuop_func op_53e0_0_ff;
extern cpuop_func op_53e8_0_nf;
extern cpuop_func op_53e8_0_ff;
extern cpuop_func op_53f0_0_nf;
extern cpuop_func op_53f0_0_ff;
extern cpuop_func op_53f8_0_nf;
extern cpuop_func op_53f8_0_ff;
extern cpuop_func op_53f9_0_nf;
extern cpuop_func op_53f9_0_ff;
extern cpuop_func op_53fa_0_nf;
extern cpuop_func op_53fa_0_ff;
extern cpuop_func op_53fb_0_nf;
extern cpuop_func op_53fb_0_ff;
extern cpuop_func op_53fc_0_nf;
extern cpuop_func op_53fc_0_ff;
extern cpuop_func op_54c0_0_nf;
extern cpuop_func op_54c0_0_ff;
extern cpuop_func op_54c8_0_nf;
extern cpuop_func op_54c8_0_ff;
extern cpuop_func op_54d0_0_nf;
extern cpuop_func op_54d0_0_ff;
extern cpuop_func op_54d8_0_nf;
extern cpuop_func op_54d8_0_ff;
extern cpuop_func op_54e0_0_nf;
extern cpuop_func op_54e0_0_ff;
extern cpuop_func op_54e8_0_nf;
extern cpuop_func op_54e8_0_ff;
extern cpuop_func op_54f0_0_nf;
extern cpuop_func op_54f0_0_ff;
extern cpuop_func op_54f8_0_nf;
extern cpuop_func op_54f8_0_ff;
extern cpuop_func op_54f9_0_nf;
extern cpuop_func op_54f9_0_ff;
extern cpuop_func op_54fa_0_nf;
extern cpuop_func op_54fa_0_ff;
extern cpuop_func op_54fb_0_nf;
extern cpuop_func op_54fb_0_ff;
extern cpuop_func op_54fc_0_nf;
extern cpuop_func op_54fc_0_ff;
extern cpuop_func op_55c0_0_nf;
extern cpuop_func op_55c0_0_ff;
extern cpuop_func op_55c8_0_nf;
extern cpuop_func op_55c8_0_ff;
extern cpuop_func op_55d0_0_nf;
extern cpuop_func op_55d0_0_ff;
extern cpuop_func op_55d8_0_nf;
extern cpuop_func op_55d8_0_ff;
extern cpuop_func op_55e0_0_nf;
extern cpuop_func op_55e0_0_ff;
extern cpuop_func op_55e8_0_nf;
extern cpuop_func op_55e8_0_ff;
extern cpuop_func op_55f0_0_nf;
extern cpuop_func op_55f0_0_ff;
extern cpuop_func op_55f8_0_nf;
extern cpuop_func op_55f8_0_ff;
extern cpuop_func op_55f9_0_nf;
extern cpuop_func op_55f9_0_ff;
extern cpuop_func op_55fa_0_nf;
extern cpuop_func op_55fa_0_ff;
extern cpuop_func op_55fb_0_nf;
extern cpuop_func op_55fb_0_ff;
extern cpuop_func op_55fc_0_nf;
extern cpuop_func op_55fc_0_ff;
extern cpuop_func op_56c0_0_nf;
extern cpuop_func op_56c0_0_ff;
extern cpuop_func op_56c8_0_nf;
extern cpuop_func op_56c8_0_ff;
extern cpuop_func op_56d0_0_nf;
extern cpuop_func op_56d0_0_ff;
extern cpuop_func op_56d8_0_nf;
extern cpuop_func op_56d8_0_ff;
extern cpuop_func op_56e0_0_nf;
extern cpuop_func op_56e0_0_ff;
extern cpuop_func op_56e8_0_nf;
extern cpuop_func op_56e8_0_ff;
extern cpuop_func op_56f0_0_nf;
extern cpuop_func op_56f0_0_ff;
extern cpuop_func op_56f8_0_nf;
extern cpuop_func op_56f8_0_ff;
extern cpuop_func op_56f9_0_nf;
extern cpuop_func op_56f9_0_ff;
extern cpuop_func op_56fa_0_nf;
extern cpuop_func op_56fa_0_ff;
extern cpuop_func op_56fb_0_nf;
extern cpuop_func op_56fb_0_ff;
extern cpuop_func op_56fc_0_nf;
extern cpuop_func op_56fc_0_ff;
extern cpuop_func op_57c0_0_nf;
extern cpuop_func op_57c0_0_ff;
extern cpuop_func op_57c8_0_nf;
extern cpuop_func op_57c8_0_ff;
extern cpuop_func op_57d0_0_nf;
extern cpuop_func op_57d0_0_ff;
extern cpuop_func op_57d8_0_nf;
extern cpuop_func op_57d8_0_ff;
extern cpuop_func op_57e0_0_nf;
extern cpuop_func op_57e0_0_ff;
extern cpuop_func op_57e8_0_nf;
extern cpuop_func op_57e8_0_ff;
extern cpuop_func op_57f0_0_nf;
extern cpuop_func op_57f0_0_ff;
extern cpuop_func op_57f8_0_nf;
extern cpuop_func op_57f8_0_ff;
extern cpuop_func op_57f9_0_nf;
extern cpuop_func op_57f9_0_ff;
extern cpuop_func op_57fa_0_nf;
extern cpuop_func op_57fa_0_ff;
extern cpuop_func op_57fb_0_nf;
extern cpuop_func op_57fb_0_ff;
extern cpuop_func op_57fc_0_nf;
extern cpuop_func op_57fc_0_ff;
extern cpuop_func op_58c0_0_nf;
extern cpuop_func op_58c0_0_ff;
extern cpuop_func op_58c8_0_nf;
extern cpuop_func op_58c8_0_ff;
extern cpuop_func op_58d0_0_nf;
extern cpuop_func op_58d0_0_ff;
extern cpuop_func op_58d8_0_nf;
extern cpuop_func op_58d8_0_ff;
extern cpuop_func op_58e0_0_nf;
extern cpuop_func op_58e0_0_ff;
extern cpuop_func op_58e8_0_nf;
extern cpuop_func op_58e8_0_ff;
extern cpuop_func op_58f0_0_nf;
extern cpuop_func op_58f0_0_ff;
extern cpuop_func op_58f8_0_nf;
extern cpuop_func op_58f8_0_ff;
extern cpuop_func op_58f9_0_nf;
extern cpuop_func op_58f9_0_ff;
extern cpuop_func op_58fa_0_nf;
extern cpuop_func op_58fa_0_ff;
extern cpuop_func op_58fb_0_nf;
extern cpuop_func op_58fb_0_ff;
extern cpuop_func op_58fc_0_nf;
extern cpuop_func op_58fc_0_ff;
extern cpuop_func op_59c0_0_nf;
extern cpuop_func op_59c0_0_ff;
extern cpuop_func op_59c8_0_nf;
extern cpuop_func op_59c8_0_ff;
extern cpuop_func op_59d0_0_nf;
extern cpuop_func op_59d0_0_ff;
extern cpuop_func op_59d8_0_nf;
extern cpuop_func op_59d8_0_ff;
extern cpuop_func op_59e0_0_nf;
extern cpuop_func op_59e0_0_ff;
extern cpuop_func op_59e8_0_nf;
extern cpuop_func op_59e8_0_ff;
extern cpuop_func op_59f0_0_nf;
extern cpuop_func op_59f0_0_ff;
extern cpuop_func op_59f8_0_nf;
extern cpuop_func op_59f8_0_ff;
extern cpuop_func op_59f9_0_nf;
extern cpuop_func op_59f9_0_ff;
extern cpuop_func op_59fa_0_nf;
extern cpuop_func op_59fa_0_ff;
extern cpuop_func op_59fb_0_nf;
extern cpuop_func op_59fb_0_ff;
extern cpuop_func op_59fc_0_nf;
extern cpuop_func op_59fc_0_ff;
extern cpuop_func op_5ac0_0_nf;
extern cpuop_func op_5ac0_0_ff;
extern cpuop_func op_5ac8_0_nf;
extern cpuop_func op_5ac8_0_ff;
extern cpuop_func op_5ad0_0_nf;
extern cpuop_func op_5ad0_0_ff;
extern cpuop_func op_5ad8_0_nf;
extern cpuop_func op_5ad8_0_ff;
extern cpuop_func op_5ae0_0_nf;
extern cpuop_func op_5ae0_0_ff;
extern cpuop_func op_5ae8_0_nf;
extern cpuop_func op_5ae8_0_ff;
extern cpuop_func op_5af0_0_nf;
extern cpuop_func op_5af0_0_ff;
extern cpuop_func op_5af8_0_nf;
extern cpuop_func op_5af8_0_ff;
extern cpuop_func op_5af9_0_nf;
extern cpuop_func op_5af9_0_ff;
extern cpuop_func op_5afa_0_nf;
extern cpuop_func op_5afa_0_ff;
extern cpuop_func op_5afb_0_nf;
extern cpuop_func op_5afb_0_ff;
extern cpuop_func op_5afc_0_nf;
extern cpuop_func op_5afc_0_ff;
extern cpuop_func op_5bc0_0_nf;
extern cpuop_func op_5bc0_0_ff;
extern cpuop_func op_5bc8_0_nf;
extern cpuop_func op_5bc8_0_ff;
extern cpuop_func op_5bd0_0_nf;
extern cpuop_func op_5bd0_0_ff;
extern cpuop_func op_5bd8_0_nf;
extern cpuop_func op_5bd8_0_ff;
extern cpuop_func op_5be0_0_nf;
extern cpuop_func op_5be0_0_ff;
extern cpuop_func op_5be8_0_nf;
extern cpuop_func op_5be8_0_ff;
extern cpuop_func op_5bf0_0_nf;
extern cpuop_func op_5bf0_0_ff;
extern cpuop_func op_5bf8_0_nf;
extern cpuop_func op_5bf8_0_ff;
extern cpuop_func op_5bf9_0_nf;
extern cpuop_func op_5bf9_0_ff;
extern cpuop_func op_5bfa_0_nf;
extern cpuop_func op_5bfa_0_ff;
extern cpuop_func op_5bfb_0_nf;
extern cpuop_func op_5bfb_0_ff;
extern cpuop_func op_5bfc_0_nf;
extern cpuop_func op_5bfc_0_ff;
extern cpuop_func op_5cc0_0_nf;
extern cpuop_func op_5cc0_0_ff;
extern cpuop_func op_5cc8_0_nf;
extern cpuop_func op_5cc8_0_ff;
extern cpuop_func op_5cd0_0_nf;
extern cpuop_func op_5cd0_0_ff;
extern cpuop_func op_5cd8_0_nf;
extern cpuop_func op_5cd8_0_ff;
extern cpuop_func op_5ce0_0_nf;
extern cpuop_func op_5ce0_0_ff;
extern cpuop_func op_5ce8_0_nf;
extern cpuop_func op_5ce8_0_ff;
extern cpuop_func op_5cf0_0_nf;
extern cpuop_func op_5cf0_0_ff;
extern cpuop_func op_5cf8_0_nf;
extern cpuop_func op_5cf8_0_ff;
extern cpuop_func op_5cf9_0_nf;
extern cpuop_func op_5cf9_0_ff;
extern cpuop_func op_5cfa_0_nf;
extern cpuop_func op_5cfa_0_ff;
extern cpuop_func op_5cfb_0_nf;
extern cpuop_func op_5cfb_0_ff;
extern cpuop_func op_5cfc_0_nf;
extern cpuop_func op_5cfc_0_ff;
extern cpuop_func op_5dc0_0_nf;
extern cpuop_func op_5dc0_0_ff;
extern cpuop_func op_5dc8_0_nf;
extern cpuop_func op_5dc8_0_ff;
extern cpuop_func op_5dd0_0_nf;
extern cpuop_func op_5dd0_0_ff;
extern cpuop_func op_5dd8_0_nf;
extern cpuop_func op_5dd8_0_ff;
extern cpuop_func op_5de0_0_nf;
extern cpuop_func op_5de0_0_ff;
extern cpuop_func op_5de8_0_nf;
extern cpuop_func op_5de8_0_ff;
extern cpuop_func op_5df0_0_nf;
extern cpuop_func op_5df0_0_ff;
extern cpuop_func op_5df8_0_nf;
extern cpuop_func op_5df8_0_ff;
extern cpuop_func op_5df9_0_nf;
extern cpuop_func op_5df9_0_ff;
extern cpuop_func op_5dfa_0_nf;
extern cpuop_func op_5dfa_0_ff;
extern cpuop_func op_5dfb_0_nf;
extern cpuop_func op_5dfb_0_ff;
extern cpuop_func op_5dfc_0_nf;
extern cpuop_func op_5dfc_0_ff;
extern cpuop_func op_5ec0_0_nf;
extern cpuop_func op_5ec0_0_ff;
extern cpuop_func op_5ec8_0_nf;
extern cpuop_func op_5ec8_0_ff;
extern cpuop_func op_5ed0_0_nf;
extern cpuop_func op_5ed0_0_ff;
extern cpuop_func op_5ed8_0_nf;
extern cpuop_func op_5ed8_0_ff;
extern cpuop_func op_5ee0_0_nf;
extern cpuop_func op_5ee0_0_ff;
extern cpuop_func op_5ee8_0_nf;
extern cpuop_func op_5ee8_0_ff;
extern cpuop_func op_5ef0_0_nf;
extern cpuop_func op_5ef0_0_ff;
extern cpuop_func op_5ef8_0_nf;
extern cpuop_func op_5ef8_0_ff;
extern cpuop_func op_5ef9_0_nf;
extern cpuop_func op_5ef9_0_ff;
extern cpuop_func op_5efa_0_nf;
extern cpuop_func op_5efa_0_ff;
extern cpuop_func op_5efb_0_nf;
extern cpuop_func op_5efb_0_ff;
extern cpuop_func op_5efc_0_nf;
extern cpuop_func op_5efc_0_ff;
extern cpuop_func op_5fc0_0_nf;
extern cpuop_func op_5fc0_0_ff;
extern cpuop_func op_5fc8_0_nf;
extern cpuop_func op_5fc8_0_ff;
extern cpuop_func op_5fd0_0_nf;
extern cpuop_func op_5fd0_0_ff;
extern cpuop_func op_5fd8_0_nf;
extern cpuop_func op_5fd8_0_ff;
extern cpuop_func op_5fe0_0_nf;
extern cpuop_func op_5fe0_0_ff;
extern cpuop_func op_5fe8_0_nf;
extern cpuop_func op_5fe8_0_ff;
extern cpuop_func op_5ff0_0_nf;
extern cpuop_func op_5ff0_0_ff;
extern cpuop_func op_5ff8_0_nf;
extern cpuop_func op_5ff8_0_ff;
extern cpuop_func op_5ff9_0_nf;
extern cpuop_func op_5ff9_0_ff;
extern cpuop_func op_5ffa_0_nf;
extern cpuop_func op_5ffa_0_ff;
extern cpuop_func op_5ffb_0_nf;
extern cpuop_func op_5ffb_0_ff;
extern cpuop_func op_5ffc_0_nf;
extern cpuop_func op_5ffc_0_ff;
extern cpuop_func op_6000_0_nf;
extern cpuop_func op_6000_0_ff;
extern cpuop_func op_6001_0_nf;
extern cpuop_func op_6001_0_ff;
extern cpuop_func op_60ff_0_nf;
extern cpuop_func op_60ff_0_ff;
extern cpuop_func op_6100_0_nf;
extern cpuop_func op_6100_0_ff;
extern cpuop_func op_6101_0_nf;
extern cpuop_func op_6101_0_ff;
extern cpuop_func op_61ff_0_nf;
extern cpuop_func op_61ff_0_ff;
extern cpuop_func op_6200_0_nf;
extern cpuop_func op_6200_0_ff;
extern cpuop_func op_6201_0_nf;
extern cpuop_func op_6201_0_ff;
extern cpuop_func op_62ff_0_nf;
extern cpuop_func op_62ff_0_ff;
extern cpuop_func op_6300_0_nf;
extern cpuop_func op_6300_0_ff;
extern cpuop_func op_6301_0_nf;
extern cpuop_func op_6301_0_ff;
extern cpuop_func op_63ff_0_nf;
extern cpuop_func op_63ff_0_ff;
extern cpuop_func op_6400_0_nf;
extern cpuop_func op_6400_0_ff;
extern cpuop_func op_6401_0_nf;
extern cpuop_func op_6401_0_ff;
extern cpuop_func op_64ff_0_nf;
extern cpuop_func op_64ff_0_ff;
extern cpuop_func op_6500_0_nf;
extern cpuop_func op_6500_0_ff;
extern cpuop_func op_6501_0_nf;
extern cpuop_func op_6501_0_ff;
extern cpuop_func op_65ff_0_nf;
extern cpuop_func op_65ff_0_ff;
extern cpuop_func op_6600_0_nf;
extern cpuop_func op_6600_0_ff;
extern cpuop_func op_6601_0_nf;
extern cpuop_func op_6601_0_ff;
extern cpuop_func op_66ff_0_nf;
extern cpuop_func op_66ff_0_ff;
extern cpuop_func op_6700_0_nf;
extern cpuop_func op_6700_0_ff;
extern cpuop_func op_6701_0_nf;
extern cpuop_func op_6701_0_ff;
extern cpuop_func op_67ff_0_nf;
extern cpuop_func op_67ff_0_ff;
extern cpuop_func op_6800_0_nf;
extern cpuop_func op_6800_0_ff;
extern cpuop_func op_6801_0_nf;
extern cpuop_func op_6801_0_ff;
extern cpuop_func op_68ff_0_nf;
extern cpuop_func op_68ff_0_ff;
extern cpuop_func op_6900_0_nf;
extern cpuop_func op_6900_0_ff;
extern cpuop_func op_6901_0_nf;
extern cpuop_func op_6901_0_ff;
extern cpuop_func op_69ff_0_nf;
extern cpuop_func op_69ff_0_ff;
extern cpuop_func op_6a00_0_nf;
extern cpuop_func op_6a00_0_ff;
extern cpuop_func op_6a01_0_nf;
extern cpuop_func op_6a01_0_ff;
extern cpuop_func op_6aff_0_nf;
extern cpuop_func op_6aff_0_ff;
extern cpuop_func op_6b00_0_nf;
extern cpuop_func op_6b00_0_ff;
extern cpuop_func op_6b01_0_nf;
extern cpuop_func op_6b01_0_ff;
extern cpuop_func op_6bff_0_nf;
extern cpuop_func op_6bff_0_ff;
extern cpuop_func op_6c00_0_nf;
extern cpuop_func op_6c00_0_ff;
extern cpuop_func op_6c01_0_nf;
extern cpuop_func op_6c01_0_ff;
extern cpuop_func op_6cff_0_nf;
extern cpuop_func op_6cff_0_ff;
extern cpuop_func op_6d00_0_nf;
extern cpuop_func op_6d00_0_ff;
extern cpuop_func op_6d01_0_nf;
extern cpuop_func op_6d01_0_ff;
extern cpuop_func op_6dff_0_nf;
extern cpuop_func op_6dff_0_ff;
extern cpuop_func op_6e00_0_nf;
extern cpuop_func op_6e00_0_ff;
extern cpuop_func op_6e01_0_nf;
extern cpuop_func op_6e01_0_ff;
extern cpuop_func op_6eff_0_nf;
extern cpuop_func op_6eff_0_ff;
extern cpuop_func op_6f00_0_nf;
extern cpuop_func op_6f00_0_ff;
extern cpuop_func op_6f01_0_nf;
extern cpuop_func op_6f01_0_ff;
extern cpuop_func op_6fff_0_nf;
extern cpuop_func op_6fff_0_ff;
extern cpuop_func op_7000_0_nf;
extern cpuop_func op_7000_0_ff;
extern cpuop_func op_8000_0_nf;
extern cpuop_func op_8000_0_ff;
extern cpuop_func op_8010_0_nf;
extern cpuop_func op_8010_0_ff;
extern cpuop_func op_8018_0_nf;
extern cpuop_func op_8018_0_ff;
extern cpuop_func op_8020_0_nf;
extern cpuop_func op_8020_0_ff;
extern cpuop_func op_8028_0_nf;
extern cpuop_func op_8028_0_ff;
extern cpuop_func op_8030_0_nf;
extern cpuop_func op_8030_0_ff;
extern cpuop_func op_8038_0_nf;
extern cpuop_func op_8038_0_ff;
extern cpuop_func op_8039_0_nf;
extern cpuop_func op_8039_0_ff;
extern cpuop_func op_803a_0_nf;
extern cpuop_func op_803a_0_ff;
extern cpuop_func op_803b_0_nf;
extern cpuop_func op_803b_0_ff;
extern cpuop_func op_803c_0_nf;
extern cpuop_func op_803c_0_ff;
extern cpuop_func op_8040_0_nf;
extern cpuop_func op_8040_0_ff;
extern cpuop_func op_8050_0_nf;
extern cpuop_func op_8050_0_ff;
extern cpuop_func op_8058_0_nf;
extern cpuop_func op_8058_0_ff;
extern cpuop_func op_8060_0_nf;
extern cpuop_func op_8060_0_ff;
extern cpuop_func op_8068_0_nf;
extern cpuop_func op_8068_0_ff;
extern cpuop_func op_8070_0_nf;
extern cpuop_func op_8070_0_ff;
extern cpuop_func op_8078_0_nf;
extern cpuop_func op_8078_0_ff;
extern cpuop_func op_8079_0_nf;
extern cpuop_func op_8079_0_ff;
extern cpuop_func op_807a_0_nf;
extern cpuop_func op_807a_0_ff;
extern cpuop_func op_807b_0_nf;
extern cpuop_func op_807b_0_ff;
extern cpuop_func op_807c_0_nf;
extern cpuop_func op_807c_0_ff;
extern cpuop_func op_8080_0_nf;
extern cpuop_func op_8080_0_ff;
extern cpuop_func op_8090_0_nf;
extern cpuop_func op_8090_0_ff;
extern cpuop_func op_8098_0_nf;
extern cpuop_func op_8098_0_ff;
extern cpuop_func op_80a0_0_nf;
extern cpuop_func op_80a0_0_ff;
extern cpuop_func op_80a8_0_nf;
extern cpuop_func op_80a8_0_ff;
extern cpuop_func op_80b0_0_nf;
extern cpuop_func op_80b0_0_ff;
extern cpuop_func op_80b8_0_nf;
extern cpuop_func op_80b8_0_ff;
extern cpuop_func op_80b9_0_nf;
extern cpuop_func op_80b9_0_ff;
extern cpuop_func op_80ba_0_nf;
extern cpuop_func op_80ba_0_ff;
extern cpuop_func op_80bb_0_nf;
extern cpuop_func op_80bb_0_ff;
extern cpuop_func op_80bc_0_nf;
extern cpuop_func op_80bc_0_ff;
extern cpuop_func op_80c0_0_nf;
extern cpuop_func op_80c0_0_ff;
extern cpuop_func op_80d0_0_nf;
extern cpuop_func op_80d0_0_ff;
extern cpuop_func op_80d8_0_nf;
extern cpuop_func op_80d8_0_ff;
extern cpuop_func op_80e0_0_nf;
extern cpuop_func op_80e0_0_ff;
extern cpuop_func op_80e8_0_nf;
extern cpuop_func op_80e8_0_ff;
extern cpuop_func op_80f0_0_nf;
extern cpuop_func op_80f0_0_ff;
extern cpuop_func op_80f8_0_nf;
extern cpuop_func op_80f8_0_ff;
extern cpuop_func op_80f9_0_nf;
extern cpuop_func op_80f9_0_ff;
extern cpuop_func op_80fa_0_nf;
extern cpuop_func op_80fa_0_ff;
extern cpuop_func op_80fb_0_nf;
extern cpuop_func op_80fb_0_ff;
extern cpuop_func op_80fc_0_nf;
extern cpuop_func op_80fc_0_ff;
extern cpuop_func op_8100_0_nf;
extern cpuop_func op_8100_0_ff;
extern cpuop_func op_8108_0_nf;
extern cpuop_func op_8108_0_ff;
extern cpuop_func op_8110_0_nf;
extern cpuop_func op_8110_0_ff;
extern cpuop_func op_8118_0_nf;
extern cpuop_func op_8118_0_ff;
extern cpuop_func op_8120_0_nf;
extern cpuop_func op_8120_0_ff;
extern cpuop_func op_8128_0_nf;
extern cpuop_func op_8128_0_ff;
extern cpuop_func op_8130_0_nf;
extern cpuop_func op_8130_0_ff;
extern cpuop_func op_8138_0_nf;
extern cpuop_func op_8138_0_ff;
extern cpuop_func op_8139_0_nf;
extern cpuop_func op_8139_0_ff;
extern cpuop_func op_8140_0_nf;
extern cpuop_func op_8140_0_ff;
extern cpuop_func op_8148_0_nf;
extern cpuop_func op_8148_0_ff;
extern cpuop_func op_8150_0_nf;
extern cpuop_func op_8150_0_ff;
extern cpuop_func op_8158_0_nf;
extern cpuop_func op_8158_0_ff;
extern cpuop_func op_8160_0_nf;
extern cpuop_func op_8160_0_ff;
extern cpuop_func op_8168_0_nf;
extern cpuop_func op_8168_0_ff;
extern cpuop_func op_8170_0_nf;
extern cpuop_func op_8170_0_ff;
extern cpuop_func op_8178_0_nf;
extern cpuop_func op_8178_0_ff;
extern cpuop_func op_8179_0_nf;
extern cpuop_func op_8179_0_ff;
extern cpuop_func op_8180_0_nf;
extern cpuop_func op_8180_0_ff;
extern cpuop_func op_8188_0_nf;
extern cpuop_func op_8188_0_ff;
extern cpuop_func op_8190_0_nf;
extern cpuop_func op_8190_0_ff;
extern cpuop_func op_8198_0_nf;
extern cpuop_func op_8198_0_ff;
extern cpuop_func op_81a0_0_nf;
extern cpuop_func op_81a0_0_ff;
extern cpuop_func op_81a8_0_nf;
extern cpuop_func op_81a8_0_ff;
extern cpuop_func op_81b0_0_nf;
extern cpuop_func op_81b0_0_ff;
extern cpuop_func op_81b8_0_nf;
extern cpuop_func op_81b8_0_ff;
extern cpuop_func op_81b9_0_nf;
extern cpuop_func op_81b9_0_ff;
extern cpuop_func op_81c0_0_nf;
extern cpuop_func op_81c0_0_ff;
extern cpuop_func op_81d0_0_nf;
extern cpuop_func op_81d0_0_ff;
extern cpuop_func op_81d8_0_nf;
extern cpuop_func op_81d8_0_ff;
extern cpuop_func op_81e0_0_nf;
extern cpuop_func op_81e0_0_ff;
extern cpuop_func op_81e8_0_nf;
extern cpuop_func op_81e8_0_ff;
extern cpuop_func op_81f0_0_nf;
extern cpuop_func op_81f0_0_ff;
extern cpuop_func op_81f8_0_nf;
extern cpuop_func op_81f8_0_ff;
extern cpuop_func op_81f9_0_nf;
extern cpuop_func op_81f9_0_ff;
extern cpuop_func op_81fa_0_nf;
extern cpuop_func op_81fa_0_ff;
extern cpuop_func op_81fb_0_nf;
extern cpuop_func op_81fb_0_ff;
extern cpuop_func op_81fc_0_nf;
extern cpuop_func op_81fc_0_ff;
extern cpuop_func op_9000_0_nf;
extern cpuop_func op_9000_0_ff;
extern cpuop_func op_9010_0_nf;
extern cpuop_func op_9010_0_ff;
extern cpuop_func op_9018_0_nf;
extern cpuop_func op_9018_0_ff;
extern cpuop_func op_9020_0_nf;
extern cpuop_func op_9020_0_ff;
extern cpuop_func op_9028_0_nf;
extern cpuop_func op_9028_0_ff;
extern cpuop_func op_9030_0_nf;
extern cpuop_func op_9030_0_ff;
extern cpuop_func op_9038_0_nf;
extern cpuop_func op_9038_0_ff;
extern cpuop_func op_9039_0_nf;
extern cpuop_func op_9039_0_ff;
extern cpuop_func op_903a_0_nf;
extern cpuop_func op_903a_0_ff;
extern cpuop_func op_903b_0_nf;
extern cpuop_func op_903b_0_ff;
extern cpuop_func op_903c_0_nf;
extern cpuop_func op_903c_0_ff;
extern cpuop_func op_9040_0_nf;
extern cpuop_func op_9040_0_ff;
extern cpuop_func op_9048_0_nf;
extern cpuop_func op_9048_0_ff;
extern cpuop_func op_9050_0_nf;
extern cpuop_func op_9050_0_ff;
extern cpuop_func op_9058_0_nf;
extern cpuop_func op_9058_0_ff;
extern cpuop_func op_9060_0_nf;
extern cpuop_func op_9060_0_ff;
extern cpuop_func op_9068_0_nf;
extern cpuop_func op_9068_0_ff;
extern cpuop_func op_9070_0_nf;
extern cpuop_func op_9070_0_ff;
extern cpuop_func op_9078_0_nf;
extern cpuop_func op_9078_0_ff;
extern cpuop_func op_9079_0_nf;
extern cpuop_func op_9079_0_ff;
extern cpuop_func op_907a_0_nf;
extern cpuop_func op_907a_0_ff;
extern cpuop_func op_907b_0_nf;
extern cpuop_func op_907b_0_ff;
extern cpuop_func op_907c_0_nf;
extern cpuop_func op_907c_0_ff;
extern cpuop_func op_9080_0_nf;
extern cpuop_func op_9080_0_ff;
extern cpuop_func op_9088_0_nf;
extern cpuop_func op_9088_0_ff;
extern cpuop_func op_9090_0_nf;
extern cpuop_func op_9090_0_ff;
extern cpuop_func op_9098_0_nf;
extern cpuop_func op_9098_0_ff;
extern cpuop_func op_90a0_0_nf;
extern cpuop_func op_90a0_0_ff;
extern cpuop_func op_90a8_0_nf;
extern cpuop_func op_90a8_0_ff;
extern cpuop_func op_90b0_0_nf;
extern cpuop_func op_90b0_0_ff;
extern cpuop_func op_90b8_0_nf;
extern cpuop_func op_90b8_0_ff;
extern cpuop_func op_90b9_0_nf;
extern cpuop_func op_90b9_0_ff;
extern cpuop_func op_90ba_0_nf;
extern cpuop_func op_90ba_0_ff;
extern cpuop_func op_90bb_0_nf;
extern cpuop_func op_90bb_0_ff;
extern cpuop_func op_90bc_0_nf;
extern cpuop_func op_90bc_0_ff;
extern cpuop_func op_90c0_0_nf;
extern cpuop_func op_90c0_0_ff;
extern cpuop_func op_90c8_0_nf;
extern cpuop_func op_90c8_0_ff;
extern cpuop_func op_90d0_0_nf;
extern cpuop_func op_90d0_0_ff;
extern cpuop_func op_90d8_0_nf;
extern cpuop_func op_90d8_0_ff;
extern cpuop_func op_90e0_0_nf;
extern cpuop_func op_90e0_0_ff;
extern cpuop_func op_90e8_0_nf;
extern cpuop_func op_90e8_0_ff;
extern cpuop_func op_90f0_0_nf;
extern cpuop_func op_90f0_0_ff;
extern cpuop_func op_90f8_0_nf;
extern cpuop_func op_90f8_0_ff;
extern cpuop_func op_90f9_0_nf;
extern cpuop_func op_90f9_0_ff;
extern cpuop_func op_90fa_0_nf;
extern cpuop_func op_90fa_0_ff;
extern cpuop_func op_90fb_0_nf;
extern cpuop_func op_90fb_0_ff;
extern cpuop_func op_90fc_0_nf;
extern cpuop_func op_90fc_0_ff;
extern cpuop_func op_9100_0_nf;
extern cpuop_func op_9100_0_ff;
extern cpuop_func op_9108_0_nf;
extern cpuop_func op_9108_0_ff;
extern cpuop_func op_9110_0_nf;
extern cpuop_func op_9110_0_ff;
extern cpuop_func op_9118_0_nf;
extern cpuop_func op_9118_0_ff;
extern cpuop_func op_9120_0_nf;
extern cpuop_func op_9120_0_ff;
extern cpuop_func op_9128_0_nf;
extern cpuop_func op_9128_0_ff;
extern cpuop_func op_9130_0_nf;
extern cpuop_func op_9130_0_ff;
extern cpuop_func op_9138_0_nf;
extern cpuop_func op_9138_0_ff;
extern cpuop_func op_9139_0_nf;
extern cpuop_func op_9139_0_ff;
extern cpuop_func op_9140_0_nf;
extern cpuop_func op_9140_0_ff;
extern cpuop_func op_9148_0_nf;
extern cpuop_func op_9148_0_ff;
extern cpuop_func op_9150_0_nf;
extern cpuop_func op_9150_0_ff;
extern cpuop_func op_9158_0_nf;
extern cpuop_func op_9158_0_ff;
extern cpuop_func op_9160_0_nf;
extern cpuop_func op_9160_0_ff;
extern cpuop_func op_9168_0_nf;
extern cpuop_func op_9168_0_ff;
extern cpuop_func op_9170_0_nf;
extern cpuop_func op_9170_0_ff;
extern cpuop_func op_9178_0_nf;
extern cpuop_func op_9178_0_ff;
extern cpuop_func op_9179_0_nf;
extern cpuop_func op_9179_0_ff;
extern cpuop_func op_9180_0_nf;
extern cpuop_func op_9180_0_ff;
extern cpuop_func op_9188_0_nf;
extern cpuop_func op_9188_0_ff;
extern cpuop_func op_9190_0_nf;
extern cpuop_func op_9190_0_ff;
extern cpuop_func op_9198_0_nf;
extern cpuop_func op_9198_0_ff;
extern cpuop_func op_91a0_0_nf;
extern cpuop_func op_91a0_0_ff;
extern cpuop_func op_91a8_0_nf;
extern cpuop_func op_91a8_0_ff;
extern cpuop_func op_91b0_0_nf;
extern cpuop_func op_91b0_0_ff;
extern cpuop_func op_91b8_0_nf;
extern cpuop_func op_91b8_0_ff;
extern cpuop_func op_91b9_0_nf;
extern cpuop_func op_91b9_0_ff;
extern cpuop_func op_91c0_0_nf;
extern cpuop_func op_91c0_0_ff;
extern cpuop_func op_91c8_0_nf;
extern cpuop_func op_91c8_0_ff;
extern cpuop_func op_91d0_0_nf;
extern cpuop_func op_91d0_0_ff;
extern cpuop_func op_91d8_0_nf;
extern cpuop_func op_91d8_0_ff;
extern cpuop_func op_91e0_0_nf;
extern cpuop_func op_91e0_0_ff;
extern cpuop_func op_91e8_0_nf;
extern cpuop_func op_91e8_0_ff;
extern cpuop_func op_91f0_0_nf;
extern cpuop_func op_91f0_0_ff;
extern cpuop_func op_91f8_0_nf;
extern cpuop_func op_91f8_0_ff;
extern cpuop_func op_91f9_0_nf;
extern cpuop_func op_91f9_0_ff;
extern cpuop_func op_91fa_0_nf;
extern cpuop_func op_91fa_0_ff;
extern cpuop_func op_91fb_0_nf;
extern cpuop_func op_91fb_0_ff;
extern cpuop_func op_91fc_0_nf;
extern cpuop_func op_91fc_0_ff;
extern cpuop_func op_b000_0_nf;
extern cpuop_func op_b000_0_ff;
extern cpuop_func op_b010_0_nf;
extern cpuop_func op_b010_0_ff;
extern cpuop_func op_b018_0_nf;
extern cpuop_func op_b018_0_ff;
extern cpuop_func op_b020_0_nf;
extern cpuop_func op_b020_0_ff;
extern cpuop_func op_b028_0_nf;
extern cpuop_func op_b028_0_ff;
extern cpuop_func op_b030_0_nf;
extern cpuop_func op_b030_0_ff;
extern cpuop_func op_b038_0_nf;
extern cpuop_func op_b038_0_ff;
extern cpuop_func op_b039_0_nf;
extern cpuop_func op_b039_0_ff;
extern cpuop_func op_b03a_0_nf;
extern cpuop_func op_b03a_0_ff;
extern cpuop_func op_b03b_0_nf;
extern cpuop_func op_b03b_0_ff;
extern cpuop_func op_b03c_0_nf;
extern cpuop_func op_b03c_0_ff;
extern cpuop_func op_b040_0_nf;
extern cpuop_func op_b040_0_ff;
extern cpuop_func op_b048_0_nf;
extern cpuop_func op_b048_0_ff;
extern cpuop_func op_b050_0_nf;
extern cpuop_func op_b050_0_ff;
extern cpuop_func op_b058_0_nf;
extern cpuop_func op_b058_0_ff;
extern cpuop_func op_b060_0_nf;
extern cpuop_func op_b060_0_ff;
extern cpuop_func op_b068_0_nf;
extern cpuop_func op_b068_0_ff;
extern cpuop_func op_b070_0_nf;
extern cpuop_func op_b070_0_ff;
extern cpuop_func op_b078_0_nf;
extern cpuop_func op_b078_0_ff;
extern cpuop_func op_b079_0_nf;
extern cpuop_func op_b079_0_ff;
extern cpuop_func op_b07a_0_nf;
extern cpuop_func op_b07a_0_ff;
extern cpuop_func op_b07b_0_nf;
extern cpuop_func op_b07b_0_ff;
extern cpuop_func op_b07c_0_nf;
extern cpuop_func op_b07c_0_ff;
extern cpuop_func op_b080_0_nf;
extern cpuop_func op_b080_0_ff;
extern cpuop_func op_b088_0_nf;
extern cpuop_func op_b088_0_ff;
extern cpuop_func op_b090_0_nf;
extern cpuop_func op_b090_0_ff;
extern cpuop_func op_b098_0_nf;
extern cpuop_func op_b098_0_ff;
extern cpuop_func op_b0a0_0_nf;
extern cpuop_func op_b0a0_0_ff;
extern cpuop_func op_b0a8_0_nf;
extern cpuop_func op_b0a8_0_ff;
extern cpuop_func op_b0b0_0_nf;
extern cpuop_func op_b0b0_0_ff;
extern cpuop_func op_b0b8_0_nf;
extern cpuop_func op_b0b8_0_ff;
extern cpuop_func op_b0b9_0_nf;
extern cpuop_func op_b0b9_0_ff;
extern cpuop_func op_b0ba_0_nf;
extern cpuop_func op_b0ba_0_ff;
extern cpuop_func op_b0bb_0_nf;
extern cpuop_func op_b0bb_0_ff;
extern cpuop_func op_b0bc_0_nf;
extern cpuop_func op_b0bc_0_ff;
extern cpuop_func op_b0c0_0_nf;
extern cpuop_func op_b0c0_0_ff;
extern cpuop_func op_b0c8_0_nf;
extern cpuop_func op_b0c8_0_ff;
extern cpuop_func op_b0d0_0_nf;
extern cpuop_func op_b0d0_0_ff;
extern cpuop_func op_b0d8_0_nf;
extern cpuop_func op_b0d8_0_ff;
extern cpuop_func op_b0e0_0_nf;
extern cpuop_func op_b0e0_0_ff;
extern cpuop_func op_b0e8_0_nf;
extern cpuop_func op_b0e8_0_ff;
extern cpuop_func op_b0f0_0_nf;
extern cpuop_func op_b0f0_0_ff;
extern cpuop_func op_b0f8_0_nf;
extern cpuop_func op_b0f8_0_ff;
extern cpuop_func op_b0f9_0_nf;
extern cpuop_func op_b0f9_0_ff;
extern cpuop_func op_b0fa_0_nf;
extern cpuop_func op_b0fa_0_ff;
extern cpuop_func op_b0fb_0_nf;
extern cpuop_func op_b0fb_0_ff;
extern cpuop_func op_b0fc_0_nf;
extern cpuop_func op_b0fc_0_ff;
extern cpuop_func op_b100_0_nf;
extern cpuop_func op_b100_0_ff;
extern cpuop_func op_b108_0_nf;
extern cpuop_func op_b108_0_ff;
extern cpuop_func op_b110_0_nf;
extern cpuop_func op_b110_0_ff;
extern cpuop_func op_b118_0_nf;
extern cpuop_func op_b118_0_ff;
extern cpuop_func op_b120_0_nf;
extern cpuop_func op_b120_0_ff;
extern cpuop_func op_b128_0_nf;
extern cpuop_func op_b128_0_ff;
extern cpuop_func op_b130_0_nf;
extern cpuop_func op_b130_0_ff;
extern cpuop_func op_b138_0_nf;
extern cpuop_func op_b138_0_ff;
extern cpuop_func op_b139_0_nf;
extern cpuop_func op_b139_0_ff;
extern cpuop_func op_b140_0_nf;
extern cpuop_func op_b140_0_ff;
extern cpuop_func op_b148_0_nf;
extern cpuop_func op_b148_0_ff;
extern cpuop_func op_b150_0_nf;
extern cpuop_func op_b150_0_ff;
extern cpuop_func op_b158_0_nf;
extern cpuop_func op_b158_0_ff;
extern cpuop_func op_b160_0_nf;
extern cpuop_func op_b160_0_ff;
extern cpuop_func op_b168_0_nf;
extern cpuop_func op_b168_0_ff;
extern cpuop_func op_b170_0_nf;
extern cpuop_func op_b170_0_ff;
extern cpuop_func op_b178_0_nf;
extern cpuop_func op_b178_0_ff;
extern cpuop_func op_b179_0_nf;
extern cpuop_func op_b179_0_ff;
extern cpuop_func op_b180_0_nf;
extern cpuop_func op_b180_0_ff;
extern cpuop_func op_b188_0_nf;
extern cpuop_func op_b188_0_ff;
extern cpuop_func op_b190_0_nf;
extern cpuop_func op_b190_0_ff;
extern cpuop_func op_b198_0_nf;
extern cpuop_func op_b198_0_ff;
extern cpuop_func op_b1a0_0_nf;
extern cpuop_func op_b1a0_0_ff;
extern cpuop_func op_b1a8_0_nf;
extern cpuop_func op_b1a8_0_ff;
extern cpuop_func op_b1b0_0_nf;
extern cpuop_func op_b1b0_0_ff;
extern cpuop_func op_b1b8_0_nf;
extern cpuop_func op_b1b8_0_ff;
extern cpuop_func op_b1b9_0_nf;
extern cpuop_func op_b1b9_0_ff;
extern cpuop_func op_b1c0_0_nf;
extern cpuop_func op_b1c0_0_ff;
extern cpuop_func op_b1c8_0_nf;
extern cpuop_func op_b1c8_0_ff;
extern cpuop_func op_b1d0_0_nf;
extern cpuop_func op_b1d0_0_ff;
extern cpuop_func op_b1d8_0_nf;
extern cpuop_func op_b1d8_0_ff;
extern cpuop_func op_b1e0_0_nf;
extern cpuop_func op_b1e0_0_ff;
extern cpuop_func op_b1e8_0_nf;
extern cpuop_func op_b1e8_0_ff;
extern cpuop_func op_b1f0_0_nf;
extern cpuop_func op_b1f0_0_ff;
extern cpuop_func op_b1f8_0_nf;
extern cpuop_func op_b1f8_0_ff;
extern cpuop_func op_b1f9_0_nf;
extern cpuop_func op_b1f9_0_ff;
extern cpuop_func op_b1fa_0_nf;
extern cpuop_func op_b1fa_0_ff;
extern cpuop_func op_b1fb_0_nf;
extern cpuop_func op_b1fb_0_ff;
extern cpuop_func op_b1fc_0_nf;
extern cpuop_func op_b1fc_0_ff;
extern cpuop_func op_c000_0_nf;
extern cpuop_func op_c000_0_ff;
extern cpuop_func op_c010_0_nf;
extern cpuop_func op_c010_0_ff;
extern cpuop_func op_c018_0_nf;
extern cpuop_func op_c018_0_ff;
extern cpuop_func op_c020_0_nf;
extern cpuop_func op_c020_0_ff;
extern cpuop_func op_c028_0_nf;
extern cpuop_func op_c028_0_ff;
extern cpuop_func op_c030_0_nf;
extern cpuop_func op_c030_0_ff;
extern cpuop_func op_c038_0_nf;
extern cpuop_func op_c038_0_ff;
extern cpuop_func op_c039_0_nf;
extern cpuop_func op_c039_0_ff;
extern cpuop_func op_c03a_0_nf;
extern cpuop_func op_c03a_0_ff;
extern cpuop_func op_c03b_0_nf;
extern cpuop_func op_c03b_0_ff;
extern cpuop_func op_c03c_0_nf;
extern cpuop_func op_c03c_0_ff;
extern cpuop_func op_c040_0_nf;
extern cpuop_func op_c040_0_ff;
extern cpuop_func op_c050_0_nf;
extern cpuop_func op_c050_0_ff;
extern cpuop_func op_c058_0_nf;
extern cpuop_func op_c058_0_ff;
extern cpuop_func op_c060_0_nf;
extern cpuop_func op_c060_0_ff;
extern cpuop_func op_c068_0_nf;
extern cpuop_func op_c068_0_ff;
extern cpuop_func op_c070_0_nf;
extern cpuop_func op_c070_0_ff;
extern cpuop_func op_c078_0_nf;
extern cpuop_func op_c078_0_ff;
extern cpuop_func op_c079_0_nf;
extern cpuop_func op_c079_0_ff;
extern cpuop_func op_c07a_0_nf;
extern cpuop_func op_c07a_0_ff;
extern cpuop_func op_c07b_0_nf;
extern cpuop_func op_c07b_0_ff;
extern cpuop_func op_c07c_0_nf;
extern cpuop_func op_c07c_0_ff;
extern cpuop_func op_c080_0_nf;
extern cpuop_func op_c080_0_ff;
extern cpuop_func op_c090_0_nf;
extern cpuop_func op_c090_0_ff;
extern cpuop_func op_c098_0_nf;
extern cpuop_func op_c098_0_ff;
extern cpuop_func op_c0a0_0_nf;
extern cpuop_func op_c0a0_0_ff;
extern cpuop_func op_c0a8_0_nf;
extern cpuop_func op_c0a8_0_ff;
extern cpuop_func op_c0b0_0_nf;
extern cpuop_func op_c0b0_0_ff;
extern cpuop_func op_c0b8_0_nf;
extern cpuop_func op_c0b8_0_ff;
extern cpuop_func op_c0b9_0_nf;
extern cpuop_func op_c0b9_0_ff;
extern cpuop_func op_c0ba_0_nf;
extern cpuop_func op_c0ba_0_ff;
extern cpuop_func op_c0bb_0_nf;
extern cpuop_func op_c0bb_0_ff;
extern cpuop_func op_c0bc_0_nf;
extern cpuop_func op_c0bc_0_ff;
extern cpuop_func op_c0c0_0_nf;
extern cpuop_func op_c0c0_0_ff;
extern cpuop_func op_c0d0_0_nf;
extern cpuop_func op_c0d0_0_ff;
extern cpuop_func op_c0d8_0_nf;
extern cpuop_func op_c0d8_0_ff;
extern cpuop_func op_c0e0_0_nf;
extern cpuop_func op_c0e0_0_ff;
extern cpuop_func op_c0e8_0_nf;
extern cpuop_func op_c0e8_0_ff;
extern cpuop_func op_c0f0_0_nf;
extern cpuop_func op_c0f0_0_ff;
extern cpuop_func op_c0f8_0_nf;
extern cpuop_func op_c0f8_0_ff;
extern cpuop_func op_c0f9_0_nf;
extern cpuop_func op_c0f9_0_ff;
extern cpuop_func op_c0fa_0_nf;
extern cpuop_func op_c0fa_0_ff;
extern cpuop_func op_c0fb_0_nf;
extern cpuop_func op_c0fb_0_ff;
extern cpuop_func op_c0fc_0_nf;
extern cpuop_func op_c0fc_0_ff;
extern cpuop_func op_c100_0_nf;
extern cpuop_func op_c100_0_ff;
extern cpuop_func op_c108_0_nf;
extern cpuop_func op_c108_0_ff;
extern cpuop_func op_c110_0_nf;
extern cpuop_func op_c110_0_ff;
extern cpuop_func op_c118_0_nf;
extern cpuop_func op_c118_0_ff;
extern cpuop_func op_c120_0_nf;
extern cpuop_func op_c120_0_ff;
extern cpuop_func op_c128_0_nf;
extern cpuop_func op_c128_0_ff;
extern cpuop_func op_c130_0_nf;
extern cpuop_func op_c130_0_ff;
extern cpuop_func op_c138_0_nf;
extern cpuop_func op_c138_0_ff;
extern cpuop_func op_c139_0_nf;
extern cpuop_func op_c139_0_ff;
extern cpuop_func op_c140_0_nf;
extern cpuop_func op_c140_0_ff;
extern cpuop_func op_c148_0_nf;
extern cpuop_func op_c148_0_ff;
extern cpuop_func op_c150_0_nf;
extern cpuop_func op_c150_0_ff;
extern cpuop_func op_c158_0_nf;
extern cpuop_func op_c158_0_ff;
extern cpuop_func op_c160_0_nf;
extern cpuop_func op_c160_0_ff;
extern cpuop_func op_c168_0_nf;
extern cpuop_func op_c168_0_ff;
extern cpuop_func op_c170_0_nf;
extern cpuop_func op_c170_0_ff;
extern cpuop_func op_c178_0_nf;
extern cpuop_func op_c178_0_ff;
extern cpuop_func op_c179_0_nf;
extern cpuop_func op_c179_0_ff;
extern cpuop_func op_c188_0_nf;
extern cpuop_func op_c188_0_ff;
extern cpuop_func op_c190_0_nf;
extern cpuop_func op_c190_0_ff;
extern cpuop_func op_c198_0_nf;
extern cpuop_func op_c198_0_ff;
extern cpuop_func op_c1a0_0_nf;
extern cpuop_func op_c1a0_0_ff;
extern cpuop_func op_c1a8_0_nf;
extern cpuop_func op_c1a8_0_ff;
extern cpuop_func op_c1b0_0_nf;
extern cpuop_func op_c1b0_0_ff;
extern cpuop_func op_c1b8_0_nf;
extern cpuop_func op_c1b8_0_ff;
extern cpuop_func op_c1b9_0_nf;
extern cpuop_func op_c1b9_0_ff;
extern cpuop_func op_c1c0_0_nf;
extern cpuop_func op_c1c0_0_ff;
extern cpuop_func op_c1d0_0_nf;
extern cpuop_func op_c1d0_0_ff;
extern cpuop_func op_c1d8_0_nf;
extern cpuop_func op_c1d8_0_ff;
extern cpuop_func op_c1e0_0_nf;
extern cpuop_func op_c1e0_0_ff;
extern cpuop_func op_c1e8_0_nf;
extern cpuop_func op_c1e8_0_ff;
extern cpuop_func op_c1f0_0_nf;
extern cpuop_func op_c1f0_0_ff;
extern cpuop_func op_c1f8_0_nf;
extern cpuop_func op_c1f8_0_ff;
extern cpuop_func op_c1f9_0_nf;
extern cpuop_func op_c1f9_0_ff;
extern cpuop_func op_c1fa_0_nf;
extern cpuop_func op_c1fa_0_ff;
extern cpuop_func op_c1fb_0_nf;
extern cpuop_func op_c1fb_0_ff;
extern cpuop_func op_c1fc_0_nf;
extern cpuop_func op_c1fc_0_ff;
extern cpuop_func op_d000_0_nf;
extern cpuop_func op_d000_0_ff;
extern cpuop_func op_d010_0_nf;
extern cpuop_func op_d010_0_ff;
extern cpuop_func op_d018_0_nf;
extern cpuop_func op_d018_0_ff;
extern cpuop_func op_d020_0_nf;
extern cpuop_func op_d020_0_ff;
extern cpuop_func op_d028_0_nf;
extern cpuop_func op_d028_0_ff;
extern cpuop_func op_d030_0_nf;
extern cpuop_func op_d030_0_ff;
extern cpuop_func op_d038_0_nf;
extern cpuop_func op_d038_0_ff;
extern cpuop_func op_d039_0_nf;
extern cpuop_func op_d039_0_ff;
extern cpuop_func op_d03a_0_nf;
extern cpuop_func op_d03a_0_ff;
extern cpuop_func op_d03b_0_nf;
extern cpuop_func op_d03b_0_ff;
extern cpuop_func op_d03c_0_nf;
extern cpuop_func op_d03c_0_ff;
extern cpuop_func op_d040_0_nf;
extern cpuop_func op_d040_0_ff;
extern cpuop_func op_d048_0_nf;
extern cpuop_func op_d048_0_ff;
extern cpuop_func op_d050_0_nf;
extern cpuop_func op_d050_0_ff;
extern cpuop_func op_d058_0_nf;
extern cpuop_func op_d058_0_ff;
extern cpuop_func op_d060_0_nf;
extern cpuop_func op_d060_0_ff;
extern cpuop_func op_d068_0_nf;
extern cpuop_func op_d068_0_ff;
extern cpuop_func op_d070_0_nf;
extern cpuop_func op_d070_0_ff;
extern cpuop_func op_d078_0_nf;
extern cpuop_func op_d078_0_ff;
extern cpuop_func op_d079_0_nf;
extern cpuop_func op_d079_0_ff;
extern cpuop_func op_d07a_0_nf;
extern cpuop_func op_d07a_0_ff;
extern cpuop_func op_d07b_0_nf;
extern cpuop_func op_d07b_0_ff;
extern cpuop_func op_d07c_0_nf;
extern cpuop_func op_d07c_0_ff;
extern cpuop_func op_d080_0_nf;
extern cpuop_func op_d080_0_ff;
extern cpuop_func op_d088_0_nf;
extern cpuop_func op_d088_0_ff;
extern cpuop_func op_d090_0_nf;
extern cpuop_func op_d090_0_ff;
extern cpuop_func op_d098_0_nf;
extern cpuop_func op_d098_0_ff;
extern cpuop_func op_d0a0_0_nf;
extern cpuop_func op_d0a0_0_ff;
extern cpuop_func op_d0a8_0_nf;
extern cpuop_func op_d0a8_0_ff;
extern cpuop_func op_d0b0_0_nf;
extern cpuop_func op_d0b0_0_ff;
extern cpuop_func op_d0b8_0_nf;
extern cpuop_func op_d0b8_0_ff;
extern cpuop_func op_d0b9_0_nf;
extern cpuop_func op_d0b9_0_ff;
extern cpuop_func op_d0ba_0_nf;
extern cpuop_func op_d0ba_0_ff;
extern cpuop_func op_d0bb_0_nf;
extern cpuop_func op_d0bb_0_ff;
extern cpuop_func op_d0bc_0_nf;
extern cpuop_func op_d0bc_0_ff;
extern cpuop_func op_d0c0_0_nf;
extern cpuop_func op_d0c0_0_ff;
extern cpuop_func op_d0c8_0_nf;
extern cpuop_func op_d0c8_0_ff;
extern cpuop_func op_d0d0_0_nf;
extern cpuop_func op_d0d0_0_ff;
extern cpuop_func op_d0d8_0_nf;
extern cpuop_func op_d0d8_0_ff;
extern cpuop_func op_d0e0_0_nf;
extern cpuop_func op_d0e0_0_ff;
extern cpuop_func op_d0e8_0_nf;
extern cpuop_func op_d0e8_0_ff;
extern cpuop_func op_d0f0_0_nf;
extern cpuop_func op_d0f0_0_ff;
extern cpuop_func op_d0f8_0_nf;
extern cpuop_func op_d0f8_0_ff;
extern cpuop_func op_d0f9_0_nf;
extern cpuop_func op_d0f9_0_ff;
extern cpuop_func op_d0fa_0_nf;
extern cpuop_func op_d0fa_0_ff;
extern cpuop_func op_d0fb_0_nf;
extern cpuop_func op_d0fb_0_ff;
extern cpuop_func op_d0fc_0_nf;
extern cpuop_func op_d0fc_0_ff;
extern cpuop_func op_d100_0_nf;
extern cpuop_func op_d100_0_ff;
extern cpuop_func op_d108_0_nf;
extern cpuop_func op_d108_0_ff;
extern cpuop_func op_d110_0_nf;
extern cpuop_func op_d110_0_ff;
extern cpuop_func op_d118_0_nf;
extern cpuop_func op_d118_0_ff;
extern cpuop_func op_d120_0_nf;
extern cpuop_func op_d120_0_ff;
extern cpuop_func op_d128_0_nf;
extern cpuop_func op_d128_0_ff;
extern cpuop_func op_d130_0_nf;
extern cpuop_func op_d130_0_ff;
extern cpuop_func op_d138_0_nf;
extern cpuop_func op_d138_0_ff;
extern cpuop_func op_d139_0_nf;
extern cpuop_func op_d139_0_ff;
extern cpuop_func op_d140_0_nf;
extern cpuop_func op_d140_0_ff;
extern cpuop_func op_d148_0_nf;
extern cpuop_func op_d148_0_ff;
extern cpuop_func op_d150_0_nf;
extern cpuop_func op_d150_0_ff;
extern cpuop_func op_d158_0_nf;
extern cpuop_func op_d158_0_ff;
extern cpuop_func op_d160_0_nf;
extern cpuop_func op_d160_0_ff;
extern cpuop_func op_d168_0_nf;
extern cpuop_func op_d168_0_ff;
extern cpuop_func op_d170_0_nf;
extern cpuop_func op_d170_0_ff;
extern cpuop_func op_d178_0_nf;
extern cpuop_func op_d178_0_ff;
extern cpuop_func op_d179_0_nf;
extern cpuop_func op_d179_0_ff;
extern cpuop_func op_d180_0_nf;
extern cpuop_func op_d180_0_ff;
extern cpuop_func op_d188_0_nf;
extern cpuop_func op_d188_0_ff;
extern cpuop_func op_d190_0_nf;
extern cpuop_func op_d190_0_ff;
extern cpuop_func op_d198_0_nf;
extern cpuop_func op_d198_0_ff;
extern cpuop_func op_d1a0_0_nf;
extern cpuop_func op_d1a0_0_ff;
extern cpuop_func op_d1a8_0_nf;
extern cpuop_func op_d1a8_0_ff;
extern cpuop_func op_d1b0_0_nf;
extern cpuop_func op_d1b0_0_ff;
extern cpuop_func op_d1b8_0_nf;
extern cpuop_func op_d1b8_0_ff;
extern cpuop_func op_d1b9_0_nf;
extern cpuop_func op_d1b9_0_ff;
extern cpuop_func op_d1c0_0_nf;
extern cpuop_func op_d1c0_0_ff;
extern cpuop_func op_d1c8_0_nf;
extern cpuop_func op_d1c8_0_ff;
extern cpuop_func op_d1d0_0_nf;
extern cpuop_func op_d1d0_0_ff;
extern cpuop_func op_d1d8_0_nf;
extern cpuop_func op_d1d8_0_ff;
extern cpuop_func op_d1e0_0_nf;
extern cpuop_func op_d1e0_0_ff;
extern cpuop_func op_d1e8_0_nf;
extern cpuop_func op_d1e8_0_ff;
extern cpuop_func op_d1f0_0_nf;
extern cpuop_func op_d1f0_0_ff;
extern cpuop_func op_d1f8_0_nf;
extern cpuop_func op_d1f8_0_ff;
extern cpuop_func op_d1f9_0_nf;
extern cpuop_func op_d1f9_0_ff;
extern cpuop_func op_d1fa_0_nf;
extern cpuop_func op_d1fa_0_ff;
extern cpuop_func op_d1fb_0_nf;
extern cpuop_func op_d1fb_0_ff;
extern cpuop_func op_d1fc_0_nf;
extern cpuop_func op_d1fc_0_ff;
extern cpuop_func op_e000_0_nf;
extern cpuop_func op_e000_0_ff;
extern cpuop_func op_e008_0_nf;
extern cpuop_func op_e008_0_ff;
extern cpuop_func op_e010_0_nf;
extern cpuop_func op_e010_0_ff;
extern cpuop_func op_e018_0_nf;
extern cpuop_func op_e018_0_ff;
extern cpuop_func op_e020_0_nf;
extern cpuop_func op_e020_0_ff;
extern cpuop_func op_e028_0_nf;
extern cpuop_func op_e028_0_ff;
extern cpuop_func op_e030_0_nf;
extern cpuop_func op_e030_0_ff;
extern cpuop_func op_e038_0_nf;
extern cpuop_func op_e038_0_ff;
extern cpuop_func op_e040_0_nf;
extern cpuop_func op_e040_0_ff;
extern cpuop_func op_e048_0_nf;
extern cpuop_func op_e048_0_ff;
extern cpuop_func op_e050_0_nf;
extern cpuop_func op_e050_0_ff;
extern cpuop_func op_e058_0_nf;
extern cpuop_func op_e058_0_ff;
extern cpuop_func op_e060_0_nf;
extern cpuop_func op_e060_0_ff;
extern cpuop_func op_e068_0_nf;
extern cpuop_func op_e068_0_ff;
extern cpuop_func op_e070_0_nf;
extern cpuop_func op_e070_0_ff;
extern cpuop_func op_e078_0_nf;
extern cpuop_func op_e078_0_ff;
extern cpuop_func op_e080_0_nf;
extern cpuop_func op_e080_0_ff;
extern cpuop_func op_e088_0_nf;
extern cpuop_func op_e088_0_ff;
extern cpuop_func op_e090_0_nf;
extern cpuop_func op_e090_0_ff;
extern cpuop_func op_e098_0_nf;
extern cpuop_func op_e098_0_ff;
extern cpuop_func op_e0a0_0_nf;
extern cpuop_func op_e0a0_0_ff;
extern cpuop_func op_e0a8_0_nf;
extern cpuop_func op_e0a8_0_ff;
extern cpuop_func op_e0b0_0_nf;
extern cpuop_func op_e0b0_0_ff;
extern cpuop_func op_e0b8_0_nf;
extern cpuop_func op_e0b8_0_ff;
extern cpuop_func op_e0d0_0_nf;
extern cpuop_func op_e0d0_0_ff;
extern cpuop_func op_e0d8_0_nf;
extern cpuop_func op_e0d8_0_ff;
extern cpuop_func op_e0e0_0_nf;
extern cpuop_func op_e0e0_0_ff;
extern cpuop_func op_e0e8_0_nf;
extern cpuop_func op_e0e8_0_ff;
extern cpuop_func op_e0f0_0_nf;
extern cpuop_func op_e0f0_0_ff;
extern cpuop_func op_e0f8_0_nf;
extern cpuop_func op_e0f8_0_ff;
extern cpuop_func op_e0f9_0_nf;
extern cpuop_func op_e0f9_0_ff;
extern cpuop_func op_e100_0_nf;
extern cpuop_func op_e100_0_ff;
extern cpuop_func op_e108_0_nf;
extern cpuop_func op_e108_0_ff;
extern cpuop_func op_e110_0_nf;
extern cpuop_func op_e110_0_ff;
extern cpuop_func op_e118_0_nf;
extern cpuop_func op_e118_0_ff;
extern cpuop_func op_e120_0_nf;
extern cpuop_func op_e120_0_ff;
extern cpuop_func op_e128_0_nf;
extern cpuop_func op_e128_0_ff;
extern cpuop_func op_e130_0_nf;
extern cpuop_func op_e130_0_ff;
extern cpuop_func op_e138_0_nf;
extern cpuop_func op_e138_0_ff;
extern cpuop_func op_e140_0_nf;
extern cpuop_func op_e140_0_ff;
extern cpuop_func op_e148_0_nf;
extern cpuop_func op_e148_0_ff;
extern cpuop_func op_e150_0_nf;
extern cpuop_func op_e150_0_ff;
extern cpuop_func op_e158_0_nf;
extern cpuop_func op_e158_0_ff;
extern cpuop_func op_e160_0_nf;
extern cpuop_func op_e160_0_ff;
extern cpuop_func op_e168_0_nf;
extern cpuop_func op_e168_0_ff;
extern cpuop_func op_e170_0_nf;
extern cpuop_func op_e170_0_ff;
extern cpuop_func op_e178_0_nf;
extern cpuop_func op_e178_0_ff;
extern cpuop_func op_e180_0_nf;
extern cpuop_func op_e180_0_ff;
extern cpuop_func op_e188_0_nf;
extern cpuop_func op_e188_0_ff;
extern cpuop_func op_e190_0_nf;
extern cpuop_func op_e190_0_ff;
extern cpuop_func op_e198_0_nf;
extern cpuop_func op_e198_0_ff;
extern cpuop_func op_e1a0_0_nf;
extern cpuop_func op_e1a0_0_ff;
extern cpuop_func op_e1a8_0_nf;
extern cpuop_func op_e1a8_0_ff;
extern cpuop_func op_e1b0_0_nf;
extern cpuop_func op_e1b0_0_ff;
extern cpuop_func op_e1b8_0_nf;
extern cpuop_func op_e1b8_0_ff;
extern cpuop_func op_e1d0_0_nf;
extern cpuop_func op_e1d0_0_ff;
extern cpuop_func op_e1d8_0_nf;
extern cpuop_func op_e1d8_0_ff;
extern cpuop_func op_e1e0_0_nf;
extern cpuop_func op_e1e0_0_ff;
extern cpuop_func op_e1e8_0_nf;
extern cpuop_func op_e1e8_0_ff;
extern cpuop_func op_e1f0_0_nf;
extern cpuop_func op_e1f0_0_ff;
extern cpuop_func op_e1f8_0_nf;
extern cpuop_func op_e1f8_0_ff;
extern cpuop_func op_e1f9_0_nf;
extern cpuop_func op_e1f9_0_ff;
extern cpuop_func op_e2d0_0_nf;
extern cpuop_func op_e2d0_0_ff;
extern cpuop_func op_e2d8_0_nf;
extern cpuop_func op_e2d8_0_ff;
extern cpuop_func op_e2e0_0_nf;
extern cpuop_func op_e2e0_0_ff;
extern cpuop_func op_e2e8_0_nf;
extern cpuop_func op_e2e8_0_ff;
extern cpuop_func op_e2f0_0_nf;
extern cpuop_func op_e2f0_0_ff;
extern cpuop_func op_e2f8_0_nf;
extern cpuop_func op_e2f8_0_ff;
extern cpuop_func op_e2f9_0_nf;
extern cpuop_func op_e2f9_0_ff;
extern cpuop_func op_e3d0_0_nf;
extern cpuop_func op_e3d0_0_ff;
extern cpuop_func op_e3d8_0_nf;
extern cpuop_func op_e3d8_0_ff;
extern cpuop_func op_e3e0_0_nf;
extern cpuop_func op_e3e0_0_ff;
extern cpuop_func op_e3e8_0_nf;
extern cpuop_func op_e3e8_0_ff;
extern cpuop_func op_e3f0_0_nf;
extern cpuop_func op_e3f0_0_ff;
extern cpuop_func op_e3f8_0_nf;
extern cpuop_func op_e3f8_0_ff;
extern cpuop_func op_e3f9_0_nf;
extern cpuop_func op_e3f9_0_ff;
extern cpuop_func op_e4d0_0_nf;
extern cpuop_func op_e4d0_0_ff;
extern cpuop_func op_e4d8_0_nf;
extern cpuop_func op_e4d8_0_ff;
extern cpuop_func op_e4e0_0_nf;
extern cpuop_func op_e4e0_0_ff;
extern cpuop_func op_e4e8_0_nf;
extern cpuop_func op_e4e8_0_ff;
extern cpuop_func op_e4f0_0_nf;
extern cpuop_func op_e4f0_0_ff;
extern cpuop_func op_e4f8_0_nf;
extern cpuop_func op_e4f8_0_ff;
extern cpuop_func op_e4f9_0_nf;
extern cpuop_func op_e4f9_0_ff;
extern cpuop_func op_e5d0_0_nf;
extern cpuop_func op_e5d0_0_ff;
extern cpuop_func op_e5d8_0_nf;
extern cpuop_func op_e5d8_0_ff;
extern cpuop_func op_e5e0_0_nf;
extern cpuop_func op_e5e0_0_ff;
extern cpuop_func op_e5e8_0_nf;
extern cpuop_func op_e5e8_0_ff;
extern cpuop_func op_e5f0_0_nf;
extern cpuop_func op_e5f0_0_ff;
extern cpuop_func op_e5f8_0_nf;
extern cpuop_func op_e5f8_0_ff;
extern cpuop_func op_e5f9_0_nf;
extern cpuop_func op_e5f9_0_ff;
extern cpuop_func op_e6d0_0_nf;
extern cpuop_func op_e6d0_0_ff;
extern cpuop_func op_e6d8_0_nf;
extern cpuop_func op_e6d8_0_ff;
extern cpuop_func op_e6e0_0_nf;
extern cpuop_func op_e6e0_0_ff;
extern cpuop_func op_e6e8_0_nf;
extern cpuop_func op_e6e8_0_ff;
extern cpuop_func op_e6f0_0_nf;
extern cpuop_func op_e6f0_0_ff;
extern cpuop_func op_e6f8_0_nf;
extern cpuop_func op_e6f8_0_ff;
extern cpuop_func op_e6f9_0_nf;
extern cpuop_func op_e6f9_0_ff;
extern cpuop_func op_e7d0_0_nf;
extern cpuop_func op_e7d0_0_ff;
extern cpuop_func op_e7d8_0_nf;
extern cpuop_func op_e7d8_0_ff;
extern cpuop_func op_e7e0_0_nf;
extern cpuop_func op_e7e0_0_ff;
extern cpuop_func op_e7e8_0_nf;
extern cpuop_func op_e7e8_0_ff;
extern cpuop_func op_e7f0_0_nf;
extern cpuop_func op_e7f0_0_ff;
extern cpuop_func op_e7f8_0_nf;
extern cpuop_func op_e7f8_0_ff;
extern cpuop_func op_e7f9_0_nf;
extern cpuop_func op_e7f9_0_ff;
extern cpuop_func op_e8c0_0_nf;
extern cpuop_func op_e8c0_0_ff;
extern cpuop_func op_e8d0_0_nf;
extern cpuop_func op_e8d0_0_ff;
extern cpuop_func op_e8e8_0_nf;
extern cpuop_func op_e8e8_0_ff;
extern cpuop_func op_e8f0_0_nf;
extern cpuop_func op_e8f0_0_ff;
extern cpuop_func op_e8f8_0_nf;
extern cpuop_func op_e8f8_0_ff;
extern cpuop_func op_e8f9_0_nf;
extern cpuop_func op_e8f9_0_ff;
extern cpuop_func op_e8fa_0_nf;
extern cpuop_func op_e8fa_0_ff;
extern cpuop_func op_e8fb_0_nf;
extern cpuop_func op_e8fb_0_ff;
extern cpuop_func op_e9c0_0_nf;
extern cpuop_func op_e9c0_0_ff;
extern cpuop_func op_e9d0_0_nf;
extern cpuop_func op_e9d0_0_ff;
extern cpuop_func op_e9e8_0_nf;
extern cpuop_func op_e9e8_0_ff;
extern cpuop_func op_e9f0_0_nf;
extern cpuop_func op_e9f0_0_ff;
extern cpuop_func op_e9f8_0_nf;
extern cpuop_func op_e9f8_0_ff;
extern cpuop_func op_e9f9_0_nf;
extern cpuop_func op_e9f9_0_ff;
extern cpuop_func op_e9fa_0_nf;
extern cpuop_func op_e9fa_0_ff;
extern cpuop_func op_e9fb_0_nf;
extern cpuop_func op_e9fb_0_ff;
extern cpuop_func op_eac0_0_nf;
extern cpuop_func op_eac0_0_ff;
extern cpuop_func op_ead0_0_nf;
extern cpuop_func op_ead0_0_ff;
extern cpuop_func op_eae8_0_nf;
extern cpuop_func op_eae8_0_ff;
extern cpuop_func op_eaf0_0_nf;
extern cpuop_func op_eaf0_0_ff;
extern cpuop_func op_eaf8_0_nf;
extern cpuop_func op_eaf8_0_ff;
extern cpuop_func op_eaf9_0_nf;
extern cpuop_func op_eaf9_0_ff;
extern cpuop_func op_ebc0_0_nf;
extern cpuop_func op_ebc0_0_ff;
extern cpuop_func op_ebd0_0_nf;
extern cpuop_func op_ebd0_0_ff;
extern cpuop_func op_ebe8_0_nf;
extern cpuop_func op_ebe8_0_ff;
extern cpuop_func op_ebf0_0_nf;
extern cpuop_func op_ebf0_0_ff;
extern cpuop_func op_ebf8_0_nf;
extern cpuop_func op_ebf8_0_ff;
extern cpuop_func op_ebf9_0_nf;
extern cpuop_func op_ebf9_0_ff;
extern cpuop_func op_ebfa_0_nf;
extern cpuop_func op_ebfa_0_ff;
extern cpuop_func op_ebfb_0_nf;
extern cpuop_func op_ebfb_0_ff;
extern cpuop_func op_ecc0_0_nf;
extern cpuop_func op_ecc0_0_ff;
extern cpuop_func op_ecd0_0_nf;
extern cpuop_func op_ecd0_0_ff;
extern cpuop_func op_ece8_0_nf;
extern cpuop_func op_ece8_0_ff;
extern cpuop_func op_ecf0_0_nf;
extern cpuop_func op_ecf0_0_ff;
extern cpuop_func op_ecf8_0_nf;
extern cpuop_func op_ecf8_0_ff;
extern cpuop_func op_ecf9_0_nf;
extern cpuop_func op_ecf9_0_ff;
extern cpuop_func op_edc0_0_nf;
extern cpuop_func op_edc0_0_ff;
extern cpuop_func op_edd0_0_nf;
extern cpuop_func op_edd0_0_ff;
extern cpuop_func op_ede8_0_nf;
extern cpuop_func op_ede8_0_ff;
extern cpuop_func op_edf0_0_nf;
extern cpuop_func op_edf0_0_ff;
extern cpuop_func op_edf8_0_nf;
extern cpuop_func op_edf8_0_ff;
extern cpuop_func op_edf9_0_nf;
extern cpuop_func op_edf9_0_ff;
extern cpuop_func op_edfa_0_nf;
extern cpuop_func op_edfa_0_ff;
extern cpuop_func op_edfb_0_nf;
extern cpuop_func op_edfb_0_ff;
extern cpuop_func op_eec0_0_nf;
extern cpuop_func op_eec0_0_ff;
extern cpuop_func op_eed0_0_nf;
extern cpuop_func op_eed0_0_ff;
extern cpuop_func op_eee8_0_nf;
extern cpuop_func op_eee8_0_ff;
extern cpuop_func op_eef0_0_nf;
extern cpuop_func op_eef0_0_ff;
extern cpuop_func op_eef8_0_nf;
extern cpuop_func op_eef8_0_ff;
extern cpuop_func op_eef9_0_nf;
extern cpuop_func op_eef9_0_ff;
extern cpuop_func op_efc0_0_nf;
extern cpuop_func op_efc0_0_ff;
extern cpuop_func op_efd0_0_nf;
extern cpuop_func op_efd0_0_ff;
extern cpuop_func op_efe8_0_nf;
extern cpuop_func op_efe8_0_ff;
extern cpuop_func op_eff0_0_nf;
extern cpuop_func op_eff0_0_ff;
extern cpuop_func op_eff8_0_nf;
extern cpuop_func op_eff8_0_ff;
extern cpuop_func op_eff9_0_nf;
extern cpuop_func op_eff9_0_ff;
extern cpuop_func op_f000_0_nf;
extern cpuop_func op_f000_0_ff;
extern cpuop_func op_f008_0_nf;
extern cpuop_func op_f008_0_ff;
extern cpuop_func op_f010_0_nf;
extern cpuop_func op_f010_0_ff;
extern cpuop_func op_f018_0_nf;
extern cpuop_func op_f018_0_ff;
extern cpuop_func op_f020_0_nf;
extern cpuop_func op_f020_0_ff;
extern cpuop_func op_f028_0_nf;
extern cpuop_func op_f028_0_ff;
extern cpuop_func op_f030_0_nf;
extern cpuop_func op_f030_0_ff;
extern cpuop_func op_f038_0_nf;
extern cpuop_func op_f038_0_ff;
extern cpuop_func op_f039_0_nf;
extern cpuop_func op_f039_0_ff;
extern cpuop_func op_f200_0_nf;
extern cpuop_func op_f200_0_ff;
extern cpuop_func op_f208_0_nf;
extern cpuop_func op_f208_0_ff;
extern cpuop_func op_f210_0_nf;
extern cpuop_func op_f210_0_ff;
extern cpuop_func op_f218_0_nf;
extern cpuop_func op_f218_0_ff;
extern cpuop_func op_f220_0_nf;
extern cpuop_func op_f220_0_ff;
extern cpuop_func op_f228_0_nf;
extern cpuop_func op_f228_0_ff;
extern cpuop_func op_f230_0_nf;
extern cpuop_func op_f230_0_ff;
extern cpuop_func op_f238_0_nf;
extern cpuop_func op_f238_0_ff;
extern cpuop_func op_f239_0_nf;
extern cpuop_func op_f239_0_ff;
extern cpuop_func op_f23a_0_nf;
extern cpuop_func op_f23a_0_ff;
extern cpuop_func op_f23b_0_nf;
extern cpuop_func op_f23b_0_ff;
extern cpuop_func op_f23c_0_nf;
extern cpuop_func op_f23c_0_ff;
extern cpuop_func op_f240_0_nf;
extern cpuop_func op_f240_0_ff;
extern cpuop_func op_f248_0_nf;
extern cpuop_func op_f248_0_ff;
extern cpuop_func op_f250_0_nf;
extern cpuop_func op_f250_0_ff;
extern cpuop_func op_f258_0_nf;
extern cpuop_func op_f258_0_ff;
extern cpuop_func op_f260_0_nf;
extern cpuop_func op_f260_0_ff;
extern cpuop_func op_f268_0_nf;
extern cpuop_func op_f268_0_ff;
extern cpuop_func op_f270_0_nf;
extern cpuop_func op_f270_0_ff;
extern cpuop_func op_f278_0_nf;
extern cpuop_func op_f278_0_ff;
extern cpuop_func op_f279_0_nf;
extern cpuop_func op_f279_0_ff;
extern cpuop_func op_f27a_0_nf;
extern cpuop_func op_f27a_0_ff;
extern cpuop_func op_f27b_0_nf;
extern cpuop_func op_f27b_0_ff;
extern cpuop_func op_f27c_0_nf;
extern cpuop_func op_f27c_0_ff;
extern cpuop_func op_f280_0_nf;
extern cpuop_func op_f280_0_ff;
extern cpuop_func op_f2c0_0_nf;
extern cpuop_func op_f2c0_0_ff;
extern cpuop_func op_f310_0_nf;
extern cpuop_func op_f310_0_ff;
extern cpuop_func op_f320_0_nf;
extern cpuop_func op_f320_0_ff;
extern cpuop_func op_f328_0_nf;
extern cpuop_func op_f328_0_ff;
extern cpuop_func op_f330_0_nf;
extern cpuop_func op_f330_0_ff;
extern cpuop_func op_f338_0_nf;
extern cpuop_func op_f338_0_ff;
extern cpuop_func op_f339_0_nf;
extern cpuop_func op_f339_0_ff;
extern cpuop_func op_f350_0_nf;
extern cpuop_func op_f350_0_ff;
extern cpuop_func op_f358_0_nf;
extern cpuop_func op_f358_0_ff;
extern cpuop_func op_f368_0_nf;
extern cpuop_func op_f368_0_ff;
extern cpuop_func op_f370_0_nf;
extern cpuop_func op_f370_0_ff;
extern cpuop_func op_f378_0_nf;
extern cpuop_func op_f378_0_ff;
extern cpuop_func op_f379_0_nf;
extern cpuop_func op_f379_0_ff;
extern cpuop_func op_f37a_0_nf;
extern cpuop_func op_f37a_0_ff;
extern cpuop_func op_f37b_0_nf;
extern cpuop_func op_f37b_0_ff;
extern cpuop_func op_f408_0_nf;
extern cpuop_func op_f408_0_ff;
extern cpuop_func op_f410_0_nf;
extern cpuop_func op_f410_0_ff;
extern cpuop_func op_f418_0_nf;
extern cpuop_func op_f418_0_ff;
extern cpuop_func op_f419_0_nf;
extern cpuop_func op_f419_0_ff;
extern cpuop_func op_f41a_0_nf;
extern cpuop_func op_f41a_0_ff;
extern cpuop_func op_f41b_0_nf;
extern cpuop_func op_f41b_0_ff;
extern cpuop_func op_f41c_0_nf;
extern cpuop_func op_f41c_0_ff;
extern cpuop_func op_f41d_0_nf;
extern cpuop_func op_f41d_0_ff;
extern cpuop_func op_f41e_0_nf;
extern cpuop_func op_f41e_0_ff;
extern cpuop_func op_f41f_0_nf;
extern cpuop_func op_f41f_0_ff;
extern cpuop_func op_f428_0_nf;
extern cpuop_func op_f428_0_ff;
extern cpuop_func op_f430_0_nf;
extern cpuop_func op_f430_0_ff;
extern cpuop_func op_f438_0_nf;
extern cpuop_func op_f438_0_ff;
extern cpuop_func op_f439_0_nf;
extern cpuop_func op_f439_0_ff;
extern cpuop_func op_f43a_0_nf;
extern cpuop_func op_f43a_0_ff;
extern cpuop_func op_f43b_0_nf;
extern cpuop_func op_f43b_0_ff;
extern cpuop_func op_f43c_0_nf;
extern cpuop_func op_f43c_0_ff;
extern cpuop_func op_f43d_0_nf;
extern cpuop_func op_f43d_0_ff;
extern cpuop_func op_f43e_0_nf;
extern cpuop_func op_f43e_0_ff;
extern cpuop_func op_f43f_0_nf;
extern cpuop_func op_f43f_0_ff;
extern cpuop_func op_f500_0_nf;
extern cpuop_func op_f500_0_ff;
extern cpuop_func op_f508_0_nf;
extern cpuop_func op_f508_0_ff;
extern cpuop_func op_f510_0_nf;
extern cpuop_func op_f510_0_ff;
extern cpuop_func op_f518_0_nf;
extern cpuop_func op_f518_0_ff;
extern cpuop_func op_f548_0_nf;
extern cpuop_func op_f548_0_ff;
extern cpuop_func op_f568_0_nf;
extern cpuop_func op_f568_0_ff;
extern cpuop_func op_f588_0_nf;
extern cpuop_func op_f588_0_ff;
extern cpuop_func op_f5c8_0_nf;
extern cpuop_func op_f5c8_0_ff;
extern cpuop_func op_f600_0_nf;
extern cpuop_func op_f600_0_ff;
extern cpuop_func op_f608_0_nf;
extern cpuop_func op_f608_0_ff;
extern cpuop_func op_f610_0_nf;
extern cpuop_func op_f610_0_ff;
extern cpuop_func op_f618_0_nf;
extern cpuop_func op_f618_0_ff;
extern cpuop_func op_f620_0_nf;
extern cpuop_func op_f620_0_ff;
extern cpuop_func op_f800_0_nf;
extern cpuop_func op_f800_0_ff;
extern cpuop_func op_4800_2_nf;
extern cpuop_func op_4800_2_ff;
extern cpuop_func op_4810_2_nf;
extern cpuop_func op_4810_2_ff;
extern cpuop_func op_4818_2_nf;
extern cpuop_func op_4818_2_ff;
extern cpuop_func op_4820_2_nf;
extern cpuop_func op_4820_2_ff;
extern cpuop_func op_4828_2_nf;
extern cpuop_func op_4828_2_ff;
extern cpuop_func op_4830_2_nf;
extern cpuop_func op_4830_2_ff;
extern cpuop_func op_4838_2_nf;
extern cpuop_func op_4838_2_ff;
extern cpuop_func op_4839_2_nf;
extern cpuop_func op_4839_2_ff;
extern cpuop_func op_8100_2_nf;
extern cpuop_func op_8100_2_ff;
extern cpuop_func op_8108_2_nf;
extern cpuop_func op_8108_2_ff;
extern cpuop_func op_c100_2_nf;
extern cpuop_func op_c100_2_ff;
extern cpuop_func op_c108_2_nf;
extern cpuop_func op_c108_2_ff;
extern cpuop_func op_0000_4_nf;
extern cpuop_func op_0000_4_ff;
extern cpuop_func op_0010_4_nf;
extern cpuop_func op_0010_4_ff;
extern cpuop_func op_0018_4_nf;
extern cpuop_func op_0018_4_ff;
extern cpuop_func op_0020_4_nf;
extern cpuop_func op_0020_4_ff;
extern cpuop_func op_0028_4_nf;
extern cpuop_func op_0028_4_ff;
extern cpuop_func op_0030_4_nf;
extern cpuop_func op_0030_4_ff;
extern cpuop_func op_0038_4_nf;
extern cpuop_func op_0038_4_ff;
extern cpuop_func op_0039_4_nf;
extern cpuop_func op_0039_4_ff;
extern cpuop_func op_003c_4_nf;
extern cpuop_func op_003c_4_ff;
extern cpuop_func op_0040_4_nf;
extern cpuop_func op_0040_4_ff;
extern cpuop_func op_0050_4_nf;
extern cpuop_func op_0050_4_ff;
extern cpuop_func op_0058_4_nf;
extern cpuop_func op_0058_4_ff;
extern cpuop_func op_0060_4_nf;
extern cpuop_func op_0060_4_ff;
extern cpuop_func op_0068_4_nf;
extern cpuop_func op_0068_4_ff;
extern cpuop_func op_0070_4_nf;
extern cpuop_func op_0070_4_ff;
extern cpuop_func op_0078_4_nf;
extern cpuop_func op_0078_4_ff;
extern cpuop_func op_0079_4_nf;
extern cpuop_func op_0079_4_ff;
extern cpuop_func op_007c_4_nf;
extern cpuop_func op_007c_4_ff;
extern cpuop_func op_0080_4_nf;
extern cpuop_func op_0080_4_ff;
extern cpuop_func op_0090_4_nf;
extern cpuop_func op_0090_4_ff;
extern cpuop_func op_0098_4_nf;
extern cpuop_func op_0098_4_ff;
extern cpuop_func op_00a0_4_nf;
extern cpuop_func op_00a0_4_ff;
extern cpuop_func op_00a8_4_nf;
extern cpuop_func op_00a8_4_ff;
extern cpuop_func op_00b0_4_nf;
extern cpuop_func op_00b0_4_ff;
extern cpuop_func op_00b8_4_nf;
extern cpuop_func op_00b8_4_ff;
extern cpuop_func op_00b9_4_nf;
extern cpuop_func op_00b9_4_ff;
extern cpuop_func op_0100_4_nf;
extern cpuop_func op_0100_4_ff;
extern cpuop_func op_0108_4_nf;
extern cpuop_func op_0108_4_ff;
extern cpuop_func op_0110_4_nf;
extern cpuop_func op_0110_4_ff;
extern cpuop_func op_0118_4_nf;
extern cpuop_func op_0118_4_ff;
extern cpuop_func op_0120_4_nf;
extern cpuop_func op_0120_4_ff;
extern cpuop_func op_0128_4_nf;
extern cpuop_func op_0128_4_ff;
extern cpuop_func op_0130_4_nf;
extern cpuop_func op_0130_4_ff;
extern cpuop_func op_0138_4_nf;
extern cpuop_func op_0138_4_ff;
extern cpuop_func op_0139_4_nf;
extern cpuop_func op_0139_4_ff;
extern cpuop_func op_013a_4_nf;
extern cpuop_func op_013a_4_ff;
extern cpuop_func op_013b_4_nf;
extern cpuop_func op_013b_4_ff;
extern cpuop_func op_013c_4_nf;
extern cpuop_func op_013c_4_ff;
extern cpuop_func op_0140_4_nf;
extern cpuop_func op_0140_4_ff;
extern cpuop_func op_0148_4_nf;
extern cpuop_func op_0148_4_ff;
extern cpuop_func op_0150_4_nf;
extern cpuop_func op_0150_4_ff;
extern cpuop_func op_0158_4_nf;
extern cpuop_func op_0158_4_ff;
extern cpuop_func op_0160_4_nf;
extern cpuop_func op_0160_4_ff;
extern cpuop_func op_0168_4_nf;
extern cpuop_func op_0168_4_ff;
extern cpuop_func op_0170_4_nf;
extern cpuop_func op_0170_4_ff;
extern cpuop_func op_0178_4_nf;
extern cpuop_func op_0178_4_ff;
extern cpuop_func op_0179_4_nf;
extern cpuop_func op_0179_4_ff;
extern cpuop_func op_0180_4_nf;
extern cpuop_func op_0180_4_ff;
extern cpuop_func op_0188_4_nf;
extern cpuop_func op_0188_4_ff;
extern cpuop_func op_0190_4_nf;
extern cpuop_func op_0190_4_ff;
extern cpuop_func op_0198_4_nf;
extern cpuop_func op_0198_4_ff;
extern cpuop_func op_01a0_4_nf;
extern cpuop_func op_01a0_4_ff;
extern cpuop_func op_01a8_4_nf;
extern cpuop_func op_01a8_4_ff;
extern cpuop_func op_01b0_4_nf;
extern cpuop_func op_01b0_4_ff;
extern cpuop_func op_01b8_4_nf;
extern cpuop_func op_01b8_4_ff;
extern cpuop_func op_01b9_4_nf;
extern cpuop_func op_01b9_4_ff;
extern cpuop_func op_01c0_4_nf;
extern cpuop_func op_01c0_4_ff;
extern cpuop_func op_01c8_4_nf;
extern cpuop_func op_01c8_4_ff;
extern cpuop_func op_01d0_4_nf;
extern cpuop_func op_01d0_4_ff;
extern cpuop_func op_01d8_4_nf;
extern cpuop_func op_01d8_4_ff;
extern cpuop_func op_01e0_4_nf;
extern cpuop_func op_01e0_4_ff;
extern cpuop_func op_01e8_4_nf;
extern cpuop_func op_01e8_4_ff;
extern cpuop_func op_01f0_4_nf;
extern cpuop_func op_01f0_4_ff;
extern cpuop_func op_01f8_4_nf;
extern cpuop_func op_01f8_4_ff;
extern cpuop_func op_01f9_4_nf;
extern cpuop_func op_01f9_4_ff;
extern cpuop_func op_0200_4_nf;
extern cpuop_func op_0200_4_ff;
extern cpuop_func op_0210_4_nf;
extern cpuop_func op_0210_4_ff;
extern cpuop_func op_0218_4_nf;
extern cpuop_func op_0218_4_ff;
extern cpuop_func op_0220_4_nf;
extern cpuop_func op_0220_4_ff;
extern cpuop_func op_0228_4_nf;
extern cpuop_func op_0228_4_ff;
extern cpuop_func op_0230_4_nf;
extern cpuop_func op_0230_4_ff;
extern cpuop_func op_0238_4_nf;
extern cpuop_func op_0238_4_ff;
extern cpuop_func op_0239_4_nf;
extern cpuop_func op_0239_4_ff;
extern cpuop_func op_023c_4_nf;
extern cpuop_func op_023c_4_ff;
extern cpuop_func op_0240_4_nf;
extern cpuop_func op_0240_4_ff;
extern cpuop_func op_0250_4_nf;
extern cpuop_func op_0250_4_ff;
extern cpuop_func op_0258_4_nf;
extern cpuop_func op_0258_4_ff;
extern cpuop_func op_0260_4_nf;
extern cpuop_func op_0260_4_ff;
extern cpuop_func op_0268_4_nf;
extern cpuop_func op_0268_4_ff;
extern cpuop_func op_0270_4_nf;
extern cpuop_func op_0270_4_ff;
extern cpuop_func op_0278_4_nf;
extern cpuop_func op_0278_4_ff;
extern cpuop_func op_0279_4_nf;
extern cpuop_func op_0279_4_ff;
extern cpuop_func op_027c_4_nf;
extern cpuop_func op_027c_4_ff;
extern cpuop_func op_0280_4_nf;
extern cpuop_func op_0280_4_ff;
extern cpuop_func op_0290_4_nf;
extern cpuop_func op_0290_4_ff;
extern cpuop_func op_0298_4_nf;
extern cpuop_func op_0298_4_ff;
extern cpuop_func op_02a0_4_nf;
extern cpuop_func op_02a0_4_ff;
extern cpuop_func op_02a8_4_nf;
extern cpuop_func op_02a8_4_ff;
extern cpuop_func op_02b0_4_nf;
extern cpuop_func op_02b0_4_ff;
extern cpuop_func op_02b8_4_nf;
extern cpuop_func op_02b8_4_ff;
extern cpuop_func op_02b9_4_nf;
extern cpuop_func op_02b9_4_ff;
extern cpuop_func op_0400_4_nf;
extern cpuop_func op_0400_4_ff;
extern cpuop_func op_0410_4_nf;
extern cpuop_func op_0410_4_ff;
extern cpuop_func op_0418_4_nf;
extern cpuop_func op_0418_4_ff;
extern cpuop_func op_0420_4_nf;
extern cpuop_func op_0420_4_ff;
extern cpuop_func op_0428_4_nf;
extern cpuop_func op_0428_4_ff;
extern cpuop_func op_0430_4_nf;
extern cpuop_func op_0430_4_ff;
extern cpuop_func op_0438_4_nf;
extern cpuop_func op_0438_4_ff;
extern cpuop_func op_0439_4_nf;
extern cpuop_func op_0439_4_ff;
extern cpuop_func op_0440_4_nf;
extern cpuop_func op_0440_4_ff;
extern cpuop_func op_0450_4_nf;
extern cpuop_func op_0450_4_ff;
extern cpuop_func op_0458_4_nf;
extern cpuop_func op_0458_4_ff;
extern cpuop_func op_0460_4_nf;
extern cpuop_func op_0460_4_ff;
extern cpuop_func op_0468_4_nf;
extern cpuop_func op_0468_4_ff;
extern cpuop_func op_0470_4_nf;
extern cpuop_func op_0470_4_ff;
extern cpuop_func op_0478_4_nf;
extern cpuop_func op_0478_4_ff;
extern cpuop_func op_0479_4_nf;
extern cpuop_func op_0479_4_ff;
extern cpuop_func op_0480_4_nf;
extern cpuop_func op_0480_4_ff;
extern cpuop_func op_0490_4_nf;
extern cpuop_func op_0490_4_ff;
extern cpuop_func op_0498_4_nf;
extern cpuop_func op_0498_4_ff;
extern cpuop_func op_04a0_4_nf;
extern cpuop_func op_04a0_4_ff;
extern cpuop_func op_04a8_4_nf;
extern cpuop_func op_04a8_4_ff;
extern cpuop_func op_04b0_4_nf;
extern cpuop_func op_04b0_4_ff;
extern cpuop_func op_04b8_4_nf;
extern cpuop_func op_04b8_4_ff;
extern cpuop_func op_04b9_4_nf;
extern cpuop_func op_04b9_4_ff;
extern cpuop_func op_0600_4_nf;
extern cpuop_func op_0600_4_ff;
extern cpuop_func op_0610_4_nf;
extern cpuop_func op_0610_4_ff;
extern cpuop_func op_0618_4_nf;
extern cpuop_func op_0618_4_ff;
extern cpuop_func op_0620_4_nf;
extern cpuop_func op_0620_4_ff;
extern cpuop_func op_0628_4_nf;
extern cpuop_func op_0628_4_ff;
extern cpuop_func op_0630_4_nf;
extern cpuop_func op_0630_4_ff;
extern cpuop_func op_0638_4_nf;
extern cpuop_func op_0638_4_ff;
extern cpuop_func op_0639_4_nf;
extern cpuop_func op_0639_4_ff;
extern cpuop_func op_0640_4_nf;
extern cpuop_func op_0640_4_ff;
extern cpuop_func op_0650_4_nf;
extern cpuop_func op_0650_4_ff;
extern cpuop_func op_0658_4_nf;
extern cpuop_func op_0658_4_ff;
extern cpuop_func op_0660_4_nf;
extern cpuop_func op_0660_4_ff;
extern cpuop_func op_0668_4_nf;
extern cpuop_func op_0668_4_ff;
extern cpuop_func op_0670_4_nf;
extern cpuop_func op_0670_4_ff;
extern cpuop_func op_0678_4_nf;
extern cpuop_func op_0678_4_ff;
extern cpuop_func op_0679_4_nf;
extern cpuop_func op_0679_4_ff;
extern cpuop_func op_0680_4_nf;
extern cpuop_func op_0680_4_ff;
extern cpuop_func op_0690_4_nf;
extern cpuop_func op_0690_4_ff;
extern cpuop_func op_0698_4_nf;
extern cpuop_func op_0698_4_ff;
extern cpuop_func op_06a0_4_nf;
extern cpuop_func op_06a0_4_ff;
extern cpuop_func op_06a8_4_nf;
extern cpuop_func op_06a8_4_ff;
extern cpuop_func op_06b0_4_nf;
extern cpuop_func op_06b0_4_ff;
extern cpuop_func op_06b8_4_nf;
extern cpuop_func op_06b8_4_ff;
extern cpuop_func op_06b9_4_nf;
extern cpuop_func op_06b9_4_ff;
extern cpuop_func op_0800_4_nf;
extern cpuop_func op_0800_4_ff;
extern cpuop_func op_0810_4_nf;
extern cpuop_func op_0810_4_ff;
extern cpuop_func op_0818_4_nf;
extern cpuop_func op_0818_4_ff;
extern cpuop_func op_0820_4_nf;
extern cpuop_func op_0820_4_ff;
extern cpuop_func op_0828_4_nf;
extern cpuop_func op_0828_4_ff;
extern cpuop_func op_0830_4_nf;
extern cpuop_func op_0830_4_ff;
extern cpuop_func op_0838_4_nf;
extern cpuop_func op_0838_4_ff;
extern cpuop_func op_0839_4_nf;
extern cpuop_func op_0839_4_ff;
extern cpuop_func op_083a_4_nf;
extern cpuop_func op_083a_4_ff;
extern cpuop_func op_083b_4_nf;
extern cpuop_func op_083b_4_ff;
extern cpuop_func op_0840_4_nf;
extern cpuop_func op_0840_4_ff;
extern cpuop_func op_0850_4_nf;
extern cpuop_func op_0850_4_ff;
extern cpuop_func op_0858_4_nf;
extern cpuop_func op_0858_4_ff;
extern cpuop_func op_0860_4_nf;
extern cpuop_func op_0860_4_ff;
extern cpuop_func op_0868_4_nf;
extern cpuop_func op_0868_4_ff;
extern cpuop_func op_0870_4_nf;
extern cpuop_func op_0870_4_ff;
extern cpuop_func op_0878_4_nf;
extern cpuop_func op_0878_4_ff;
extern cpuop_func op_0879_4_nf;
extern cpuop_func op_0879_4_ff;
extern cpuop_func op_0880_4_nf;
extern cpuop_func op_0880_4_ff;
extern cpuop_func op_0890_4_nf;
extern cpuop_func op_0890_4_ff;
extern cpuop_func op_0898_4_nf;
extern cpuop_func op_0898_4_ff;
extern cpuop_func op_08a0_4_nf;
extern cpuop_func op_08a0_4_ff;
extern cpuop_func op_08a8_4_nf;
extern cpuop_func op_08a8_4_ff;
extern cpuop_func op_08b0_4_nf;
extern cpuop_func op_08b0_4_ff;
extern cpuop_func op_08b8_4_nf;
extern cpuop_func op_08b8_4_ff;
extern cpuop_func op_08b9_4_nf;
extern cpuop_func op_08b9_4_ff;
extern cpuop_func op_08c0_4_nf;
extern cpuop_func op_08c0_4_ff;
extern cpuop_func op_08d0_4_nf;
extern cpuop_func op_08d0_4_ff;
extern cpuop_func op_08d8_4_nf;
extern cpuop_func op_08d8_4_ff;
extern cpuop_func op_08e0_4_nf;
extern cpuop_func op_08e0_4_ff;
extern cpuop_func op_08e8_4_nf;
extern cpuop_func op_08e8_4_ff;
extern cpuop_func op_08f0_4_nf;
extern cpuop_func op_08f0_4_ff;
extern cpuop_func op_08f8_4_nf;
extern cpuop_func op_08f8_4_ff;
extern cpuop_func op_08f9_4_nf;
extern cpuop_func op_08f9_4_ff;
extern cpuop_func op_0a00_4_nf;
extern cpuop_func op_0a00_4_ff;
extern cpuop_func op_0a10_4_nf;
extern cpuop_func op_0a10_4_ff;
extern cpuop_func op_0a18_4_nf;
extern cpuop_func op_0a18_4_ff;
extern cpuop_func op_0a20_4_nf;
extern cpuop_func op_0a20_4_ff;
extern cpuop_func op_0a28_4_nf;
extern cpuop_func op_0a28_4_ff;
extern cpuop_func op_0a30_4_nf;
extern cpuop_func op_0a30_4_ff;
extern cpuop_func op_0a38_4_nf;
extern cpuop_func op_0a38_4_ff;
extern cpuop_func op_0a39_4_nf;
extern cpuop_func op_0a39_4_ff;
extern cpuop_func op_0a3c_4_nf;
extern cpuop_func op_0a3c_4_ff;
extern cpuop_func op_0a40_4_nf;
extern cpuop_func op_0a40_4_ff;
extern cpuop_func op_0a50_4_nf;
extern cpuop_func op_0a50_4_ff;
extern cpuop_func op_0a58_4_nf;
extern cpuop_func op_0a58_4_ff;
extern cpuop_func op_0a60_4_nf;
extern cpuop_func op_0a60_4_ff;
extern cpuop_func op_0a68_4_nf;
extern cpuop_func op_0a68_4_ff;
extern cpuop_func op_0a70_4_nf;
extern cpuop_func op_0a70_4_ff;
extern cpuop_func op_0a78_4_nf;
extern cpuop_func op_0a78_4_ff;
extern cpuop_func op_0a79_4_nf;
extern cpuop_func op_0a79_4_ff;
extern cpuop_func op_0a7c_4_nf;
extern cpuop_func op_0a7c_4_ff;
extern cpuop_func op_0a80_4_nf;
extern cpuop_func op_0a80_4_ff;
extern cpuop_func op_0a90_4_nf;
extern cpuop_func op_0a90_4_ff;
extern cpuop_func op_0a98_4_nf;
extern cpuop_func op_0a98_4_ff;
extern cpuop_func op_0aa0_4_nf;
extern cpuop_func op_0aa0_4_ff;
extern cpuop_func op_0aa8_4_nf;
extern cpuop_func op_0aa8_4_ff;
extern cpuop_func op_0ab0_4_nf;
extern cpuop_func op_0ab0_4_ff;
extern cpuop_func op_0ab8_4_nf;
extern cpuop_func op_0ab8_4_ff;
extern cpuop_func op_0ab9_4_nf;
extern cpuop_func op_0ab9_4_ff;
extern cpuop_func op_0c00_4_nf;
extern cpuop_func op_0c00_4_ff;
extern cpuop_func op_0c10_4_nf;
extern cpuop_func op_0c10_4_ff;
extern cpuop_func op_0c18_4_nf;
extern cpuop_func op_0c18_4_ff;
extern cpuop_func op_0c20_4_nf;
extern cpuop_func op_0c20_4_ff;
extern cpuop_func op_0c28_4_nf;
extern cpuop_func op_0c28_4_ff;
extern cpuop_func op_0c30_4_nf;
extern cpuop_func op_0c30_4_ff;
extern cpuop_func op_0c38_4_nf;
extern cpuop_func op_0c38_4_ff;
extern cpuop_func op_0c39_4_nf;
extern cpuop_func op_0c39_4_ff;
extern cpuop_func op_0c40_4_nf;
extern cpuop_func op_0c40_4_ff;
extern cpuop_func op_0c50_4_nf;
extern cpuop_func op_0c50_4_ff;
extern cpuop_func op_0c58_4_nf;
extern cpuop_func op_0c58_4_ff;
extern cpuop_func op_0c60_4_nf;
extern cpuop_func op_0c60_4_ff;
extern cpuop_func op_0c68_4_nf;
extern cpuop_func op_0c68_4_ff;
extern cpuop_func op_0c70_4_nf;
extern cpuop_func op_0c70_4_ff;
extern cpuop_func op_0c78_4_nf;
extern cpuop_func op_0c78_4_ff;
extern cpuop_func op_0c79_4_nf;
extern cpuop_func op_0c79_4_ff;
extern cpuop_func op_0c80_4_nf;
extern cpuop_func op_0c80_4_ff;
extern cpuop_func op_0c90_4_nf;
extern cpuop_func op_0c90_4_ff;
extern cpuop_func op_0c98_4_nf;
extern cpuop_func op_0c98_4_ff;
extern cpuop_func op_0ca0_4_nf;
extern cpuop_func op_0ca0_4_ff;
extern cpuop_func op_0ca8_4_nf;
extern cpuop_func op_0ca8_4_ff;
extern cpuop_func op_0cb0_4_nf;
extern cpuop_func op_0cb0_4_ff;
extern cpuop_func op_0cb8_4_nf;
extern cpuop_func op_0cb8_4_ff;
extern cpuop_func op_0cb9_4_nf;
extern cpuop_func op_0cb9_4_ff;
extern cpuop_func op_1000_4_nf;
extern cpuop_func op_1000_4_ff;
extern cpuop_func op_1010_4_nf;
extern cpuop_func op_1010_4_ff;
extern cpuop_func op_1018_4_nf;
extern cpuop_func op_1018_4_ff;
extern cpuop_func op_1020_4_nf;
extern cpuop_func op_1020_4_ff;
extern cpuop_func op_1028_4_nf;
extern cpuop_func op_1028_4_ff;
extern cpuop_func op_1030_4_nf;
extern cpuop_func op_1030_4_ff;
extern cpuop_func op_1038_4_nf;
extern cpuop_func op_1038_4_ff;
extern cpuop_func op_1039_4_nf;
extern cpuop_func op_1039_4_ff;
extern cpuop_func op_103a_4_nf;
extern cpuop_func op_103a_4_ff;
extern cpuop_func op_103b_4_nf;
extern cpuop_func op_103b_4_ff;
extern cpuop_func op_103c_4_nf;
extern cpuop_func op_103c_4_ff;
extern cpuop_func op_1080_4_nf;
extern cpuop_func op_1080_4_ff;
extern cpuop_func op_1090_4_nf;
extern cpuop_func op_1090_4_ff;
extern cpuop_func op_1098_4_nf;
extern cpuop_func op_1098_4_ff;
extern cpuop_func op_10a0_4_nf;
extern cpuop_func op_10a0_4_ff;
extern cpuop_func op_10a8_4_nf;
extern cpuop_func op_10a8_4_ff;
extern cpuop_func op_10b0_4_nf;
extern cpuop_func op_10b0_4_ff;
extern cpuop_func op_10b8_4_nf;
extern cpuop_func op_10b8_4_ff;
extern cpuop_func op_10b9_4_nf;
extern cpuop_func op_10b9_4_ff;
extern cpuop_func op_10ba_4_nf;
extern cpuop_func op_10ba_4_ff;
extern cpuop_func op_10bb_4_nf;
extern cpuop_func op_10bb_4_ff;
extern cpuop_func op_10bc_4_nf;
extern cpuop_func op_10bc_4_ff;
extern cpuop_func op_10c0_4_nf;
extern cpuop_func op_10c0_4_ff;
extern cpuop_func op_10d0_4_nf;
extern cpuop_func op_10d0_4_ff;
extern cpuop_func op_10d8_4_nf;
extern cpuop_func op_10d8_4_ff;
extern cpuop_func op_10e0_4_nf;
extern cpuop_func op_10e0_4_ff;
extern cpuop_func op_10e8_4_nf;
extern cpuop_func op_10e8_4_ff;
extern cpuop_func op_10f0_4_nf;
extern cpuop_func op_10f0_4_ff;
extern cpuop_func op_10f8_4_nf;
extern cpuop_func op_10f8_4_ff;
extern cpuop_func op_10f9_4_nf;
extern cpuop_func op_10f9_4_ff;
extern cpuop_func op_10fa_4_nf;
extern cpuop_func op_10fa_4_ff;
extern cpuop_func op_10fb_4_nf;
extern cpuop_func op_10fb_4_ff;
extern cpuop_func op_10fc_4_nf;
extern cpuop_func op_10fc_4_ff;
extern cpuop_func op_1100_4_nf;
extern cpuop_func op_1100_4_ff;
extern cpuop_func op_1110_4_nf;
extern cpuop_func op_1110_4_ff;
extern cpuop_func op_1118_4_nf;
extern cpuop_func op_1118_4_ff;
extern cpuop_func op_1120_4_nf;
extern cpuop_func op_1120_4_ff;
extern cpuop_func op_1128_4_nf;
extern cpuop_func op_1128_4_ff;
extern cpuop_func op_1130_4_nf;
extern cpuop_func op_1130_4_ff;
extern cpuop_func op_1138_4_nf;
extern cpuop_func op_1138_4_ff;
extern cpuop_func op_1139_4_nf;
extern cpuop_func op_1139_4_ff;
extern cpuop_func op_113a_4_nf;
extern cpuop_func op_113a_4_ff;
extern cpuop_func op_113b_4_nf;
extern cpuop_func op_113b_4_ff;
extern cpuop_func op_113c_4_nf;
extern cpuop_func op_113c_4_ff;
extern cpuop_func op_1140_4_nf;
extern cpuop_func op_1140_4_ff;
extern cpuop_func op_1150_4_nf;
extern cpuop_func op_1150_4_ff;
extern cpuop_func op_1158_4_nf;
extern cpuop_func op_1158_4_ff;
extern cpuop_func op_1160_4_nf;
extern cpuop_func op_1160_4_ff;
extern cpuop_func op_1168_4_nf;
extern cpuop_func op_1168_4_ff;
extern cpuop_func op_1170_4_nf;
extern cpuop_func op_1170_4_ff;
extern cpuop_func op_1178_4_nf;
extern cpuop_func op_1178_4_ff;
extern cpuop_func op_1179_4_nf;
extern cpuop_func op_1179_4_ff;
extern cpuop_func op_117a_4_nf;
extern cpuop_func op_117a_4_ff;
extern cpuop_func op_117b_4_nf;
extern cpuop_func op_117b_4_ff;
extern cpuop_func op_117c_4_nf;
extern cpuop_func op_117c_4_ff;
extern cpuop_func op_1180_4_nf;
extern cpuop_func op_1180_4_ff;
extern cpuop_func op_1190_4_nf;
extern cpuop_func op_1190_4_ff;
extern cpuop_func op_1198_4_nf;
extern cpuop_func op_1198_4_ff;
extern cpuop_func op_11a0_4_nf;
extern cpuop_func op_11a0_4_ff;
extern cpuop_func op_11a8_4_nf;
extern cpuop_func op_11a8_4_ff;
extern cpuop_func op_11b0_4_nf;
extern cpuop_func op_11b0_4_ff;
extern cpuop_func op_11b8_4_nf;
extern cpuop_func op_11b8_4_ff;
extern cpuop_func op_11b9_4_nf;
extern cpuop_func op_11b9_4_ff;
extern cpuop_func op_11ba_4_nf;
extern cpuop_func op_11ba_4_ff;
extern cpuop_func op_11bb_4_nf;
extern cpuop_func op_11bb_4_ff;
extern cpuop_func op_11bc_4_nf;
extern cpuop_func op_11bc_4_ff;
extern cpuop_func op_11c0_4_nf;
extern cpuop_func op_11c0_4_ff;
extern cpuop_func op_11d0_4_nf;
extern cpuop_func op_11d0_4_ff;
extern cpuop_func op_11d8_4_nf;
extern cpuop_func op_11d8_4_ff;
extern cpuop_func op_11e0_4_nf;
extern cpuop_func op_11e0_4_ff;
extern cpuop_func op_11e8_4_nf;
extern cpuop_func op_11e8_4_ff;
extern cpuop_func op_11f0_4_nf;
extern cpuop_func op_11f0_4_ff;
extern cpuop_func op_11f8_4_nf;
extern cpuop_func op_11f8_4_ff;
extern cpuop_func op_11f9_4_nf;
extern cpuop_func op_11f9_4_ff;
extern cpuop_func op_11fa_4_nf;
extern cpuop_func op_11fa_4_ff;
extern cpuop_func op_11fb_4_nf;
extern cpuop_func op_11fb_4_ff;
extern cpuop_func op_11fc_4_nf;
extern cpuop_func op_11fc_4_ff;
extern cpuop_func op_13c0_4_nf;
extern cpuop_func op_13c0_4_ff;
extern cpuop_func op_13d0_4_nf;
extern cpuop_func op_13d0_4_ff;
extern cpuop_func op_13d8_4_nf;
extern cpuop_func op_13d8_4_ff;
extern cpuop_func op_13e0_4_nf;
extern cpuop_func op_13e0_4_ff;
extern cpuop_func op_13e8_4_nf;
extern cpuop_func op_13e8_4_ff;
extern cpuop_func op_13f0_4_nf;
extern cpuop_func op_13f0_4_ff;
extern cpuop_func op_13f8_4_nf;
extern cpuop_func op_13f8_4_ff;
extern cpuop_func op_13f9_4_nf;
extern cpuop_func op_13f9_4_ff;
extern cpuop_func op_13fa_4_nf;
extern cpuop_func op_13fa_4_ff;
extern cpuop_func op_13fb_4_nf;
extern cpuop_func op_13fb_4_ff;
extern cpuop_func op_13fc_4_nf;
extern cpuop_func op_13fc_4_ff;
extern cpuop_func op_2000_4_nf;
extern cpuop_func op_2000_4_ff;
extern cpuop_func op_2008_4_nf;
extern cpuop_func op_2008_4_ff;
extern cpuop_func op_2010_4_nf;
extern cpuop_func op_2010_4_ff;
extern cpuop_func op_2018_4_nf;
extern cpuop_func op_2018_4_ff;
extern cpuop_func op_2020_4_nf;
extern cpuop_func op_2020_4_ff;
extern cpuop_func op_2028_4_nf;
extern cpuop_func op_2028_4_ff;
extern cpuop_func op_2030_4_nf;
extern cpuop_func op_2030_4_ff;
extern cpuop_func op_2038_4_nf;
extern cpuop_func op_2038_4_ff;
extern cpuop_func op_2039_4_nf;
extern cpuop_func op_2039_4_ff;
extern cpuop_func op_203a_4_nf;
extern cpuop_func op_203a_4_ff;
extern cpuop_func op_203b_4_nf;
extern cpuop_func op_203b_4_ff;
extern cpuop_func op_203c_4_nf;
extern cpuop_func op_203c_4_ff;
extern cpuop_func op_2040_4_nf;
extern cpuop_func op_2040_4_ff;
extern cpuop_func op_2048_4_nf;
extern cpuop_func op_2048_4_ff;
extern cpuop_func op_2050_4_nf;
extern cpuop_func op_2050_4_ff;
extern cpuop_func op_2058_4_nf;
extern cpuop_func op_2058_4_ff;
extern cpuop_func op_2060_4_nf;
extern cpuop_func op_2060_4_ff;
extern cpuop_func op_2068_4_nf;
extern cpuop_func op_2068_4_ff;
extern cpuop_func op_2070_4_nf;
extern cpuop_func op_2070_4_ff;
extern cpuop_func op_2078_4_nf;
extern cpuop_func op_2078_4_ff;
extern cpuop_func op_2079_4_nf;
extern cpuop_func op_2079_4_ff;
extern cpuop_func op_207a_4_nf;
extern cpuop_func op_207a_4_ff;
extern cpuop_func op_207b_4_nf;
extern cpuop_func op_207b_4_ff;
extern cpuop_func op_207c_4_nf;
extern cpuop_func op_207c_4_ff;
extern cpuop_func op_2080_4_nf;
extern cpuop_func op_2080_4_ff;
extern cpuop_func op_2088_4_nf;
extern cpuop_func op_2088_4_ff;
extern cpuop_func op_2090_4_nf;
extern cpuop_func op_2090_4_ff;
extern cpuop_func op_2098_4_nf;
extern cpuop_func op_2098_4_ff;
extern cpuop_func op_20a0_4_nf;
extern cpuop_func op_20a0_4_ff;
extern cpuop_func op_20a8_4_nf;
extern cpuop_func op_20a8_4_ff;
extern cpuop_func op_20b0_4_nf;
extern cpuop_func op_20b0_4_ff;
extern cpuop_func op_20b8_4_nf;
extern cpuop_func op_20b8_4_ff;
extern cpuop_func op_20b9_4_nf;
extern cpuop_func op_20b9_4_ff;
extern cpuop_func op_20ba_4_nf;
extern cpuop_func op_20ba_4_ff;
extern cpuop_func op_20bb_4_nf;
extern cpuop_func op_20bb_4_ff;
extern cpuop_func op_20bc_4_nf;
extern cpuop_func op_20bc_4_ff;
extern cpuop_func op_20c0_4_nf;
extern cpuop_func op_20c0_4_ff;
extern cpuop_func op_20c8_4_nf;
extern cpuop_func op_20c8_4_ff;
extern cpuop_func op_20d0_4_nf;
extern cpuop_func op_20d0_4_ff;
extern cpuop_func op_20d8_4_nf;
extern cpuop_func op_20d8_4_ff;
extern cpuop_func op_20e0_4_nf;
extern cpuop_func op_20e0_4_ff;
extern cpuop_func op_20e8_4_nf;
extern cpuop_func op_20e8_4_ff;
extern cpuop_func op_20f0_4_nf;
extern cpuop_func op_20f0_4_ff;
extern cpuop_func op_20f8_4_nf;
extern cpuop_func op_20f8_4_ff;
extern cpuop_func op_20f9_4_nf;
extern cpuop_func op_20f9_4_ff;
extern cpuop_func op_20fa_4_nf;
extern cpuop_func op_20fa_4_ff;
extern cpuop_func op_20fb_4_nf;
extern cpuop_func op_20fb_4_ff;
extern cpuop_func op_20fc_4_nf;
extern cpuop_func op_20fc_4_ff;
extern cpuop_func op_2100_4_nf;
extern cpuop_func op_2100_4_ff;
extern cpuop_func op_2108_4_nf;
extern cpuop_func op_2108_4_ff;
extern cpuop_func op_2110_4_nf;
extern cpuop_func op_2110_4_ff;
extern cpuop_func op_2118_4_nf;
extern cpuop_func op_2118_4_ff;
extern cpuop_func op_2120_4_nf;
extern cpuop_func op_2120_4_ff;
extern cpuop_func op_2128_4_nf;
extern cpuop_func op_2128_4_ff;
extern cpuop_func op_2130_4_nf;
extern cpuop_func op_2130_4_ff;
extern cpuop_func op_2138_4_nf;
extern cpuop_func op_2138_4_ff;
extern cpuop_func op_2139_4_nf;
extern cpuop_func op_2139_4_ff;
extern cpuop_func op_213a_4_nf;
extern cpuop_func op_213a_4_ff;
extern cpuop_func op_213b_4_nf;
extern cpuop_func op_213b_4_ff;
extern cpuop_func op_213c_4_nf;
extern cpuop_func op_213c_4_ff;
extern cpuop_func op_2140_4_nf;
extern cpuop_func op_2140_4_ff;
extern cpuop_func op_2148_4_nf;
extern cpuop_func op_2148_4_ff;
extern cpuop_func op_2150_4_nf;
extern cpuop_func op_2150_4_ff;
extern cpuop_func op_2158_4_nf;
extern cpuop_func op_2158_4_ff;
extern cpuop_func op_2160_4_nf;
extern cpuop_func op_2160_4_ff;
extern cpuop_func op_2168_4_nf;
extern cpuop_func op_2168_4_ff;
extern cpuop_func op_2170_4_nf;
extern cpuop_func op_2170_4_ff;
extern cpuop_func op_2178_4_nf;
extern cpuop_func op_2178_4_ff;
extern cpuop_func op_2179_4_nf;
extern cpuop_func op_2179_4_ff;
extern cpuop_func op_217a_4_nf;
extern cpuop_func op_217a_4_ff;
extern cpuop_func op_217b_4_nf;
extern cpuop_func op_217b_4_ff;
extern cpuop_func op_217c_4_nf;
extern cpuop_func op_217c_4_ff;
extern cpuop_func op_2180_4_nf;
extern cpuop_func op_2180_4_ff;
extern cpuop_func op_2188_4_nf;
extern cpuop_func op_2188_4_ff;
extern cpuop_func op_2190_4_nf;
extern cpuop_func op_2190_4_ff;
extern cpuop_func op_2198_4_nf;
extern cpuop_func op_2198_4_ff;
extern cpuop_func op_21a0_4_nf;
extern cpuop_func op_21a0_4_ff;
extern cpuop_func op_21a8_4_nf;
extern cpuop_func op_21a8_4_ff;
extern cpuop_func op_21b0_4_nf;
extern cpuop_func op_21b0_4_ff;
extern cpuop_func op_21b8_4_nf;
extern cpuop_func op_21b8_4_ff;
extern cpuop_func op_21b9_4_nf;
extern cpuop_func op_21b9_4_ff;
extern cpuop_func op_21ba_4_nf;
extern cpuop_func op_21ba_4_ff;
extern cpuop_func op_21bb_4_nf;
extern cpuop_func op_21bb_4_ff;
extern cpuop_func op_21bc_4_nf;
extern cpuop_func op_21bc_4_ff;
extern cpuop_func op_21c0_4_nf;
extern cpuop_func op_21c0_4_ff;
extern cpuop_func op_21c8_4_nf;
extern cpuop_func op_21c8_4_ff;
extern cpuop_func op_21d0_4_nf;
extern cpuop_func op_21d0_4_ff;
extern cpuop_func op_21d8_4_nf;
extern cpuop_func op_21d8_4_ff;
extern cpuop_func op_21e0_4_nf;
extern cpuop_func op_21e0_4_ff;
extern cpuop_func op_21e8_4_nf;
extern cpuop_func op_21e8_4_ff;
extern cpuop_func op_21f0_4_nf;
extern cpuop_func op_21f0_4_ff;
extern cpuop_func op_21f8_4_nf;
extern cpuop_func op_21f8_4_ff;
extern cpuop_func op_21f9_4_nf;
extern cpuop_func op_21f9_4_ff;
extern cpuop_func op_21fa_4_nf;
extern cpuop_func op_21fa_4_ff;
extern cpuop_func op_21fb_4_nf;
extern cpuop_func op_21fb_4_ff;
extern cpuop_func op_21fc_4_nf;
extern cpuop_func op_21fc_4_ff;
extern cpuop_func op_23c0_4_nf;
extern cpuop_func op_23c0_4_ff;
extern cpuop_func op_23c8_4_nf;
extern cpuop_func op_23c8_4_ff;
extern cpuop_func op_23d0_4_nf;
extern cpuop_func op_23d0_4_ff;
extern cpuop_func op_23d8_4_nf;
extern cpuop_func op_23d8_4_ff;
extern cpuop_func op_23e0_4_nf;
extern cpuop_func op_23e0_4_ff;
extern cpuop_func op_23e8_4_nf;
extern cpuop_func op_23e8_4_ff;
extern cpuop_func op_23f0_4_nf;
extern cpuop_func op_23f0_4_ff;
extern cpuop_func op_23f8_4_nf;
extern cpuop_func op_23f8_4_ff;
extern cpuop_func op_23f9_4_nf;
extern cpuop_func op_23f9_4_ff;
extern cpuop_func op_23fa_4_nf;
extern cpuop_func op_23fa_4_ff;
extern cpuop_func op_23fb_4_nf;
extern cpuop_func op_23fb_4_ff;
extern cpuop_func op_23fc_4_nf;
extern cpuop_func op_23fc_4_ff;
extern cpuop_func op_3000_4_nf;
extern cpuop_func op_3000_4_ff;
extern cpuop_func op_3008_4_nf;
extern cpuop_func op_3008_4_ff;
extern cpuop_func op_3010_4_nf;
extern cpuop_func op_3010_4_ff;
extern cpuop_func op_3018_4_nf;
extern cpuop_func op_3018_4_ff;
extern cpuop_func op_3020_4_nf;
extern cpuop_func op_3020_4_ff;
extern cpuop_func op_3028_4_nf;
extern cpuop_func op_3028_4_ff;
extern cpuop_func op_3030_4_nf;
extern cpuop_func op_3030_4_ff;
extern cpuop_func op_3038_4_nf;
extern cpuop_func op_3038_4_ff;
extern cpuop_func op_3039_4_nf;
extern cpuop_func op_3039_4_ff;
extern cpuop_func op_303a_4_nf;
extern cpuop_func op_303a_4_ff;
extern cpuop_func op_303b_4_nf;
extern cpuop_func op_303b_4_ff;
extern cpuop_func op_303c_4_nf;
extern cpuop_func op_303c_4_ff;
extern cpuop_func op_3040_4_nf;
extern cpuop_func op_3040_4_ff;
extern cpuop_func op_3048_4_nf;
extern cpuop_func op_3048_4_ff;
extern cpuop_func op_3050_4_nf;
extern cpuop_func op_3050_4_ff;
extern cpuop_func op_3058_4_nf;
extern cpuop_func op_3058_4_ff;
extern cpuop_func op_3060_4_nf;
extern cpuop_func op_3060_4_ff;
extern cpuop_func op_3068_4_nf;
extern cpuop_func op_3068_4_ff;
extern cpuop_func op_3070_4_nf;
extern cpuop_func op_3070_4_ff;
extern cpuop_func op_3078_4_nf;
extern cpuop_func op_3078_4_ff;
extern cpuop_func op_3079_4_nf;
extern cpuop_func op_3079_4_ff;
extern cpuop_func op_307a_4_nf;
extern cpuop_func op_307a_4_ff;
extern cpuop_func op_307b_4_nf;
extern cpuop_func op_307b_4_ff;
extern cpuop_func op_307c_4_nf;
extern cpuop_func op_307c_4_ff;
extern cpuop_func op_3080_4_nf;
extern cpuop_func op_3080_4_ff;
extern cpuop_func op_3088_4_nf;
extern cpuop_func op_3088_4_ff;
extern cpuop_func op_3090_4_nf;
extern cpuop_func op_3090_4_ff;
extern cpuop_func op_3098_4_nf;
extern cpuop_func op_3098_4_ff;
extern cpuop_func op_30a0_4_nf;
extern cpuop_func op_30a0_4_ff;
extern cpuop_func op_30a8_4_nf;
extern cpuop_func op_30a8_4_ff;
extern cpuop_func op_30b0_4_nf;
extern cpuop_func op_30b0_4_ff;
extern cpuop_func op_30b8_4_nf;
extern cpuop_func op_30b8_4_ff;
extern cpuop_func op_30b9_4_nf;
extern cpuop_func op_30b9_4_ff;
extern cpuop_func op_30ba_4_nf;
extern cpuop_func op_30ba_4_ff;
extern cpuop_func op_30bb_4_nf;
extern cpuop_func op_30bb_4_ff;
extern cpuop_func op_30bc_4_nf;
extern cpuop_func op_30bc_4_ff;
extern cpuop_func op_30c0_4_nf;
extern cpuop_func op_30c0_4_ff;
extern cpuop_func op_30c8_4_nf;
extern cpuop_func op_30c8_4_ff;
extern cpuop_func op_30d0_4_nf;
extern cpuop_func op_30d0_4_ff;
extern cpuop_func op_30d8_4_nf;
extern cpuop_func op_30d8_4_ff;
extern cpuop_func op_30e0_4_nf;
extern cpuop_func op_30e0_4_ff;
extern cpuop_func op_30e8_4_nf;
extern cpuop_func op_30e8_4_ff;
extern cpuop_func op_30f0_4_nf;
extern cpuop_func op_30f0_4_ff;
extern cpuop_func op_30f8_4_nf;
extern cpuop_func op_30f8_4_ff;
extern cpuop_func op_30f9_4_nf;
extern cpuop_func op_30f9_4_ff;
extern cpuop_func op_30fa_4_nf;
extern cpuop_func op_30fa_4_ff;
extern cpuop_func op_30fb_4_nf;
extern cpuop_func op_30fb_4_ff;
extern cpuop_func op_30fc_4_nf;
extern cpuop_func op_30fc_4_ff;
extern cpuop_func op_3100_4_nf;
extern cpuop_func op_3100_4_ff;
extern cpuop_func op_3108_4_nf;
extern cpuop_func op_3108_4_ff;
extern cpuop_func op_3110_4_nf;
extern cpuop_func op_3110_4_ff;
extern cpuop_func op_3118_4_nf;
extern cpuop_func op_3118_4_ff;
extern cpuop_func op_3120_4_nf;
extern cpuop_func op_3120_4_ff;
extern cpuop_func op_3128_4_nf;
extern cpuop_func op_3128_4_ff;
extern cpuop_func op_3130_4_nf;
extern cpuop_func op_3130_4_ff;
extern cpuop_func op_3138_4_nf;
extern cpuop_func op_3138_4_ff;
extern cpuop_func op_3139_4_nf;
extern cpuop_func op_3139_4_ff;
extern cpuop_func op_313a_4_nf;
extern cpuop_func op_313a_4_ff;
extern cpuop_func op_313b_4_nf;
extern cpuop_func op_313b_4_ff;
extern cpuop_func op_313c_4_nf;
extern cpuop_func op_313c_4_ff;
extern cpuop_func op_3140_4_nf;
extern cpuop_func op_3140_4_ff;
extern cpuop_func op_3148_4_nf;
extern cpuop_func op_3148_4_ff;
extern cpuop_func op_3150_4_nf;
extern cpuop_func op_3150_4_ff;
extern cpuop_func op_3158_4_nf;
extern cpuop_func op_3158_4_ff;
extern cpuop_func op_3160_4_nf;
extern cpuop_func op_3160_4_ff;
extern cpuop_func op_3168_4_nf;
extern cpuop_func op_3168_4_ff;
extern cpuop_func op_3170_4_nf;
extern cpuop_func op_3170_4_ff;
extern cpuop_func op_3178_4_nf;
extern cpuop_func op_3178_4_ff;
extern cpuop_func op_3179_4_nf;
extern cpuop_func op_3179_4_ff;
extern cpuop_func op_317a_4_nf;
extern cpuop_func op_317a_4_ff;
extern cpuop_func op_317b_4_nf;
extern cpuop_func op_317b_4_ff;
extern cpuop_func op_317c_4_nf;
extern cpuop_func op_317c_4_ff;
extern cpuop_func op_3180_4_nf;
extern cpuop_func op_3180_4_ff;
extern cpuop_func op_3188_4_nf;
extern cpuop_func op_3188_4_ff;
extern cpuop_func op_3190_4_nf;
extern cpuop_func op_3190_4_ff;
extern cpuop_func op_3198_4_nf;
extern cpuop_func op_3198_4_ff;
extern cpuop_func op_31a0_4_nf;
extern cpuop_func op_31a0_4_ff;
extern cpuop_func op_31a8_4_nf;
extern cpuop_func op_31a8_4_ff;
extern cpuop_func op_31b0_4_nf;
extern cpuop_func op_31b0_4_ff;
extern cpuop_func op_31b8_4_nf;
extern cpuop_func op_31b8_4_ff;
extern cpuop_func op_31b9_4_nf;
extern cpuop_func op_31b9_4_ff;
extern cpuop_func op_31ba_4_nf;
extern cpuop_func op_31ba_4_ff;
extern cpuop_func op_31bb_4_nf;
extern cpuop_func op_31bb_4_ff;
extern cpuop_func op_31bc_4_nf;
extern cpuop_func op_31bc_4_ff;
extern cpuop_func op_31c0_4_nf;
extern cpuop_func op_31c0_4_ff;
extern cpuop_func op_31c8_4_nf;
extern cpuop_func op_31c8_4_ff;
extern cpuop_func op_31d0_4_nf;
extern cpuop_func op_31d0_4_ff;
extern cpuop_func op_31d8_4_nf;
extern cpuop_func op_31d8_4_ff;
extern cpuop_func op_31e0_4_nf;
extern cpuop_func op_31e0_4_ff;
extern cpuop_func op_31e8_4_nf;
extern cpuop_func op_31e8_4_ff;
extern cpuop_func op_31f0_4_nf;
extern cpuop_func op_31f0_4_ff;
extern cpuop_func op_31f8_4_nf;
extern cpuop_func op_31f8_4_ff;
extern cpuop_func op_31f9_4_nf;
extern cpuop_func op_31f9_4_ff;
extern cpuop_func op_31fa_4_nf;
extern cpuop_func op_31fa_4_ff;
extern cpuop_func op_31fb_4_nf;
extern cpuop_func op_31fb_4_ff;
extern cpuop_func op_31fc_4_nf;
extern cpuop_func op_31fc_4_ff;
extern cpuop_func op_33c0_4_nf;
extern cpuop_func op_33c0_4_ff;
extern cpuop_func op_33c8_4_nf;
extern cpuop_func op_33c8_4_ff;
extern cpuop_func op_33d0_4_nf;
extern cpuop_func op_33d0_4_ff;
extern cpuop_func op_33d8_4_nf;
extern cpuop_func op_33d8_4_ff;
extern cpuop_func op_33e0_4_nf;
extern cpuop_func op_33e0_4_ff;
extern cpuop_func op_33e8_4_nf;
extern cpuop_func op_33e8_4_ff;
extern cpuop_func op_33f0_4_nf;
extern cpuop_func op_33f0_4_ff;
extern cpuop_func op_33f8_4_nf;
extern cpuop_func op_33f8_4_ff;
extern cpuop_func op_33f9_4_nf;
extern cpuop_func op_33f9_4_ff;
extern cpuop_func op_33fa_4_nf;
extern cpuop_func op_33fa_4_ff;
extern cpuop_func op_33fb_4_nf;
extern cpuop_func op_33fb_4_ff;
extern cpuop_func op_33fc_4_nf;
extern cpuop_func op_33fc_4_ff;
extern cpuop_func op_4000_4_nf;
extern cpuop_func op_4000_4_ff;
extern cpuop_func op_4010_4_nf;
extern cpuop_func op_4010_4_ff;
extern cpuop_func op_4018_4_nf;
extern cpuop_func op_4018_4_ff;
extern cpuop_func op_4020_4_nf;
extern cpuop_func op_4020_4_ff;
extern cpuop_func op_4028_4_nf;
extern cpuop_func op_4028_4_ff;
extern cpuop_func op_4030_4_nf;
extern cpuop_func op_4030_4_ff;
extern cpuop_func op_4038_4_nf;
extern cpuop_func op_4038_4_ff;
extern cpuop_func op_4039_4_nf;
extern cpuop_func op_4039_4_ff;
extern cpuop_func op_4040_4_nf;
extern cpuop_func op_4040_4_ff;
extern cpuop_func op_4050_4_nf;
extern cpuop_func op_4050_4_ff;
extern cpuop_func op_4058_4_nf;
extern cpuop_func op_4058_4_ff;
extern cpuop_func op_4060_4_nf;
extern cpuop_func op_4060_4_ff;
extern cpuop_func op_4068_4_nf;
extern cpuop_func op_4068_4_ff;
extern cpuop_func op_4070_4_nf;
extern cpuop_func op_4070_4_ff;
extern cpuop_func op_4078_4_nf;
extern cpuop_func op_4078_4_ff;
extern cpuop_func op_4079_4_nf;
extern cpuop_func op_4079_4_ff;
extern cpuop_func op_4080_4_nf;
extern cpuop_func op_4080_4_ff;
extern cpuop_func op_4090_4_nf;
extern cpuop_func op_4090_4_ff;
extern cpuop_func op_4098_4_nf;
extern cpuop_func op_4098_4_ff;
extern cpuop_func op_40a0_4_nf;
extern cpuop_func op_40a0_4_ff;
extern cpuop_func op_40a8_4_nf;
extern cpuop_func op_40a8_4_ff;
extern cpuop_func op_40b0_4_nf;
extern cpuop_func op_40b0_4_ff;
extern cpuop_func op_40b8_4_nf;
extern cpuop_func op_40b8_4_ff;
extern cpuop_func op_40b9_4_nf;
extern cpuop_func op_40b9_4_ff;
extern cpuop_func op_40c0_4_nf;
extern cpuop_func op_40c0_4_ff;
extern cpuop_func op_40d0_4_nf;
extern cpuop_func op_40d0_4_ff;
extern cpuop_func op_40d8_4_nf;
extern cpuop_func op_40d8_4_ff;
extern cpuop_func op_40e0_4_nf;
extern cpuop_func op_40e0_4_ff;
extern cpuop_func op_40e8_4_nf;
extern cpuop_func op_40e8_4_ff;
extern cpuop_func op_40f0_4_nf;
extern cpuop_func op_40f0_4_ff;
extern cpuop_func op_40f8_4_nf;
extern cpuop_func op_40f8_4_ff;
extern cpuop_func op_40f9_4_nf;
extern cpuop_func op_40f9_4_ff;
extern cpuop_func op_4180_4_nf;
extern cpuop_func op_4180_4_ff;
extern cpuop_func op_4190_4_nf;
extern cpuop_func op_4190_4_ff;
extern cpuop_func op_4198_4_nf;
extern cpuop_func op_4198_4_ff;
extern cpuop_func op_41a0_4_nf;
extern cpuop_func op_41a0_4_ff;
extern cpuop_func op_41a8_4_nf;
extern cpuop_func op_41a8_4_ff;
extern cpuop_func op_41b0_4_nf;
extern cpuop_func op_41b0_4_ff;
extern cpuop_func op_41b8_4_nf;
extern cpuop_func op_41b8_4_ff;
extern cpuop_func op_41b9_4_nf;
extern cpuop_func op_41b9_4_ff;
extern cpuop_func op_41ba_4_nf;
extern cpuop_func op_41ba_4_ff;
extern cpuop_func op_41bb_4_nf;
extern cpuop_func op_41bb_4_ff;
extern cpuop_func op_41bc_4_nf;
extern cpuop_func op_41bc_4_ff;
extern cpuop_func op_41d0_4_nf;
extern cpuop_func op_41d0_4_ff;
extern cpuop_func op_41e8_4_nf;
extern cpuop_func op_41e8_4_ff;
extern cpuop_func op_41f0_4_nf;
extern cpuop_func op_41f0_4_ff;
extern cpuop_func op_41f8_4_nf;
extern cpuop_func op_41f8_4_ff;
extern cpuop_func op_41f9_4_nf;
extern cpuop_func op_41f9_4_ff;
extern cpuop_func op_41fa_4_nf;
extern cpuop_func op_41fa_4_ff;
extern cpuop_func op_41fb_4_nf;
extern cpuop_func op_41fb_4_ff;
extern cpuop_func op_4200_4_nf;
extern cpuop_func op_4200_4_ff;
extern cpuop_func op_4210_4_nf;
extern cpuop_func op_4210_4_ff;
extern cpuop_func op_4218_4_nf;
extern cpuop_func op_4218_4_ff;
extern cpuop_func op_4220_4_nf;
extern cpuop_func op_4220_4_ff;
extern cpuop_func op_4228_4_nf;
extern cpuop_func op_4228_4_ff;
extern cpuop_func op_4230_4_nf;
extern cpuop_func op_4230_4_ff;
extern cpuop_func op_4238_4_nf;
extern cpuop_func op_4238_4_ff;
extern cpuop_func op_4239_4_nf;
extern cpuop_func op_4239_4_ff;
extern cpuop_func op_4240_4_nf;
extern cpuop_func op_4240_4_ff;
extern cpuop_func op_4250_4_nf;
extern cpuop_func op_4250_4_ff;
extern cpuop_func op_4258_4_nf;
extern cpuop_func op_4258_4_ff;
extern cpuop_func op_4260_4_nf;
extern cpuop_func op_4260_4_ff;
extern cpuop_func op_4268_4_nf;
extern cpuop_func op_4268_4_ff;
extern cpuop_func op_4270_4_nf;
extern cpuop_func op_4270_4_ff;
extern cpuop_func op_4278_4_nf;
extern cpuop_func op_4278_4_ff;
extern cpuop_func op_4279_4_nf;
extern cpuop_func op_4279_4_ff;
extern cpuop_func op_4280_4_nf;
extern cpuop_func op_4280_4_ff;
extern cpuop_func op_4290_4_nf;
extern cpuop_func op_4290_4_ff;
extern cpuop_func op_4298_4_nf;
extern cpuop_func op_4298_4_ff;
extern cpuop_func op_42a0_4_nf;
extern cpuop_func op_42a0_4_ff;
extern cpuop_func op_42a8_4_nf;
extern cpuop_func op_42a8_4_ff;
extern cpuop_func op_42b0_4_nf;
extern cpuop_func op_42b0_4_ff;
extern cpuop_func op_42b8_4_nf;
extern cpuop_func op_42b8_4_ff;
extern cpuop_func op_42b9_4_nf;
extern cpuop_func op_42b9_4_ff;
extern cpuop_func op_42c0_4_nf;
extern cpuop_func op_42c0_4_ff;
extern cpuop_func op_42d0_4_nf;
extern cpuop_func op_42d0_4_ff;
extern cpuop_func op_42d8_4_nf;
extern cpuop_func op_42d8_4_ff;
extern cpuop_func op_42e0_4_nf;
extern cpuop_func op_42e0_4_ff;
extern cpuop_func op_42e8_4_nf;
extern cpuop_func op_42e8_4_ff;
extern cpuop_func op_42f0_4_nf;
extern cpuop_func op_42f0_4_ff;
extern cpuop_func op_42f8_4_nf;
extern cpuop_func op_42f8_4_ff;
extern cpuop_func op_42f9_4_nf;
extern cpuop_func op_42f9_4_ff;
extern cpuop_func op_4400_4_nf;
extern cpuop_func op_4400_4_ff;
extern cpuop_func op_4410_4_nf;
extern cpuop_func op_4410_4_ff;
extern cpuop_func op_4418_4_nf;
extern cpuop_func op_4418_4_ff;
extern cpuop_func op_4420_4_nf;
extern cpuop_func op_4420_4_ff;
extern cpuop_func op_4428_4_nf;
extern cpuop_func op_4428_4_ff;
extern cpuop_func op_4430_4_nf;
extern cpuop_func op_4430_4_ff;
extern cpuop_func op_4438_4_nf;
extern cpuop_func op_4438_4_ff;
extern cpuop_func op_4439_4_nf;
extern cpuop_func op_4439_4_ff;
extern cpuop_func op_4440_4_nf;
extern cpuop_func op_4440_4_ff;
extern cpuop_func op_4450_4_nf;
extern cpuop_func op_4450_4_ff;
extern cpuop_func op_4458_4_nf;
extern cpuop_func op_4458_4_ff;
extern cpuop_func op_4460_4_nf;
extern cpuop_func op_4460_4_ff;
extern cpuop_func op_4468_4_nf;
extern cpuop_func op_4468_4_ff;
extern cpuop_func op_4470_4_nf;
extern cpuop_func op_4470_4_ff;
extern cpuop_func op_4478_4_nf;
extern cpuop_func op_4478_4_ff;
extern cpuop_func op_4479_4_nf;
extern cpuop_func op_4479_4_ff;
extern cpuop_func op_4480_4_nf;
extern cpuop_func op_4480_4_ff;
extern cpuop_func op_4490_4_nf;
extern cpuop_func op_4490_4_ff;
extern cpuop_func op_4498_4_nf;
extern cpuop_func op_4498_4_ff;
extern cpuop_func op_44a0_4_nf;
extern cpuop_func op_44a0_4_ff;
extern cpuop_func op_44a8_4_nf;
extern cpuop_func op_44a8_4_ff;
extern cpuop_func op_44b0_4_nf;
extern cpuop_func op_44b0_4_ff;
extern cpuop_func op_44b8_4_nf;
extern cpuop_func op_44b8_4_ff;
extern cpuop_func op_44b9_4_nf;
extern cpuop_func op_44b9_4_ff;
extern cpuop_func op_44c0_4_nf;
extern cpuop_func op_44c0_4_ff;
extern cpuop_func op_44d0_4_nf;
extern cpuop_func op_44d0_4_ff;
extern cpuop_func op_44d8_4_nf;
extern cpuop_func op_44d8_4_ff;
extern cpuop_func op_44e0_4_nf;
extern cpuop_func op_44e0_4_ff;
extern cpuop_func op_44e8_4_nf;
extern cpuop_func op_44e8_4_ff;
extern cpuop_func op_44f0_4_nf;
extern cpuop_func op_44f0_4_ff;
extern cpuop_func op_44f8_4_nf;
extern cpuop_func op_44f8_4_ff;
extern cpuop_func op_44f9_4_nf;
extern cpuop_func op_44f9_4_ff;
extern cpuop_func op_44fa_4_nf;
extern cpuop_func op_44fa_4_ff;
extern cpuop_func op_44fb_4_nf;
extern cpuop_func op_44fb_4_ff;
extern cpuop_func op_44fc_4_nf;
extern cpuop_func op_44fc_4_ff;
extern cpuop_func op_4600_4_nf;
extern cpuop_func op_4600_4_ff;
extern cpuop_func op_4610_4_nf;
extern cpuop_func op_4610_4_ff;
extern cpuop_func op_4618_4_nf;
extern cpuop_func op_4618_4_ff;
extern cpuop_func op_4620_4_nf;
extern cpuop_func op_4620_4_ff;
extern cpuop_func op_4628_4_nf;
extern cpuop_func op_4628_4_ff;
extern cpuop_func op_4630_4_nf;
extern cpuop_func op_4630_4_ff;
extern cpuop_func op_4638_4_nf;
extern cpuop_func op_4638_4_ff;
extern cpuop_func op_4639_4_nf;
extern cpuop_func op_4639_4_ff;
extern cpuop_func op_4640_4_nf;
extern cpuop_func op_4640_4_ff;
extern cpuop_func op_4650_4_nf;
extern cpuop_func op_4650_4_ff;
extern cpuop_func op_4658_4_nf;
extern cpuop_func op_4658_4_ff;
extern cpuop_func op_4660_4_nf;
extern cpuop_func op_4660_4_ff;
extern cpuop_func op_4668_4_nf;
extern cpuop_func op_4668_4_ff;
extern cpuop_func op_4670_4_nf;
extern cpuop_func op_4670_4_ff;
extern cpuop_func op_4678_4_nf;
extern cpuop_func op_4678_4_ff;
extern cpuop_func op_4679_4_nf;
extern cpuop_func op_4679_4_ff;
extern cpuop_func op_4680_4_nf;
extern cpuop_func op_4680_4_ff;
extern cpuop_func op_4690_4_nf;
extern cpuop_func op_4690_4_ff;
extern cpuop_func op_4698_4_nf;
extern cpuop_func op_4698_4_ff;
extern cpuop_func op_46a0_4_nf;
extern cpuop_func op_46a0_4_ff;
extern cpuop_func op_46a8_4_nf;
extern cpuop_func op_46a8_4_ff;
extern cpuop_func op_46b0_4_nf;
extern cpuop_func op_46b0_4_ff;
extern cpuop_func op_46b8_4_nf;
extern cpuop_func op_46b8_4_ff;
extern cpuop_func op_46b9_4_nf;
extern cpuop_func op_46b9_4_ff;
extern cpuop_func op_46c0_4_nf;
extern cpuop_func op_46c0_4_ff;
extern cpuop_func op_46d0_4_nf;
extern cpuop_func op_46d0_4_ff;
extern cpuop_func op_46d8_4_nf;
extern cpuop_func op_46d8_4_ff;
extern cpuop_func op_46e0_4_nf;
extern cpuop_func op_46e0_4_ff;
extern cpuop_func op_46e8_4_nf;
extern cpuop_func op_46e8_4_ff;
extern cpuop_func op_46f0_4_nf;
extern cpuop_func op_46f0_4_ff;
extern cpuop_func op_46f8_4_nf;
extern cpuop_func op_46f8_4_ff;
extern cpuop_func op_46f9_4_nf;
extern cpuop_func op_46f9_4_ff;
extern cpuop_func op_46fa_4_nf;
extern cpuop_func op_46fa_4_ff;
extern cpuop_func op_46fb_4_nf;
extern cpuop_func op_46fb_4_ff;
extern cpuop_func op_46fc_4_nf;
extern cpuop_func op_46fc_4_ff;
extern cpuop_func op_4800_4_nf;
extern cpuop_func op_4800_4_ff;
extern cpuop_func op_4810_4_nf;
extern cpuop_func op_4810_4_ff;
extern cpuop_func op_4818_4_nf;
extern cpuop_func op_4818_4_ff;
extern cpuop_func op_4820_4_nf;
extern cpuop_func op_4820_4_ff;
extern cpuop_func op_4828_4_nf;
extern cpuop_func op_4828_4_ff;
extern cpuop_func op_4830_4_nf;
extern cpuop_func op_4830_4_ff;
extern cpuop_func op_4838_4_nf;
extern cpuop_func op_4838_4_ff;
extern cpuop_func op_4839_4_nf;
extern cpuop_func op_4839_4_ff;
extern cpuop_func op_4840_4_nf;
extern cpuop_func op_4840_4_ff;
extern cpuop_func op_4850_4_nf;
extern cpuop_func op_4850_4_ff;
extern cpuop_func op_4868_4_nf;
extern cpuop_func op_4868_4_ff;
extern cpuop_func op_4870_4_nf;
extern cpuop_func op_4870_4_ff;
extern cpuop_func op_4878_4_nf;
extern cpuop_func op_4878_4_ff;
extern cpuop_func op_4879_4_nf;
extern cpuop_func op_4879_4_ff;
extern cpuop_func op_487a_4_nf;
extern cpuop_func op_487a_4_ff;
extern cpuop_func op_487b_4_nf;
extern cpuop_func op_487b_4_ff;
extern cpuop_func op_4880_4_nf;
extern cpuop_func op_4880_4_ff;
extern cpuop_func op_4890_4_nf;
extern cpuop_func op_4890_4_ff;
extern cpuop_func op_48a0_4_nf;
extern cpuop_func op_48a0_4_ff;
extern cpuop_func op_48a8_4_nf;
extern cpuop_func op_48a8_4_ff;
extern cpuop_func op_48b0_4_nf;
extern cpuop_func op_48b0_4_ff;
extern cpuop_func op_48b8_4_nf;
extern cpuop_func op_48b8_4_ff;
extern cpuop_func op_48b9_4_nf;
extern cpuop_func op_48b9_4_ff;
extern cpuop_func op_48c0_4_nf;
extern cpuop_func op_48c0_4_ff;
extern cpuop_func op_48d0_4_nf;
extern cpuop_func op_48d0_4_ff;
extern cpuop_func op_48e0_4_nf;
extern cpuop_func op_48e0_4_ff;
extern cpuop_func op_48e8_4_nf;
extern cpuop_func op_48e8_4_ff;
extern cpuop_func op_48f0_4_nf;
extern cpuop_func op_48f0_4_ff;
extern cpuop_func op_48f8_4_nf;
extern cpuop_func op_48f8_4_ff;
extern cpuop_func op_48f9_4_nf;
extern cpuop_func op_48f9_4_ff;
extern cpuop_func op_4a00_4_nf;
extern cpuop_func op_4a00_4_ff;
extern cpuop_func op_4a10_4_nf;
extern cpuop_func op_4a10_4_ff;
extern cpuop_func op_4a18_4_nf;
extern cpuop_func op_4a18_4_ff;
extern cpuop_func op_4a20_4_nf;
extern cpuop_func op_4a20_4_ff;
extern cpuop_func op_4a28_4_nf;
extern cpuop_func op_4a28_4_ff;
extern cpuop_func op_4a30_4_nf;
extern cpuop_func op_4a30_4_ff;
extern cpuop_func op_4a38_4_nf;
extern cpuop_func op_4a38_4_ff;
extern cpuop_func op_4a39_4_nf;
extern cpuop_func op_4a39_4_ff;
extern cpuop_func op_4a40_4_nf;
extern cpuop_func op_4a40_4_ff;
extern cpuop_func op_4a50_4_nf;
extern cpuop_func op_4a50_4_ff;
extern cpuop_func op_4a58_4_nf;
extern cpuop_func op_4a58_4_ff;
extern cpuop_func op_4a60_4_nf;
extern cpuop_func op_4a60_4_ff;
extern cpuop_func op_4a68_4_nf;
extern cpuop_func op_4a68_4_ff;
extern cpuop_func op_4a70_4_nf;
extern cpuop_func op_4a70_4_ff;
extern cpuop_func op_4a78_4_nf;
extern cpuop_func op_4a78_4_ff;
extern cpuop_func op_4a79_4_nf;
extern cpuop_func op_4a79_4_ff;
extern cpuop_func op_4a80_4_nf;
extern cpuop_func op_4a80_4_ff;
extern cpuop_func op_4a90_4_nf;
extern cpuop_func op_4a90_4_ff;
extern cpuop_func op_4a98_4_nf;
extern cpuop_func op_4a98_4_ff;
extern cpuop_func op_4aa0_4_nf;
extern cpuop_func op_4aa0_4_ff;
extern cpuop_func op_4aa8_4_nf;
extern cpuop_func op_4aa8_4_ff;
extern cpuop_func op_4ab0_4_nf;
extern cpuop_func op_4ab0_4_ff;
extern cpuop_func op_4ab8_4_nf;
extern cpuop_func op_4ab8_4_ff;
extern cpuop_func op_4ab9_4_nf;
extern cpuop_func op_4ab9_4_ff;
extern cpuop_func op_4ac0_4_nf;
extern cpuop_func op_4ac0_4_ff;
extern cpuop_func op_4ad0_4_nf;
extern cpuop_func op_4ad0_4_ff;
extern cpuop_func op_4ad8_4_nf;
extern cpuop_func op_4ad8_4_ff;
extern cpuop_func op_4ae0_4_nf;
extern cpuop_func op_4ae0_4_ff;
extern cpuop_func op_4ae8_4_nf;
extern cpuop_func op_4ae8_4_ff;
extern cpuop_func op_4af0_4_nf;
extern cpuop_func op_4af0_4_ff;
extern cpuop_func op_4af8_4_nf;
extern cpuop_func op_4af8_4_ff;
extern cpuop_func op_4af9_4_nf;
extern cpuop_func op_4af9_4_ff;
extern cpuop_func op_4c90_4_nf;
extern cpuop_func op_4c90_4_ff;
extern cpuop_func op_4c98_4_nf;
extern cpuop_func op_4c98_4_ff;
extern cpuop_func op_4ca8_4_nf;
extern cpuop_func op_4ca8_4_ff;
extern cpuop_func op_4cb0_4_nf;
extern cpuop_func op_4cb0_4_ff;
extern cpuop_func op_4cb8_4_nf;
extern cpuop_func op_4cb8_4_ff;
extern cpuop_func op_4cb9_4_nf;
extern cpuop_func op_4cb9_4_ff;
extern cpuop_func op_4cba_4_nf;
extern cpuop_func op_4cba_4_ff;
extern cpuop_func op_4cbb_4_nf;
extern cpuop_func op_4cbb_4_ff;
extern cpuop_func op_4cd0_4_nf;
extern cpuop_func op_4cd0_4_ff;
extern cpuop_func op_4cd8_4_nf;
extern cpuop_func op_4cd8_4_ff;
extern cpuop_func op_4ce8_4_nf;
extern cpuop_func op_4ce8_4_ff;
extern cpuop_func op_4cf0_4_nf;
extern cpuop_func op_4cf0_4_ff;
extern cpuop_func op_4cf8_4_nf;
extern cpuop_func op_4cf8_4_ff;
extern cpuop_func op_4cf9_4_nf;
extern cpuop_func op_4cf9_4_ff;
extern cpuop_func op_4cfa_4_nf;
extern cpuop_func op_4cfa_4_ff;
extern cpuop_func op_4cfb_4_nf;
extern cpuop_func op_4cfb_4_ff;
extern cpuop_func op_4e40_4_nf;
extern cpuop_func op_4e40_4_ff;
extern cpuop_func op_4e50_4_nf;
extern cpuop_func op_4e50_4_ff;
extern cpuop_func op_4e58_4_nf;
extern cpuop_func op_4e58_4_ff;
extern cpuop_func op_4e60_4_nf;
extern cpuop_func op_4e60_4_ff;
extern cpuop_func op_4e68_4_nf;
extern cpuop_func op_4e68_4_ff;
extern cpuop_func op_4e70_4_nf;
extern cpuop_func op_4e70_4_ff;
extern cpuop_func op_4e71_4_nf;
extern cpuop_func op_4e71_4_ff;
extern cpuop_func op_4e72_4_nf;
extern cpuop_func op_4e72_4_ff;
extern cpuop_func op_4e73_4_nf;
extern cpuop_func op_4e73_4_ff;
extern cpuop_func op_4e74_4_nf;
extern cpuop_func op_4e74_4_ff;
extern cpuop_func op_4e75_4_nf;
extern cpuop_func op_4e75_4_ff;
extern cpuop_func op_4e76_4_nf;
extern cpuop_func op_4e76_4_ff;
extern cpuop_func op_4e77_4_nf;
extern cpuop_func op_4e77_4_ff;
extern cpuop_func op_4e7a_4_nf;
extern cpuop_func op_4e7a_4_ff;
extern cpuop_func op_4e7b_4_nf;
extern cpuop_func op_4e7b_4_ff;
extern cpuop_func op_4e90_4_nf;
extern cpuop_func op_4e90_4_ff;
extern cpuop_func op_4ea8_4_nf;
extern cpuop_func op_4ea8_4_ff;
extern cpuop_func op_4eb0_4_nf;
extern cpuop_func op_4eb0_4_ff;
extern cpuop_func op_4eb8_4_nf;
extern cpuop_func op_4eb8_4_ff;
extern cpuop_func op_4eb9_4_nf;
extern cpuop_func op_4eb9_4_ff;
extern cpuop_func op_4eba_4_nf;
extern cpuop_func op_4eba_4_ff;
extern cpuop_func op_4ebb_4_nf;
extern cpuop_func op_4ebb_4_ff;
extern cpuop_func op_4ed0_4_nf;
extern cpuop_func op_4ed0_4_ff;
extern cpuop_func op_4ee8_4_nf;
extern cpuop_func op_4ee8_4_ff;
extern cpuop_func op_4ef0_4_nf;
extern cpuop_func op_4ef0_4_ff;
extern cpuop_func op_4ef8_4_nf;
extern cpuop_func op_4ef8_4_ff;
extern cpuop_func op_4ef9_4_nf;
extern cpuop_func op_4ef9_4_ff;
extern cpuop_func op_4efa_4_nf;
extern cpuop_func op_4efa_4_ff;
extern cpuop_func op_4efb_4_nf;
extern cpuop_func op_4efb_4_ff;
extern cpuop_func op_5000_4_nf;
extern cpuop_func op_5000_4_ff;
extern cpuop_func op_5010_4_nf;
extern cpuop_func op_5010_4_ff;
extern cpuop_func op_5018_4_nf;
extern cpuop_func op_5018_4_ff;
extern cpuop_func op_5020_4_nf;
extern cpuop_func op_5020_4_ff;
extern cpuop_func op_5028_4_nf;
extern cpuop_func op_5028_4_ff;
extern cpuop_func op_5030_4_nf;
extern cpuop_func op_5030_4_ff;
extern cpuop_func op_5038_4_nf;
extern cpuop_func op_5038_4_ff;
extern cpuop_func op_5039_4_nf;
extern cpuop_func op_5039_4_ff;
extern cpuop_func op_5040_4_nf;
extern cpuop_func op_5040_4_ff;
extern cpuop_func op_5048_4_nf;
extern cpuop_func op_5048_4_ff;
extern cpuop_func op_5050_4_nf;
extern cpuop_func op_5050_4_ff;
extern cpuop_func op_5058_4_nf;
extern cpuop_func op_5058_4_ff;
extern cpuop_func op_5060_4_nf;
extern cpuop_func op_5060_4_ff;
extern cpuop_func op_5068_4_nf;
extern cpuop_func op_5068_4_ff;
extern cpuop_func op_5070_4_nf;
extern cpuop_func op_5070_4_ff;
extern cpuop_func op_5078_4_nf;
extern cpuop_func op_5078_4_ff;
extern cpuop_func op_5079_4_nf;
extern cpuop_func op_5079_4_ff;
extern cpuop_func op_5080_4_nf;
extern cpuop_func op_5080_4_ff;
extern cpuop_func op_5088_4_nf;
extern cpuop_func op_5088_4_ff;
extern cpuop_func op_5090_4_nf;
extern cpuop_func op_5090_4_ff;
extern cpuop_func op_5098_4_nf;
extern cpuop_func op_5098_4_ff;
extern cpuop_func op_50a0_4_nf;
extern cpuop_func op_50a0_4_ff;
extern cpuop_func op_50a8_4_nf;
extern cpuop_func op_50a8_4_ff;
extern cpuop_func op_50b0_4_nf;
extern cpuop_func op_50b0_4_ff;
extern cpuop_func op_50b8_4_nf;
extern cpuop_func op_50b8_4_ff;
extern cpuop_func op_50b9_4_nf;
extern cpuop_func op_50b9_4_ff;
extern cpuop_func op_50c0_4_nf;
extern cpuop_func op_50c0_4_ff;
extern cpuop_func op_50c8_4_nf;
extern cpuop_func op_50c8_4_ff;
extern cpuop_func op_50d0_4_nf;
extern cpuop_func op_50d0_4_ff;
extern cpuop_func op_50d8_4_nf;
extern cpuop_func op_50d8_4_ff;
extern cpuop_func op_50e0_4_nf;
extern cpuop_func op_50e0_4_ff;
extern cpuop_func op_50e8_4_nf;
extern cpuop_func op_50e8_4_ff;
extern cpuop_func op_50f0_4_nf;
extern cpuop_func op_50f0_4_ff;
extern cpuop_func op_50f8_4_nf;
extern cpuop_func op_50f8_4_ff;
extern cpuop_func op_50f9_4_nf;
extern cpuop_func op_50f9_4_ff;
extern cpuop_func op_5100_4_nf;
extern cpuop_func op_5100_4_ff;
extern cpuop_func op_5110_4_nf;
extern cpuop_func op_5110_4_ff;
extern cpuop_func op_5118_4_nf;
extern cpuop_func op_5118_4_ff;
extern cpuop_func op_5120_4_nf;
extern cpuop_func op_5120_4_ff;
extern cpuop_func op_5128_4_nf;
extern cpuop_func op_5128_4_ff;
extern cpuop_func op_5130_4_nf;
extern cpuop_func op_5130_4_ff;
extern cpuop_func op_5138_4_nf;
extern cpuop_func op_5138_4_ff;
extern cpuop_func op_5139_4_nf;
extern cpuop_func op_5139_4_ff;
extern cpuop_func op_5140_4_nf;
extern cpuop_func op_5140_4_ff;
extern cpuop_func op_5148_4_nf;
extern cpuop_func op_5148_4_ff;
extern cpuop_func op_5150_4_nf;
extern cpuop_func op_5150_4_ff;
extern cpuop_func op_5158_4_nf;
extern cpuop_func op_5158_4_ff;
extern cpuop_func op_5160_4_nf;
extern cpuop_func op_5160_4_ff;
extern cpuop_func op_5168_4_nf;
extern cpuop_func op_5168_4_ff;
extern cpuop_func op_5170_4_nf;
extern cpuop_func op_5170_4_ff;
extern cpuop_func op_5178_4_nf;
extern cpuop_func op_5178_4_ff;
extern cpuop_func op_5179_4_nf;
extern cpuop_func op_5179_4_ff;
extern cpuop_func op_5180_4_nf;
extern cpuop_func op_5180_4_ff;
extern cpuop_func op_5188_4_nf;
extern cpuop_func op_5188_4_ff;
extern cpuop_func op_5190_4_nf;
extern cpuop_func op_5190_4_ff;
extern cpuop_func op_5198_4_nf;
extern cpuop_func op_5198_4_ff;
extern cpuop_func op_51a0_4_nf;
extern cpuop_func op_51a0_4_ff;
extern cpuop_func op_51a8_4_nf;
extern cpuop_func op_51a8_4_ff;
extern cpuop_func op_51b0_4_nf;
extern cpuop_func op_51b0_4_ff;
extern cpuop_func op_51b8_4_nf;
extern cpuop_func op_51b8_4_ff;
extern cpuop_func op_51b9_4_nf;
extern cpuop_func op_51b9_4_ff;
extern cpuop_func op_51c0_4_nf;
extern cpuop_func op_51c0_4_ff;
extern cpuop_func op_51c8_4_nf;
extern cpuop_func op_51c8_4_ff;
extern cpuop_func op_51d0_4_nf;
extern cpuop_func op_51d0_4_ff;
extern cpuop_func op_51d8_4_nf;
extern cpuop_func op_51d8_4_ff;
extern cpuop_func op_51e0_4_nf;
extern cpuop_func op_51e0_4_ff;
extern cpuop_func op_51e8_4_nf;
extern cpuop_func op_51e8_4_ff;
extern cpuop_func op_51f0_4_nf;
extern cpuop_func op_51f0_4_ff;
extern cpuop_func op_51f8_4_nf;
extern cpuop_func op_51f8_4_ff;
extern cpuop_func op_51f9_4_nf;
extern cpuop_func op_51f9_4_ff;
extern cpuop_func op_52c0_4_nf;
extern cpuop_func op_52c0_4_ff;
extern cpuop_func op_52c8_4_nf;
extern cpuop_func op_52c8_4_ff;
extern cpuop_func op_52d0_4_nf;
extern cpuop_func op_52d0_4_ff;
extern cpuop_func op_52d8_4_nf;
extern cpuop_func op_52d8_4_ff;
extern cpuop_func op_52e0_4_nf;
extern cpuop_func op_52e0_4_ff;
extern cpuop_func op_52e8_4_nf;
extern cpuop_func op_52e8_4_ff;
extern cpuop_func op_52f0_4_nf;
extern cpuop_func op_52f0_4_ff;
extern cpuop_func op_52f8_4_nf;
extern cpuop_func op_52f8_4_ff;
extern cpuop_func op_52f9_4_nf;
extern cpuop_func op_52f9_4_ff;
extern cpuop_func op_53c0_4_nf;
extern cpuop_func op_53c0_4_ff;
extern cpuop_func op_53c8_4_nf;
extern cpuop_func op_53c8_4_ff;
extern cpuop_func op_53d0_4_nf;
extern cpuop_func op_53d0_4_ff;
extern cpuop_func op_53d8_4_nf;
extern cpuop_func op_53d8_4_ff;
extern cpuop_func op_53e0_4_nf;
extern cpuop_func op_53e0_4_ff;
extern cpuop_func op_53e8_4_nf;
extern cpuop_func op_53e8_4_ff;
extern cpuop_func op_53f0_4_nf;
extern cpuop_func op_53f0_4_ff;
extern cpuop_func op_53f8_4_nf;
extern cpuop_func op_53f8_4_ff;
extern cpuop_func op_53f9_4_nf;
extern cpuop_func op_53f9_4_ff;
extern cpuop_func op_54c0_4_nf;
extern cpuop_func op_54c0_4_ff;
extern cpuop_func op_54c8_4_nf;
extern cpuop_func op_54c8_4_ff;
extern cpuop_func op_54d0_4_nf;
extern cpuop_func op_54d0_4_ff;
extern cpuop_func op_54d8_4_nf;
extern cpuop_func op_54d8_4_ff;
extern cpuop_func op_54e0_4_nf;
extern cpuop_func op_54e0_4_ff;
extern cpuop_func op_54e8_4_nf;
extern cpuop_func op_54e8_4_ff;
extern cpuop_func op_54f0_4_nf;
extern cpuop_func op_54f0_4_ff;
extern cpuop_func op_54f8_4_nf;
extern cpuop_func op_54f8_4_ff;
extern cpuop_func op_54f9_4_nf;
extern cpuop_func op_54f9_4_ff;
extern cpuop_func op_55c0_4_nf;
extern cpuop_func op_55c0_4_ff;
extern cpuop_func op_55c8_4_nf;
extern cpuop_func op_55c8_4_ff;
extern cpuop_func op_55d0_4_nf;
extern cpuop_func op_55d0_4_ff;
extern cpuop_func op_55d8_4_nf;
extern cpuop_func op_55d8_4_ff;
extern cpuop_func op_55e0_4_nf;
extern cpuop_func op_55e0_4_ff;
extern cpuop_func op_55e8_4_nf;
extern cpuop_func op_55e8_4_ff;
extern cpuop_func op_55f0_4_nf;
extern cpuop_func op_55f0_4_ff;
extern cpuop_func op_55f8_4_nf;
extern cpuop_func op_55f8_4_ff;
extern cpuop_func op_55f9_4_nf;
extern cpuop_func op_55f9_4_ff;
extern cpuop_func op_56c0_4_nf;
extern cpuop_func op_56c0_4_ff;
extern cpuop_func op_56c8_4_nf;
extern cpuop_func op_56c8_4_ff;
extern cpuop_func op_56d0_4_nf;
extern cpuop_func op_56d0_4_ff;
extern cpuop_func op_56d8_4_nf;
extern cpuop_func op_56d8_4_ff;
extern cpuop_func op_56e0_4_nf;
extern cpuop_func op_56e0_4_ff;
extern cpuop_func op_56e8_4_nf;
extern cpuop_func op_56e8_4_ff;
extern cpuop_func op_56f0_4_nf;
extern cpuop_func op_56f0_4_ff;
extern cpuop_func op_56f8_4_nf;
extern cpuop_func op_56f8_4_ff;
extern cpuop_func op_56f9_4_nf;
extern cpuop_func op_56f9_4_ff;
extern cpuop_func op_57c0_4_nf;
extern cpuop_func op_57c0_4_ff;
extern cpuop_func op_57c8_4_nf;
extern cpuop_func op_57c8_4_ff;
extern cpuop_func op_57d0_4_nf;
extern cpuop_func op_57d0_4_ff;
extern cpuop_func op_57d8_4_nf;
extern cpuop_func op_57d8_4_ff;
extern cpuop_func op_57e0_4_nf;
extern cpuop_func op_57e0_4_ff;
extern cpuop_func op_57e8_4_nf;
extern cpuop_func op_57e8_4_ff;
extern cpuop_func op_57f0_4_nf;
extern cpuop_func op_57f0_4_ff;
extern cpuop_func op_57f8_4_nf;
extern cpuop_func op_57f8_4_ff;
extern cpuop_func op_57f9_4_nf;
extern cpuop_func op_57f9_4_ff;
extern cpuop_func op_58c0_4_nf;
extern cpuop_func op_58c0_4_ff;
extern cpuop_func op_58c8_4_nf;
extern cpuop_func op_58c8_4_ff;
extern cpuop_func op_58d0_4_nf;
extern cpuop_func op_58d0_4_ff;
extern cpuop_func op_58d8_4_nf;
extern cpuop_func op_58d8_4_ff;
extern cpuop_func op_58e0_4_nf;
extern cpuop_func op_58e0_4_ff;
extern cpuop_func op_58e8_4_nf;
extern cpuop_func op_58e8_4_ff;
extern cpuop_func op_58f0_4_nf;
extern cpuop_func op_58f0_4_ff;
extern cpuop_func op_58f8_4_nf;
extern cpuop_func op_58f8_4_ff;
extern cpuop_func op_58f9_4_nf;
extern cpuop_func op_58f9_4_ff;
extern cpuop_func op_59c0_4_nf;
extern cpuop_func op_59c0_4_ff;
extern cpuop_func op_59c8_4_nf;
extern cpuop_func op_59c8_4_ff;
extern cpuop_func op_59d0_4_nf;
extern cpuop_func op_59d0_4_ff;
extern cpuop_func op_59d8_4_nf;
extern cpuop_func op_59d8_4_ff;
extern cpuop_func op_59e0_4_nf;
extern cpuop_func op_59e0_4_ff;
extern cpuop_func op_59e8_4_nf;
extern cpuop_func op_59e8_4_ff;
extern cpuop_func op_59f0_4_nf;
extern cpuop_func op_59f0_4_ff;
extern cpuop_func op_59f8_4_nf;
extern cpuop_func op_59f8_4_ff;
extern cpuop_func op_59f9_4_nf;
extern cpuop_func op_59f9_4_ff;
extern cpuop_func op_5ac0_4_nf;
extern cpuop_func op_5ac0_4_ff;
extern cpuop_func op_5ac8_4_nf;
extern cpuop_func op_5ac8_4_ff;
extern cpuop_func op_5ad0_4_nf;
extern cpuop_func op_5ad0_4_ff;
extern cpuop_func op_5ad8_4_nf;
extern cpuop_func op_5ad8_4_ff;
extern cpuop_func op_5ae0_4_nf;
extern cpuop_func op_5ae0_4_ff;
extern cpuop_func op_5ae8_4_nf;
extern cpuop_func op_5ae8_4_ff;
extern cpuop_func op_5af0_4_nf;
extern cpuop_func op_5af0_4_ff;
extern cpuop_func op_5af8_4_nf;
extern cpuop_func op_5af8_4_ff;
extern cpuop_func op_5af9_4_nf;
extern cpuop_func op_5af9_4_ff;
extern cpuop_func op_5bc0_4_nf;
extern cpuop_func op_5bc0_4_ff;
extern cpuop_func op_5bc8_4_nf;
extern cpuop_func op_5bc8_4_ff;
extern cpuop_func op_5bd0_4_nf;
extern cpuop_func op_5bd0_4_ff;
extern cpuop_func op_5bd8_4_nf;
extern cpuop_func op_5bd8_4_ff;
extern cpuop_func op_5be0_4_nf;
extern cpuop_func op_5be0_4_ff;
extern cpuop_func op_5be8_4_nf;
extern cpuop_func op_5be8_4_ff;
extern cpuop_func op_5bf0_4_nf;
extern cpuop_func op_5bf0_4_ff;
extern cpuop_func op_5bf8_4_nf;
extern cpuop_func op_5bf8_4_ff;
extern cpuop_func op_5bf9_4_nf;
extern cpuop_func op_5bf9_4_ff;
extern cpuop_func op_5cc0_4_nf;
extern cpuop_func op_5cc0_4_ff;
extern cpuop_func op_5cc8_4_nf;
extern cpuop_func op_5cc8_4_ff;
extern cpuop_func op_5cd0_4_nf;
extern cpuop_func op_5cd0_4_ff;
extern cpuop_func op_5cd8_4_nf;
extern cpuop_func op_5cd8_4_ff;
extern cpuop_func op_5ce0_4_nf;
extern cpuop_func op_5ce0_4_ff;
extern cpuop_func op_5ce8_4_nf;
extern cpuop_func op_5ce8_4_ff;
extern cpuop_func op_5cf0_4_nf;
extern cpuop_func op_5cf0_4_ff;
extern cpuop_func op_5cf8_4_nf;
extern cpuop_func op_5cf8_4_ff;
extern cpuop_func op_5cf9_4_nf;
extern cpuop_func op_5cf9_4_ff;
extern cpuop_func op_5dc0_4_nf;
extern cpuop_func op_5dc0_4_ff;
extern cpuop_func op_5dc8_4_nf;
extern cpuop_func op_5dc8_4_ff;
extern cpuop_func op_5dd0_4_nf;
extern cpuop_func op_5dd0_4_ff;
extern cpuop_func op_5dd8_4_nf;
extern cpuop_func op_5dd8_4_ff;
extern cpuop_func op_5de0_4_nf;
extern cpuop_func op_5de0_4_ff;
extern cpuop_func op_5de8_4_nf;
extern cpuop_func op_5de8_4_ff;
extern cpuop_func op_5df0_4_nf;
extern cpuop_func op_5df0_4_ff;
extern cpuop_func op_5df8_4_nf;
extern cpuop_func op_5df8_4_ff;
extern cpuop_func op_5df9_4_nf;
extern cpuop_func op_5df9_4_ff;
extern cpuop_func op_5ec0_4_nf;
extern cpuop_func op_5ec0_4_ff;
extern cpuop_func op_5ec8_4_nf;
extern cpuop_func op_5ec8_4_ff;
extern cpuop_func op_5ed0_4_nf;
extern cpuop_func op_5ed0_4_ff;
extern cpuop_func op_5ed8_4_nf;
extern cpuop_func op_5ed8_4_ff;
extern cpuop_func op_5ee0_4_nf;
extern cpuop_func op_5ee0_4_ff;
extern cpuop_func op_5ee8_4_nf;
extern cpuop_func op_5ee8_4_ff;
extern cpuop_func op_5ef0_4_nf;
extern cpuop_func op_5ef0_4_ff;
extern cpuop_func op_5ef8_4_nf;
extern cpuop_func op_5ef8_4_ff;
extern cpuop_func op_5ef9_4_nf;
extern cpuop_func op_5ef9_4_ff;
extern cpuop_func op_5fc0_4_nf;
extern cpuop_func op_5fc0_4_ff;
extern cpuop_func op_5fc8_4_nf;
extern cpuop_func op_5fc8_4_ff;
extern cpuop_func op_5fd0_4_nf;
extern cpuop_func op_5fd0_4_ff;
extern cpuop_func op_5fd8_4_nf;
extern cpuop_func op_5fd8_4_ff;
extern cpuop_func op_5fe0_4_nf;
extern cpuop_func op_5fe0_4_ff;
extern cpuop_func op_5fe8_4_nf;
extern cpuop_func op_5fe8_4_ff;
extern cpuop_func op_5ff0_4_nf;
extern cpuop_func op_5ff0_4_ff;
extern cpuop_func op_5ff8_4_nf;
extern cpuop_func op_5ff8_4_ff;
extern cpuop_func op_5ff9_4_nf;
extern cpuop_func op_5ff9_4_ff;
extern cpuop_func op_6000_4_nf;
extern cpuop_func op_6000_4_ff;
extern cpuop_func op_6001_4_nf;
extern cpuop_func op_6001_4_ff;
extern cpuop_func op_60ff_4_nf;
extern cpuop_func op_60ff_4_ff;
extern cpuop_func op_6100_4_nf;
extern cpuop_func op_6100_4_ff;
extern cpuop_func op_6101_4_nf;
extern cpuop_func op_6101_4_ff;
extern cpuop_func op_61ff_4_nf;
extern cpuop_func op_61ff_4_ff;
extern cpuop_func op_6200_4_nf;
extern cpuop_func op_6200_4_ff;
extern cpuop_func op_6201_4_nf;
extern cpuop_func op_6201_4_ff;
extern cpuop_func op_62ff_4_nf;
extern cpuop_func op_62ff_4_ff;
extern cpuop_func op_6300_4_nf;
extern cpuop_func op_6300_4_ff;
extern cpuop_func op_6301_4_nf;
extern cpuop_func op_6301_4_ff;
extern cpuop_func op_63ff_4_nf;
extern cpuop_func op_63ff_4_ff;
extern cpuop_func op_6400_4_nf;
extern cpuop_func op_6400_4_ff;
extern cpuop_func op_6401_4_nf;
extern cpuop_func op_6401_4_ff;
extern cpuop_func op_64ff_4_nf;
extern cpuop_func op_64ff_4_ff;
extern cpuop_func op_6500_4_nf;
extern cpuop_func op_6500_4_ff;
extern cpuop_func op_6501_4_nf;
extern cpuop_func op_6501_4_ff;
extern cpuop_func op_65ff_4_nf;
extern cpuop_func op_65ff_4_ff;
extern cpuop_func op_6600_4_nf;
extern cpuop_func op_6600_4_ff;
extern cpuop_func op_6601_4_nf;
extern cpuop_func op_6601_4_ff;
extern cpuop_func op_66ff_4_nf;
extern cpuop_func op_66ff_4_ff;
extern cpuop_func op_6700_4_nf;
extern cpuop_func op_6700_4_ff;
extern cpuop_func op_6701_4_nf;
extern cpuop_func op_6701_4_ff;
extern cpuop_func op_67ff_4_nf;
extern cpuop_func op_67ff_4_ff;
extern cpuop_func op_6800_4_nf;
extern cpuop_func op_6800_4_ff;
extern cpuop_func op_6801_4_nf;
extern cpuop_func op_6801_4_ff;
extern cpuop_func op_68ff_4_nf;
extern cpuop_func op_68ff_4_ff;
extern cpuop_func op_6900_4_nf;
extern cpuop_func op_6900_4_ff;
extern cpuop_func op_6901_4_nf;
extern cpuop_func op_6901_4_ff;
extern cpuop_func op_69ff_4_nf;
extern cpuop_func op_69ff_4_ff;
extern cpuop_func op_6a00_4_nf;
extern cpuop_func op_6a00_4_ff;
extern cpuop_func op_6a01_4_nf;
extern cpuop_func op_6a01_4_ff;
extern cpuop_func op_6aff_4_nf;
extern cpuop_func op_6aff_4_ff;
extern cpuop_func op_6b00_4_nf;
extern cpuop_func op_6b00_4_ff;
extern cpuop_func op_6b01_4_nf;
extern cpuop_func op_6b01_4_ff;
extern cpuop_func op_6bff_4_nf;
extern cpuop_func op_6bff_4_ff;
extern cpuop_func op_6c00_4_nf;
extern cpuop_func op_6c00_4_ff;
extern cpuop_func op_6c01_4_nf;
extern cpuop_func op_6c01_4_ff;
extern cpuop_func op_6cff_4_nf;
extern cpuop_func op_6cff_4_ff;
extern cpuop_func op_6d00_4_nf;
extern cpuop_func op_6d00_4_ff;
extern cpuop_func op_6d01_4_nf;
extern cpuop_func op_6d01_4_ff;
extern cpuop_func op_6dff_4_nf;
extern cpuop_func op_6dff_4_ff;
extern cpuop_func op_6e00_4_nf;
extern cpuop_func op_6e00_4_ff;
extern cpuop_func op_6e01_4_nf;
extern cpuop_func op_6e01_4_ff;
extern cpuop_func op_6eff_4_nf;
extern cpuop_func op_6eff_4_ff;
extern cpuop_func op_6f00_4_nf;
extern cpuop_func op_6f00_4_ff;
extern cpuop_func op_6f01_4_nf;
extern cpuop_func op_6f01_4_ff;
extern cpuop_func op_6fff_4_nf;
extern cpuop_func op_6fff_4_ff;
extern cpuop_func op_7000_4_nf;
extern cpuop_func op_7000_4_ff;
extern cpuop_func op_8000_4_nf;
extern cpuop_func op_8000_4_ff;
extern cpuop_func op_8010_4_nf;
extern cpuop_func op_8010_4_ff;
extern cpuop_func op_8018_4_nf;
extern cpuop_func op_8018_4_ff;
extern cpuop_func op_8020_4_nf;
extern cpuop_func op_8020_4_ff;
extern cpuop_func op_8028_4_nf;
extern cpuop_func op_8028_4_ff;
extern cpuop_func op_8030_4_nf;
extern cpuop_func op_8030_4_ff;
extern cpuop_func op_8038_4_nf;
extern cpuop_func op_8038_4_ff;
extern cpuop_func op_8039_4_nf;
extern cpuop_func op_8039_4_ff;
extern cpuop_func op_803a_4_nf;
extern cpuop_func op_803a_4_ff;
extern cpuop_func op_803b_4_nf;
extern cpuop_func op_803b_4_ff;
extern cpuop_func op_803c_4_nf;
extern cpuop_func op_803c_4_ff;
extern cpuop_func op_8040_4_nf;
extern cpuop_func op_8040_4_ff;
extern cpuop_func op_8050_4_nf;
extern cpuop_func op_8050_4_ff;
extern cpuop_func op_8058_4_nf;
extern cpuop_func op_8058_4_ff;
extern cpuop_func op_8060_4_nf;
extern cpuop_func op_8060_4_ff;
extern cpuop_func op_8068_4_nf;
extern cpuop_func op_8068_4_ff;
extern cpuop_func op_8070_4_nf;
extern cpuop_func op_8070_4_ff;
extern cpuop_func op_8078_4_nf;
extern cpuop_func op_8078_4_ff;
extern cpuop_func op_8079_4_nf;
extern cpuop_func op_8079_4_ff;
extern cpuop_func op_807a_4_nf;
extern cpuop_func op_807a_4_ff;
extern cpuop_func op_807b_4_nf;
extern cpuop_func op_807b_4_ff;
extern cpuop_func op_807c_4_nf;
extern cpuop_func op_807c_4_ff;
extern cpuop_func op_8080_4_nf;
extern cpuop_func op_8080_4_ff;
extern cpuop_func op_8090_4_nf;
extern cpuop_func op_8090_4_ff;
extern cpuop_func op_8098_4_nf;
extern cpuop_func op_8098_4_ff;
extern cpuop_func op_80a0_4_nf;
extern cpuop_func op_80a0_4_ff;
extern cpuop_func op_80a8_4_nf;
extern cpuop_func op_80a8_4_ff;
extern cpuop_func op_80b0_4_nf;
extern cpuop_func op_80b0_4_ff;
extern cpuop_func op_80b8_4_nf;
extern cpuop_func op_80b8_4_ff;
extern cpuop_func op_80b9_4_nf;
extern cpuop_func op_80b9_4_ff;
extern cpuop_func op_80ba_4_nf;
extern cpuop_func op_80ba_4_ff;
extern cpuop_func op_80bb_4_nf;
extern cpuop_func op_80bb_4_ff;
extern cpuop_func op_80bc_4_nf;
extern cpuop_func op_80bc_4_ff;
extern cpuop_func op_80c0_4_nf;
extern cpuop_func op_80c0_4_ff;
extern cpuop_func op_80d0_4_nf;
extern cpuop_func op_80d0_4_ff;
extern cpuop_func op_80d8_4_nf;
extern cpuop_func op_80d8_4_ff;
extern cpuop_func op_80e0_4_nf;
extern cpuop_func op_80e0_4_ff;
extern cpuop_func op_80e8_4_nf;
extern cpuop_func op_80e8_4_ff;
extern cpuop_func op_80f0_4_nf;
extern cpuop_func op_80f0_4_ff;
extern cpuop_func op_80f8_4_nf;
extern cpuop_func op_80f8_4_ff;
extern cpuop_func op_80f9_4_nf;
extern cpuop_func op_80f9_4_ff;
extern cpuop_func op_80fa_4_nf;
extern cpuop_func op_80fa_4_ff;
extern cpuop_func op_80fb_4_nf;
extern cpuop_func op_80fb_4_ff;
extern cpuop_func op_80fc_4_nf;
extern cpuop_func op_80fc_4_ff;
extern cpuop_func op_8100_4_nf;
extern cpuop_func op_8100_4_ff;
extern cpuop_func op_8108_4_nf;
extern cpuop_func op_8108_4_ff;
extern cpuop_func op_8110_4_nf;
extern cpuop_func op_8110_4_ff;
extern cpuop_func op_8118_4_nf;
extern cpuop_func op_8118_4_ff;
extern cpuop_func op_8120_4_nf;
extern cpuop_func op_8120_4_ff;
extern cpuop_func op_8128_4_nf;
extern cpuop_func op_8128_4_ff;
extern cpuop_func op_8130_4_nf;
extern cpuop_func op_8130_4_ff;
extern cpuop_func op_8138_4_nf;
extern cpuop_func op_8138_4_ff;
extern cpuop_func op_8139_4_nf;
extern cpuop_func op_8139_4_ff;
extern cpuop_func op_8150_4_nf;
extern cpuop_func op_8150_4_ff;
extern cpuop_func op_8158_4_nf;
extern cpuop_func op_8158_4_ff;
extern cpuop_func op_8160_4_nf;
extern cpuop_func op_8160_4_ff;
extern cpuop_func op_8168_4_nf;
extern cpuop_func op_8168_4_ff;
extern cpuop_func op_8170_4_nf;
extern cpuop_func op_8170_4_ff;
extern cpuop_func op_8178_4_nf;
extern cpuop_func op_8178_4_ff;
extern cpuop_func op_8179_4_nf;
extern cpuop_func op_8179_4_ff;
extern cpuop_func op_8190_4_nf;
extern cpuop_func op_8190_4_ff;
extern cpuop_func op_8198_4_nf;
extern cpuop_func op_8198_4_ff;
extern cpuop_func op_81a0_4_nf;
extern cpuop_func op_81a0_4_ff;
extern cpuop_func op_81a8_4_nf;
extern cpuop_func op_81a8_4_ff;
extern cpuop_func op_81b0_4_nf;
extern cpuop_func op_81b0_4_ff;
extern cpuop_func op_81b8_4_nf;
extern cpuop_func op_81b8_4_ff;
extern cpuop_func op_81b9_4_nf;
extern cpuop_func op_81b9_4_ff;
extern cpuop_func op_81c0_4_nf;
extern cpuop_func op_81c0_4_ff;
extern cpuop_func op_81d0_4_nf;
extern cpuop_func op_81d0_4_ff;
extern cpuop_func op_81d8_4_nf;
extern cpuop_func op_81d8_4_ff;
extern cpuop_func op_81e0_4_nf;
extern cpuop_func op_81e0_4_ff;
extern cpuop_func op_81e8_4_nf;
extern cpuop_func op_81e8_4_ff;
extern cpuop_func op_81f0_4_nf;
extern cpuop_func op_81f0_4_ff;
extern cpuop_func op_81f8_4_nf;
extern cpuop_func op_81f8_4_ff;
extern cpuop_func op_81f9_4_nf;
extern cpuop_func op_81f9_4_ff;
extern cpuop_func op_81fa_4_nf;
extern cpuop_func op_81fa_4_ff;
extern cpuop_func op_81fb_4_nf;
extern cpuop_func op_81fb_4_ff;
extern cpuop_func op_81fc_4_nf;
extern cpuop_func op_81fc_4_ff;
extern cpuop_func op_9000_4_nf;
extern cpuop_func op_9000_4_ff;
extern cpuop_func op_9010_4_nf;
extern cpuop_func op_9010_4_ff;
extern cpuop_func op_9018_4_nf;
extern cpuop_func op_9018_4_ff;
extern cpuop_func op_9020_4_nf;
extern cpuop_func op_9020_4_ff;
extern cpuop_func op_9028_4_nf;
extern cpuop_func op_9028_4_ff;
extern cpuop_func op_9030_4_nf;
extern cpuop_func op_9030_4_ff;
extern cpuop_func op_9038_4_nf;
extern cpuop_func op_9038_4_ff;
extern cpuop_func op_9039_4_nf;
extern cpuop_func op_9039_4_ff;
extern cpuop_func op_903a_4_nf;
extern cpuop_func op_903a_4_ff;
extern cpuop_func op_903b_4_nf;
extern cpuop_func op_903b_4_ff;
extern cpuop_func op_903c_4_nf;
extern cpuop_func op_903c_4_ff;
extern cpuop_func op_9040_4_nf;
extern cpuop_func op_9040_4_ff;
extern cpuop_func op_9048_4_nf;
extern cpuop_func op_9048_4_ff;
extern cpuop_func op_9050_4_nf;
extern cpuop_func op_9050_4_ff;
extern cpuop_func op_9058_4_nf;
extern cpuop_func op_9058_4_ff;
extern cpuop_func op_9060_4_nf;
extern cpuop_func op_9060_4_ff;
extern cpuop_func op_9068_4_nf;
extern cpuop_func op_9068_4_ff;
extern cpuop_func op_9070_4_nf;
extern cpuop_func op_9070_4_ff;
extern cpuop_func op_9078_4_nf;
extern cpuop_func op_9078_4_ff;
extern cpuop_func op_9079_4_nf;
extern cpuop_func op_9079_4_ff;
extern cpuop_func op_907a_4_nf;
extern cpuop_func op_907a_4_ff;
extern cpuop_func op_907b_4_nf;
extern cpuop_func op_907b_4_ff;
extern cpuop_func op_907c_4_nf;
extern cpuop_func op_907c_4_ff;
extern cpuop_func op_9080_4_nf;
extern cpuop_func op_9080_4_ff;
extern cpuop_func op_9088_4_nf;
extern cpuop_func op_9088_4_ff;
extern cpuop_func op_9090_4_nf;
extern cpuop_func op_9090_4_ff;
extern cpuop_func op_9098_4_nf;
extern cpuop_func op_9098_4_ff;
extern cpuop_func op_90a0_4_nf;
extern cpuop_func op_90a0_4_ff;
extern cpuop_func op_90a8_4_nf;
extern cpuop_func op_90a8_4_ff;
extern cpuop_func op_90b0_4_nf;
extern cpuop_func op_90b0_4_ff;
extern cpuop_func op_90b8_4_nf;
extern cpuop_func op_90b8_4_ff;
extern cpuop_func op_90b9_4_nf;
extern cpuop_func op_90b9_4_ff;
extern cpuop_func op_90ba_4_nf;
extern cpuop_func op_90ba_4_ff;
extern cpuop_func op_90bb_4_nf;
extern cpuop_func op_90bb_4_ff;
extern cpuop_func op_90bc_4_nf;
extern cpuop_func op_90bc_4_ff;
extern cpuop_func op_90c0_4_nf;
extern cpuop_func op_90c0_4_ff;
extern cpuop_func op_90c8_4_nf;
extern cpuop_func op_90c8_4_ff;
extern cpuop_func op_90d0_4_nf;
extern cpuop_func op_90d0_4_ff;
extern cpuop_func op_90d8_4_nf;
extern cpuop_func op_90d8_4_ff;
extern cpuop_func op_90e0_4_nf;
extern cpuop_func op_90e0_4_ff;
extern cpuop_func op_90e8_4_nf;
extern cpuop_func op_90e8_4_ff;
extern cpuop_func op_90f0_4_nf;
extern cpuop_func op_90f0_4_ff;
extern cpuop_func op_90f8_4_nf;
extern cpuop_func op_90f8_4_ff;
extern cpuop_func op_90f9_4_nf;
extern cpuop_func op_90f9_4_ff;
extern cpuop_func op_90fa_4_nf;
extern cpuop_func op_90fa_4_ff;
extern cpuop_func op_90fb_4_nf;
extern cpuop_func op_90fb_4_ff;
extern cpuop_func op_90fc_4_nf;
extern cpuop_func op_90fc_4_ff;
extern cpuop_func op_9100_4_nf;
extern cpuop_func op_9100_4_ff;
extern cpuop_func op_9108_4_nf;
extern cpuop_func op_9108_4_ff;
extern cpuop_func op_9110_4_nf;
extern cpuop_func op_9110_4_ff;
extern cpuop_func op_9118_4_nf;
extern cpuop_func op_9118_4_ff;
extern cpuop_func op_9120_4_nf;
extern cpuop_func op_9120_4_ff;
extern cpuop_func op_9128_4_nf;
extern cpuop_func op_9128_4_ff;
extern cpuop_func op_9130_4_nf;
extern cpuop_func op_9130_4_ff;
extern cpuop_func op_9138_4_nf;
extern cpuop_func op_9138_4_ff;
extern cpuop_func op_9139_4_nf;
extern cpuop_func op_9139_4_ff;
extern cpuop_func op_9140_4_nf;
extern cpuop_func op_9140_4_ff;
extern cpuop_func op_9148_4_nf;
extern cpuop_func op_9148_4_ff;
extern cpuop_func op_9150_4_nf;
extern cpuop_func op_9150_4_ff;
extern cpuop_func op_9158_4_nf;
extern cpuop_func op_9158_4_ff;
extern cpuop_func op_9160_4_nf;
extern cpuop_func op_9160_4_ff;
extern cpuop_func op_9168_4_nf;
extern cpuop_func op_9168_4_ff;
extern cpuop_func op_9170_4_nf;
extern cpuop_func op_9170_4_ff;
extern cpuop_func op_9178_4_nf;
extern cpuop_func op_9178_4_ff;
extern cpuop_func op_9179_4_nf;
extern cpuop_func op_9179_4_ff;
extern cpuop_func op_9180_4_nf;
extern cpuop_func op_9180_4_ff;
extern cpuop_func op_9188_4_nf;
extern cpuop_func op_9188_4_ff;
extern cpuop_func op_9190_4_nf;
extern cpuop_func op_9190_4_ff;
extern cpuop_func op_9198_4_nf;
extern cpuop_func op_9198_4_ff;
extern cpuop_func op_91a0_4_nf;
extern cpuop_func op_91a0_4_ff;
extern cpuop_func op_91a8_4_nf;
extern cpuop_func op_91a8_4_ff;
extern cpuop_func op_91b0_4_nf;
extern cpuop_func op_91b0_4_ff;
extern cpuop_func op_91b8_4_nf;
extern cpuop_func op_91b8_4_ff;
extern cpuop_func op_91b9_4_nf;
extern cpuop_func op_91b9_4_ff;
extern cpuop_func op_91c0_4_nf;
extern cpuop_func op_91c0_4_ff;
extern cpuop_func op_91c8_4_nf;
extern cpuop_func op_91c8_4_ff;
extern cpuop_func op_91d0_4_nf;
extern cpuop_func op_91d0_4_ff;
extern cpuop_func op_91d8_4_nf;
extern cpuop_func op_91d8_4_ff;
extern cpuop_func op_91e0_4_nf;
extern cpuop_func op_91e0_4_ff;
extern cpuop_func op_91e8_4_nf;
extern cpuop_func op_91e8_4_ff;
extern cpuop_func op_91f0_4_nf;
extern cpuop_func op_91f0_4_ff;
extern cpuop_func op_91f8_4_nf;
extern cpuop_func op_91f8_4_ff;
extern cpuop_func op_91f9_4_nf;
extern cpuop_func op_91f9_4_ff;
extern cpuop_func op_91fa_4_nf;
extern cpuop_func op_91fa_4_ff;
extern cpuop_func op_91fb_4_nf;
extern cpuop_func op_91fb_4_ff;
extern cpuop_func op_91fc_4_nf;
extern cpuop_func op_91fc_4_ff;
extern cpuop_func op_b000_4_nf;
extern cpuop_func op_b000_4_ff;
extern cpuop_func op_b010_4_nf;
extern cpuop_func op_b010_4_ff;
extern cpuop_func op_b018_4_nf;
extern cpuop_func op_b018_4_ff;
extern cpuop_func op_b020_4_nf;
extern cpuop_func op_b020_4_ff;
extern cpuop_func op_b028_4_nf;
extern cpuop_func op_b028_4_ff;
extern cpuop_func op_b030_4_nf;
extern cpuop_func op_b030_4_ff;
extern cpuop_func op_b038_4_nf;
extern cpuop_func op_b038_4_ff;
extern cpuop_func op_b039_4_nf;
extern cpuop_func op_b039_4_ff;
extern cpuop_func op_b03a_4_nf;
extern cpuop_func op_b03a_4_ff;
extern cpuop_func op_b03b_4_nf;
extern cpuop_func op_b03b_4_ff;
extern cpuop_func op_b03c_4_nf;
extern cpuop_func op_b03c_4_ff;
extern cpuop_func op_b040_4_nf;
extern cpuop_func op_b040_4_ff;
extern cpuop_func op_b048_4_nf;
extern cpuop_func op_b048_4_ff;
extern cpuop_func op_b050_4_nf;
extern cpuop_func op_b050_4_ff;
extern cpuop_func op_b058_4_nf;
extern cpuop_func op_b058_4_ff;
extern cpuop_func op_b060_4_nf;
extern cpuop_func op_b060_4_ff;
extern cpuop_func op_b068_4_nf;
extern cpuop_func op_b068_4_ff;
extern cpuop_func op_b070_4_nf;
extern cpuop_func op_b070_4_ff;
extern cpuop_func op_b078_4_nf;
extern cpuop_func op_b078_4_ff;
extern cpuop_func op_b079_4_nf;
extern cpuop_func op_b079_4_ff;
extern cpuop_func op_b07a_4_nf;
extern cpuop_func op_b07a_4_ff;
extern cpuop_func op_b07b_4_nf;
extern cpuop_func op_b07b_4_ff;
extern cpuop_func op_b07c_4_nf;
extern cpuop_func op_b07c_4_ff;
extern cpuop_func op_b080_4_nf;
extern cpuop_func op_b080_4_ff;
extern cpuop_func op_b088_4_nf;
extern cpuop_func op_b088_4_ff;
extern cpuop_func op_b090_4_nf;
extern cpuop_func op_b090_4_ff;
extern cpuop_func op_b098_4_nf;
extern cpuop_func op_b098_4_ff;
extern cpuop_func op_b0a0_4_nf;
extern cpuop_func op_b0a0_4_ff;
extern cpuop_func op_b0a8_4_nf;
extern cpuop_func op_b0a8_4_ff;
extern cpuop_func op_b0b0_4_nf;
extern cpuop_func op_b0b0_4_ff;
extern cpuop_func op_b0b8_4_nf;
extern cpuop_func op_b0b8_4_ff;
extern cpuop_func op_b0b9_4_nf;
extern cpuop_func op_b0b9_4_ff;
extern cpuop_func op_b0ba_4_nf;
extern cpuop_func op_b0ba_4_ff;
extern cpuop_func op_b0bb_4_nf;
extern cpuop_func op_b0bb_4_ff;
extern cpuop_func op_b0bc_4_nf;
extern cpuop_func op_b0bc_4_ff;
extern cpuop_func op_b0c0_4_nf;
extern cpuop_func op_b0c0_4_ff;
extern cpuop_func op_b0c8_4_nf;
extern cpuop_func op_b0c8_4_ff;
extern cpuop_func op_b0d0_4_nf;
extern cpuop_func op_b0d0_4_ff;
extern cpuop_func op_b0d8_4_nf;
extern cpuop_func op_b0d8_4_ff;
extern cpuop_func op_b0e0_4_nf;
extern cpuop_func op_b0e0_4_ff;
extern cpuop_func op_b0e8_4_nf;
extern cpuop_func op_b0e8_4_ff;
extern cpuop_func op_b0f0_4_nf;
extern cpuop_func op_b0f0_4_ff;
extern cpuop_func op_b0f8_4_nf;
extern cpuop_func op_b0f8_4_ff;
extern cpuop_func op_b0f9_4_nf;
extern cpuop_func op_b0f9_4_ff;
extern cpuop_func op_b0fa_4_nf;
extern cpuop_func op_b0fa_4_ff;
extern cpuop_func op_b0fb_4_nf;
extern cpuop_func op_b0fb_4_ff;
extern cpuop_func op_b0fc_4_nf;
extern cpuop_func op_b0fc_4_ff;
extern cpuop_func op_b100_4_nf;
extern cpuop_func op_b100_4_ff;
extern cpuop_func op_b108_4_nf;
extern cpuop_func op_b108_4_ff;
extern cpuop_func op_b110_4_nf;
extern cpuop_func op_b110_4_ff;
extern cpuop_func op_b118_4_nf;
extern cpuop_func op_b118_4_ff;
extern cpuop_func op_b120_4_nf;
extern cpuop_func op_b120_4_ff;
extern cpuop_func op_b128_4_nf;
extern cpuop_func op_b128_4_ff;
extern cpuop_func op_b130_4_nf;
extern cpuop_func op_b130_4_ff;
extern cpuop_func op_b138_4_nf;
extern cpuop_func op_b138_4_ff;
extern cpuop_func op_b139_4_nf;
extern cpuop_func op_b139_4_ff;
extern cpuop_func op_b140_4_nf;
extern cpuop_func op_b140_4_ff;
extern cpuop_func op_b148_4_nf;
extern cpuop_func op_b148_4_ff;
extern cpuop_func op_b150_4_nf;
extern cpuop_func op_b150_4_ff;
extern cpuop_func op_b158_4_nf;
extern cpuop_func op_b158_4_ff;
extern cpuop_func op_b160_4_nf;
extern cpuop_func op_b160_4_ff;
extern cpuop_func op_b168_4_nf;
extern cpuop_func op_b168_4_ff;
extern cpuop_func op_b170_4_nf;
extern cpuop_func op_b170_4_ff;
extern cpuop_func op_b178_4_nf;
extern cpuop_func op_b178_4_ff;
extern cpuop_func op_b179_4_nf;
extern cpuop_func op_b179_4_ff;
extern cpuop_func op_b180_4_nf;
extern cpuop_func op_b180_4_ff;
extern cpuop_func op_b188_4_nf;
extern cpuop_func op_b188_4_ff;
extern cpuop_func op_b190_4_nf;
extern cpuop_func op_b190_4_ff;
extern cpuop_func op_b198_4_nf;
extern cpuop_func op_b198_4_ff;
extern cpuop_func op_b1a0_4_nf;
extern cpuop_func op_b1a0_4_ff;
extern cpuop_func op_b1a8_4_nf;
extern cpuop_func op_b1a8_4_ff;
extern cpuop_func op_b1b0_4_nf;
extern cpuop_func op_b1b0_4_ff;
extern cpuop_func op_b1b8_4_nf;
extern cpuop_func op_b1b8_4_ff;
extern cpuop_func op_b1b9_4_nf;
extern cpuop_func op_b1b9_4_ff;
extern cpuop_func op_b1c0_4_nf;
extern cpuop_func op_b1c0_4_ff;
extern cpuop_func op_b1c8_4_nf;
extern cpuop_func op_b1c8_4_ff;
extern cpuop_func op_b1d0_4_nf;
extern cpuop_func op_b1d0_4_ff;
extern cpuop_func op_b1d8_4_nf;
extern cpuop_func op_b1d8_4_ff;
extern cpuop_func op_b1e0_4_nf;
extern cpuop_func op_b1e0_4_ff;
extern cpuop_func op_b1e8_4_nf;
extern cpuop_func op_b1e8_4_ff;
extern cpuop_func op_b1f0_4_nf;
extern cpuop_func op_b1f0_4_ff;
extern cpuop_func op_b1f8_4_nf;
extern cpuop_func op_b1f8_4_ff;
extern cpuop_func op_b1f9_4_nf;
extern cpuop_func op_b1f9_4_ff;
extern cpuop_func op_b1fa_4_nf;
extern cpuop_func op_b1fa_4_ff;
extern cpuop_func op_b1fb_4_nf;
extern cpuop_func op_b1fb_4_ff;
extern cpuop_func op_b1fc_4_nf;
extern cpuop_func op_b1fc_4_ff;
extern cpuop_func op_c000_4_nf;
extern cpuop_func op_c000_4_ff;
extern cpuop_func op_c010_4_nf;
extern cpuop_func op_c010_4_ff;
extern cpuop_func op_c018_4_nf;
extern cpuop_func op_c018_4_ff;
extern cpuop_func op_c020_4_nf;
extern cpuop_func op_c020_4_ff;
extern cpuop_func op_c028_4_nf;
extern cpuop_func op_c028_4_ff;
extern cpuop_func op_c030_4_nf;
extern cpuop_func op_c030_4_ff;
extern cpuop_func op_c038_4_nf;
extern cpuop_func op_c038_4_ff;
extern cpuop_func op_c039_4_nf;
extern cpuop_func op_c039_4_ff;
extern cpuop_func op_c03a_4_nf;
extern cpuop_func op_c03a_4_ff;
extern cpuop_func op_c03b_4_nf;
extern cpuop_func op_c03b_4_ff;
extern cpuop_func op_c03c_4_nf;
extern cpuop_func op_c03c_4_ff;
extern cpuop_func op_c040_4_nf;
extern cpuop_func op_c040_4_ff;
extern cpuop_func op_c050_4_nf;
extern cpuop_func op_c050_4_ff;
extern cpuop_func op_c058_4_nf;
extern cpuop_func op_c058_4_ff;
extern cpuop_func op_c060_4_nf;
extern cpuop_func op_c060_4_ff;
extern cpuop_func op_c068_4_nf;
extern cpuop_func op_c068_4_ff;
extern cpuop_func op_c070_4_nf;
extern cpuop_func op_c070_4_ff;
extern cpuop_func op_c078_4_nf;
extern cpuop_func op_c078_4_ff;
extern cpuop_func op_c079_4_nf;
extern cpuop_func op_c079_4_ff;
extern cpuop_func op_c07a_4_nf;
extern cpuop_func op_c07a_4_ff;
extern cpuop_func op_c07b_4_nf;
extern cpuop_func op_c07b_4_ff;
extern cpuop_func op_c07c_4_nf;
extern cpuop_func op_c07c_4_ff;
extern cpuop_func op_c080_4_nf;
extern cpuop_func op_c080_4_ff;
extern cpuop_func op_c090_4_nf;
extern cpuop_func op_c090_4_ff;
extern cpuop_func op_c098_4_nf;
extern cpuop_func op_c098_4_ff;
extern cpuop_func op_c0a0_4_nf;
extern cpuop_func op_c0a0_4_ff;
extern cpuop_func op_c0a8_4_nf;
extern cpuop_func op_c0a8_4_ff;
extern cpuop_func op_c0b0_4_nf;
extern cpuop_func op_c0b0_4_ff;
extern cpuop_func op_c0b8_4_nf;
extern cpuop_func op_c0b8_4_ff;
extern cpuop_func op_c0b9_4_nf;
extern cpuop_func op_c0b9_4_ff;
extern cpuop_func op_c0ba_4_nf;
extern cpuop_func op_c0ba_4_ff;
extern cpuop_func op_c0bb_4_nf;
extern cpuop_func op_c0bb_4_ff;
extern cpuop_func op_c0bc_4_nf;
extern cpuop_func op_c0bc_4_ff;
extern cpuop_func op_c0c0_4_nf;
extern cpuop_func op_c0c0_4_ff;
extern cpuop_func op_c0d0_4_nf;
extern cpuop_func op_c0d0_4_ff;
extern cpuop_func op_c0d8_4_nf;
extern cpuop_func op_c0d8_4_ff;
extern cpuop_func op_c0e0_4_nf;
extern cpuop_func op_c0e0_4_ff;
extern cpuop_func op_c0e8_4_nf;
extern cpuop_func op_c0e8_4_ff;
extern cpuop_func op_c0f0_4_nf;
extern cpuop_func op_c0f0_4_ff;
extern cpuop_func op_c0f8_4_nf;
extern cpuop_func op_c0f8_4_ff;
extern cpuop_func op_c0f9_4_nf;
extern cpuop_func op_c0f9_4_ff;
extern cpuop_func op_c0fa_4_nf;
extern cpuop_func op_c0fa_4_ff;
extern cpuop_func op_c0fb_4_nf;
extern cpuop_func op_c0fb_4_ff;
extern cpuop_func op_c0fc_4_nf;
extern cpuop_func op_c0fc_4_ff;
extern cpuop_func op_c100_4_nf;
extern cpuop_func op_c100_4_ff;
extern cpuop_func op_c108_4_nf;
extern cpuop_func op_c108_4_ff;
extern cpuop_func op_c110_4_nf;
extern cpuop_func op_c110_4_ff;
extern cpuop_func op_c118_4_nf;
extern cpuop_func op_c118_4_ff;
extern cpuop_func op_c120_4_nf;
extern cpuop_func op_c120_4_ff;
extern cpuop_func op_c128_4_nf;
extern cpuop_func op_c128_4_ff;
extern cpuop_func op_c130_4_nf;
extern cpuop_func op_c130_4_ff;
extern cpuop_func op_c138_4_nf;
extern cpuop_func op_c138_4_ff;
extern cpuop_func op_c139_4_nf;
extern cpuop_func op_c139_4_ff;
extern cpuop_func op_c140_4_nf;
extern cpuop_func op_c140_4_ff;
extern cpuop_func op_c148_4_nf;
extern cpuop_func op_c148_4_ff;
extern cpuop_func op_c150_4_nf;
extern cpuop_func op_c150_4_ff;
extern cpuop_func op_c158_4_nf;
extern cpuop_func op_c158_4_ff;
extern cpuop_func op_c160_4_nf;
extern cpuop_func op_c160_4_ff;
extern cpuop_func op_c168_4_nf;
extern cpuop_func op_c168_4_ff;
extern cpuop_func op_c170_4_nf;
extern cpuop_func op_c170_4_ff;
extern cpuop_func op_c178_4_nf;
extern cpuop_func op_c178_4_ff;
extern cpuop_func op_c179_4_nf;
extern cpuop_func op_c179_4_ff;
extern cpuop_func op_c188_4_nf;
extern cpuop_func op_c188_4_ff;
extern cpuop_func op_c190_4_nf;
extern cpuop_func op_c190_4_ff;
extern cpuop_func op_c198_4_nf;
extern cpuop_func op_c198_4_ff;
extern cpuop_func op_c1a0_4_nf;
extern cpuop_func op_c1a0_4_ff;
extern cpuop_func op_c1a8_4_nf;
extern cpuop_func op_c1a8_4_ff;
extern cpuop_func op_c1b0_4_nf;
extern cpuop_func op_c1b0_4_ff;
extern cpuop_func op_c1b8_4_nf;
extern cpuop_func op_c1b8_4_ff;
extern cpuop_func op_c1b9_4_nf;
extern cpuop_func op_c1b9_4_ff;
extern cpuop_func op_c1c0_4_nf;
extern cpuop_func op_c1c0_4_ff;
extern cpuop_func op_c1d0_4_nf;
extern cpuop_func op_c1d0_4_ff;
extern cpuop_func op_c1d8_4_nf;
extern cpuop_func op_c1d8_4_ff;
extern cpuop_func op_c1e0_4_nf;
extern cpuop_func op_c1e0_4_ff;
extern cpuop_func op_c1e8_4_nf;
extern cpuop_func op_c1e8_4_ff;
extern cpuop_func op_c1f0_4_nf;
extern cpuop_func op_c1f0_4_ff;
extern cpuop_func op_c1f8_4_nf;
extern cpuop_func op_c1f8_4_ff;
extern cpuop_func op_c1f9_4_nf;
extern cpuop_func op_c1f9_4_ff;
extern cpuop_func op_c1fa_4_nf;
extern cpuop_func op_c1fa_4_ff;
extern cpuop_func op_c1fb_4_nf;
extern cpuop_func op_c1fb_4_ff;
extern cpuop_func op_c1fc_4_nf;
extern cpuop_func op_c1fc_4_ff;
extern cpuop_func op_d000_4_nf;
extern cpuop_func op_d000_4_ff;
extern cpuop_func op_d010_4_nf;
extern cpuop_func op_d010_4_ff;
extern cpuop_func op_d018_4_nf;
extern cpuop_func op_d018_4_ff;
extern cpuop_func op_d020_4_nf;
extern cpuop_func op_d020_4_ff;
extern cpuop_func op_d028_4_nf;
extern cpuop_func op_d028_4_ff;
extern cpuop_func op_d030_4_nf;
extern cpuop_func op_d030_4_ff;
extern cpuop_func op_d038_4_nf;
extern cpuop_func op_d038_4_ff;
extern cpuop_func op_d039_4_nf;
extern cpuop_func op_d039_4_ff;
extern cpuop_func op_d03a_4_nf;
extern cpuop_func op_d03a_4_ff;
extern cpuop_func op_d03b_4_nf;
extern cpuop_func op_d03b_4_ff;
extern cpuop_func op_d03c_4_nf;
extern cpuop_func op_d03c_4_ff;
extern cpuop_func op_d040_4_nf;
extern cpuop_func op_d040_4_ff;
extern cpuop_func op_d048_4_nf;
extern cpuop_func op_d048_4_ff;
extern cpuop_func op_d050_4_nf;
extern cpuop_func op_d050_4_ff;
extern cpuop_func op_d058_4_nf;
extern cpuop_func op_d058_4_ff;
extern cpuop_func op_d060_4_nf;
extern cpuop_func op_d060_4_ff;
extern cpuop_func op_d068_4_nf;
extern cpuop_func op_d068_4_ff;
extern cpuop_func op_d070_4_nf;
extern cpuop_func op_d070_4_ff;
extern cpuop_func op_d078_4_nf;
extern cpuop_func op_d078_4_ff;
extern cpuop_func op_d079_4_nf;
extern cpuop_func op_d079_4_ff;
extern cpuop_func op_d07a_4_nf;
extern cpuop_func op_d07a_4_ff;
extern cpuop_func op_d07b_4_nf;
extern cpuop_func op_d07b_4_ff;
extern cpuop_func op_d07c_4_nf;
extern cpuop_func op_d07c_4_ff;
extern cpuop_func op_d080_4_nf;
extern cpuop_func op_d080_4_ff;
extern cpuop_func op_d088_4_nf;
extern cpuop_func op_d088_4_ff;
extern cpuop_func op_d090_4_nf;
extern cpuop_func op_d090_4_ff;
extern cpuop_func op_d098_4_nf;
extern cpuop_func op_d098_4_ff;
extern cpuop_func op_d0a0_4_nf;
extern cpuop_func op_d0a0_4_ff;
extern cpuop_func op_d0a8_4_nf;
extern cpuop_func op_d0a8_4_ff;
extern cpuop_func op_d0b0_4_nf;
extern cpuop_func op_d0b0_4_ff;
extern cpuop_func op_d0b8_4_nf;
extern cpuop_func op_d0b8_4_ff;
extern cpuop_func op_d0b9_4_nf;
extern cpuop_func op_d0b9_4_ff;
extern cpuop_func op_d0ba_4_nf;
extern cpuop_func op_d0ba_4_ff;
extern cpuop_func op_d0bb_4_nf;
extern cpuop_func op_d0bb_4_ff;
extern cpuop_func op_d0bc_4_nf;
extern cpuop_func op_d0bc_4_ff;
extern cpuop_func op_d0c0_4_nf;
extern cpuop_func op_d0c0_4_ff;
extern cpuop_func op_d0c8_4_nf;
extern cpuop_func op_d0c8_4_ff;
extern cpuop_func op_d0d0_4_nf;
extern cpuop_func op_d0d0_4_ff;
extern cpuop_func op_d0d8_4_nf;
extern cpuop_func op_d0d8_4_ff;
extern cpuop_func op_d0e0_4_nf;
extern cpuop_func op_d0e0_4_ff;
extern cpuop_func op_d0e8_4_nf;
extern cpuop_func op_d0e8_4_ff;
extern cpuop_func op_d0f0_4_nf;
extern cpuop_func op_d0f0_4_ff;
extern cpuop_func op_d0f8_4_nf;
extern cpuop_func op_d0f8_4_ff;
extern cpuop_func op_d0f9_4_nf;
extern cpuop_func op_d0f9_4_ff;
extern cpuop_func op_d0fa_4_nf;
extern cpuop_func op_d0fa_4_ff;
extern cpuop_func op_d0fb_4_nf;
extern cpuop_func op_d0fb_4_ff;
extern cpuop_func op_d0fc_4_nf;
extern cpuop_func op_d0fc_4_ff;
extern cpuop_func op_d100_4_nf;
extern cpuop_func op_d100_4_ff;
extern cpuop_func op_d108_4_nf;
extern cpuop_func op_d108_4_ff;
extern cpuop_func op_d110_4_nf;
extern cpuop_func op_d110_4_ff;
extern cpuop_func op_d118_4_nf;
extern cpuop_func op_d118_4_ff;
extern cpuop_func op_d120_4_nf;
extern cpuop_func op_d120_4_ff;
extern cpuop_func op_d128_4_nf;
extern cpuop_func op_d128_4_ff;
extern cpuop_func op_d130_4_nf;
extern cpuop_func op_d130_4_ff;
extern cpuop_func op_d138_4_nf;
extern cpuop_func op_d138_4_ff;
extern cpuop_func op_d139_4_nf;
extern cpuop_func op_d139_4_ff;
extern cpuop_func op_d140_4_nf;
extern cpuop_func op_d140_4_ff;
extern cpuop_func op_d148_4_nf;
extern cpuop_func op_d148_4_ff;
extern cpuop_func op_d150_4_nf;
extern cpuop_func op_d150_4_ff;
extern cpuop_func op_d158_4_nf;
extern cpuop_func op_d158_4_ff;
extern cpuop_func op_d160_4_nf;
extern cpuop_func op_d160_4_ff;
extern cpuop_func op_d168_4_nf;
extern cpuop_func op_d168_4_ff;
extern cpuop_func op_d170_4_nf;
extern cpuop_func op_d170_4_ff;
extern cpuop_func op_d178_4_nf;
extern cpuop_func op_d178_4_ff;
extern cpuop_func op_d179_4_nf;
extern cpuop_func op_d179_4_ff;
extern cpuop_func op_d180_4_nf;
extern cpuop_func op_d180_4_ff;
extern cpuop_func op_d188_4_nf;
extern cpuop_func op_d188_4_ff;
extern cpuop_func op_d190_4_nf;
extern cpuop_func op_d190_4_ff;
extern cpuop_func op_d198_4_nf;
extern cpuop_func op_d198_4_ff;
extern cpuop_func op_d1a0_4_nf;
extern cpuop_func op_d1a0_4_ff;
extern cpuop_func op_d1a8_4_nf;
extern cpuop_func op_d1a8_4_ff;
extern cpuop_func op_d1b0_4_nf;
extern cpuop_func op_d1b0_4_ff;
extern cpuop_func op_d1b8_4_nf;
extern cpuop_func op_d1b8_4_ff;
extern cpuop_func op_d1b9_4_nf;
extern cpuop_func op_d1b9_4_ff;
extern cpuop_func op_d1c0_4_nf;
extern cpuop_func op_d1c0_4_ff;
extern cpuop_func op_d1c8_4_nf;
extern cpuop_func op_d1c8_4_ff;
extern cpuop_func op_d1d0_4_nf;
extern cpuop_func op_d1d0_4_ff;
extern cpuop_func op_d1d8_4_nf;
extern cpuop_func op_d1d8_4_ff;
extern cpuop_func op_d1e0_4_nf;
extern cpuop_func op_d1e0_4_ff;
extern cpuop_func op_d1e8_4_nf;
extern cpuop_func op_d1e8_4_ff;
extern cpuop_func op_d1f0_4_nf;
extern cpuop_func op_d1f0_4_ff;
extern cpuop_func op_d1f8_4_nf;
extern cpuop_func op_d1f8_4_ff;
extern cpuop_func op_d1f9_4_nf;
extern cpuop_func op_d1f9_4_ff;
extern cpuop_func op_d1fa_4_nf;
extern cpuop_func op_d1fa_4_ff;
extern cpuop_func op_d1fb_4_nf;
extern cpuop_func op_d1fb_4_ff;
extern cpuop_func op_d1fc_4_nf;
extern cpuop_func op_d1fc_4_ff;
extern cpuop_func op_e000_4_nf;
extern cpuop_func op_e000_4_ff;
extern cpuop_func op_e008_4_nf;
extern cpuop_func op_e008_4_ff;
extern cpuop_func op_e010_4_nf;
extern cpuop_func op_e010_4_ff;
extern cpuop_func op_e018_4_nf;
extern cpuop_func op_e018_4_ff;
extern cpuop_func op_e020_4_nf;
extern cpuop_func op_e020_4_ff;
extern cpuop_func op_e028_4_nf;
extern cpuop_func op_e028_4_ff;
extern cpuop_func op_e030_4_nf;
extern cpuop_func op_e030_4_ff;
extern cpuop_func op_e038_4_nf;
extern cpuop_func op_e038_4_ff;
extern cpuop_func op_e040_4_nf;
extern cpuop_func op_e040_4_ff;
extern cpuop_func op_e048_4_nf;
extern cpuop_func op_e048_4_ff;
extern cpuop_func op_e050_4_nf;
extern cpuop_func op_e050_4_ff;
extern cpuop_func op_e058_4_nf;
extern cpuop_func op_e058_4_ff;
extern cpuop_func op_e060_4_nf;
extern cpuop_func op_e060_4_ff;
extern cpuop_func op_e068_4_nf;
extern cpuop_func op_e068_4_ff;
extern cpuop_func op_e070_4_nf;
extern cpuop_func op_e070_4_ff;
extern cpuop_func op_e078_4_nf;
extern cpuop_func op_e078_4_ff;
extern cpuop_func op_e080_4_nf;
extern cpuop_func op_e080_4_ff;
extern cpuop_func op_e088_4_nf;
extern cpuop_func op_e088_4_ff;
extern cpuop_func op_e090_4_nf;
extern cpuop_func op_e090_4_ff;
extern cpuop_func op_e098_4_nf;
extern cpuop_func op_e098_4_ff;
extern cpuop_func op_e0a0_4_nf;
extern cpuop_func op_e0a0_4_ff;
extern cpuop_func op_e0a8_4_nf;
extern cpuop_func op_e0a8_4_ff;
extern cpuop_func op_e0b0_4_nf;
extern cpuop_func op_e0b0_4_ff;
extern cpuop_func op_e0b8_4_nf;
extern cpuop_func op_e0b8_4_ff;
extern cpuop_func op_e0d0_4_nf;
extern cpuop_func op_e0d0_4_ff;
extern cpuop_func op_e0d8_4_nf;
extern cpuop_func op_e0d8_4_ff;
extern cpuop_func op_e0e0_4_nf;
extern cpuop_func op_e0e0_4_ff;
extern cpuop_func op_e0e8_4_nf;
extern cpuop_func op_e0e8_4_ff;
extern cpuop_func op_e0f0_4_nf;
extern cpuop_func op_e0f0_4_ff;
extern cpuop_func op_e0f8_4_nf;
extern cpuop_func op_e0f8_4_ff;
extern cpuop_func op_e0f9_4_nf;
extern cpuop_func op_e0f9_4_ff;
extern cpuop_func op_e100_4_nf;
extern cpuop_func op_e100_4_ff;
extern cpuop_func op_e108_4_nf;
extern cpuop_func op_e108_4_ff;
extern cpuop_func op_e110_4_nf;
extern cpuop_func op_e110_4_ff;
extern cpuop_func op_e118_4_nf;
extern cpuop_func op_e118_4_ff;
extern cpuop_func op_e120_4_nf;
extern cpuop_func op_e120_4_ff;
extern cpuop_func op_e128_4_nf;
extern cpuop_func op_e128_4_ff;
extern cpuop_func op_e130_4_nf;
extern cpuop_func op_e130_4_ff;
extern cpuop_func op_e138_4_nf;
extern cpuop_func op_e138_4_ff;
extern cpuop_func op_e140_4_nf;
extern cpuop_func op_e140_4_ff;
extern cpuop_func op_e148_4_nf;
extern cpuop_func op_e148_4_ff;
extern cpuop_func op_e150_4_nf;
extern cpuop_func op_e150_4_ff;
extern cpuop_func op_e158_4_nf;
extern cpuop_func op_e158_4_ff;
extern cpuop_func op_e160_4_nf;
extern cpuop_func op_e160_4_ff;
extern cpuop_func op_e168_4_nf;
extern cpuop_func op_e168_4_ff;
extern cpuop_func op_e170_4_nf;
extern cpuop_func op_e170_4_ff;
extern cpuop_func op_e178_4_nf;
extern cpuop_func op_e178_4_ff;
extern cpuop_func op_e180_4_nf;
extern cpuop_func op_e180_4_ff;
extern cpuop_func op_e188_4_nf;
extern cpuop_func op_e188_4_ff;
extern cpuop_func op_e190_4_nf;
extern cpuop_func op_e190_4_ff;
extern cpuop_func op_e198_4_nf;
extern cpuop_func op_e198_4_ff;
extern cpuop_func op_e1a0_4_nf;
extern cpuop_func op_e1a0_4_ff;
extern cpuop_func op_e1a8_4_nf;
extern cpuop_func op_e1a8_4_ff;
extern cpuop_func op_e1b0_4_nf;
extern cpuop_func op_e1b0_4_ff;
extern cpuop_func op_e1b8_4_nf;
extern cpuop_func op_e1b8_4_ff;
extern cpuop_func op_e1d0_4_nf;
extern cpuop_func op_e1d0_4_ff;
extern cpuop_func op_e1d8_4_nf;
extern cpuop_func op_e1d8_4_ff;
extern cpuop_func op_e1e0_4_nf;
extern cpuop_func op_e1e0_4_ff;
extern cpuop_func op_e1e8_4_nf;
extern cpuop_func op_e1e8_4_ff;
extern cpuop_func op_e1f0_4_nf;
extern cpuop_func op_e1f0_4_ff;
extern cpuop_func op_e1f8_4_nf;
extern cpuop_func op_e1f8_4_ff;
extern cpuop_func op_e1f9_4_nf;
extern cpuop_func op_e1f9_4_ff;
extern cpuop_func op_e2d0_4_nf;
extern cpuop_func op_e2d0_4_ff;
extern cpuop_func op_e2d8_4_nf;
extern cpuop_func op_e2d8_4_ff;
extern cpuop_func op_e2e0_4_nf;
extern cpuop_func op_e2e0_4_ff;
extern cpuop_func op_e2e8_4_nf;
extern cpuop_func op_e2e8_4_ff;
extern cpuop_func op_e2f0_4_nf;
extern cpuop_func op_e2f0_4_ff;
extern cpuop_func op_e2f8_4_nf;
extern cpuop_func op_e2f8_4_ff;
extern cpuop_func op_e2f9_4_nf;
extern cpuop_func op_e2f9_4_ff;
extern cpuop_func op_e3d0_4_nf;
extern cpuop_func op_e3d0_4_ff;
extern cpuop_func op_e3d8_4_nf;
extern cpuop_func op_e3d8_4_ff;
extern cpuop_func op_e3e0_4_nf;
extern cpuop_func op_e3e0_4_ff;
extern cpuop_func op_e3e8_4_nf;
extern cpuop_func op_e3e8_4_ff;
extern cpuop_func op_e3f0_4_nf;
extern cpuop_func op_e3f0_4_ff;
extern cpuop_func op_e3f8_4_nf;
extern cpuop_func op_e3f8_4_ff;
extern cpuop_func op_e3f9_4_nf;
extern cpuop_func op_e3f9_4_ff;
extern cpuop_func op_e4d0_4_nf;
extern cpuop_func op_e4d0_4_ff;
extern cpuop_func op_e4d8_4_nf;
extern cpuop_func op_e4d8_4_ff;
extern cpuop_func op_e4e0_4_nf;
extern cpuop_func op_e4e0_4_ff;
extern cpuop_func op_e4e8_4_nf;
extern cpuop_func op_e4e8_4_ff;
extern cpuop_func op_e4f0_4_nf;
extern cpuop_func op_e4f0_4_ff;
extern cpuop_func op_e4f8_4_nf;
extern cpuop_func op_e4f8_4_ff;
extern cpuop_func op_e4f9_4_nf;
extern cpuop_func op_e4f9_4_ff;
extern cpuop_func op_e5d0_4_nf;
extern cpuop_func op_e5d0_4_ff;
extern cpuop_func op_e5d8_4_nf;
extern cpuop_func op_e5d8_4_ff;
extern cpuop_func op_e5e0_4_nf;
extern cpuop_func op_e5e0_4_ff;
extern cpuop_func op_e5e8_4_nf;
extern cpuop_func op_e5e8_4_ff;
extern cpuop_func op_e5f0_4_nf;
extern cpuop_func op_e5f0_4_ff;
extern cpuop_func op_e5f8_4_nf;
extern cpuop_func op_e5f8_4_ff;
extern cpuop_func op_e5f9_4_nf;
extern cpuop_func op_e5f9_4_ff;
extern cpuop_func op_e6d0_4_nf;
extern cpuop_func op_e6d0_4_ff;
extern cpuop_func op_e6d8_4_nf;
extern cpuop_func op_e6d8_4_ff;
extern cpuop_func op_e6e0_4_nf;
extern cpuop_func op_e6e0_4_ff;
extern cpuop_func op_e6e8_4_nf;
extern cpuop_func op_e6e8_4_ff;
extern cpuop_func op_e6f0_4_nf;
extern cpuop_func op_e6f0_4_ff;
extern cpuop_func op_e6f8_4_nf;
extern cpuop_func op_e6f8_4_ff;
extern cpuop_func op_e6f9_4_nf;
extern cpuop_func op_e6f9_4_ff;
extern cpuop_func op_e7d0_4_nf;
extern cpuop_func op_e7d0_4_ff;
extern cpuop_func op_e7d8_4_nf;
extern cpuop_func op_e7d8_4_ff;
extern cpuop_func op_e7e0_4_nf;
extern cpuop_func op_e7e0_4_ff;
extern cpuop_func op_e7e8_4_nf;
extern cpuop_func op_e7e8_4_ff;
extern cpuop_func op_e7f0_4_nf;
extern cpuop_func op_e7f0_4_ff;
extern cpuop_func op_e7f8_4_nf;
extern cpuop_func op_e7f8_4_ff;
extern cpuop_func op_e7f9_4_nf;
extern cpuop_func op_e7f9_4_ff;
extern cpuop_func op_40c0_5_nf;
extern cpuop_func op_40c0_5_ff;
extern cpuop_func op_40d0_5_nf;
extern cpuop_func op_40d0_5_ff;
extern cpuop_func op_40d8_5_nf;
extern cpuop_func op_40d8_5_ff;
extern cpuop_func op_40e0_5_nf;
extern cpuop_func op_40e0_5_ff;
extern cpuop_func op_40e8_5_nf;
extern cpuop_func op_40e8_5_ff;
extern cpuop_func op_40f0_5_nf;
extern cpuop_func op_40f0_5_ff;
extern cpuop_func op_40f8_5_nf;
extern cpuop_func op_40f8_5_ff;
extern cpuop_func op_40f9_5_nf;
extern cpuop_func op_40f9_5_ff;
extern cpuop_func op_4200_5_nf;
extern cpuop_func op_4200_5_ff;
extern cpuop_func op_4210_5_nf;
extern cpuop_func op_4210_5_ff;
extern cpuop_func op_4218_5_nf;
extern cpuop_func op_4218_5_ff;
extern cpuop_func op_4220_5_nf;
extern cpuop_func op_4220_5_ff;
extern cpuop_func op_4228_5_nf;
extern cpuop_func op_4228_5_ff;
extern cpuop_func op_4230_5_nf;
extern cpuop_func op_4230_5_ff;
extern cpuop_func op_4238_5_nf;
extern cpuop_func op_4238_5_ff;
extern cpuop_func op_4239_5_nf;
extern cpuop_func op_4239_5_ff;
extern cpuop_func op_4240_5_nf;
extern cpuop_func op_4240_5_ff;
extern cpuop_func op_4250_5_nf;
extern cpuop_func op_4250_5_ff;
extern cpuop_func op_4258_5_nf;
extern cpuop_func op_4258_5_ff;
extern cpuop_func op_4260_5_nf;
extern cpuop_func op_4260_5_ff;
extern cpuop_func op_4268_5_nf;
extern cpuop_func op_4268_5_ff;
extern cpuop_func op_4270_5_nf;
extern cpuop_func op_4270_5_ff;
extern cpuop_func op_4278_5_nf;
extern cpuop_func op_4278_5_ff;
extern cpuop_func op_4279_5_nf;
extern cpuop_func op_4279_5_ff;
extern cpuop_func op_4280_5_nf;
extern cpuop_func op_4280_5_ff;
extern cpuop_func op_4290_5_nf;
extern cpuop_func op_4290_5_ff;
extern cpuop_func op_4298_5_nf;
extern cpuop_func op_4298_5_ff;
extern cpuop_func op_42a0_5_nf;
extern cpuop_func op_42a0_5_ff;
extern cpuop_func op_42a8_5_nf;
extern cpuop_func op_42a8_5_ff;
extern cpuop_func op_42b0_5_nf;
extern cpuop_func op_42b0_5_ff;
extern cpuop_func op_42b8_5_nf;
extern cpuop_func op_42b8_5_ff;
extern cpuop_func op_42b9_5_nf;
extern cpuop_func op_42b9_5_ff;
extern cpuop_func op_4e73_5_nf;
extern cpuop_func op_4e73_5_ff;
extern cpuop_func op_50c0_5_nf;
extern cpuop_func op_50c0_5_ff;
extern cpuop_func op_50d0_5_nf;
extern cpuop_func op_50d0_5_ff;
extern cpuop_func op_50d8_5_nf;
extern cpuop_func op_50d8_5_ff;
extern cpuop_func op_50e0_5_nf;
extern cpuop_func op_50e0_5_ff;
extern cpuop_func op_50e8_5_nf;
extern cpuop_func op_50e8_5_ff;
extern cpuop_func op_50f0_5_nf;
extern cpuop_func op_50f0_5_ff;
extern cpuop_func op_50f8_5_nf;
extern cpuop_func op_50f8_5_ff;
extern cpuop_func op_50f9_5_nf;
extern cpuop_func op_50f9_5_ff;
extern cpuop_func op_51c0_5_nf;
extern cpuop_func op_51c0_5_ff;
extern cpuop_func op_51d0_5_nf;
extern cpuop_func op_51d0_5_ff;
extern cpuop_func op_51d8_5_nf;
extern cpuop_func op_51d8_5_ff;
extern cpuop_func op_51e0_5_nf;
extern cpuop_func op_51e0_5_ff;
extern cpuop_func op_51e8_5_nf;
extern cpuop_func op_51e8_5_ff;
extern cpuop_func op_51f0_5_nf;
extern cpuop_func op_51f0_5_ff;
extern cpuop_func op_51f8_5_nf;
extern cpuop_func op_51f8_5_ff;
extern cpuop_func op_51f9_5_nf;
extern cpuop_func op_51f9_5_ff;
extern cpuop_func op_52c0_5_nf;
extern cpuop_func op_52c0_5_ff;
extern cpuop_func op_52d0_5_nf;
extern cpuop_func op_52d0_5_ff;
extern cpuop_func op_52d8_5_nf;
extern cpuop_func op_52d8_5_ff;
extern cpuop_func op_52e0_5_nf;
extern cpuop_func op_52e0_5_ff;
extern cpuop_func op_52e8_5_nf;
extern cpuop_func op_52e8_5_ff;
extern cpuop_func op_52f0_5_nf;
extern cpuop_func op_52f0_5_ff;
extern cpuop_func op_52f8_5_nf;
extern cpuop_func op_52f8_5_ff;
extern cpuop_func op_52f9_5_nf;
extern cpuop_func op_52f9_5_ff;
extern cpuop_func op_53c0_5_nf;
extern cpuop_func op_53c0_5_ff;
extern cpuop_func op_53d0_5_nf;
extern cpuop_func op_53d0_5_ff;
extern cpuop_func op_53d8_5_nf;
extern cpuop_func op_53d8_5_ff;
extern cpuop_func op_53e0_5_nf;
extern cpuop_func op_53e0_5_ff;
extern cpuop_func op_53e8_5_nf;
extern cpuop_func op_53e8_5_ff;
extern cpuop_func op_53f0_5_nf;
extern cpuop_func op_53f0_5_ff;
extern cpuop_func op_53f8_5_nf;
extern cpuop_func op_53f8_5_ff;
extern cpuop_func op_53f9_5_nf;
extern cpuop_func op_53f9_5_ff;
extern cpuop_func op_54c0_5_nf;
extern cpuop_func op_54c0_5_ff;
extern cpuop_func op_54d0_5_nf;
extern cpuop_func op_54d0_5_ff;
extern cpuop_func op_54d8_5_nf;
extern cpuop_func op_54d8_5_ff;
extern cpuop_func op_54e0_5_nf;
extern cpuop_func op_54e0_5_ff;
extern cpuop_func op_54e8_5_nf;
extern cpuop_func op_54e8_5_ff;
extern cpuop_func op_54f0_5_nf;
extern cpuop_func op_54f0_5_ff;
extern cpuop_func op_54f8_5_nf;
extern cpuop_func op_54f8_5_ff;
extern cpuop_func op_54f9_5_nf;
extern cpuop_func op_54f9_5_ff;
extern cpuop_func op_55c0_5_nf;
extern cpuop_func op_55c0_5_ff;
extern cpuop_func op_55d0_5_nf;
extern cpuop_func op_55d0_5_ff;
extern cpuop_func op_55d8_5_nf;
extern cpuop_func op_55d8_5_ff;
extern cpuop_func op_55e0_5_nf;
extern cpuop_func op_55e0_5_ff;
extern cpuop_func op_55e8_5_nf;
extern cpuop_func op_55e8_5_ff;
extern cpuop_func op_55f0_5_nf;
extern cpuop_func op_55f0_5_ff;
extern cpuop_func op_55f8_5_nf;
extern cpuop_func op_55f8_5_ff;
extern cpuop_func op_55f9_5_nf;
extern cpuop_func op_55f9_5_ff;
extern cpuop_func op_56c0_5_nf;
extern cpuop_func op_56c0_5_ff;
extern cpuop_func op_56d0_5_nf;
extern cpuop_func op_56d0_5_ff;
extern cpuop_func op_56d8_5_nf;
extern cpuop_func op_56d8_5_ff;
extern cpuop_func op_56e0_5_nf;
extern cpuop_func op_56e0_5_ff;
extern cpuop_func op_56e8_5_nf;
extern cpuop_func op_56e8_5_ff;
extern cpuop_func op_56f0_5_nf;
extern cpuop_func op_56f0_5_ff;
extern cpuop_func op_56f8_5_nf;
extern cpuop_func op_56f8_5_ff;
extern cpuop_func op_56f9_5_nf;
extern cpuop_func op_56f9_5_ff;
extern cpuop_func op_57c0_5_nf;
extern cpuop_func op_57c0_5_ff;
extern cpuop_func op_57d0_5_nf;
extern cpuop_func op_57d0_5_ff;
extern cpuop_func op_57d8_5_nf;
extern cpuop_func op_57d8_5_ff;
extern cpuop_func op_57e0_5_nf;
extern cpuop_func op_57e0_5_ff;
extern cpuop_func op_57e8_5_nf;
extern cpuop_func op_57e8_5_ff;
extern cpuop_func op_57f0_5_nf;
extern cpuop_func op_57f0_5_ff;
extern cpuop_func op_57f8_5_nf;
extern cpuop_func op_57f8_5_ff;
extern cpuop_func op_57f9_5_nf;
extern cpuop_func op_57f9_5_ff;
extern cpuop_func op_58c0_5_nf;
extern cpuop_func op_58c0_5_ff;
extern cpuop_func op_58d0_5_nf;
extern cpuop_func op_58d0_5_ff;
extern cpuop_func op_58d8_5_nf;
extern cpuop_func op_58d8_5_ff;
extern cpuop_func op_58e0_5_nf;
extern cpuop_func op_58e0_5_ff;
extern cpuop_func op_58e8_5_nf;
extern cpuop_func op_58e8_5_ff;
extern cpuop_func op_58f0_5_nf;
extern cpuop_func op_58f0_5_ff;
extern cpuop_func op_58f8_5_nf;
extern cpuop_func op_58f8_5_ff;
extern cpuop_func op_58f9_5_nf;
extern cpuop_func op_58f9_5_ff;
extern cpuop_func op_59c0_5_nf;
extern cpuop_func op_59c0_5_ff;
extern cpuop_func op_59d0_5_nf;
extern cpuop_func op_59d0_5_ff;
extern cpuop_func op_59d8_5_nf;
extern cpuop_func op_59d8_5_ff;
extern cpuop_func op_59e0_5_nf;
extern cpuop_func op_59e0_5_ff;
extern cpuop_func op_59e8_5_nf;
extern cpuop_func op_59e8_5_ff;
extern cpuop_func op_59f0_5_nf;
extern cpuop_func op_59f0_5_ff;
extern cpuop_func op_59f8_5_nf;
extern cpuop_func op_59f8_5_ff;
extern cpuop_func op_59f9_5_nf;
extern cpuop_func op_59f9_5_ff;
extern cpuop_func op_5ac0_5_nf;
extern cpuop_func op_5ac0_5_ff;
extern cpuop_func op_5ad0_5_nf;
extern cpuop_func op_5ad0_5_ff;
extern cpuop_func op_5ad8_5_nf;
extern cpuop_func op_5ad8_5_ff;
extern cpuop_func op_5ae0_5_nf;
extern cpuop_func op_5ae0_5_ff;
extern cpuop_func op_5ae8_5_nf;
extern cpuop_func op_5ae8_5_ff;
extern cpuop_func op_5af0_5_nf;
extern cpuop_func op_5af0_5_ff;
extern cpuop_func op_5af8_5_nf;
extern cpuop_func op_5af8_5_ff;
extern cpuop_func op_5af9_5_nf;
extern cpuop_func op_5af9_5_ff;
extern cpuop_func op_5bc0_5_nf;
extern cpuop_func op_5bc0_5_ff;
extern cpuop_func op_5bd0_5_nf;
extern cpuop_func op_5bd0_5_ff;
extern cpuop_func op_5bd8_5_nf;
extern cpuop_func op_5bd8_5_ff;
extern cpuop_func op_5be0_5_nf;
extern cpuop_func op_5be0_5_ff;
extern cpuop_func op_5be8_5_nf;
extern cpuop_func op_5be8_5_ff;
extern cpuop_func op_5bf0_5_nf;
extern cpuop_func op_5bf0_5_ff;
extern cpuop_func op_5bf8_5_nf;
extern cpuop_func op_5bf8_5_ff;
extern cpuop_func op_5bf9_5_nf;
extern cpuop_func op_5bf9_5_ff;
extern cpuop_func op_5cc0_5_nf;
extern cpuop_func op_5cc0_5_ff;
extern cpuop_func op_5cd0_5_nf;
extern cpuop_func op_5cd0_5_ff;
extern cpuop_func op_5cd8_5_nf;
extern cpuop_func op_5cd8_5_ff;
extern cpuop_func op_5ce0_5_nf;
extern cpuop_func op_5ce0_5_ff;
extern cpuop_func op_5ce8_5_nf;
extern cpuop_func op_5ce8_5_ff;
extern cpuop_func op_5cf0_5_nf;
extern cpuop_func op_5cf0_5_ff;
extern cpuop_func op_5cf8_5_nf;
extern cpuop_func op_5cf8_5_ff;
extern cpuop_func op_5cf9_5_nf;
extern cpuop_func op_5cf9_5_ff;
extern cpuop_func op_5dc0_5_nf;
extern cpuop_func op_5dc0_5_ff;
extern cpuop_func op_5dd0_5_nf;
extern cpuop_func op_5dd0_5_ff;
extern cpuop_func op_5dd8_5_nf;
extern cpuop_func op_5dd8_5_ff;
extern cpuop_func op_5de0_5_nf;
extern cpuop_func op_5de0_5_ff;
extern cpuop_func op_5de8_5_nf;
extern cpuop_func op_5de8_5_ff;
extern cpuop_func op_5df0_5_nf;
extern cpuop_func op_5df0_5_ff;
extern cpuop_func op_5df8_5_nf;
extern cpuop_func op_5df8_5_ff;
extern cpuop_func op_5df9_5_nf;
extern cpuop_func op_5df9_5_ff;
extern cpuop_func op_5ec0_5_nf;
extern cpuop_func op_5ec0_5_ff;
extern cpuop_func op_5ed0_5_nf;
extern cpuop_func op_5ed0_5_ff;
extern cpuop_func op_5ed8_5_nf;
extern cpuop_func op_5ed8_5_ff;
extern cpuop_func op_5ee0_5_nf;
extern cpuop_func op_5ee0_5_ff;
extern cpuop_func op_5ee8_5_nf;
extern cpuop_func op_5ee8_5_ff;
extern cpuop_func op_5ef0_5_nf;
extern cpuop_func op_5ef0_5_ff;
extern cpuop_func op_5ef8_5_nf;
extern cpuop_func op_5ef8_5_ff;
extern cpuop_func op_5ef9_5_nf;
extern cpuop_func op_5ef9_5_ff;
extern cpuop_func op_5fc0_5_nf;
extern cpuop_func op_5fc0_5_ff;
extern cpuop_func op_5fd0_5_nf;
extern cpuop_func op_5fd0_5_ff;
extern cpuop_func op_5fd8_5_nf;
extern cpuop_func op_5fd8_5_ff;
extern cpuop_func op_5fe0_5_nf;
extern cpuop_func op_5fe0_5_ff;
extern cpuop_func op_5fe8_5_nf;
extern cpuop_func op_5fe8_5_ff;
extern cpuop_func op_5ff0_5_nf;
extern cpuop_func op_5ff0_5_ff;
extern cpuop_func op_5ff8_5_nf;
extern cpuop_func op_5ff8_5_ff;
extern cpuop_func op_5ff9_5_nf;
extern cpuop_func op_5ff9_5_ff;
extern cpuop_func op_0000_11_nf;
extern cpuop_func op_0000_11_ff;
extern cpuop_func op_0010_11_nf;
extern cpuop_func op_0010_11_ff;
extern cpuop_func op_0018_11_nf;
extern cpuop_func op_0018_11_ff;
extern cpuop_func op_0020_11_nf;
extern cpuop_func op_0020_11_ff;
extern cpuop_func op_0028_11_nf;
extern cpuop_func op_0028_11_ff;
extern cpuop_func op_0030_11_nf;
extern cpuop_func op_0030_11_ff;
extern cpuop_func op_0038_11_nf;
extern cpuop_func op_0038_11_ff;
extern cpuop_func op_0039_11_nf;
extern cpuop_func op_0039_11_ff;
extern cpuop_func op_003c_11_nf;
extern cpuop_func op_003c_11_ff;
extern cpuop_func op_0040_11_nf;
extern cpuop_func op_0040_11_ff;
extern cpuop_func op_0050_11_nf;
extern cpuop_func op_0050_11_ff;
extern cpuop_func op_0058_11_nf;
extern cpuop_func op_0058_11_ff;
extern cpuop_func op_0060_11_nf;
extern cpuop_func op_0060_11_ff;
extern cpuop_func op_0068_11_nf;
extern cpuop_func op_0068_11_ff;
extern cpuop_func op_0070_11_nf;
extern cpuop_func op_0070_11_ff;
extern cpuop_func op_0078_11_nf;
extern cpuop_func op_0078_11_ff;
extern cpuop_func op_0079_11_nf;
extern cpuop_func op_0079_11_ff;
extern cpuop_func op_007c_11_nf;
extern cpuop_func op_007c_11_ff;
extern cpuop_func op_0080_11_nf;
extern cpuop_func op_0080_11_ff;
extern cpuop_func op_0090_11_nf;
extern cpuop_func op_0090_11_ff;
extern cpuop_func op_0098_11_nf;
extern cpuop_func op_0098_11_ff;
extern cpuop_func op_00a0_11_nf;
extern cpuop_func op_00a0_11_ff;
extern cpuop_func op_00a8_11_nf;
extern cpuop_func op_00a8_11_ff;
extern cpuop_func op_00b0_11_nf;
extern cpuop_func op_00b0_11_ff;
extern cpuop_func op_00b8_11_nf;
extern cpuop_func op_00b8_11_ff;
extern cpuop_func op_00b9_11_nf;
extern cpuop_func op_00b9_11_ff;
extern cpuop_func op_0100_11_nf;
extern cpuop_func op_0100_11_ff;
extern cpuop_func op_0108_11_nf;
extern cpuop_func op_0108_11_ff;
extern cpuop_func op_0110_11_nf;
extern cpuop_func op_0110_11_ff;
extern cpuop_func op_0118_11_nf;
extern cpuop_func op_0118_11_ff;
extern cpuop_func op_0120_11_nf;
extern cpuop_func op_0120_11_ff;
extern cpuop_func op_0128_11_nf;
extern cpuop_func op_0128_11_ff;
extern cpuop_func op_0130_11_nf;
extern cpuop_func op_0130_11_ff;
extern cpuop_func op_0138_11_nf;
extern cpuop_func op_0138_11_ff;
extern cpuop_func op_0139_11_nf;
extern cpuop_func op_0139_11_ff;
extern cpuop_func op_013a_11_nf;
extern cpuop_func op_013a_11_ff;
extern cpuop_func op_013b_11_nf;
extern cpuop_func op_013b_11_ff;
extern cpuop_func op_013c_11_nf;
extern cpuop_func op_013c_11_ff;
extern cpuop_func op_0140_11_nf;
extern cpuop_func op_0140_11_ff;
extern cpuop_func op_0148_11_nf;
extern cpuop_func op_0148_11_ff;
extern cpuop_func op_0150_11_nf;
extern cpuop_func op_0150_11_ff;
extern cpuop_func op_0158_11_nf;
extern cpuop_func op_0158_11_ff;
extern cpuop_func op_0160_11_nf;
extern cpuop_func op_0160_11_ff;
extern cpuop_func op_0168_11_nf;
extern cpuop_func op_0168_11_ff;
extern cpuop_func op_0170_11_nf;
extern cpuop_func op_0170_11_ff;
extern cpuop_func op_0178_11_nf;
extern cpuop_func op_0178_11_ff;
extern cpuop_func op_0179_11_nf;
extern cpuop_func op_0179_11_ff;
extern cpuop_func op_0180_11_nf;
extern cpuop_func op_0180_11_ff;
extern cpuop_func op_0188_11_nf;
extern cpuop_func op_0188_11_ff;
extern cpuop_func op_0190_11_nf;
extern cpuop_func op_0190_11_ff;
extern cpuop_func op_0198_11_nf;
extern cpuop_func op_0198_11_ff;
extern cpuop_func op_01a0_11_nf;
extern cpuop_func op_01a0_11_ff;
extern cpuop_func op_01a8_11_nf;
extern cpuop_func op_01a8_11_ff;
extern cpuop_func op_01b0_11_nf;
extern cpuop_func op_01b0_11_ff;
extern cpuop_func op_01b8_11_nf;
extern cpuop_func op_01b8_11_ff;
extern cpuop_func op_01b9_11_nf;
extern cpuop_func op_01b9_11_ff;
extern cpuop_func op_01c0_11_nf;
extern cpuop_func op_01c0_11_ff;
extern cpuop_func op_01c8_11_nf;
extern cpuop_func op_01c8_11_ff;
extern cpuop_func op_01d0_11_nf;
extern cpuop_func op_01d0_11_ff;
extern cpuop_func op_01d8_11_nf;
extern cpuop_func op_01d8_11_ff;
extern cpuop_func op_01e0_11_nf;
extern cpuop_func op_01e0_11_ff;
extern cpuop_func op_01e8_11_nf;
extern cpuop_func op_01e8_11_ff;
extern cpuop_func op_01f0_11_nf;
extern cpuop_func op_01f0_11_ff;
extern cpuop_func op_01f8_11_nf;
extern cpuop_func op_01f8_11_ff;
extern cpuop_func op_01f9_11_nf;
extern cpuop_func op_01f9_11_ff;
extern cpuop_func op_0200_11_nf;
extern cpuop_func op_0200_11_ff;
extern cpuop_func op_0210_11_nf;
extern cpuop_func op_0210_11_ff;
extern cpuop_func op_0218_11_nf;
extern cpuop_func op_0218_11_ff;
extern cpuop_func op_0220_11_nf;
extern cpuop_func op_0220_11_ff;
extern cpuop_func op_0228_11_nf;
extern cpuop_func op_0228_11_ff;
extern cpuop_func op_0230_11_nf;
extern cpuop_func op_0230_11_ff;
extern cpuop_func op_0238_11_nf;
extern cpuop_func op_0238_11_ff;
extern cpuop_func op_0239_11_nf;
extern cpuop_func op_0239_11_ff;
extern cpuop_func op_023c_11_nf;
extern cpuop_func op_023c_11_ff;
extern cpuop_func op_0240_11_nf;
extern cpuop_func op_0240_11_ff;
extern cpuop_func op_0250_11_nf;
extern cpuop_func op_0250_11_ff;
extern cpuop_func op_0258_11_nf;
extern cpuop_func op_0258_11_ff;
extern cpuop_func op_0260_11_nf;
extern cpuop_func op_0260_11_ff;
extern cpuop_func op_0268_11_nf;
extern cpuop_func op_0268_11_ff;
extern cpuop_func op_0270_11_nf;
extern cpuop_func op_0270_11_ff;
extern cpuop_func op_0278_11_nf;
extern cpuop_func op_0278_11_ff;
extern cpuop_func op_0279_11_nf;
extern cpuop_func op_0279_11_ff;
extern cpuop_func op_027c_11_nf;
extern cpuop_func op_027c_11_ff;
extern cpuop_func op_0280_11_nf;
extern cpuop_func op_0280_11_ff;
extern cpuop_func op_0290_11_nf;
extern cpuop_func op_0290_11_ff;
extern cpuop_func op_0298_11_nf;
extern cpuop_func op_0298_11_ff;
extern cpuop_func op_02a0_11_nf;
extern cpuop_func op_02a0_11_ff;
extern cpuop_func op_02a8_11_nf;
extern cpuop_func op_02a8_11_ff;
extern cpuop_func op_02b0_11_nf;
extern cpuop_func op_02b0_11_ff;
extern cpuop_func op_02b8_11_nf;
extern cpuop_func op_02b8_11_ff;
extern cpuop_func op_02b9_11_nf;
extern cpuop_func op_02b9_11_ff;
extern cpuop_func op_0400_11_nf;
extern cpuop_func op_0400_11_ff;
extern cpuop_func op_0410_11_nf;
extern cpuop_func op_0410_11_ff;
extern cpuop_func op_0418_11_nf;
extern cpuop_func op_0418_11_ff;
extern cpuop_func op_0420_11_nf;
extern cpuop_func op_0420_11_ff;
extern cpuop_func op_0428_11_nf;
extern cpuop_func op_0428_11_ff;
extern cpuop_func op_0430_11_nf;
extern cpuop_func op_0430_11_ff;
extern cpuop_func op_0438_11_nf;
extern cpuop_func op_0438_11_ff;
extern cpuop_func op_0439_11_nf;
extern cpuop_func op_0439_11_ff;
extern cpuop_func op_0440_11_nf;
extern cpuop_func op_0440_11_ff;
extern cpuop_func op_0450_11_nf;
extern cpuop_func op_0450_11_ff;
extern cpuop_func op_0458_11_nf;
extern cpuop_func op_0458_11_ff;
extern cpuop_func op_0460_11_nf;
extern cpuop_func op_0460_11_ff;
extern cpuop_func op_0468_11_nf;
extern cpuop_func op_0468_11_ff;
extern cpuop_func op_0470_11_nf;
extern cpuop_func op_0470_11_ff;
extern cpuop_func op_0478_11_nf;
extern cpuop_func op_0478_11_ff;
extern cpuop_func op_0479_11_nf;
extern cpuop_func op_0479_11_ff;
extern cpuop_func op_0480_11_nf;
extern cpuop_func op_0480_11_ff;
extern cpuop_func op_0490_11_nf;
extern cpuop_func op_0490_11_ff;
extern cpuop_func op_0498_11_nf;
extern cpuop_func op_0498_11_ff;
extern cpuop_func op_04a0_11_nf;
extern cpuop_func op_04a0_11_ff;
extern cpuop_func op_04a8_11_nf;
extern cpuop_func op_04a8_11_ff;
extern cpuop_func op_04b0_11_nf;
extern cpuop_func op_04b0_11_ff;
extern cpuop_func op_04b8_11_nf;
extern cpuop_func op_04b8_11_ff;
extern cpuop_func op_04b9_11_nf;
extern cpuop_func op_04b9_11_ff;
extern cpuop_func op_0600_11_nf;
extern cpuop_func op_0600_11_ff;
extern cpuop_func op_0610_11_nf;
extern cpuop_func op_0610_11_ff;
extern cpuop_func op_0618_11_nf;
extern cpuop_func op_0618_11_ff;
extern cpuop_func op_0620_11_nf;
extern cpuop_func op_0620_11_ff;
extern cpuop_func op_0628_11_nf;
extern cpuop_func op_0628_11_ff;
extern cpuop_func op_0630_11_nf;
extern cpuop_func op_0630_11_ff;
extern cpuop_func op_0638_11_nf;
extern cpuop_func op_0638_11_ff;
extern cpuop_func op_0639_11_nf;
extern cpuop_func op_0639_11_ff;
extern cpuop_func op_0640_11_nf;
extern cpuop_func op_0640_11_ff;
extern cpuop_func op_0650_11_nf;
extern cpuop_func op_0650_11_ff;
extern cpuop_func op_0658_11_nf;
extern cpuop_func op_0658_11_ff;
extern cpuop_func op_0660_11_nf;
extern cpuop_func op_0660_11_ff;
extern cpuop_func op_0668_11_nf;
extern cpuop_func op_0668_11_ff;
extern cpuop_func op_0670_11_nf;
extern cpuop_func op_0670_11_ff;
extern cpuop_func op_0678_11_nf;
extern cpuop_func op_0678_11_ff;
extern cpuop_func op_0679_11_nf;
extern cpuop_func op_0679_11_ff;
extern cpuop_func op_0680_11_nf;
extern cpuop_func op_0680_11_ff;
extern cpuop_func op_0690_11_nf;
extern cpuop_func op_0690_11_ff;
extern cpuop_func op_0698_11_nf;
extern cpuop_func op_0698_11_ff;
extern cpuop_func op_06a0_11_nf;
extern cpuop_func op_06a0_11_ff;
extern cpuop_func op_06a8_11_nf;
extern cpuop_func op_06a8_11_ff;
extern cpuop_func op_06b0_11_nf;
extern cpuop_func op_06b0_11_ff;
extern cpuop_func op_06b8_11_nf;
extern cpuop_func op_06b8_11_ff;
extern cpuop_func op_06b9_11_nf;
extern cpuop_func op_06b9_11_ff;
extern cpuop_func op_0800_11_nf;
extern cpuop_func op_0800_11_ff;
extern cpuop_func op_0810_11_nf;
extern cpuop_func op_0810_11_ff;
extern cpuop_func op_0818_11_nf;
extern cpuop_func op_0818_11_ff;
extern cpuop_func op_0820_11_nf;
extern cpuop_func op_0820_11_ff;
extern cpuop_func op_0828_11_nf;
extern cpuop_func op_0828_11_ff;
extern cpuop_func op_0830_11_nf;
extern cpuop_func op_0830_11_ff;
extern cpuop_func op_0838_11_nf;
extern cpuop_func op_0838_11_ff;
extern cpuop_func op_0839_11_nf;
extern cpuop_func op_0839_11_ff;
extern cpuop_func op_083a_11_nf;
extern cpuop_func op_083a_11_ff;
extern cpuop_func op_083b_11_nf;
extern cpuop_func op_083b_11_ff;
extern cpuop_func op_0840_11_nf;
extern cpuop_func op_0840_11_ff;
extern cpuop_func op_0850_11_nf;
extern cpuop_func op_0850_11_ff;
extern cpuop_func op_0858_11_nf;
extern cpuop_func op_0858_11_ff;
extern cpuop_func op_0860_11_nf;
extern cpuop_func op_0860_11_ff;
extern cpuop_func op_0868_11_nf;
extern cpuop_func op_0868_11_ff;
extern cpuop_func op_0870_11_nf;
extern cpuop_func op_0870_11_ff;
extern cpuop_func op_0878_11_nf;
extern cpuop_func op_0878_11_ff;
extern cpuop_func op_0879_11_nf;
extern cpuop_func op_0879_11_ff;
extern cpuop_func op_0880_11_nf;
extern cpuop_func op_0880_11_ff;
extern cpuop_func op_0890_11_nf;
extern cpuop_func op_0890_11_ff;
extern cpuop_func op_0898_11_nf;
extern cpuop_func op_0898_11_ff;
extern cpuop_func op_08a0_11_nf;
extern cpuop_func op_08a0_11_ff;
extern cpuop_func op_08a8_11_nf;
extern cpuop_func op_08a8_11_ff;
extern cpuop_func op_08b0_11_nf;
extern cpuop_func op_08b0_11_ff;
extern cpuop_func op_08b8_11_nf;
extern cpuop_func op_08b8_11_ff;
extern cpuop_func op_08b9_11_nf;
extern cpuop_func op_08b9_11_ff;
extern cpuop_func op_08c0_11_nf;
extern cpuop_func op_08c0_11_ff;
extern cpuop_func op_08d0_11_nf;
extern cpuop_func op_08d0_11_ff;
extern cpuop_func op_08d8_11_nf;
extern cpuop_func op_08d8_11_ff;
extern cpuop_func op_08e0_11_nf;
extern cpuop_func op_08e0_11_ff;
extern cpuop_func op_08e8_11_nf;
extern cpuop_func op_08e8_11_ff;
extern cpuop_func op_08f0_11_nf;
extern cpuop_func op_08f0_11_ff;
extern cpuop_func op_08f8_11_nf;
extern cpuop_func op_08f8_11_ff;
extern cpuop_func op_08f9_11_nf;
extern cpuop_func op_08f9_11_ff;
extern cpuop_func op_0a00_11_nf;
extern cpuop_func op_0a00_11_ff;
extern cpuop_func op_0a10_11_nf;
extern cpuop_func op_0a10_11_ff;
extern cpuop_func op_0a18_11_nf;
extern cpuop_func op_0a18_11_ff;
extern cpuop_func op_0a20_11_nf;
extern cpuop_func op_0a20_11_ff;
extern cpuop_func op_0a28_11_nf;
extern cpuop_func op_0a28_11_ff;
extern cpuop_func op_0a30_11_nf;
extern cpuop_func op_0a30_11_ff;
extern cpuop_func op_0a38_11_nf;
extern cpuop_func op_0a38_11_ff;
extern cpuop_func op_0a39_11_nf;
extern cpuop_func op_0a39_11_ff;
extern cpuop_func op_0a3c_11_nf;
extern cpuop_func op_0a3c_11_ff;
extern cpuop_func op_0a40_11_nf;
extern cpuop_func op_0a40_11_ff;
extern cpuop_func op_0a50_11_nf;
extern cpuop_func op_0a50_11_ff;
extern cpuop_func op_0a58_11_nf;
extern cpuop_func op_0a58_11_ff;
extern cpuop_func op_0a60_11_nf;
extern cpuop_func op_0a60_11_ff;
extern cpuop_func op_0a68_11_nf;
extern cpuop_func op_0a68_11_ff;
extern cpuop_func op_0a70_11_nf;
extern cpuop_func op_0a70_11_ff;
extern cpuop_func op_0a78_11_nf;
extern cpuop_func op_0a78_11_ff;
extern cpuop_func op_0a79_11_nf;
extern cpuop_func op_0a79_11_ff;
extern cpuop_func op_0a7c_11_nf;
extern cpuop_func op_0a7c_11_ff;
extern cpuop_func op_0a80_11_nf;
extern cpuop_func op_0a80_11_ff;
extern cpuop_func op_0a90_11_nf;
extern cpuop_func op_0a90_11_ff;
extern cpuop_func op_0a98_11_nf;
extern cpuop_func op_0a98_11_ff;
extern cpuop_func op_0aa0_11_nf;
extern cpuop_func op_0aa0_11_ff;
extern cpuop_func op_0aa8_11_nf;
extern cpuop_func op_0aa8_11_ff;
extern cpuop_func op_0ab0_11_nf;
extern cpuop_func op_0ab0_11_ff;
extern cpuop_func op_0ab8_11_nf;
extern cpuop_func op_0ab8_11_ff;
extern cpuop_func op_0ab9_11_nf;
extern cpuop_func op_0ab9_11_ff;
extern cpuop_func op_0c00_11_nf;
extern cpuop_func op_0c00_11_ff;
extern cpuop_func op_0c10_11_nf;
extern cpuop_func op_0c10_11_ff;
extern cpuop_func op_0c18_11_nf;
extern cpuop_func op_0c18_11_ff;
extern cpuop_func op_0c20_11_nf;
extern cpuop_func op_0c20_11_ff;
extern cpuop_func op_0c28_11_nf;
extern cpuop_func op_0c28_11_ff;
extern cpuop_func op_0c30_11_nf;
extern cpuop_func op_0c30_11_ff;
extern cpuop_func op_0c38_11_nf;
extern cpuop_func op_0c38_11_ff;
extern cpuop_func op_0c39_11_nf;
extern cpuop_func op_0c39_11_ff;
extern cpuop_func op_0c40_11_nf;
extern cpuop_func op_0c40_11_ff;
extern cpuop_func op_0c50_11_nf;
extern cpuop_func op_0c50_11_ff;
extern cpuop_func op_0c58_11_nf;
extern cpuop_func op_0c58_11_ff;
extern cpuop_func op_0c60_11_nf;
extern cpuop_func op_0c60_11_ff;
extern cpuop_func op_0c68_11_nf;
extern cpuop_func op_0c68_11_ff;
extern cpuop_func op_0c70_11_nf;
extern cpuop_func op_0c70_11_ff;
extern cpuop_func op_0c78_11_nf;
extern cpuop_func op_0c78_11_ff;
extern cpuop_func op_0c79_11_nf;
extern cpuop_func op_0c79_11_ff;
extern cpuop_func op_0c80_11_nf;
extern cpuop_func op_0c80_11_ff;
extern cpuop_func op_0c90_11_nf;
extern cpuop_func op_0c90_11_ff;
extern cpuop_func op_0c98_11_nf;
extern cpuop_func op_0c98_11_ff;
extern cpuop_func op_0ca0_11_nf;
extern cpuop_func op_0ca0_11_ff;
extern cpuop_func op_0ca8_11_nf;
extern cpuop_func op_0ca8_11_ff;
extern cpuop_func op_0cb0_11_nf;
extern cpuop_func op_0cb0_11_ff;
extern cpuop_func op_0cb8_11_nf;
extern cpuop_func op_0cb8_11_ff;
extern cpuop_func op_0cb9_11_nf;
extern cpuop_func op_0cb9_11_ff;
extern cpuop_func op_1000_11_nf;
extern cpuop_func op_1000_11_ff;
extern cpuop_func op_1010_11_nf;
extern cpuop_func op_1010_11_ff;
extern cpuop_func op_1018_11_nf;
extern cpuop_func op_1018_11_ff;
extern cpuop_func op_1020_11_nf;
extern cpuop_func op_1020_11_ff;
extern cpuop_func op_1028_11_nf;
extern cpuop_func op_1028_11_ff;
extern cpuop_func op_1030_11_nf;
extern cpuop_func op_1030_11_ff;
extern cpuop_func op_1038_11_nf;
extern cpuop_func op_1038_11_ff;
extern cpuop_func op_1039_11_nf;
extern cpuop_func op_1039_11_ff;
extern cpuop_func op_103a_11_nf;
extern cpuop_func op_103a_11_ff;
extern cpuop_func op_103b_11_nf;
extern cpuop_func op_103b_11_ff;
extern cpuop_func op_103c_11_nf;
extern cpuop_func op_103c_11_ff;
extern cpuop_func op_1080_11_nf;
extern cpuop_func op_1080_11_ff;
extern cpuop_func op_1090_11_nf;
extern cpuop_func op_1090_11_ff;
extern cpuop_func op_1098_11_nf;
extern cpuop_func op_1098_11_ff;
extern cpuop_func op_10a0_11_nf;
extern cpuop_func op_10a0_11_ff;
extern cpuop_func op_10a8_11_nf;
extern cpuop_func op_10a8_11_ff;
extern cpuop_func op_10b0_11_nf;
extern cpuop_func op_10b0_11_ff;
extern cpuop_func op_10b8_11_nf;
extern cpuop_func op_10b8_11_ff;
extern cpuop_func op_10b9_11_nf;
extern cpuop_func op_10b9_11_ff;
extern cpuop_func op_10ba_11_nf;
extern cpuop_func op_10ba_11_ff;
extern cpuop_func op_10bb_11_nf;
extern cpuop_func op_10bb_11_ff;
extern cpuop_func op_10bc_11_nf;
extern cpuop_func op_10bc_11_ff;
extern cpuop_func op_10c0_11_nf;
extern cpuop_func op_10c0_11_ff;
extern cpuop_func op_10d0_11_nf;
extern cpuop_func op_10d0_11_ff;
extern cpuop_func op_10d8_11_nf;
extern cpuop_func op_10d8_11_ff;
extern cpuop_func op_10e0_11_nf;
extern cpuop_func op_10e0_11_ff;
extern cpuop_func op_10e8_11_nf;
extern cpuop_func op_10e8_11_ff;
extern cpuop_func op_10f0_11_nf;
extern cpuop_func op_10f0_11_ff;
extern cpuop_func op_10f8_11_nf;
extern cpuop_func op_10f8_11_ff;
extern cpuop_func op_10f9_11_nf;
extern cpuop_func op_10f9_11_ff;
extern cpuop_func op_10fa_11_nf;
extern cpuop_func op_10fa_11_ff;
extern cpuop_func op_10fb_11_nf;
extern cpuop_func op_10fb_11_ff;
extern cpuop_func op_10fc_11_nf;
extern cpuop_func op_10fc_11_ff;
extern cpuop_func op_1100_11_nf;
extern cpuop_func op_1100_11_ff;
extern cpuop_func op_1110_11_nf;
extern cpuop_func op_1110_11_ff;
extern cpuop_func op_1118_11_nf;
extern cpuop_func op_1118_11_ff;
extern cpuop_func op_1120_11_nf;
extern cpuop_func op_1120_11_ff;
extern cpuop_func op_1128_11_nf;
extern cpuop_func op_1128_11_ff;
extern cpuop_func op_1130_11_nf;
extern cpuop_func op_1130_11_ff;
extern cpuop_func op_1138_11_nf;
extern cpuop_func op_1138_11_ff;
extern cpuop_func op_1139_11_nf;
extern cpuop_func op_1139_11_ff;
extern cpuop_func op_113a_11_nf;
extern cpuop_func op_113a_11_ff;
extern cpuop_func op_113b_11_nf;
extern cpuop_func op_113b_11_ff;
extern cpuop_func op_113c_11_nf;
extern cpuop_func op_113c_11_ff;
extern cpuop_func op_1140_11_nf;
extern cpuop_func op_1140_11_ff;
extern cpuop_func op_1150_11_nf;
extern cpuop_func op_1150_11_ff;
extern cpuop_func op_1158_11_nf;
extern cpuop_func op_1158_11_ff;
extern cpuop_func op_1160_11_nf;
extern cpuop_func op_1160_11_ff;
extern cpuop_func op_1168_11_nf;
extern cpuop_func op_1168_11_ff;
extern cpuop_func op_1170_11_nf;
extern cpuop_func op_1170_11_ff;
extern cpuop_func op_1178_11_nf;
extern cpuop_func op_1178_11_ff;
extern cpuop_func op_1179_11_nf;
extern cpuop_func op_1179_11_ff;
extern cpuop_func op_117a_11_nf;
extern cpuop_func op_117a_11_ff;
extern cpuop_func op_117b_11_nf;
extern cpuop_func op_117b_11_ff;
extern cpuop_func op_117c_11_nf;
extern cpuop_func op_117c_11_ff;
extern cpuop_func op_1180_11_nf;
extern cpuop_func op_1180_11_ff;
extern cpuop_func op_1190_11_nf;
extern cpuop_func op_1190_11_ff;
extern cpuop_func op_1198_11_nf;
extern cpuop_func op_1198_11_ff;
extern cpuop_func op_11a0_11_nf;
extern cpuop_func op_11a0_11_ff;
extern cpuop_func op_11a8_11_nf;
extern cpuop_func op_11a8_11_ff;
extern cpuop_func op_11b0_11_nf;
extern cpuop_func op_11b0_11_ff;
extern cpuop_func op_11b8_11_nf;
extern cpuop_func op_11b8_11_ff;
extern cpuop_func op_11b9_11_nf;
extern cpuop_func op_11b9_11_ff;
extern cpuop_func op_11ba_11_nf;
extern cpuop_func op_11ba_11_ff;
extern cpuop_func op_11bb_11_nf;
extern cpuop_func op_11bb_11_ff;
extern cpuop_func op_11bc_11_nf;
extern cpuop_func op_11bc_11_ff;
extern cpuop_func op_11c0_11_nf;
extern cpuop_func op_11c0_11_ff;
extern cpuop_func op_11d0_11_nf;
extern cpuop_func op_11d0_11_ff;
extern cpuop_func op_11d8_11_nf;
extern cpuop_func op_11d8_11_ff;
extern cpuop_func op_11e0_11_nf;
extern cpuop_func op_11e0_11_ff;
extern cpuop_func op_11e8_11_nf;
extern cpuop_func op_11e8_11_ff;
extern cpuop_func op_11f0_11_nf;
extern cpuop_func op_11f0_11_ff;
extern cpuop_func op_11f8_11_nf;
extern cpuop_func op_11f8_11_ff;
extern cpuop_func op_11f9_11_nf;
extern cpuop_func op_11f9_11_ff;
extern cpuop_func op_11fa_11_nf;
extern cpuop_func op_11fa_11_ff;
extern cpuop_func op_11fb_11_nf;
extern cpuop_func op_11fb_11_ff;
extern cpuop_func op_11fc_11_nf;
extern cpuop_func op_11fc_11_ff;
extern cpuop_func op_13c0_11_nf;
extern cpuop_func op_13c0_11_ff;
extern cpuop_func op_13d0_11_nf;
extern cpuop_func op_13d0_11_ff;
extern cpuop_func op_13d8_11_nf;
extern cpuop_func op_13d8_11_ff;
extern cpuop_func op_13e0_11_nf;
extern cpuop_func op_13e0_11_ff;
extern cpuop_func op_13e8_11_nf;
extern cpuop_func op_13e8_11_ff;
extern cpuop_func op_13f0_11_nf;
extern cpuop_func op_13f0_11_ff;
extern cpuop_func op_13f8_11_nf;
extern cpuop_func op_13f8_11_ff;
extern cpuop_func op_13f9_11_nf;
extern cpuop_func op_13f9_11_ff;
extern cpuop_func op_13fa_11_nf;
extern cpuop_func op_13fa_11_ff;
extern cpuop_func op_13fb_11_nf;
extern cpuop_func op_13fb_11_ff;
extern cpuop_func op_13fc_11_nf;
extern cpuop_func op_13fc_11_ff;
extern cpuop_func op_2000_11_nf;
extern cpuop_func op_2000_11_ff;
extern cpuop_func op_2008_11_nf;
extern cpuop_func op_2008_11_ff;
extern cpuop_func op_2010_11_nf;
extern cpuop_func op_2010_11_ff;
extern cpuop_func op_2018_11_nf;
extern cpuop_func op_2018_11_ff;
extern cpuop_func op_2020_11_nf;
extern cpuop_func op_2020_11_ff;
extern cpuop_func op_2028_11_nf;
extern cpuop_func op_2028_11_ff;
extern cpuop_func op_2030_11_nf;
extern cpuop_func op_2030_11_ff;
extern cpuop_func op_2038_11_nf;
extern cpuop_func op_2038_11_ff;
extern cpuop_func op_2039_11_nf;
extern cpuop_func op_2039_11_ff;
extern cpuop_func op_203a_11_nf;
extern cpuop_func op_203a_11_ff;
extern cpuop_func op_203b_11_nf;
extern cpuop_func op_203b_11_ff;
extern cpuop_func op_203c_11_nf;
extern cpuop_func op_203c_11_ff;
extern cpuop_func op_2040_11_nf;
extern cpuop_func op_2040_11_ff;
extern cpuop_func op_2048_11_nf;
extern cpuop_func op_2048_11_ff;
extern cpuop_func op_2050_11_nf;
extern cpuop_func op_2050_11_ff;
extern cpuop_func op_2058_11_nf;
extern cpuop_func op_2058_11_ff;
extern cpuop_func op_2060_11_nf;
extern cpuop_func op_2060_11_ff;
extern cpuop_func op_2068_11_nf;
extern cpuop_func op_2068_11_ff;
extern cpuop_func op_2070_11_nf;
extern cpuop_func op_2070_11_ff;
extern cpuop_func op_2078_11_nf;
extern cpuop_func op_2078_11_ff;
extern cpuop_func op_2079_11_nf;
extern cpuop_func op_2079_11_ff;
extern cpuop_func op_207a_11_nf;
extern cpuop_func op_207a_11_ff;
extern cpuop_func op_207b_11_nf;
extern cpuop_func op_207b_11_ff;
extern cpuop_func op_207c_11_nf;
extern cpuop_func op_207c_11_ff;
extern cpuop_func op_2080_11_nf;
extern cpuop_func op_2080_11_ff;
extern cpuop_func op_2088_11_nf;
extern cpuop_func op_2088_11_ff;
extern cpuop_func op_2090_11_nf;
extern cpuop_func op_2090_11_ff;
extern cpuop_func op_2098_11_nf;
extern cpuop_func op_2098_11_ff;
extern cpuop_func op_20a0_11_nf;
extern cpuop_func op_20a0_11_ff;
extern cpuop_func op_20a8_11_nf;
extern cpuop_func op_20a8_11_ff;
extern cpuop_func op_20b0_11_nf;
extern cpuop_func op_20b0_11_ff;
extern cpuop_func op_20b8_11_nf;
extern cpuop_func op_20b8_11_ff;
extern cpuop_func op_20b9_11_nf;
extern cpuop_func op_20b9_11_ff;
extern cpuop_func op_20ba_11_nf;
extern cpuop_func op_20ba_11_ff;
extern cpuop_func op_20bb_11_nf;
extern cpuop_func op_20bb_11_ff;
extern cpuop_func op_20bc_11_nf;
extern cpuop_func op_20bc_11_ff;
extern cpuop_func op_20c0_11_nf;
extern cpuop_func op_20c0_11_ff;
extern cpuop_func op_20c8_11_nf;
extern cpuop_func op_20c8_11_ff;
extern cpuop_func op_20d0_11_nf;
extern cpuop_func op_20d0_11_ff;
extern cpuop_func op_20d8_11_nf;
extern cpuop_func op_20d8_11_ff;
extern cpuop_func op_20e0_11_nf;
extern cpuop_func op_20e0_11_ff;
extern cpuop_func op_20e8_11_nf;
extern cpuop_func op_20e8_11_ff;
extern cpuop_func op_20f0_11_nf;
extern cpuop_func op_20f0_11_ff;
extern cpuop_func op_20f8_11_nf;
extern cpuop_func op_20f8_11_ff;
extern cpuop_func op_20f9_11_nf;
extern cpuop_func op_20f9_11_ff;
extern cpuop_func op_20fa_11_nf;
extern cpuop_func op_20fa_11_ff;
extern cpuop_func op_20fb_11_nf;
extern cpuop_func op_20fb_11_ff;
extern cpuop_func op_20fc_11_nf;
extern cpuop_func op_20fc_11_ff;
extern cpuop_func op_2100_11_nf;
extern cpuop_func op_2100_11_ff;
extern cpuop_func op_2108_11_nf;
extern cpuop_func op_2108_11_ff;
extern cpuop_func op_2110_11_nf;
extern cpuop_func op_2110_11_ff;
extern cpuop_func op_2118_11_nf;
extern cpuop_func op_2118_11_ff;
extern cpuop_func op_2120_11_nf;
extern cpuop_func op_2120_11_ff;
extern cpuop_func op_2128_11_nf;
extern cpuop_func op_2128_11_ff;
extern cpuop_func op_2130_11_nf;
extern cpuop_func op_2130_11_ff;
extern cpuop_func op_2138_11_nf;
extern cpuop_func op_2138_11_ff;
extern cpuop_func op_2139_11_nf;
extern cpuop_func op_2139_11_ff;
extern cpuop_func op_213a_11_nf;
extern cpuop_func op_213a_11_ff;
extern cpuop_func op_213b_11_nf;
extern cpuop_func op_213b_11_ff;
extern cpuop_func op_213c_11_nf;
extern cpuop_func op_213c_11_ff;
extern cpuop_func op_2140_11_nf;
extern cpuop_func op_2140_11_ff;
extern cpuop_func op_2148_11_nf;
extern cpuop_func op_2148_11_ff;
extern cpuop_func op_2150_11_nf;
extern cpuop_func op_2150_11_ff;
extern cpuop_func op_2158_11_nf;
extern cpuop_func op_2158_11_ff;
extern cpuop_func op_2160_11_nf;
extern cpuop_func op_2160_11_ff;
extern cpuop_func op_2168_11_nf;
extern cpuop_func op_2168_11_ff;
extern cpuop_func op_2170_11_nf;
extern cpuop_func op_2170_11_ff;
extern cpuop_func op_2178_11_nf;
extern cpuop_func op_2178_11_ff;
extern cpuop_func op_2179_11_nf;
extern cpuop_func op_2179_11_ff;
extern cpuop_func op_217a_11_nf;
extern cpuop_func op_217a_11_ff;
extern cpuop_func op_217b_11_nf;
extern cpuop_func op_217b_11_ff;
extern cpuop_func op_217c_11_nf;
extern cpuop_func op_217c_11_ff;
extern cpuop_func op_2180_11_nf;
extern cpuop_func op_2180_11_ff;
extern cpuop_func op_2188_11_nf;
extern cpuop_func op_2188_11_ff;
extern cpuop_func op_2190_11_nf;
extern cpuop_func op_2190_11_ff;
extern cpuop_func op_2198_11_nf;
extern cpuop_func op_2198_11_ff;
extern cpuop_func op_21a0_11_nf;
extern cpuop_func op_21a0_11_ff;
extern cpuop_func op_21a8_11_nf;
extern cpuop_func op_21a8_11_ff;
extern cpuop_func op_21b0_11_nf;
extern cpuop_func op_21b0_11_ff;
extern cpuop_func op_21b8_11_nf;
extern cpuop_func op_21b8_11_ff;
extern cpuop_func op_21b9_11_nf;
extern cpuop_func op_21b9_11_ff;
extern cpuop_func op_21ba_11_nf;
extern cpuop_func op_21ba_11_ff;
extern cpuop_func op_21bb_11_nf;
extern cpuop_func op_21bb_11_ff;
extern cpuop_func op_21bc_11_nf;
extern cpuop_func op_21bc_11_ff;
extern cpuop_func op_21c0_11_nf;
extern cpuop_func op_21c0_11_ff;
extern cpuop_func op_21c8_11_nf;
extern cpuop_func op_21c8_11_ff;
extern cpuop_func op_21d0_11_nf;
extern cpuop_func op_21d0_11_ff;
extern cpuop_func op_21d8_11_nf;
extern cpuop_func op_21d8_11_ff;
extern cpuop_func op_21e0_11_nf;
extern cpuop_func op_21e0_11_ff;
extern cpuop_func op_21e8_11_nf;
extern cpuop_func op_21e8_11_ff;
extern cpuop_func op_21f0_11_nf;
extern cpuop_func op_21f0_11_ff;
extern cpuop_func op_21f8_11_nf;
extern cpuop_func op_21f8_11_ff;
extern cpuop_func op_21f9_11_nf;
extern cpuop_func op_21f9_11_ff;
extern cpuop_func op_21fa_11_nf;
extern cpuop_func op_21fa_11_ff;
extern cpuop_func op_21fb_11_nf;
extern cpuop_func op_21fb_11_ff;
extern cpuop_func op_21fc_11_nf;
extern cpuop_func op_21fc_11_ff;
extern cpuop_func op_23c0_11_nf;
extern cpuop_func op_23c0_11_ff;
extern cpuop_func op_23c8_11_nf;
extern cpuop_func op_23c8_11_ff;
extern cpuop_func op_23d0_11_nf;
extern cpuop_func op_23d0_11_ff;
extern cpuop_func op_23d8_11_nf;
extern cpuop_func op_23d8_11_ff;
extern cpuop_func op_23e0_11_nf;
extern cpuop_func op_23e0_11_ff;
extern cpuop_func op_23e8_11_nf;
extern cpuop_func op_23e8_11_ff;
extern cpuop_func op_23f0_11_nf;
extern cpuop_func op_23f0_11_ff;
extern cpuop_func op_23f8_11_nf;
extern cpuop_func op_23f8_11_ff;
extern cpuop_func op_23f9_11_nf;
extern cpuop_func op_23f9_11_ff;
extern cpuop_func op_23fa_11_nf;
extern cpuop_func op_23fa_11_ff;
extern cpuop_func op_23fb_11_nf;
extern cpuop_func op_23fb_11_ff;
extern cpuop_func op_23fc_11_nf;
extern cpuop_func op_23fc_11_ff;
extern cpuop_func op_3000_11_nf;
extern cpuop_func op_3000_11_ff;
extern cpuop_func op_3008_11_nf;
extern cpuop_func op_3008_11_ff;
extern cpuop_func op_3010_11_nf;
extern cpuop_func op_3010_11_ff;
extern cpuop_func op_3018_11_nf;
extern cpuop_func op_3018_11_ff;
extern cpuop_func op_3020_11_nf;
extern cpuop_func op_3020_11_ff;
extern cpuop_func op_3028_11_nf;
extern cpuop_func op_3028_11_ff;
extern cpuop_func op_3030_11_nf;
extern cpuop_func op_3030_11_ff;
extern cpuop_func op_3038_11_nf;
extern cpuop_func op_3038_11_ff;
extern cpuop_func op_3039_11_nf;
extern cpuop_func op_3039_11_ff;
extern cpuop_func op_303a_11_nf;
extern cpuop_func op_303a_11_ff;
extern cpuop_func op_303b_11_nf;
extern cpuop_func op_303b_11_ff;
extern cpuop_func op_303c_11_nf;
extern cpuop_func op_303c_11_ff;
extern cpuop_func op_3040_11_nf;
extern cpuop_func op_3040_11_ff;
extern cpuop_func op_3048_11_nf;
extern cpuop_func op_3048_11_ff;
extern cpuop_func op_3050_11_nf;
extern cpuop_func op_3050_11_ff;
extern cpuop_func op_3058_11_nf;
extern cpuop_func op_3058_11_ff;
extern cpuop_func op_3060_11_nf;
extern cpuop_func op_3060_11_ff;
extern cpuop_func op_3068_11_nf;
extern cpuop_func op_3068_11_ff;
extern cpuop_func op_3070_11_nf;
extern cpuop_func op_3070_11_ff;
extern cpuop_func op_3078_11_nf;
extern cpuop_func op_3078_11_ff;
extern cpuop_func op_3079_11_nf;
extern cpuop_func op_3079_11_ff;
extern cpuop_func op_307a_11_nf;
extern cpuop_func op_307a_11_ff;
extern cpuop_func op_307b_11_nf;
extern cpuop_func op_307b_11_ff;
extern cpuop_func op_307c_11_nf;
extern cpuop_func op_307c_11_ff;
extern cpuop_func op_3080_11_nf;
extern cpuop_func op_3080_11_ff;
extern cpuop_func op_3088_11_nf;
extern cpuop_func op_3088_11_ff;
extern cpuop_func op_3090_11_nf;
extern cpuop_func op_3090_11_ff;
extern cpuop_func op_3098_11_nf;
extern cpuop_func op_3098_11_ff;
extern cpuop_func op_30a0_11_nf;
extern cpuop_func op_30a0_11_ff;
extern cpuop_func op_30a8_11_nf;
extern cpuop_func op_30a8_11_ff;
extern cpuop_func op_30b0_11_nf;
extern cpuop_func op_30b0_11_ff;
extern cpuop_func op_30b8_11_nf;
extern cpuop_func op_30b8_11_ff;
extern cpuop_func op_30b9_11_nf;
extern cpuop_func op_30b9_11_ff;
extern cpuop_func op_30ba_11_nf;
extern cpuop_func op_30ba_11_ff;
extern cpuop_func op_30bb_11_nf;
extern cpuop_func op_30bb_11_ff;
extern cpuop_func op_30bc_11_nf;
extern cpuop_func op_30bc_11_ff;
extern cpuop_func op_30c0_11_nf;
extern cpuop_func op_30c0_11_ff;
extern cpuop_func op_30c8_11_nf;
extern cpuop_func op_30c8_11_ff;
extern cpuop_func op_30d0_11_nf;
extern cpuop_func op_30d0_11_ff;
extern cpuop_func op_30d8_11_nf;
extern cpuop_func op_30d8_11_ff;
extern cpuop_func op_30e0_11_nf;
extern cpuop_func op_30e0_11_ff;
extern cpuop_func op_30e8_11_nf;
extern cpuop_func op_30e8_11_ff;
extern cpuop_func op_30f0_11_nf;
extern cpuop_func op_30f0_11_ff;
extern cpuop_func op_30f8_11_nf;
extern cpuop_func op_30f8_11_ff;
extern cpuop_func op_30f9_11_nf;
extern cpuop_func op_30f9_11_ff;
extern cpuop_func op_30fa_11_nf;
extern cpuop_func op_30fa_11_ff;
extern cpuop_func op_30fb_11_nf;
extern cpuop_func op_30fb_11_ff;
extern cpuop_func op_30fc_11_nf;
extern cpuop_func op_30fc_11_ff;
extern cpuop_func op_3100_11_nf;
extern cpuop_func op_3100_11_ff;
extern cpuop_func op_3108_11_nf;
extern cpuop_func op_3108_11_ff;
extern cpuop_func op_3110_11_nf;
extern cpuop_func op_3110_11_ff;
extern cpuop_func op_3118_11_nf;
extern cpuop_func op_3118_11_ff;
extern cpuop_func op_3120_11_nf;
extern cpuop_func op_3120_11_ff;
extern cpuop_func op_3128_11_nf;
extern cpuop_func op_3128_11_ff;
extern cpuop_func op_3130_11_nf;
extern cpuop_func op_3130_11_ff;
extern cpuop_func op_3138_11_nf;
extern cpuop_func op_3138_11_ff;
extern cpuop_func op_3139_11_nf;
extern cpuop_func op_3139_11_ff;
extern cpuop_func op_313a_11_nf;
extern cpuop_func op_313a_11_ff;
extern cpuop_func op_313b_11_nf;
extern cpuop_func op_313b_11_ff;
extern cpuop_func op_313c_11_nf;
extern cpuop_func op_313c_11_ff;
extern cpuop_func op_3140_11_nf;
extern cpuop_func op_3140_11_ff;
extern cpuop_func op_3148_11_nf;
extern cpuop_func op_3148_11_ff;
extern cpuop_func op_3150_11_nf;
extern cpuop_func op_3150_11_ff;
extern cpuop_func op_3158_11_nf;
extern cpuop_func op_3158_11_ff;
extern cpuop_func op_3160_11_nf;
extern cpuop_func op_3160_11_ff;
extern cpuop_func op_3168_11_nf;
extern cpuop_func op_3168_11_ff;
extern cpuop_func op_3170_11_nf;
extern cpuop_func op_3170_11_ff;
extern cpuop_func op_3178_11_nf;
extern cpuop_func op_3178_11_ff;
extern cpuop_func op_3179_11_nf;
extern cpuop_func op_3179_11_ff;
extern cpuop_func op_317a_11_nf;
extern cpuop_func op_317a_11_ff;
extern cpuop_func op_317b_11_nf;
extern cpuop_func op_317b_11_ff;
extern cpuop_func op_317c_11_nf;
extern cpuop_func op_317c_11_ff;
extern cpuop_func op_3180_11_nf;
extern cpuop_func op_3180_11_ff;
extern cpuop_func op_3188_11_nf;
extern cpuop_func op_3188_11_ff;
extern cpuop_func op_3190_11_nf;
extern cpuop_func op_3190_11_ff;
extern cpuop_func op_3198_11_nf;
extern cpuop_func op_3198_11_ff;
extern cpuop_func op_31a0_11_nf;
extern cpuop_func op_31a0_11_ff;
extern cpuop_func op_31a8_11_nf;
extern cpuop_func op_31a8_11_ff;
extern cpuop_func op_31b0_11_nf;
extern cpuop_func op_31b0_11_ff;
extern cpuop_func op_31b8_11_nf;
extern cpuop_func op_31b8_11_ff;
extern cpuop_func op_31b9_11_nf;
extern cpuop_func op_31b9_11_ff;
extern cpuop_func op_31ba_11_nf;
extern cpuop_func op_31ba_11_ff;
extern cpuop_func op_31bb_11_nf;
extern cpuop_func op_31bb_11_ff;
extern cpuop_func op_31bc_11_nf;
extern cpuop_func op_31bc_11_ff;
extern cpuop_func op_31c0_11_nf;
extern cpuop_func op_31c0_11_ff;
extern cpuop_func op_31c8_11_nf;
extern cpuop_func op_31c8_11_ff;
extern cpuop_func op_31d0_11_nf;
extern cpuop_func op_31d0_11_ff;
extern cpuop_func op_31d8_11_nf;
extern cpuop_func op_31d8_11_ff;
extern cpuop_func op_31e0_11_nf;
extern cpuop_func op_31e0_11_ff;
extern cpuop_func op_31e8_11_nf;
extern cpuop_func op_31e8_11_ff;
extern cpuop_func op_31f0_11_nf;
extern cpuop_func op_31f0_11_ff;
extern cpuop_func op_31f8_11_nf;
extern cpuop_func op_31f8_11_ff;
extern cpuop_func op_31f9_11_nf;
extern cpuop_func op_31f9_11_ff;
extern cpuop_func op_31fa_11_nf;
extern cpuop_func op_31fa_11_ff;
extern cpuop_func op_31fb_11_nf;
extern cpuop_func op_31fb_11_ff;
extern cpuop_func op_31fc_11_nf;
extern cpuop_func op_31fc_11_ff;
extern cpuop_func op_33c0_11_nf;
extern cpuop_func op_33c0_11_ff;
extern cpuop_func op_33c8_11_nf;
extern cpuop_func op_33c8_11_ff;
extern cpuop_func op_33d0_11_nf;
extern cpuop_func op_33d0_11_ff;
extern cpuop_func op_33d8_11_nf;
extern cpuop_func op_33d8_11_ff;
extern cpuop_func op_33e0_11_nf;
extern cpuop_func op_33e0_11_ff;
extern cpuop_func op_33e8_11_nf;
extern cpuop_func op_33e8_11_ff;
extern cpuop_func op_33f0_11_nf;
extern cpuop_func op_33f0_11_ff;
extern cpuop_func op_33f8_11_nf;
extern cpuop_func op_33f8_11_ff;
extern cpuop_func op_33f9_11_nf;
extern cpuop_func op_33f9_11_ff;
extern cpuop_func op_33fa_11_nf;
extern cpuop_func op_33fa_11_ff;
extern cpuop_func op_33fb_11_nf;
extern cpuop_func op_33fb_11_ff;
extern cpuop_func op_33fc_11_nf;
extern cpuop_func op_33fc_11_ff;
extern cpuop_func op_4000_11_nf;
extern cpuop_func op_4000_11_ff;
extern cpuop_func op_4010_11_nf;
extern cpuop_func op_4010_11_ff;
extern cpuop_func op_4018_11_nf;
extern cpuop_func op_4018_11_ff;
extern cpuop_func op_4020_11_nf;
extern cpuop_func op_4020_11_ff;
extern cpuop_func op_4028_11_nf;
extern cpuop_func op_4028_11_ff;
extern cpuop_func op_4030_11_nf;
extern cpuop_func op_4030_11_ff;
extern cpuop_func op_4038_11_nf;
extern cpuop_func op_4038_11_ff;
extern cpuop_func op_4039_11_nf;
extern cpuop_func op_4039_11_ff;
extern cpuop_func op_4040_11_nf;
extern cpuop_func op_4040_11_ff;
extern cpuop_func op_4050_11_nf;
extern cpuop_func op_4050_11_ff;
extern cpuop_func op_4058_11_nf;
extern cpuop_func op_4058_11_ff;
extern cpuop_func op_4060_11_nf;
extern cpuop_func op_4060_11_ff;
extern cpuop_func op_4068_11_nf;
extern cpuop_func op_4068_11_ff;
extern cpuop_func op_4070_11_nf;
extern cpuop_func op_4070_11_ff;
extern cpuop_func op_4078_11_nf;
extern cpuop_func op_4078_11_ff;
extern cpuop_func op_4079_11_nf;
extern cpuop_func op_4079_11_ff;
extern cpuop_func op_4080_11_nf;
extern cpuop_func op_4080_11_ff;
extern cpuop_func op_4090_11_nf;
extern cpuop_func op_4090_11_ff;
extern cpuop_func op_4098_11_nf;
extern cpuop_func op_4098_11_ff;
extern cpuop_func op_40a0_11_nf;
extern cpuop_func op_40a0_11_ff;
extern cpuop_func op_40a8_11_nf;
extern cpuop_func op_40a8_11_ff;
extern cpuop_func op_40b0_11_nf;
extern cpuop_func op_40b0_11_ff;
extern cpuop_func op_40b8_11_nf;
extern cpuop_func op_40b8_11_ff;
extern cpuop_func op_40b9_11_nf;
extern cpuop_func op_40b9_11_ff;
extern cpuop_func op_40c0_11_nf;
extern cpuop_func op_40c0_11_ff;
extern cpuop_func op_40d0_11_nf;
extern cpuop_func op_40d0_11_ff;
extern cpuop_func op_40d8_11_nf;
extern cpuop_func op_40d8_11_ff;
extern cpuop_func op_40e0_11_nf;
extern cpuop_func op_40e0_11_ff;
extern cpuop_func op_40e8_11_nf;
extern cpuop_func op_40e8_11_ff;
extern cpuop_func op_40f0_11_nf;
extern cpuop_func op_40f0_11_ff;
extern cpuop_func op_40f8_11_nf;
extern cpuop_func op_40f8_11_ff;
extern cpuop_func op_40f9_11_nf;
extern cpuop_func op_40f9_11_ff;
extern cpuop_func op_4180_11_nf;
extern cpuop_func op_4180_11_ff;
extern cpuop_func op_4190_11_nf;
extern cpuop_func op_4190_11_ff;
extern cpuop_func op_4198_11_nf;
extern cpuop_func op_4198_11_ff;
extern cpuop_func op_41a0_11_nf;
extern cpuop_func op_41a0_11_ff;
extern cpuop_func op_41a8_11_nf;
extern cpuop_func op_41a8_11_ff;
extern cpuop_func op_41b0_11_nf;
extern cpuop_func op_41b0_11_ff;
extern cpuop_func op_41b8_11_nf;
extern cpuop_func op_41b8_11_ff;
extern cpuop_func op_41b9_11_nf;
extern cpuop_func op_41b9_11_ff;
extern cpuop_func op_41ba_11_nf;
extern cpuop_func op_41ba_11_ff;
extern cpuop_func op_41bb_11_nf;
extern cpuop_func op_41bb_11_ff;
extern cpuop_func op_41bc_11_nf;
extern cpuop_func op_41bc_11_ff;
extern cpuop_func op_41d0_11_nf;
extern cpuop_func op_41d0_11_ff;
extern cpuop_func op_41e8_11_nf;
extern cpuop_func op_41e8_11_ff;
extern cpuop_func op_41f0_11_nf;
extern cpuop_func op_41f0_11_ff;
extern cpuop_func op_41f8_11_nf;
extern cpuop_func op_41f8_11_ff;
extern cpuop_func op_41f9_11_nf;
extern cpuop_func op_41f9_11_ff;
extern cpuop_func op_41fa_11_nf;
extern cpuop_func op_41fa_11_ff;
extern cpuop_func op_41fb_11_nf;
extern cpuop_func op_41fb_11_ff;
extern cpuop_func op_4200_11_nf;
extern cpuop_func op_4200_11_ff;
extern cpuop_func op_4210_11_nf;
extern cpuop_func op_4210_11_ff;
extern cpuop_func op_4218_11_nf;
extern cpuop_func op_4218_11_ff;
extern cpuop_func op_4220_11_nf;
extern cpuop_func op_4220_11_ff;
extern cpuop_func op_4228_11_nf;
extern cpuop_func op_4228_11_ff;
extern cpuop_func op_4230_11_nf;
extern cpuop_func op_4230_11_ff;
extern cpuop_func op_4238_11_nf;
extern cpuop_func op_4238_11_ff;
extern cpuop_func op_4239_11_nf;
extern cpuop_func op_4239_11_ff;
extern cpuop_func op_4240_11_nf;
extern cpuop_func op_4240_11_ff;
extern cpuop_func op_4250_11_nf;
extern cpuop_func op_4250_11_ff;
extern cpuop_func op_4258_11_nf;
extern cpuop_func op_4258_11_ff;
extern cpuop_func op_4260_11_nf;
extern cpuop_func op_4260_11_ff;
extern cpuop_func op_4268_11_nf;
extern cpuop_func op_4268_11_ff;
extern cpuop_func op_4270_11_nf;
extern cpuop_func op_4270_11_ff;
extern cpuop_func op_4278_11_nf;
extern cpuop_func op_4278_11_ff;
extern cpuop_func op_4279_11_nf;
extern cpuop_func op_4279_11_ff;
extern cpuop_func op_4280_11_nf;
extern cpuop_func op_4280_11_ff;
extern cpuop_func op_4290_11_nf;
extern cpuop_func op_4290_11_ff;
extern cpuop_func op_4298_11_nf;
extern cpuop_func op_4298_11_ff;
extern cpuop_func op_42a0_11_nf;
extern cpuop_func op_42a0_11_ff;
extern cpuop_func op_42a8_11_nf;
extern cpuop_func op_42a8_11_ff;
extern cpuop_func op_42b0_11_nf;
extern cpuop_func op_42b0_11_ff;
extern cpuop_func op_42b8_11_nf;
extern cpuop_func op_42b8_11_ff;
extern cpuop_func op_42b9_11_nf;
extern cpuop_func op_42b9_11_ff;
extern cpuop_func op_42c0_11_nf;
extern cpuop_func op_42c0_11_ff;
extern cpuop_func op_42d0_11_nf;
extern cpuop_func op_42d0_11_ff;
extern cpuop_func op_42d8_11_nf;
extern cpuop_func op_42d8_11_ff;
extern cpuop_func op_42e0_11_nf;
extern cpuop_func op_42e0_11_ff;
extern cpuop_func op_42e8_11_nf;
extern cpuop_func op_42e8_11_ff;
extern cpuop_func op_42f0_11_nf;
extern cpuop_func op_42f0_11_ff;
extern cpuop_func op_42f8_11_nf;
extern cpuop_func op_42f8_11_ff;
extern cpuop_func op_42f9_11_nf;
extern cpuop_func op_42f9_11_ff;
extern cpuop_func op_4400_11_nf;
extern cpuop_func op_4400_11_ff;
extern cpuop_func op_4410_11_nf;
extern cpuop_func op_4410_11_ff;
extern cpuop_func op_4418_11_nf;
extern cpuop_func op_4418_11_ff;
extern cpuop_func op_4420_11_nf;
extern cpuop_func op_4420_11_ff;
extern cpuop_func op_4428_11_nf;
extern cpuop_func op_4428_11_ff;
extern cpuop_func op_4430_11_nf;
extern cpuop_func op_4430_11_ff;
extern cpuop_func op_4438_11_nf;
extern cpuop_func op_4438_11_ff;
extern cpuop_func op_4439_11_nf;
extern cpuop_func op_4439_11_ff;
extern cpuop_func op_4440_11_nf;
extern cpuop_func op_4440_11_ff;
extern cpuop_func op_4450_11_nf;
extern cpuop_func op_4450_11_ff;
extern cpuop_func op_4458_11_nf;
extern cpuop_func op_4458_11_ff;
extern cpuop_func op_4460_11_nf;
extern cpuop_func op_4460_11_ff;
extern cpuop_func op_4468_11_nf;
extern cpuop_func op_4468_11_ff;
extern cpuop_func op_4470_11_nf;
extern cpuop_func op_4470_11_ff;
extern cpuop_func op_4478_11_nf;
extern cpuop_func op_4478_11_ff;
extern cpuop_func op_4479_11_nf;
extern cpuop_func op_4479_11_ff;
extern cpuop_func op_4480_11_nf;
extern cpuop_func op_4480_11_ff;
extern cpuop_func op_4490_11_nf;
extern cpuop_func op_4490_11_ff;
extern cpuop_func op_4498_11_nf;
extern cpuop_func op_4498_11_ff;
extern cpuop_func op_44a0_11_nf;
extern cpuop_func op_44a0_11_ff;
extern cpuop_func op_44a8_11_nf;
extern cpuop_func op_44a8_11_ff;
extern cpuop_func op_44b0_11_nf;
extern cpuop_func op_44b0_11_ff;
extern cpuop_func op_44b8_11_nf;
extern cpuop_func op_44b8_11_ff;
extern cpuop_func op_44b9_11_nf;
extern cpuop_func op_44b9_11_ff;
extern cpuop_func op_44c0_11_nf;
extern cpuop_func op_44c0_11_ff;
extern cpuop_func op_44d0_11_nf;
extern cpuop_func op_44d0_11_ff;
extern cpuop_func op_44d8_11_nf;
extern cpuop_func op_44d8_11_ff;
extern cpuop_func op_44e0_11_nf;
extern cpuop_func op_44e0_11_ff;
extern cpuop_func op_44e8_11_nf;
extern cpuop_func op_44e8_11_ff;
extern cpuop_func op_44f0_11_nf;
extern cpuop_func op_44f0_11_ff;
extern cpuop_func op_44f8_11_nf;
extern cpuop_func op_44f8_11_ff;
extern cpuop_func op_44f9_11_nf;
extern cpuop_func op_44f9_11_ff;
extern cpuop_func op_44fa_11_nf;
extern cpuop_func op_44fa_11_ff;
extern cpuop_func op_44fb_11_nf;
extern cpuop_func op_44fb_11_ff;
extern cpuop_func op_44fc_11_nf;
extern cpuop_func op_44fc_11_ff;
extern cpuop_func op_4600_11_nf;
extern cpuop_func op_4600_11_ff;
extern cpuop_func op_4610_11_nf;
extern cpuop_func op_4610_11_ff;
extern cpuop_func op_4618_11_nf;
extern cpuop_func op_4618_11_ff;
extern cpuop_func op_4620_11_nf;
extern cpuop_func op_4620_11_ff;
extern cpuop_func op_4628_11_nf;
extern cpuop_func op_4628_11_ff;
extern cpuop_func op_4630_11_nf;
extern cpuop_func op_4630_11_ff;
extern cpuop_func op_4638_11_nf;
extern cpuop_func op_4638_11_ff;
extern cpuop_func op_4639_11_nf;
extern cpuop_func op_4639_11_ff;
extern cpuop_func op_4640_11_nf;
extern cpuop_func op_4640_11_ff;
extern cpuop_func op_4650_11_nf;
extern cpuop_func op_4650_11_ff;
extern cpuop_func op_4658_11_nf;
extern cpuop_func op_4658_11_ff;
extern cpuop_func op_4660_11_nf;
extern cpuop_func op_4660_11_ff;
extern cpuop_func op_4668_11_nf;
extern cpuop_func op_4668_11_ff;
extern cpuop_func op_4670_11_nf;
extern cpuop_func op_4670_11_ff;
extern cpuop_func op_4678_11_nf;
extern cpuop_func op_4678_11_ff;
extern cpuop_func op_4679_11_nf;
extern cpuop_func op_4679_11_ff;
extern cpuop_func op_4680_11_nf;
extern cpuop_func op_4680_11_ff;
extern cpuop_func op_4690_11_nf;
extern cpuop_func op_4690_11_ff;
extern cpuop_func op_4698_11_nf;
extern cpuop_func op_4698_11_ff;
extern cpuop_func op_46a0_11_nf;
extern cpuop_func op_46a0_11_ff;
extern cpuop_func op_46a8_11_nf;
extern cpuop_func op_46a8_11_ff;
extern cpuop_func op_46b0_11_nf;
extern cpuop_func op_46b0_11_ff;
extern cpuop_func op_46b8_11_nf;
extern cpuop_func op_46b8_11_ff;
extern cpuop_func op_46b9_11_nf;
extern cpuop_func op_46b9_11_ff;
extern cpuop_func op_46c0_11_nf;
extern cpuop_func op_46c0_11_ff;
extern cpuop_func op_46d0_11_nf;
extern cpuop_func op_46d0_11_ff;
extern cpuop_func op_46d8_11_nf;
extern cpuop_func op_46d8_11_ff;
extern cpuop_func op_46e0_11_nf;
extern cpuop_func op_46e0_11_ff;
extern cpuop_func op_46e8_11_nf;
extern cpuop_func op_46e8_11_ff;
extern cpuop_func op_46f0_11_nf;
extern cpuop_func op_46f0_11_ff;
extern cpuop_func op_46f8_11_nf;
extern cpuop_func op_46f8_11_ff;
extern cpuop_func op_46f9_11_nf;
extern cpuop_func op_46f9_11_ff;
extern cpuop_func op_46fa_11_nf;
extern cpuop_func op_46fa_11_ff;
extern cpuop_func op_46fb_11_nf;
extern cpuop_func op_46fb_11_ff;
extern cpuop_func op_46fc_11_nf;
extern cpuop_func op_46fc_11_ff;
extern cpuop_func op_4800_11_nf;
extern cpuop_func op_4800_11_ff;
extern cpuop_func op_4810_11_nf;
extern cpuop_func op_4810_11_ff;
extern cpuop_func op_4818_11_nf;
extern cpuop_func op_4818_11_ff;
extern cpuop_func op_4820_11_nf;
extern cpuop_func op_4820_11_ff;
extern cpuop_func op_4828_11_nf;
extern cpuop_func op_4828_11_ff;
extern cpuop_func op_4830_11_nf;
extern cpuop_func op_4830_11_ff;
extern cpuop_func op_4838_11_nf;
extern cpuop_func op_4838_11_ff;
extern cpuop_func op_4839_11_nf;
extern cpuop_func op_4839_11_ff;
extern cpuop_func op_4840_11_nf;
extern cpuop_func op_4840_11_ff;
extern cpuop_func op_4850_11_nf;
extern cpuop_func op_4850_11_ff;
extern cpuop_func op_4868_11_nf;
extern cpuop_func op_4868_11_ff;
extern cpuop_func op_4870_11_nf;
extern cpuop_func op_4870_11_ff;
extern cpuop_func op_4878_11_nf;
extern cpuop_func op_4878_11_ff;
extern cpuop_func op_4879_11_nf;
extern cpuop_func op_4879_11_ff;
extern cpuop_func op_487a_11_nf;
extern cpuop_func op_487a_11_ff;
extern cpuop_func op_487b_11_nf;
extern cpuop_func op_487b_11_ff;
extern cpuop_func op_4880_11_nf;
extern cpuop_func op_4880_11_ff;
extern cpuop_func op_4890_11_nf;
extern cpuop_func op_4890_11_ff;
extern cpuop_func op_48a0_11_nf;
extern cpuop_func op_48a0_11_ff;
extern cpuop_func op_48a8_11_nf;
extern cpuop_func op_48a8_11_ff;
extern cpuop_func op_48b0_11_nf;
extern cpuop_func op_48b0_11_ff;
extern cpuop_func op_48b8_11_nf;
extern cpuop_func op_48b8_11_ff;
extern cpuop_func op_48b9_11_nf;
extern cpuop_func op_48b9_11_ff;
extern cpuop_func op_48c0_11_nf;
extern cpuop_func op_48c0_11_ff;
extern cpuop_func op_48d0_11_nf;
extern cpuop_func op_48d0_11_ff;
extern cpuop_func op_48e0_11_nf;
extern cpuop_func op_48e0_11_ff;
extern cpuop_func op_48e8_11_nf;
extern cpuop_func op_48e8_11_ff;
extern cpuop_func op_48f0_11_nf;
extern cpuop_func op_48f0_11_ff;
extern cpuop_func op_48f8_11_nf;
extern cpuop_func op_48f8_11_ff;
extern cpuop_func op_48f9_11_nf;
extern cpuop_func op_48f9_11_ff;
extern cpuop_func op_4a00_11_nf;
extern cpuop_func op_4a00_11_ff;
extern cpuop_func op_4a10_11_nf;
extern cpuop_func op_4a10_11_ff;
extern cpuop_func op_4a18_11_nf;
extern cpuop_func op_4a18_11_ff;
extern cpuop_func op_4a20_11_nf;
extern cpuop_func op_4a20_11_ff;
extern cpuop_func op_4a28_11_nf;
extern cpuop_func op_4a28_11_ff;
extern cpuop_func op_4a30_11_nf;
extern cpuop_func op_4a30_11_ff;
extern cpuop_func op_4a38_11_nf;
extern cpuop_func op_4a38_11_ff;
extern cpuop_func op_4a39_11_nf;
extern cpuop_func op_4a39_11_ff;
extern cpuop_func op_4a40_11_nf;
extern cpuop_func op_4a40_11_ff;
extern cpuop_func op_4a50_11_nf;
extern cpuop_func op_4a50_11_ff;
extern cpuop_func op_4a58_11_nf;
extern cpuop_func op_4a58_11_ff;
extern cpuop_func op_4a60_11_nf;
extern cpuop_func op_4a60_11_ff;
extern cpuop_func op_4a68_11_nf;
extern cpuop_func op_4a68_11_ff;
extern cpuop_func op_4a70_11_nf;
extern cpuop_func op_4a70_11_ff;
extern cpuop_func op_4a78_11_nf;
extern cpuop_func op_4a78_11_ff;
extern cpuop_func op_4a79_11_nf;
extern cpuop_func op_4a79_11_ff;
extern cpuop_func op_4a80_11_nf;
extern cpuop_func op_4a80_11_ff;
extern cpuop_func op_4a90_11_nf;
extern cpuop_func op_4a90_11_ff;
extern cpuop_func op_4a98_11_nf;
extern cpuop_func op_4a98_11_ff;
extern cpuop_func op_4aa0_11_nf;
extern cpuop_func op_4aa0_11_ff;
extern cpuop_func op_4aa8_11_nf;
extern cpuop_func op_4aa8_11_ff;
extern cpuop_func op_4ab0_11_nf;
extern cpuop_func op_4ab0_11_ff;
extern cpuop_func op_4ab8_11_nf;
extern cpuop_func op_4ab8_11_ff;
extern cpuop_func op_4ab9_11_nf;
extern cpuop_func op_4ab9_11_ff;
extern cpuop_func op_4ac0_11_nf;
extern cpuop_func op_4ac0_11_ff;
extern cpuop_func op_4ad0_11_nf;
extern cpuop_func op_4ad0_11_ff;
extern cpuop_func op_4ad8_11_nf;
extern cpuop_func op_4ad8_11_ff;
extern cpuop_func op_4ae0_11_nf;
extern cpuop_func op_4ae0_11_ff;
extern cpuop_func op_4ae8_11_nf;
extern cpuop_func op_4ae8_11_ff;
extern cpuop_func op_4af0_11_nf;
extern cpuop_func op_4af0_11_ff;
extern cpuop_func op_4af8_11_nf;
extern cpuop_func op_4af8_11_ff;
extern cpuop_func op_4af9_11_nf;
extern cpuop_func op_4af9_11_ff;
extern cpuop_func op_4c90_11_nf;
extern cpuop_func op_4c90_11_ff;
extern cpuop_func op_4c98_11_nf;
extern cpuop_func op_4c98_11_ff;
extern cpuop_func op_4ca8_11_nf;
extern cpuop_func op_4ca8_11_ff;
extern cpuop_func op_4cb0_11_nf;
extern cpuop_func op_4cb0_11_ff;
extern cpuop_func op_4cb8_11_nf;
extern cpuop_func op_4cb8_11_ff;
extern cpuop_func op_4cb9_11_nf;
extern cpuop_func op_4cb9_11_ff;
extern cpuop_func op_4cba_11_nf;
extern cpuop_func op_4cba_11_ff;
extern cpuop_func op_4cbb_11_nf;
extern cpuop_func op_4cbb_11_ff;
extern cpuop_func op_4cd0_11_nf;
extern cpuop_func op_4cd0_11_ff;
extern cpuop_func op_4cd8_11_nf;
extern cpuop_func op_4cd8_11_ff;
extern cpuop_func op_4ce8_11_nf;
extern cpuop_func op_4ce8_11_ff;
extern cpuop_func op_4cf0_11_nf;
extern cpuop_func op_4cf0_11_ff;
extern cpuop_func op_4cf8_11_nf;
extern cpuop_func op_4cf8_11_ff;
extern cpuop_func op_4cf9_11_nf;
extern cpuop_func op_4cf9_11_ff;
extern cpuop_func op_4cfa_11_nf;
extern cpuop_func op_4cfa_11_ff;
extern cpuop_func op_4cfb_11_nf;
extern cpuop_func op_4cfb_11_ff;
extern cpuop_func op_4e40_11_nf;
extern cpuop_func op_4e40_11_ff;
extern cpuop_func op_4e50_11_nf;
extern cpuop_func op_4e50_11_ff;
extern cpuop_func op_4e58_11_nf;
extern cpuop_func op_4e58_11_ff;
extern cpuop_func op_4e60_11_nf;
extern cpuop_func op_4e60_11_ff;
extern cpuop_func op_4e68_11_nf;
extern cpuop_func op_4e68_11_ff;
extern cpuop_func op_4e70_11_nf;
extern cpuop_func op_4e70_11_ff;
extern cpuop_func op_4e71_11_nf;
extern cpuop_func op_4e71_11_ff;
extern cpuop_func op_4e72_11_nf;
extern cpuop_func op_4e72_11_ff;
extern cpuop_func op_4e73_11_nf;
extern cpuop_func op_4e73_11_ff;
extern cpuop_func op_4e74_11_nf;
extern cpuop_func op_4e74_11_ff;
extern cpuop_func op_4e75_11_nf;
extern cpuop_func op_4e75_11_ff;
extern cpuop_func op_4e76_11_nf;
extern cpuop_func op_4e76_11_ff;
extern cpuop_func op_4e77_11_nf;
extern cpuop_func op_4e77_11_ff;
extern cpuop_func op_4e7a_11_nf;
extern cpuop_func op_4e7a_11_ff;
extern cpuop_func op_4e7b_11_nf;
extern cpuop_func op_4e7b_11_ff;
extern cpuop_func op_4e90_11_nf;
extern cpuop_func op_4e90_11_ff;
extern cpuop_func op_4ea8_11_nf;
extern cpuop_func op_4ea8_11_ff;
extern cpuop_func op_4eb0_11_nf;
extern cpuop_func op_4eb0_11_ff;
extern cpuop_func op_4eb8_11_nf;
extern cpuop_func op_4eb8_11_ff;
extern cpuop_func op_4eb9_11_nf;
extern cpuop_func op_4eb9_11_ff;
extern cpuop_func op_4eba_11_nf;
extern cpuop_func op_4eba_11_ff;
extern cpuop_func op_4ebb_11_nf;
extern cpuop_func op_4ebb_11_ff;
extern cpuop_func op_4ed0_11_nf;
extern cpuop_func op_4ed0_11_ff;
extern cpuop_func op_4ee8_11_nf;
extern cpuop_func op_4ee8_11_ff;
extern cpuop_func op_4ef0_11_nf;
extern cpuop_func op_4ef0_11_ff;
extern cpuop_func op_4ef8_11_nf;
extern cpuop_func op_4ef8_11_ff;
extern cpuop_func op_4ef9_11_nf;
extern cpuop_func op_4ef9_11_ff;
extern cpuop_func op_4efa_11_nf;
extern cpuop_func op_4efa_11_ff;
extern cpuop_func op_4efb_11_nf;
extern cpuop_func op_4efb_11_ff;
extern cpuop_func op_5000_11_nf;
extern cpuop_func op_5000_11_ff;
extern cpuop_func op_5010_11_nf;
extern cpuop_func op_5010_11_ff;
extern cpuop_func op_5018_11_nf;
extern cpuop_func op_5018_11_ff;
extern cpuop_func op_5020_11_nf;
extern cpuop_func op_5020_11_ff;
extern cpuop_func op_5028_11_nf;
extern cpuop_func op_5028_11_ff;
extern cpuop_func op_5030_11_nf;
extern cpuop_func op_5030_11_ff;
extern cpuop_func op_5038_11_nf;
extern cpuop_func op_5038_11_ff;
extern cpuop_func op_5039_11_nf;
extern cpuop_func op_5039_11_ff;
extern cpuop_func op_5040_11_nf;
extern cpuop_func op_5040_11_ff;
extern cpuop_func op_5048_11_nf;
extern cpuop_func op_5048_11_ff;
extern cpuop_func op_5050_11_nf;
extern cpuop_func op_5050_11_ff;
extern cpuop_func op_5058_11_nf;
extern cpuop_func op_5058_11_ff;
extern cpuop_func op_5060_11_nf;
extern cpuop_func op_5060_11_ff;
extern cpuop_func op_5068_11_nf;
extern cpuop_func op_5068_11_ff;
extern cpuop_func op_5070_11_nf;
extern cpuop_func op_5070_11_ff;
extern cpuop_func op_5078_11_nf;
extern cpuop_func op_5078_11_ff;
extern cpuop_func op_5079_11_nf;
extern cpuop_func op_5079_11_ff;
extern cpuop_func op_5080_11_nf;
extern cpuop_func op_5080_11_ff;
extern cpuop_func op_5088_11_nf;
extern cpuop_func op_5088_11_ff;
extern cpuop_func op_5090_11_nf;
extern cpuop_func op_5090_11_ff;
extern cpuop_func op_5098_11_nf;
extern cpuop_func op_5098_11_ff;
extern cpuop_func op_50a0_11_nf;
extern cpuop_func op_50a0_11_ff;
extern cpuop_func op_50a8_11_nf;
extern cpuop_func op_50a8_11_ff;
extern cpuop_func op_50b0_11_nf;
extern cpuop_func op_50b0_11_ff;
extern cpuop_func op_50b8_11_nf;
extern cpuop_func op_50b8_11_ff;
extern cpuop_func op_50b9_11_nf;
extern cpuop_func op_50b9_11_ff;
extern cpuop_func op_50c0_11_nf;
extern cpuop_func op_50c0_11_ff;
extern cpuop_func op_50c8_11_nf;
extern cpuop_func op_50c8_11_ff;
extern cpuop_func op_50d0_11_nf;
extern cpuop_func op_50d0_11_ff;
extern cpuop_func op_50d8_11_nf;
extern cpuop_func op_50d8_11_ff;
extern cpuop_func op_50e0_11_nf;
extern cpuop_func op_50e0_11_ff;
extern cpuop_func op_50e8_11_nf;
extern cpuop_func op_50e8_11_ff;
extern cpuop_func op_50f0_11_nf;
extern cpuop_func op_50f0_11_ff;
extern cpuop_func op_50f8_11_nf;
extern cpuop_func op_50f8_11_ff;
extern cpuop_func op_50f9_11_nf;
extern cpuop_func op_50f9_11_ff;
extern cpuop_func op_5100_11_nf;
extern cpuop_func op_5100_11_ff;
extern cpuop_func op_5110_11_nf;
extern cpuop_func op_5110_11_ff;
extern cpuop_func op_5118_11_nf;
extern cpuop_func op_5118_11_ff;
extern cpuop_func op_5120_11_nf;
extern cpuop_func op_5120_11_ff;
extern cpuop_func op_5128_11_nf;
extern cpuop_func op_5128_11_ff;
extern cpuop_func op_5130_11_nf;
extern cpuop_func op_5130_11_ff;
extern cpuop_func op_5138_11_nf;
extern cpuop_func op_5138_11_ff;
extern cpuop_func op_5139_11_nf;
extern cpuop_func op_5139_11_ff;
extern cpuop_func op_5140_11_nf;
extern cpuop_func op_5140_11_ff;
extern cpuop_func op_5148_11_nf;
extern cpuop_func op_5148_11_ff;
extern cpuop_func op_5150_11_nf;
extern cpuop_func op_5150_11_ff;
extern cpuop_func op_5158_11_nf;
extern cpuop_func op_5158_11_ff;
extern cpuop_func op_5160_11_nf;
extern cpuop_func op_5160_11_ff;
extern cpuop_func op_5168_11_nf;
extern cpuop_func op_5168_11_ff;
extern cpuop_func op_5170_11_nf;
extern cpuop_func op_5170_11_ff;
extern cpuop_func op_5178_11_nf;
extern cpuop_func op_5178_11_ff;
extern cpuop_func op_5179_11_nf;
extern cpuop_func op_5179_11_ff;
extern cpuop_func op_5180_11_nf;
extern cpuop_func op_5180_11_ff;
extern cpuop_func op_5188_11_nf;
extern cpuop_func op_5188_11_ff;
extern cpuop_func op_5190_11_nf;
extern cpuop_func op_5190_11_ff;
extern cpuop_func op_5198_11_nf;
extern cpuop_func op_5198_11_ff;
extern cpuop_func op_51a0_11_nf;
extern cpuop_func op_51a0_11_ff;
extern cpuop_func op_51a8_11_nf;
extern cpuop_func op_51a8_11_ff;
extern cpuop_func op_51b0_11_nf;
extern cpuop_func op_51b0_11_ff;
extern cpuop_func op_51b8_11_nf;
extern cpuop_func op_51b8_11_ff;
extern cpuop_func op_51b9_11_nf;
extern cpuop_func op_51b9_11_ff;
extern cpuop_func op_51c0_11_nf;
extern cpuop_func op_51c0_11_ff;
extern cpuop_func op_51c8_11_nf;
extern cpuop_func op_51c8_11_ff;
extern cpuop_func op_51d0_11_nf;
extern cpuop_func op_51d0_11_ff;
extern cpuop_func op_51d8_11_nf;
extern cpuop_func op_51d8_11_ff;
extern cpuop_func op_51e0_11_nf;
extern cpuop_func op_51e0_11_ff;
extern cpuop_func op_51e8_11_nf;
extern cpuop_func op_51e8_11_ff;
extern cpuop_func op_51f0_11_nf;
extern cpuop_func op_51f0_11_ff;
extern cpuop_func op_51f8_11_nf;
extern cpuop_func op_51f8_11_ff;
extern cpuop_func op_51f9_11_nf;
extern cpuop_func op_51f9_11_ff;
extern cpuop_func op_52c0_11_nf;
extern cpuop_func op_52c0_11_ff;
extern cpuop_func op_52c8_11_nf;
extern cpuop_func op_52c8_11_ff;
extern cpuop_func op_52d0_11_nf;
extern cpuop_func op_52d0_11_ff;
extern cpuop_func op_52d8_11_nf;
extern cpuop_func op_52d8_11_ff;
extern cpuop_func op_52e0_11_nf;
extern cpuop_func op_52e0_11_ff;
extern cpuop_func op_52e8_11_nf;
extern cpuop_func op_52e8_11_ff;
extern cpuop_func op_52f0_11_nf;
extern cpuop_func op_52f0_11_ff;
extern cpuop_func op_52f8_11_nf;
extern cpuop_func op_52f8_11_ff;
extern cpuop_func op_52f9_11_nf;
extern cpuop_func op_52f9_11_ff;
extern cpuop_func op_53c0_11_nf;
extern cpuop_func op_53c0_11_ff;
extern cpuop_func op_53c8_11_nf;
extern cpuop_func op_53c8_11_ff;
extern cpuop_func op_53d0_11_nf;
extern cpuop_func op_53d0_11_ff;
extern cpuop_func op_53d8_11_nf;
extern cpuop_func op_53d8_11_ff;
extern cpuop_func op_53e0_11_nf;
extern cpuop_func op_53e0_11_ff;
extern cpuop_func op_53e8_11_nf;
extern cpuop_func op_53e8_11_ff;
extern cpuop_func op_53f0_11_nf;
extern cpuop_func op_53f0_11_ff;
extern cpuop_func op_53f8_11_nf;
extern cpuop_func op_53f8_11_ff;
extern cpuop_func op_53f9_11_nf;
extern cpuop_func op_53f9_11_ff;
extern cpuop_func op_54c0_11_nf;
extern cpuop_func op_54c0_11_ff;
extern cpuop_func op_54c8_11_nf;
extern cpuop_func op_54c8_11_ff;
extern cpuop_func op_54d0_11_nf;
extern cpuop_func op_54d0_11_ff;
extern cpuop_func op_54d8_11_nf;
extern cpuop_func op_54d8_11_ff;
extern cpuop_func op_54e0_11_nf;
extern cpuop_func op_54e0_11_ff;
extern cpuop_func op_54e8_11_nf;
extern cpuop_func op_54e8_11_ff;
extern cpuop_func op_54f0_11_nf;
extern cpuop_func op_54f0_11_ff;
extern cpuop_func op_54f8_11_nf;
extern cpuop_func op_54f8_11_ff;
extern cpuop_func op_54f9_11_nf;
extern cpuop_func op_54f9_11_ff;
extern cpuop_func op_55c0_11_nf;
extern cpuop_func op_55c0_11_ff;
extern cpuop_func op_55c8_11_nf;
extern cpuop_func op_55c8_11_ff;
extern cpuop_func op_55d0_11_nf;
extern cpuop_func op_55d0_11_ff;
extern cpuop_func op_55d8_11_nf;
extern cpuop_func op_55d8_11_ff;
extern cpuop_func op_55e0_11_nf;
extern cpuop_func op_55e0_11_ff;
extern cpuop_func op_55e8_11_nf;
extern cpuop_func op_55e8_11_ff;
extern cpuop_func op_55f0_11_nf;
extern cpuop_func op_55f0_11_ff;
extern cpuop_func op_55f8_11_nf;
extern cpuop_func op_55f8_11_ff;
extern cpuop_func op_55f9_11_nf;
extern cpuop_func op_55f9_11_ff;
extern cpuop_func op_56c0_11_nf;
extern cpuop_func op_56c0_11_ff;
extern cpuop_func op_56c8_11_nf;
extern cpuop_func op_56c8_11_ff;
extern cpuop_func op_56d0_11_nf;
extern cpuop_func op_56d0_11_ff;
extern cpuop_func op_56d8_11_nf;
extern cpuop_func op_56d8_11_ff;
extern cpuop_func op_56e0_11_nf;
extern cpuop_func op_56e0_11_ff;
extern cpuop_func op_56e8_11_nf;
extern cpuop_func op_56e8_11_ff;
extern cpuop_func op_56f0_11_nf;
extern cpuop_func op_56f0_11_ff;
extern cpuop_func op_56f8_11_nf;
extern cpuop_func op_56f8_11_ff;
extern cpuop_func op_56f9_11_nf;
extern cpuop_func op_56f9_11_ff;
extern cpuop_func op_57c0_11_nf;
extern cpuop_func op_57c0_11_ff;
extern cpuop_func op_57c8_11_nf;
extern cpuop_func op_57c8_11_ff;
extern cpuop_func op_57d0_11_nf;
extern cpuop_func op_57d0_11_ff;
extern cpuop_func op_57d8_11_nf;
extern cpuop_func op_57d8_11_ff;
extern cpuop_func op_57e0_11_nf;
extern cpuop_func op_57e0_11_ff;
extern cpuop_func op_57e8_11_nf;
extern cpuop_func op_57e8_11_ff;
extern cpuop_func op_57f0_11_nf;
extern cpuop_func op_57f0_11_ff;
extern cpuop_func op_57f8_11_nf;
extern cpuop_func op_57f8_11_ff;
extern cpuop_func op_57f9_11_nf;
extern cpuop_func op_57f9_11_ff;
extern cpuop_func op_58c0_11_nf;
extern cpuop_func op_58c0_11_ff;
extern cpuop_func op_58c8_11_nf;
extern cpuop_func op_58c8_11_ff;
extern cpuop_func op_58d0_11_nf;
extern cpuop_func op_58d0_11_ff;
extern cpuop_func op_58d8_11_nf;
extern cpuop_func op_58d8_11_ff;
extern cpuop_func op_58e0_11_nf;
extern cpuop_func op_58e0_11_ff;
extern cpuop_func op_58e8_11_nf;
extern cpuop_func op_58e8_11_ff;
extern cpuop_func op_58f0_11_nf;
extern cpuop_func op_58f0_11_ff;
extern cpuop_func op_58f8_11_nf;
extern cpuop_func op_58f8_11_ff;
extern cpuop_func op_58f9_11_nf;
extern cpuop_func op_58f9_11_ff;
extern cpuop_func op_59c0_11_nf;
extern cpuop_func op_59c0_11_ff;
extern cpuop_func op_59c8_11_nf;
extern cpuop_func op_59c8_11_ff;
extern cpuop_func op_59d0_11_nf;
extern cpuop_func op_59d0_11_ff;
extern cpuop_func op_59d8_11_nf;
extern cpuop_func op_59d8_11_ff;
extern cpuop_func op_59e0_11_nf;
extern cpuop_func op_59e0_11_ff;
extern cpuop_func op_59e8_11_nf;
extern cpuop_func op_59e8_11_ff;
extern cpuop_func op_59f0_11_nf;
extern cpuop_func op_59f0_11_ff;
extern cpuop_func op_59f8_11_nf;
extern cpuop_func op_59f8_11_ff;
extern cpuop_func op_59f9_11_nf;
extern cpuop_func op_59f9_11_ff;
extern cpuop_func op_5ac0_11_nf;
extern cpuop_func op_5ac0_11_ff;
extern cpuop_func op_5ac8_11_nf;
extern cpuop_func op_5ac8_11_ff;
extern cpuop_func op_5ad0_11_nf;
extern cpuop_func op_5ad0_11_ff;
extern cpuop_func op_5ad8_11_nf;
extern cpuop_func op_5ad8_11_ff;
extern cpuop_func op_5ae0_11_nf;
extern cpuop_func op_5ae0_11_ff;
extern cpuop_func op_5ae8_11_nf;
extern cpuop_func op_5ae8_11_ff;
extern cpuop_func op_5af0_11_nf;
extern cpuop_func op_5af0_11_ff;
extern cpuop_func op_5af8_11_nf;
extern cpuop_func op_5af8_11_ff;
extern cpuop_func op_5af9_11_nf;
extern cpuop_func op_5af9_11_ff;
extern cpuop_func op_5bc0_11_nf;
extern cpuop_func op_5bc0_11_ff;
extern cpuop_func op_5bc8_11_nf;
extern cpuop_func op_5bc8_11_ff;
extern cpuop_func op_5bd0_11_nf;
extern cpuop_func op_5bd0_11_ff;
extern cpuop_func op_5bd8_11_nf;
extern cpuop_func op_5bd8_11_ff;
extern cpuop_func op_5be0_11_nf;
extern cpuop_func op_5be0_11_ff;
extern cpuop_func op_5be8_11_nf;
extern cpuop_func op_5be8_11_ff;
extern cpuop_func op_5bf0_11_nf;
extern cpuop_func op_5bf0_11_ff;
extern cpuop_func op_5bf8_11_nf;
extern cpuop_func op_5bf8_11_ff;
extern cpuop_func op_5bf9_11_nf;
extern cpuop_func op_5bf9_11_ff;
extern cpuop_func op_5cc0_11_nf;
extern cpuop_func op_5cc0_11_ff;
extern cpuop_func op_5cc8_11_nf;
extern cpuop_func op_5cc8_11_ff;
extern cpuop_func op_5cd0_11_nf;
extern cpuop_func op_5cd0_11_ff;
extern cpuop_func op_5cd8_11_nf;
extern cpuop_func op_5cd8_11_ff;
extern cpuop_func op_5ce0_11_nf;
extern cpuop_func op_5ce0_11_ff;
extern cpuop_func op_5ce8_11_nf;
extern cpuop_func op_5ce8_11_ff;
extern cpuop_func op_5cf0_11_nf;
extern cpuop_func op_5cf0_11_ff;
extern cpuop_func op_5cf8_11_nf;
extern cpuop_func op_5cf8_11_ff;
extern cpuop_func op_5cf9_11_nf;
extern cpuop_func op_5cf9_11_ff;
extern cpuop_func op_5dc0_11_nf;
extern cpuop_func op_5dc0_11_ff;
extern cpuop_func op_5dc8_11_nf;
extern cpuop_func op_5dc8_11_ff;
extern cpuop_func op_5dd0_11_nf;
extern cpuop_func op_5dd0_11_ff;
extern cpuop_func op_5dd8_11_nf;
extern cpuop_func op_5dd8_11_ff;
extern cpuop_func op_5de0_11_nf;
extern cpuop_func op_5de0_11_ff;
extern cpuop_func op_5de8_11_nf;
extern cpuop_func op_5de8_11_ff;
extern cpuop_func op_5df0_11_nf;
extern cpuop_func op_5df0_11_ff;
extern cpuop_func op_5df8_11_nf;
extern cpuop_func op_5df8_11_ff;
extern cpuop_func op_5df9_11_nf;
extern cpuop_func op_5df9_11_ff;
extern cpuop_func op_5ec0_11_nf;
extern cpuop_func op_5ec0_11_ff;
extern cpuop_func op_5ec8_11_nf;
extern cpuop_func op_5ec8_11_ff;
extern cpuop_func op_5ed0_11_nf;
extern cpuop_func op_5ed0_11_ff;
extern cpuop_func op_5ed8_11_nf;
extern cpuop_func op_5ed8_11_ff;
extern cpuop_func op_5ee0_11_nf;
extern cpuop_func op_5ee0_11_ff;
extern cpuop_func op_5ee8_11_nf;
extern cpuop_func op_5ee8_11_ff;
extern cpuop_func op_5ef0_11_nf;
extern cpuop_func op_5ef0_11_ff;
extern cpuop_func op_5ef8_11_nf;
extern cpuop_func op_5ef8_11_ff;
extern cpuop_func op_5ef9_11_nf;
extern cpuop_func op_5ef9_11_ff;
extern cpuop_func op_5fc0_11_nf;
extern cpuop_func op_5fc0_11_ff;
extern cpuop_func op_5fc8_11_nf;
extern cpuop_func op_5fc8_11_ff;
extern cpuop_func op_5fd0_11_nf;
extern cpuop_func op_5fd0_11_ff;
extern cpuop_func op_5fd8_11_nf;
extern cpuop_func op_5fd8_11_ff;
extern cpuop_func op_5fe0_11_nf;
extern cpuop_func op_5fe0_11_ff;
extern cpuop_func op_5fe8_11_nf;
extern cpuop_func op_5fe8_11_ff;
extern cpuop_func op_5ff0_11_nf;
extern cpuop_func op_5ff0_11_ff;
extern cpuop_func op_5ff8_11_nf;
extern cpuop_func op_5ff8_11_ff;
extern cpuop_func op_5ff9_11_nf;
extern cpuop_func op_5ff9_11_ff;
extern cpuop_func op_6000_11_nf;
extern cpuop_func op_6000_11_ff;
extern cpuop_func op_6001_11_nf;
extern cpuop_func op_6001_11_ff;
extern cpuop_func op_60ff_11_nf;
extern cpuop_func op_60ff_11_ff;
extern cpuop_func op_6100_11_nf;
extern cpuop_func op_6100_11_ff;
extern cpuop_func op_6101_11_nf;
extern cpuop_func op_6101_11_ff;
extern cpuop_func op_61ff_11_nf;
extern cpuop_func op_61ff_11_ff;
extern cpuop_func op_6200_11_nf;
extern cpuop_func op_6200_11_ff;
extern cpuop_func op_6201_11_nf;
extern cpuop_func op_6201_11_ff;
extern cpuop_func op_62ff_11_nf;
extern cpuop_func op_62ff_11_ff;
extern cpuop_func op_6300_11_nf;
extern cpuop_func op_6300_11_ff;
extern cpuop_func op_6301_11_nf;
extern cpuop_func op_6301_11_ff;
extern cpuop_func op_63ff_11_nf;
extern cpuop_func op_63ff_11_ff;
extern cpuop_func op_6400_11_nf;
extern cpuop_func op_6400_11_ff;
extern cpuop_func op_6401_11_nf;
extern cpuop_func op_6401_11_ff;
extern cpuop_func op_64ff_11_nf;
extern cpuop_func op_64ff_11_ff;
extern cpuop_func op_6500_11_nf;
extern cpuop_func op_6500_11_ff;
extern cpuop_func op_6501_11_nf;
extern cpuop_func op_6501_11_ff;
extern cpuop_func op_65ff_11_nf;
extern cpuop_func op_65ff_11_ff;
extern cpuop_func op_6600_11_nf;
extern cpuop_func op_6600_11_ff;
extern cpuop_func op_6601_11_nf;
extern cpuop_func op_6601_11_ff;
extern cpuop_func op_66ff_11_nf;
extern cpuop_func op_66ff_11_ff;
extern cpuop_func op_6700_11_nf;
extern cpuop_func op_6700_11_ff;
extern cpuop_func op_6701_11_nf;
extern cpuop_func op_6701_11_ff;
extern cpuop_func op_67ff_11_nf;
extern cpuop_func op_67ff_11_ff;
extern cpuop_func op_6800_11_nf;
extern cpuop_func op_6800_11_ff;
extern cpuop_func op_6801_11_nf;
extern cpuop_func op_6801_11_ff;
extern cpuop_func op_68ff_11_nf;
extern cpuop_func op_68ff_11_ff;
extern cpuop_func op_6900_11_nf;
extern cpuop_func op_6900_11_ff;
extern cpuop_func op_6901_11_nf;
extern cpuop_func op_6901_11_ff;
extern cpuop_func op_69ff_11_nf;
extern cpuop_func op_69ff_11_ff;
extern cpuop_func op_6a00_11_nf;
extern cpuop_func op_6a00_11_ff;
extern cpuop_func op_6a01_11_nf;
extern cpuop_func op_6a01_11_ff;
extern cpuop_func op_6aff_11_nf;
extern cpuop_func op_6aff_11_ff;
extern cpuop_func op_6b00_11_nf;
extern cpuop_func op_6b00_11_ff;
extern cpuop_func op_6b01_11_nf;
extern cpuop_func op_6b01_11_ff;
extern cpuop_func op_6bff_11_nf;
extern cpuop_func op_6bff_11_ff;
extern cpuop_func op_6c00_11_nf;
extern cpuop_func op_6c00_11_ff;
extern cpuop_func op_6c01_11_nf;
extern cpuop_func op_6c01_11_ff;
extern cpuop_func op_6cff_11_nf;
extern cpuop_func op_6cff_11_ff;
extern cpuop_func op_6d00_11_nf;
extern cpuop_func op_6d00_11_ff;
extern cpuop_func op_6d01_11_nf;
extern cpuop_func op_6d01_11_ff;
extern cpuop_func op_6dff_11_nf;
extern cpuop_func op_6dff_11_ff;
extern cpuop_func op_6e00_11_nf;
extern cpuop_func op_6e00_11_ff;
extern cpuop_func op_6e01_11_nf;
extern cpuop_func op_6e01_11_ff;
extern cpuop_func op_6eff_11_nf;
extern cpuop_func op_6eff_11_ff;
extern cpuop_func op_6f00_11_nf;
extern cpuop_func op_6f00_11_ff;
extern cpuop_func op_6f01_11_nf;
extern cpuop_func op_6f01_11_ff;
extern cpuop_func op_6fff_11_nf;
extern cpuop_func op_6fff_11_ff;
extern cpuop_func op_7000_11_nf;
extern cpuop_func op_7000_11_ff;
extern cpuop_func op_8000_11_nf;
extern cpuop_func op_8000_11_ff;
extern cpuop_func op_8010_11_nf;
extern cpuop_func op_8010_11_ff;
extern cpuop_func op_8018_11_nf;
extern cpuop_func op_8018_11_ff;
extern cpuop_func op_8020_11_nf;
extern cpuop_func op_8020_11_ff;
extern cpuop_func op_8028_11_nf;
extern cpuop_func op_8028_11_ff;
extern cpuop_func op_8030_11_nf;
extern cpuop_func op_8030_11_ff;
extern cpuop_func op_8038_11_nf;
extern cpuop_func op_8038_11_ff;
extern cpuop_func op_8039_11_nf;
extern cpuop_func op_8039_11_ff;
extern cpuop_func op_803a_11_nf;
extern cpuop_func op_803a_11_ff;
extern cpuop_func op_803b_11_nf;
extern cpuop_func op_803b_11_ff;
extern cpuop_func op_803c_11_nf;
extern cpuop_func op_803c_11_ff;
extern cpuop_func op_8040_11_nf;
extern cpuop_func op_8040_11_ff;
extern cpuop_func op_8050_11_nf;
extern cpuop_func op_8050_11_ff;
extern cpuop_func op_8058_11_nf;
extern cpuop_func op_8058_11_ff;
extern cpuop_func op_8060_11_nf;
extern cpuop_func op_8060_11_ff;
extern cpuop_func op_8068_11_nf;
extern cpuop_func op_8068_11_ff;
extern cpuop_func op_8070_11_nf;
extern cpuop_func op_8070_11_ff;
extern cpuop_func op_8078_11_nf;
extern cpuop_func op_8078_11_ff;
extern cpuop_func op_8079_11_nf;
extern cpuop_func op_8079_11_ff;
extern cpuop_func op_807a_11_nf;
extern cpuop_func op_807a_11_ff;
extern cpuop_func op_807b_11_nf;
extern cpuop_func op_807b_11_ff;
extern cpuop_func op_807c_11_nf;
extern cpuop_func op_807c_11_ff;
extern cpuop_func op_8080_11_nf;
extern cpuop_func op_8080_11_ff;
extern cpuop_func op_8090_11_nf;
extern cpuop_func op_8090_11_ff;
extern cpuop_func op_8098_11_nf;
extern cpuop_func op_8098_11_ff;
extern cpuop_func op_80a0_11_nf;
extern cpuop_func op_80a0_11_ff;
extern cpuop_func op_80a8_11_nf;
extern cpuop_func op_80a8_11_ff;
extern cpuop_func op_80b0_11_nf;
extern cpuop_func op_80b0_11_ff;
extern cpuop_func op_80b8_11_nf;
extern cpuop_func op_80b8_11_ff;
extern cpuop_func op_80b9_11_nf;
extern cpuop_func op_80b9_11_ff;
extern cpuop_func op_80ba_11_nf;
extern cpuop_func op_80ba_11_ff;
extern cpuop_func op_80bb_11_nf;
extern cpuop_func op_80bb_11_ff;
extern cpuop_func op_80bc_11_nf;
extern cpuop_func op_80bc_11_ff;
extern cpuop_func op_80c0_11_nf;
extern cpuop_func op_80c0_11_ff;
extern cpuop_func op_80d0_11_nf;
extern cpuop_func op_80d0_11_ff;
extern cpuop_func op_80d8_11_nf;
extern cpuop_func op_80d8_11_ff;
extern cpuop_func op_80e0_11_nf;
extern cpuop_func op_80e0_11_ff;
extern cpuop_func op_80e8_11_nf;
extern cpuop_func op_80e8_11_ff;
extern cpuop_func op_80f0_11_nf;
extern cpuop_func op_80f0_11_ff;
extern cpuop_func op_80f8_11_nf;
extern cpuop_func op_80f8_11_ff;
extern cpuop_func op_80f9_11_nf;
extern cpuop_func op_80f9_11_ff;
extern cpuop_func op_80fa_11_nf;
extern cpuop_func op_80fa_11_ff;
extern cpuop_func op_80fb_11_nf;
extern cpuop_func op_80fb_11_ff;
extern cpuop_func op_80fc_11_nf;
extern cpuop_func op_80fc_11_ff;
extern cpuop_func op_8100_11_nf;
extern cpuop_func op_8100_11_ff;
extern cpuop_func op_8108_11_nf;
extern cpuop_func op_8108_11_ff;
extern cpuop_func op_8110_11_nf;
extern cpuop_func op_8110_11_ff;
extern cpuop_func op_8118_11_nf;
extern cpuop_func op_8118_11_ff;
extern cpuop_func op_8120_11_nf;
extern cpuop_func op_8120_11_ff;
extern cpuop_func op_8128_11_nf;
extern cpuop_func op_8128_11_ff;
extern cpuop_func op_8130_11_nf;
extern cpuop_func op_8130_11_ff;
extern cpuop_func op_8138_11_nf;
extern cpuop_func op_8138_11_ff;
extern cpuop_func op_8139_11_nf;
extern cpuop_func op_8139_11_ff;
extern cpuop_func op_8150_11_nf;
extern cpuop_func op_8150_11_ff;
extern cpuop_func op_8158_11_nf;
extern cpuop_func op_8158_11_ff;
extern cpuop_func op_8160_11_nf;
extern cpuop_func op_8160_11_ff;
extern cpuop_func op_8168_11_nf;
extern cpuop_func op_8168_11_ff;
extern cpuop_func op_8170_11_nf;
extern cpuop_func op_8170_11_ff;
extern cpuop_func op_8178_11_nf;
extern cpuop_func op_8178_11_ff;
extern cpuop_func op_8179_11_nf;
extern cpuop_func op_8179_11_ff;
extern cpuop_func op_8190_11_nf;
extern cpuop_func op_8190_11_ff;
extern cpuop_func op_8198_11_nf;
extern cpuop_func op_8198_11_ff;
extern cpuop_func op_81a0_11_nf;
extern cpuop_func op_81a0_11_ff;
extern cpuop_func op_81a8_11_nf;
extern cpuop_func op_81a8_11_ff;
extern cpuop_func op_81b0_11_nf;
extern cpuop_func op_81b0_11_ff;
extern cpuop_func op_81b8_11_nf;
extern cpuop_func op_81b8_11_ff;
extern cpuop_func op_81b9_11_nf;
extern cpuop_func op_81b9_11_ff;
extern cpuop_func op_81c0_11_nf;
extern cpuop_func op_81c0_11_ff;
extern cpuop_func op_81d0_11_nf;
extern cpuop_func op_81d0_11_ff;
extern cpuop_func op_81d8_11_nf;
extern cpuop_func op_81d8_11_ff;
extern cpuop_func op_81e0_11_nf;
extern cpuop_func op_81e0_11_ff;
extern cpuop_func op_81e8_11_nf;
extern cpuop_func op_81e8_11_ff;
extern cpuop_func op_81f0_11_nf;
extern cpuop_func op_81f0_11_ff;
extern cpuop_func op_81f8_11_nf;
extern cpuop_func op_81f8_11_ff;
extern cpuop_func op_81f9_11_nf;
extern cpuop_func op_81f9_11_ff;
extern cpuop_func op_81fa_11_nf;
extern cpuop_func op_81fa_11_ff;
extern cpuop_func op_81fb_11_nf;
extern cpuop_func op_81fb_11_ff;
extern cpuop_func op_81fc_11_nf;
extern cpuop_func op_81fc_11_ff;
extern cpuop_func op_9000_11_nf;
extern cpuop_func op_9000_11_ff;
extern cpuop_func op_9010_11_nf;
extern cpuop_func op_9010_11_ff;
extern cpuop_func op_9018_11_nf;
extern cpuop_func op_9018_11_ff;
extern cpuop_func op_9020_11_nf;
extern cpuop_func op_9020_11_ff;
extern cpuop_func op_9028_11_nf;
extern cpuop_func op_9028_11_ff;
extern cpuop_func op_9030_11_nf;
extern cpuop_func op_9030_11_ff;
extern cpuop_func op_9038_11_nf;
extern cpuop_func op_9038_11_ff;
extern cpuop_func op_9039_11_nf;
extern cpuop_func op_9039_11_ff;
extern cpuop_func op_903a_11_nf;
extern cpuop_func op_903a_11_ff;
extern cpuop_func op_903b_11_nf;
extern cpuop_func op_903b_11_ff;
extern cpuop_func op_903c_11_nf;
extern cpuop_func op_903c_11_ff;
extern cpuop_func op_9040_11_nf;
extern cpuop_func op_9040_11_ff;
extern cpuop_func op_9048_11_nf;
extern cpuop_func op_9048_11_ff;
extern cpuop_func op_9050_11_nf;
extern cpuop_func op_9050_11_ff;
extern cpuop_func op_9058_11_nf;
extern cpuop_func op_9058_11_ff;
extern cpuop_func op_9060_11_nf;
extern cpuop_func op_9060_11_ff;
extern cpuop_func op_9068_11_nf;
extern cpuop_func op_9068_11_ff;
extern cpuop_func op_9070_11_nf;
extern cpuop_func op_9070_11_ff;
extern cpuop_func op_9078_11_nf;
extern cpuop_func op_9078_11_ff;
extern cpuop_func op_9079_11_nf;
extern cpuop_func op_9079_11_ff;
extern cpuop_func op_907a_11_nf;
extern cpuop_func op_907a_11_ff;
extern cpuop_func op_907b_11_nf;
extern cpuop_func op_907b_11_ff;
extern cpuop_func op_907c_11_nf;
extern cpuop_func op_907c_11_ff;
extern cpuop_func op_9080_11_nf;
extern cpuop_func op_9080_11_ff;
extern cpuop_func op_9088_11_nf;
extern cpuop_func op_9088_11_ff;
extern cpuop_func op_9090_11_nf;
extern cpuop_func op_9090_11_ff;
extern cpuop_func op_9098_11_nf;
extern cpuop_func op_9098_11_ff;
extern cpuop_func op_90a0_11_nf;
extern cpuop_func op_90a0_11_ff;
extern cpuop_func op_90a8_11_nf;
extern cpuop_func op_90a8_11_ff;
extern cpuop_func op_90b0_11_nf;
extern cpuop_func op_90b0_11_ff;
extern cpuop_func op_90b8_11_nf;
extern cpuop_func op_90b8_11_ff;
extern cpuop_func op_90b9_11_nf;
extern cpuop_func op_90b9_11_ff;
extern cpuop_func op_90ba_11_nf;
extern cpuop_func op_90ba_11_ff;
extern cpuop_func op_90bb_11_nf;
extern cpuop_func op_90bb_11_ff;
extern cpuop_func op_90bc_11_nf;
extern cpuop_func op_90bc_11_ff;
extern cpuop_func op_90c0_11_nf;
extern cpuop_func op_90c0_11_ff;
extern cpuop_func op_90c8_11_nf;
extern cpuop_func op_90c8_11_ff;
extern cpuop_func op_90d0_11_nf;
extern cpuop_func op_90d0_11_ff;
extern cpuop_func op_90d8_11_nf;
extern cpuop_func op_90d8_11_ff;
extern cpuop_func op_90e0_11_nf;
extern cpuop_func op_90e0_11_ff;
extern cpuop_func op_90e8_11_nf;
extern cpuop_func op_90e8_11_ff;
extern cpuop_func op_90f0_11_nf;
extern cpuop_func op_90f0_11_ff;
extern cpuop_func op_90f8_11_nf;
extern cpuop_func op_90f8_11_ff;
extern cpuop_func op_90f9_11_nf;
extern cpuop_func op_90f9_11_ff;
extern cpuop_func op_90fa_11_nf;
extern cpuop_func op_90fa_11_ff;
extern cpuop_func op_90fb_11_nf;
extern cpuop_func op_90fb_11_ff;
extern cpuop_func op_90fc_11_nf;
extern cpuop_func op_90fc_11_ff;
extern cpuop_func op_9100_11_nf;
extern cpuop_func op_9100_11_ff;
extern cpuop_func op_9108_11_nf;
extern cpuop_func op_9108_11_ff;
extern cpuop_func op_9110_11_nf;
extern cpuop_func op_9110_11_ff;
extern cpuop_func op_9118_11_nf;
extern cpuop_func op_9118_11_ff;
extern cpuop_func op_9120_11_nf;
extern cpuop_func op_9120_11_ff;
extern cpuop_func op_9128_11_nf;
extern cpuop_func op_9128_11_ff;
extern cpuop_func op_9130_11_nf;
extern cpuop_func op_9130_11_ff;
extern cpuop_func op_9138_11_nf;
extern cpuop_func op_9138_11_ff;
extern cpuop_func op_9139_11_nf;
extern cpuop_func op_9139_11_ff;
extern cpuop_func op_9140_11_nf;
extern cpuop_func op_9140_11_ff;
extern cpuop_func op_9148_11_nf;
extern cpuop_func op_9148_11_ff;
extern cpuop_func op_9150_11_nf;
extern cpuop_func op_9150_11_ff;
extern cpuop_func op_9158_11_nf;
extern cpuop_func op_9158_11_ff;
extern cpuop_func op_9160_11_nf;
extern cpuop_func op_9160_11_ff;
extern cpuop_func op_9168_11_nf;
extern cpuop_func op_9168_11_ff;
extern cpuop_func op_9170_11_nf;
extern cpuop_func op_9170_11_ff;
extern cpuop_func op_9178_11_nf;
extern cpuop_func op_9178_11_ff;
extern cpuop_func op_9179_11_nf;
extern cpuop_func op_9179_11_ff;
extern cpuop_func op_9180_11_nf;
extern cpuop_func op_9180_11_ff;
extern cpuop_func op_9188_11_nf;
extern cpuop_func op_9188_11_ff;
extern cpuop_func op_9190_11_nf;
extern cpuop_func op_9190_11_ff;
extern cpuop_func op_9198_11_nf;
extern cpuop_func op_9198_11_ff;
extern cpuop_func op_91a0_11_nf;
extern cpuop_func op_91a0_11_ff;
extern cpuop_func op_91a8_11_nf;
extern cpuop_func op_91a8_11_ff;
extern cpuop_func op_91b0_11_nf;
extern cpuop_func op_91b0_11_ff;
extern cpuop_func op_91b8_11_nf;
extern cpuop_func op_91b8_11_ff;
extern cpuop_func op_91b9_11_nf;
extern cpuop_func op_91b9_11_ff;
extern cpuop_func op_91c0_11_nf;
extern cpuop_func op_91c0_11_ff;
extern cpuop_func op_91c8_11_nf;
extern cpuop_func op_91c8_11_ff;
extern cpuop_func op_91d0_11_nf;
extern cpuop_func op_91d0_11_ff;
extern cpuop_func op_91d8_11_nf;
extern cpuop_func op_91d8_11_ff;
extern cpuop_func op_91e0_11_nf;
extern cpuop_func op_91e0_11_ff;
extern cpuop_func op_91e8_11_nf;
extern cpuop_func op_91e8_11_ff;
extern cpuop_func op_91f0_11_nf;
extern cpuop_func op_91f0_11_ff;
extern cpuop_func op_91f8_11_nf;
extern cpuop_func op_91f8_11_ff;
extern cpuop_func op_91f9_11_nf;
extern cpuop_func op_91f9_11_ff;
extern cpuop_func op_91fa_11_nf;
extern cpuop_func op_91fa_11_ff;
extern cpuop_func op_91fb_11_nf;
extern cpuop_func op_91fb_11_ff;
extern cpuop_func op_91fc_11_nf;
extern cpuop_func op_91fc_11_ff;
extern cpuop_func op_b000_11_nf;
extern cpuop_func op_b000_11_ff;
extern cpuop_func op_b010_11_nf;
extern cpuop_func op_b010_11_ff;
extern cpuop_func op_b018_11_nf;
extern cpuop_func op_b018_11_ff;
extern cpuop_func op_b020_11_nf;
extern cpuop_func op_b020_11_ff;
extern cpuop_func op_b028_11_nf;
extern cpuop_func op_b028_11_ff;
extern cpuop_func op_b030_11_nf;
extern cpuop_func op_b030_11_ff;
extern cpuop_func op_b038_11_nf;
extern cpuop_func op_b038_11_ff;
extern cpuop_func op_b039_11_nf;
extern cpuop_func op_b039_11_ff;
extern cpuop_func op_b03a_11_nf;
extern cpuop_func op_b03a_11_ff;
extern cpuop_func op_b03b_11_nf;
extern cpuop_func op_b03b_11_ff;
extern cpuop_func op_b03c_11_nf;
extern cpuop_func op_b03c_11_ff;
extern cpuop_func op_b040_11_nf;
extern cpuop_func op_b040_11_ff;
extern cpuop_func op_b048_11_nf;
extern cpuop_func op_b048_11_ff;
extern cpuop_func op_b050_11_nf;
extern cpuop_func op_b050_11_ff;
extern cpuop_func op_b058_11_nf;
extern cpuop_func op_b058_11_ff;
extern cpuop_func op_b060_11_nf;
extern cpuop_func op_b060_11_ff;
extern cpuop_func op_b068_11_nf;
extern cpuop_func op_b068_11_ff;
extern cpuop_func op_b070_11_nf;
extern cpuop_func op_b070_11_ff;
extern cpuop_func op_b078_11_nf;
extern cpuop_func op_b078_11_ff;
extern cpuop_func op_b079_11_nf;
extern cpuop_func op_b079_11_ff;
extern cpuop_func op_b07a_11_nf;
extern cpuop_func op_b07a_11_ff;
extern cpuop_func op_b07b_11_nf;
extern cpuop_func op_b07b_11_ff;
extern cpuop_func op_b07c_11_nf;
extern cpuop_func op_b07c_11_ff;
extern cpuop_func op_b080_11_nf;
extern cpuop_func op_b080_11_ff;
extern cpuop_func op_b088_11_nf;
extern cpuop_func op_b088_11_ff;
extern cpuop_func op_b090_11_nf;
extern cpuop_func op_b090_11_ff;
extern cpuop_func op_b098_11_nf;
extern cpuop_func op_b098_11_ff;
extern cpuop_func op_b0a0_11_nf;
extern cpuop_func op_b0a0_11_ff;
extern cpuop_func op_b0a8_11_nf;
extern cpuop_func op_b0a8_11_ff;
extern cpuop_func op_b0b0_11_nf;
extern cpuop_func op_b0b0_11_ff;
extern cpuop_func op_b0b8_11_nf;
extern cpuop_func op_b0b8_11_ff;
extern cpuop_func op_b0b9_11_nf;
extern cpuop_func op_b0b9_11_ff;
extern cpuop_func op_b0ba_11_nf;
extern cpuop_func op_b0ba_11_ff;
extern cpuop_func op_b0bb_11_nf;
extern cpuop_func op_b0bb_11_ff;
extern cpuop_func op_b0bc_11_nf;
extern cpuop_func op_b0bc_11_ff;
extern cpuop_func op_b0c0_11_nf;
extern cpuop_func op_b0c0_11_ff;
extern cpuop_func op_b0c8_11_nf;
extern cpuop_func op_b0c8_11_ff;
extern cpuop_func op_b0d0_11_nf;
extern cpuop_func op_b0d0_11_ff;
extern cpuop_func op_b0d8_11_nf;
extern cpuop_func op_b0d8_11_ff;
extern cpuop_func op_b0e0_11_nf;
extern cpuop_func op_b0e0_11_ff;
extern cpuop_func op_b0e8_11_nf;
extern cpuop_func op_b0e8_11_ff;
extern cpuop_func op_b0f0_11_nf;
extern cpuop_func op_b0f0_11_ff;
extern cpuop_func op_b0f8_11_nf;
extern cpuop_func op_b0f8_11_ff;
extern cpuop_func op_b0f9_11_nf;
extern cpuop_func op_b0f9_11_ff;
extern cpuop_func op_b0fa_11_nf;
extern cpuop_func op_b0fa_11_ff;
extern cpuop_func op_b0fb_11_nf;
extern cpuop_func op_b0fb_11_ff;
extern cpuop_func op_b0fc_11_nf;
extern cpuop_func op_b0fc_11_ff;
extern cpuop_func op_b100_11_nf;
extern cpuop_func op_b100_11_ff;
extern cpuop_func op_b108_11_nf;
extern cpuop_func op_b108_11_ff;
extern cpuop_func op_b110_11_nf;
extern cpuop_func op_b110_11_ff;
extern cpuop_func op_b118_11_nf;
extern cpuop_func op_b118_11_ff;
extern cpuop_func op_b120_11_nf;
extern cpuop_func op_b120_11_ff;
extern cpuop_func op_b128_11_nf;
extern cpuop_func op_b128_11_ff;
extern cpuop_func op_b130_11_nf;
extern cpuop_func op_b130_11_ff;
extern cpuop_func op_b138_11_nf;
extern cpuop_func op_b138_11_ff;
extern cpuop_func op_b139_11_nf;
extern cpuop_func op_b139_11_ff;
extern cpuop_func op_b140_11_nf;
extern cpuop_func op_b140_11_ff;
extern cpuop_func op_b148_11_nf;
extern cpuop_func op_b148_11_ff;
extern cpuop_func op_b150_11_nf;
extern cpuop_func op_b150_11_ff;
extern cpuop_func op_b158_11_nf;
extern cpuop_func op_b158_11_ff;
extern cpuop_func op_b160_11_nf;
extern cpuop_func op_b160_11_ff;
extern cpuop_func op_b168_11_nf;
extern cpuop_func op_b168_11_ff;
extern cpuop_func op_b170_11_nf;
extern cpuop_func op_b170_11_ff;
extern cpuop_func op_b178_11_nf;
extern cpuop_func op_b178_11_ff;
extern cpuop_func op_b179_11_nf;
extern cpuop_func op_b179_11_ff;
extern cpuop_func op_b180_11_nf;
extern cpuop_func op_b180_11_ff;
extern cpuop_func op_b188_11_nf;
extern cpuop_func op_b188_11_ff;
extern cpuop_func op_b190_11_nf;
extern cpuop_func op_b190_11_ff;
extern cpuop_func op_b198_11_nf;
extern cpuop_func op_b198_11_ff;
extern cpuop_func op_b1a0_11_nf;
extern cpuop_func op_b1a0_11_ff;
extern cpuop_func op_b1a8_11_nf;
extern cpuop_func op_b1a8_11_ff;
extern cpuop_func op_b1b0_11_nf;
extern cpuop_func op_b1b0_11_ff;
extern cpuop_func op_b1b8_11_nf;
extern cpuop_func op_b1b8_11_ff;
extern cpuop_func op_b1b9_11_nf;
extern cpuop_func op_b1b9_11_ff;
extern cpuop_func op_b1c0_11_nf;
extern cpuop_func op_b1c0_11_ff;
extern cpuop_func op_b1c8_11_nf;
extern cpuop_func op_b1c8_11_ff;
extern cpuop_func op_b1d0_11_nf;
extern cpuop_func op_b1d0_11_ff;
extern cpuop_func op_b1d8_11_nf;
extern cpuop_func op_b1d8_11_ff;
extern cpuop_func op_b1e0_11_nf;
extern cpuop_func op_b1e0_11_ff;
extern cpuop_func op_b1e8_11_nf;
extern cpuop_func op_b1e8_11_ff;
extern cpuop_func op_b1f0_11_nf;
extern cpuop_func op_b1f0_11_ff;
extern cpuop_func op_b1f8_11_nf;
extern cpuop_func op_b1f8_11_ff;
extern cpuop_func op_b1f9_11_nf;
extern cpuop_func op_b1f9_11_ff;
extern cpuop_func op_b1fa_11_nf;
extern cpuop_func op_b1fa_11_ff;
extern cpuop_func op_b1fb_11_nf;
extern cpuop_func op_b1fb_11_ff;
extern cpuop_func op_b1fc_11_nf;
extern cpuop_func op_b1fc_11_ff;
extern cpuop_func op_c000_11_nf;
extern cpuop_func op_c000_11_ff;
extern cpuop_func op_c010_11_nf;
extern cpuop_func op_c010_11_ff;
extern cpuop_func op_c018_11_nf;
extern cpuop_func op_c018_11_ff;
extern cpuop_func op_c020_11_nf;
extern cpuop_func op_c020_11_ff;
extern cpuop_func op_c028_11_nf;
extern cpuop_func op_c028_11_ff;
extern cpuop_func op_c030_11_nf;
extern cpuop_func op_c030_11_ff;
extern cpuop_func op_c038_11_nf;
extern cpuop_func op_c038_11_ff;
extern cpuop_func op_c039_11_nf;
extern cpuop_func op_c039_11_ff;
extern cpuop_func op_c03a_11_nf;
extern cpuop_func op_c03a_11_ff;
extern cpuop_func op_c03b_11_nf;
extern cpuop_func op_c03b_11_ff;
extern cpuop_func op_c03c_11_nf;
extern cpuop_func op_c03c_11_ff;
extern cpuop_func op_c040_11_nf;
extern cpuop_func op_c040_11_ff;
extern cpuop_func op_c050_11_nf;
extern cpuop_func op_c050_11_ff;
extern cpuop_func op_c058_11_nf;
extern cpuop_func op_c058_11_ff;
extern cpuop_func op_c060_11_nf;
extern cpuop_func op_c060_11_ff;
extern cpuop_func op_c068_11_nf;
extern cpuop_func op_c068_11_ff;
extern cpuop_func op_c070_11_nf;
extern cpuop_func op_c070_11_ff;
extern cpuop_func op_c078_11_nf;
extern cpuop_func op_c078_11_ff;
extern cpuop_func op_c079_11_nf;
extern cpuop_func op_c079_11_ff;
extern cpuop_func op_c07a_11_nf;
extern cpuop_func op_c07a_11_ff;
extern cpuop_func op_c07b_11_nf;
extern cpuop_func op_c07b_11_ff;
extern cpuop_func op_c07c_11_nf;
extern cpuop_func op_c07c_11_ff;
extern cpuop_func op_c080_11_nf;
extern cpuop_func op_c080_11_ff;
extern cpuop_func op_c090_11_nf;
extern cpuop_func op_c090_11_ff;
extern cpuop_func op_c098_11_nf;
extern cpuop_func op_c098_11_ff;
extern cpuop_func op_c0a0_11_nf;
extern cpuop_func op_c0a0_11_ff;
extern cpuop_func op_c0a8_11_nf;
extern cpuop_func op_c0a8_11_ff;
extern cpuop_func op_c0b0_11_nf;
extern cpuop_func op_c0b0_11_ff;
extern cpuop_func op_c0b8_11_nf;
extern cpuop_func op_c0b8_11_ff;
extern cpuop_func op_c0b9_11_nf;
extern cpuop_func op_c0b9_11_ff;
extern cpuop_func op_c0ba_11_nf;
extern cpuop_func op_c0ba_11_ff;
extern cpuop_func op_c0bb_11_nf;
extern cpuop_func op_c0bb_11_ff;
extern cpuop_func op_c0bc_11_nf;
extern cpuop_func op_c0bc_11_ff;
extern cpuop_func op_c0c0_11_nf;
extern cpuop_func op_c0c0_11_ff;
extern cpuop_func op_c0d0_11_nf;
extern cpuop_func op_c0d0_11_ff;
extern cpuop_func op_c0d8_11_nf;
extern cpuop_func op_c0d8_11_ff;
extern cpuop_func op_c0e0_11_nf;
extern cpuop_func op_c0e0_11_ff;
extern cpuop_func op_c0e8_11_nf;
extern cpuop_func op_c0e8_11_ff;
extern cpuop_func op_c0f0_11_nf;
extern cpuop_func op_c0f0_11_ff;
extern cpuop_func op_c0f8_11_nf;
extern cpuop_func op_c0f8_11_ff;
extern cpuop_func op_c0f9_11_nf;
extern cpuop_func op_c0f9_11_ff;
extern cpuop_func op_c0fa_11_nf;
extern cpuop_func op_c0fa_11_ff;
extern cpuop_func op_c0fb_11_nf;
extern cpuop_func op_c0fb_11_ff;
extern cpuop_func op_c0fc_11_nf;
extern cpuop_func op_c0fc_11_ff;
extern cpuop_func op_c100_11_nf;
extern cpuop_func op_c100_11_ff;
extern cpuop_func op_c108_11_nf;
extern cpuop_func op_c108_11_ff;
extern cpuop_func op_c110_11_nf;
extern cpuop_func op_c110_11_ff;
extern cpuop_func op_c118_11_nf;
extern cpuop_func op_c118_11_ff;
extern cpuop_func op_c120_11_nf;
extern cpuop_func op_c120_11_ff;
extern cpuop_func op_c128_11_nf;
extern cpuop_func op_c128_11_ff;
extern cpuop_func op_c130_11_nf;
extern cpuop_func op_c130_11_ff;
extern cpuop_func op_c138_11_nf;
extern cpuop_func op_c138_11_ff;
extern cpuop_func op_c139_11_nf;
extern cpuop_func op_c139_11_ff;
extern cpuop_func op_c140_11_nf;
extern cpuop_func op_c140_11_ff;
extern cpuop_func op_c148_11_nf;
extern cpuop_func op_c148_11_ff;
extern cpuop_func op_c150_11_nf;
extern cpuop_func op_c150_11_ff;
extern cpuop_func op_c158_11_nf;
extern cpuop_func op_c158_11_ff;
extern cpuop_func op_c160_11_nf;
extern cpuop_func op_c160_11_ff;
extern cpuop_func op_c168_11_nf;
extern cpuop_func op_c168_11_ff;
extern cpuop_func op_c170_11_nf;
extern cpuop_func op_c170_11_ff;
extern cpuop_func op_c178_11_nf;
extern cpuop_func op_c178_11_ff;
extern cpuop_func op_c179_11_nf;
extern cpuop_func op_c179_11_ff;
extern cpuop_func op_c188_11_nf;
extern cpuop_func op_c188_11_ff;
extern cpuop_func op_c190_11_nf;
extern cpuop_func op_c190_11_ff;
extern cpuop_func op_c198_11_nf;
extern cpuop_func op_c198_11_ff;
extern cpuop_func op_c1a0_11_nf;
extern cpuop_func op_c1a0_11_ff;
extern cpuop_func op_c1a8_11_nf;
extern cpuop_func op_c1a8_11_ff;
extern cpuop_func op_c1b0_11_nf;
extern cpuop_func op_c1b0_11_ff;
extern cpuop_func op_c1b8_11_nf;
extern cpuop_func op_c1b8_11_ff;
extern cpuop_func op_c1b9_11_nf;
extern cpuop_func op_c1b9_11_ff;
extern cpuop_func op_c1c0_11_nf;
extern cpuop_func op_c1c0_11_ff;
extern cpuop_func op_c1d0_11_nf;
extern cpuop_func op_c1d0_11_ff;
extern cpuop_func op_c1d8_11_nf;
extern cpuop_func op_c1d8_11_ff;
extern cpuop_func op_c1e0_11_nf;
extern cpuop_func op_c1e0_11_ff;
extern cpuop_func op_c1e8_11_nf;
extern cpuop_func op_c1e8_11_ff;
extern cpuop_func op_c1f0_11_nf;
extern cpuop_func op_c1f0_11_ff;
extern cpuop_func op_c1f8_11_nf;
extern cpuop_func op_c1f8_11_ff;
extern cpuop_func op_c1f9_11_nf;
extern cpuop_func op_c1f9_11_ff;
extern cpuop_func op_c1fa_11_nf;
extern cpuop_func op_c1fa_11_ff;
extern cpuop_func op_c1fb_11_nf;
extern cpuop_func op_c1fb_11_ff;
extern cpuop_func op_c1fc_11_nf;
extern cpuop_func op_c1fc_11_ff;
extern cpuop_func op_d000_11_nf;
extern cpuop_func op_d000_11_ff;
extern cpuop_func op_d010_11_nf;
extern cpuop_func op_d010_11_ff;
extern cpuop_func op_d018_11_nf;
extern cpuop_func op_d018_11_ff;
extern cpuop_func op_d020_11_nf;
extern cpuop_func op_d020_11_ff;
extern cpuop_func op_d028_11_nf;
extern cpuop_func op_d028_11_ff;
extern cpuop_func op_d030_11_nf;
extern cpuop_func op_d030_11_ff;
extern cpuop_func op_d038_11_nf;
extern cpuop_func op_d038_11_ff;
extern cpuop_func op_d039_11_nf;
extern cpuop_func op_d039_11_ff;
extern cpuop_func op_d03a_11_nf;
extern cpuop_func op_d03a_11_ff;
extern cpuop_func op_d03b_11_nf;
extern cpuop_func op_d03b_11_ff;
extern cpuop_func op_d03c_11_nf;
extern cpuop_func op_d03c_11_ff;
extern cpuop_func op_d040_11_nf;
extern cpuop_func op_d040_11_ff;
extern cpuop_func op_d048_11_nf;
extern cpuop_func op_d048_11_ff;
extern cpuop_func op_d050_11_nf;
extern cpuop_func op_d050_11_ff;
extern cpuop_func op_d058_11_nf;
extern cpuop_func op_d058_11_ff;
extern cpuop_func op_d060_11_nf;
extern cpuop_func op_d060_11_ff;
extern cpuop_func op_d068_11_nf;
extern cpuop_func op_d068_11_ff;
extern cpuop_func op_d070_11_nf;
extern cpuop_func op_d070_11_ff;
extern cpuop_func op_d078_11_nf;
extern cpuop_func op_d078_11_ff;
extern cpuop_func op_d079_11_nf;
extern cpuop_func op_d079_11_ff;
extern cpuop_func op_d07a_11_nf;
extern cpuop_func op_d07a_11_ff;
extern cpuop_func op_d07b_11_nf;
extern cpuop_func op_d07b_11_ff;
extern cpuop_func op_d07c_11_nf;
extern cpuop_func op_d07c_11_ff;
extern cpuop_func op_d080_11_nf;
extern cpuop_func op_d080_11_ff;
extern cpuop_func op_d088_11_nf;
extern cpuop_func op_d088_11_ff;
extern cpuop_func op_d090_11_nf;
extern cpuop_func op_d090_11_ff;
extern cpuop_func op_d098_11_nf;
extern cpuop_func op_d098_11_ff;
extern cpuop_func op_d0a0_11_nf;
extern cpuop_func op_d0a0_11_ff;
extern cpuop_func op_d0a8_11_nf;
extern cpuop_func op_d0a8_11_ff;
extern cpuop_func op_d0b0_11_nf;
extern cpuop_func op_d0b0_11_ff;
extern cpuop_func op_d0b8_11_nf;
extern cpuop_func op_d0b8_11_ff;
extern cpuop_func op_d0b9_11_nf;
extern cpuop_func op_d0b9_11_ff;
extern cpuop_func op_d0ba_11_nf;
extern cpuop_func op_d0ba_11_ff;
extern cpuop_func op_d0bb_11_nf;
extern cpuop_func op_d0bb_11_ff;
extern cpuop_func op_d0bc_11_nf;
extern cpuop_func op_d0bc_11_ff;
extern cpuop_func op_d0c0_11_nf;
extern cpuop_func op_d0c0_11_ff;
extern cpuop_func op_d0c8_11_nf;
extern cpuop_func op_d0c8_11_ff;
extern cpuop_func op_d0d0_11_nf;
extern cpuop_func op_d0d0_11_ff;
extern cpuop_func op_d0d8_11_nf;
extern cpuop_func op_d0d8_11_ff;
extern cpuop_func op_d0e0_11_nf;
extern cpuop_func op_d0e0_11_ff;
extern cpuop_func op_d0e8_11_nf;
extern cpuop_func op_d0e8_11_ff;
extern cpuop_func op_d0f0_11_nf;
extern cpuop_func op_d0f0_11_ff;
extern cpuop_func op_d0f8_11_nf;
extern cpuop_func op_d0f8_11_ff;
extern cpuop_func op_d0f9_11_nf;
extern cpuop_func op_d0f9_11_ff;
extern cpuop_func op_d0fa_11_nf;
extern cpuop_func op_d0fa_11_ff;
extern cpuop_func op_d0fb_11_nf;
extern cpuop_func op_d0fb_11_ff;
extern cpuop_func op_d0fc_11_nf;
extern cpuop_func op_d0fc_11_ff;
extern cpuop_func op_d100_11_nf;
extern cpuop_func op_d100_11_ff;
extern cpuop_func op_d108_11_nf;
extern cpuop_func op_d108_11_ff;
extern cpuop_func op_d110_11_nf;
extern cpuop_func op_d110_11_ff;
extern cpuop_func op_d118_11_nf;
extern cpuop_func op_d118_11_ff;
extern cpuop_func op_d120_11_nf;
extern cpuop_func op_d120_11_ff;
extern cpuop_func op_d128_11_nf;
extern cpuop_func op_d128_11_ff;
extern cpuop_func op_d130_11_nf;
extern cpuop_func op_d130_11_ff;
extern cpuop_func op_d138_11_nf;
extern cpuop_func op_d138_11_ff;
extern cpuop_func op_d139_11_nf;
extern cpuop_func op_d139_11_ff;
extern cpuop_func op_d140_11_nf;
extern cpuop_func op_d140_11_ff;
extern cpuop_func op_d148_11_nf;
extern cpuop_func op_d148_11_ff;
extern cpuop_func op_d150_11_nf;
extern cpuop_func op_d150_11_ff;
extern cpuop_func op_d158_11_nf;
extern cpuop_func op_d158_11_ff;
extern cpuop_func op_d160_11_nf;
extern cpuop_func op_d160_11_ff;
extern cpuop_func op_d168_11_nf;
extern cpuop_func op_d168_11_ff;
extern cpuop_func op_d170_11_nf;
extern cpuop_func op_d170_11_ff;
extern cpuop_func op_d178_11_nf;
extern cpuop_func op_d178_11_ff;
extern cpuop_func op_d179_11_nf;
extern cpuop_func op_d179_11_ff;
extern cpuop_func op_d180_11_nf;
extern cpuop_func op_d180_11_ff;
extern cpuop_func op_d188_11_nf;
extern cpuop_func op_d188_11_ff;
extern cpuop_func op_d190_11_nf;
extern cpuop_func op_d190_11_ff;
extern cpuop_func op_d198_11_nf;
extern cpuop_func op_d198_11_ff;
extern cpuop_func op_d1a0_11_nf;
extern cpuop_func op_d1a0_11_ff;
extern cpuop_func op_d1a8_11_nf;
extern cpuop_func op_d1a8_11_ff;
extern cpuop_func op_d1b0_11_nf;
extern cpuop_func op_d1b0_11_ff;
extern cpuop_func op_d1b8_11_nf;
extern cpuop_func op_d1b8_11_ff;
extern cpuop_func op_d1b9_11_nf;
extern cpuop_func op_d1b9_11_ff;
extern cpuop_func op_d1c0_11_nf;
extern cpuop_func op_d1c0_11_ff;
extern cpuop_func op_d1c8_11_nf;
extern cpuop_func op_d1c8_11_ff;
extern cpuop_func op_d1d0_11_nf;
extern cpuop_func op_d1d0_11_ff;
extern cpuop_func op_d1d8_11_nf;
extern cpuop_func op_d1d8_11_ff;
extern cpuop_func op_d1e0_11_nf;
extern cpuop_func op_d1e0_11_ff;
extern cpuop_func op_d1e8_11_nf;
extern cpuop_func op_d1e8_11_ff;
extern cpuop_func op_d1f0_11_nf;
extern cpuop_func op_d1f0_11_ff;
extern cpuop_func op_d1f8_11_nf;
extern cpuop_func op_d1f8_11_ff;
extern cpuop_func op_d1f9_11_nf;
extern cpuop_func op_d1f9_11_ff;
extern cpuop_func op_d1fa_11_nf;
extern cpuop_func op_d1fa_11_ff;
extern cpuop_func op_d1fb_11_nf;
extern cpuop_func op_d1fb_11_ff;
extern cpuop_func op_d1fc_11_nf;
extern cpuop_func op_d1fc_11_ff;
extern cpuop_func op_e000_11_nf;
extern cpuop_func op_e000_11_ff;
extern cpuop_func op_e008_11_nf;
extern cpuop_func op_e008_11_ff;
extern cpuop_func op_e010_11_nf;
extern cpuop_func op_e010_11_ff;
extern cpuop_func op_e018_11_nf;
extern cpuop_func op_e018_11_ff;
extern cpuop_func op_e020_11_nf;
extern cpuop_func op_e020_11_ff;
extern cpuop_func op_e028_11_nf;
extern cpuop_func op_e028_11_ff;
extern cpuop_func op_e030_11_nf;
extern cpuop_func op_e030_11_ff;
extern cpuop_func op_e038_11_nf;
extern cpuop_func op_e038_11_ff;
extern cpuop_func op_e040_11_nf;
extern cpuop_func op_e040_11_ff;
extern cpuop_func op_e048_11_nf;
extern cpuop_func op_e048_11_ff;
extern cpuop_func op_e050_11_nf;
extern cpuop_func op_e050_11_ff;
extern cpuop_func op_e058_11_nf;
extern cpuop_func op_e058_11_ff;
extern cpuop_func op_e060_11_nf;
extern cpuop_func op_e060_11_ff;
extern cpuop_func op_e068_11_nf;
extern cpuop_func op_e068_11_ff;
extern cpuop_func op_e070_11_nf;
extern cpuop_func op_e070_11_ff;
extern cpuop_func op_e078_11_nf;
extern cpuop_func op_e078_11_ff;
extern cpuop_func op_e080_11_nf;
extern cpuop_func op_e080_11_ff;
extern cpuop_func op_e088_11_nf;
extern cpuop_func op_e088_11_ff;
extern cpuop_func op_e090_11_nf;
extern cpuop_func op_e090_11_ff;
extern cpuop_func op_e098_11_nf;
extern cpuop_func op_e098_11_ff;
extern cpuop_func op_e0a0_11_nf;
extern cpuop_func op_e0a0_11_ff;
extern cpuop_func op_e0a8_11_nf;
extern cpuop_func op_e0a8_11_ff;
extern cpuop_func op_e0b0_11_nf;
extern cpuop_func op_e0b0_11_ff;
extern cpuop_func op_e0b8_11_nf;
extern cpuop_func op_e0b8_11_ff;
extern cpuop_func op_e0d0_11_nf;
extern cpuop_func op_e0d0_11_ff;
extern cpuop_func op_e0d8_11_nf;
extern cpuop_func op_e0d8_11_ff;
extern cpuop_func op_e0e0_11_nf;
extern cpuop_func op_e0e0_11_ff;
extern cpuop_func op_e0e8_11_nf;
extern cpuop_func op_e0e8_11_ff;
extern cpuop_func op_e0f0_11_nf;
extern cpuop_func op_e0f0_11_ff;
extern cpuop_func op_e0f8_11_nf;
extern cpuop_func op_e0f8_11_ff;
extern cpuop_func op_e0f9_11_nf;
extern cpuop_func op_e0f9_11_ff;
extern cpuop_func op_e100_11_nf;
extern cpuop_func op_e100_11_ff;
extern cpuop_func op_e108_11_nf;
extern cpuop_func op_e108_11_ff;
extern cpuop_func op_e110_11_nf;
extern cpuop_func op_e110_11_ff;
extern cpuop_func op_e118_11_nf;
extern cpuop_func op_e118_11_ff;
extern cpuop_func op_e120_11_nf;
extern cpuop_func op_e120_11_ff;
extern cpuop_func op_e128_11_nf;
extern cpuop_func op_e128_11_ff;
extern cpuop_func op_e130_11_nf;
extern cpuop_func op_e130_11_ff;
extern cpuop_func op_e138_11_nf;
extern cpuop_func op_e138_11_ff;
extern cpuop_func op_e140_11_nf;
extern cpuop_func op_e140_11_ff;
extern cpuop_func op_e148_11_nf;
extern cpuop_func op_e148_11_ff;
extern cpuop_func op_e150_11_nf;
extern cpuop_func op_e150_11_ff;
extern cpuop_func op_e158_11_nf;
extern cpuop_func op_e158_11_ff;
extern cpuop_func op_e160_11_nf;
extern cpuop_func op_e160_11_ff;
extern cpuop_func op_e168_11_nf;
extern cpuop_func op_e168_11_ff;
extern cpuop_func op_e170_11_nf;
extern cpuop_func op_e170_11_ff;
extern cpuop_func op_e178_11_nf;
extern cpuop_func op_e178_11_ff;
extern cpuop_func op_e180_11_nf;
extern cpuop_func op_e180_11_ff;
extern cpuop_func op_e188_11_nf;
extern cpuop_func op_e188_11_ff;
extern cpuop_func op_e190_11_nf;
extern cpuop_func op_e190_11_ff;
extern cpuop_func op_e198_11_nf;
extern cpuop_func op_e198_11_ff;
extern cpuop_func op_e1a0_11_nf;
extern cpuop_func op_e1a0_11_ff;
extern cpuop_func op_e1a8_11_nf;
extern cpuop_func op_e1a8_11_ff;
extern cpuop_func op_e1b0_11_nf;
extern cpuop_func op_e1b0_11_ff;
extern cpuop_func op_e1b8_11_nf;
extern cpuop_func op_e1b8_11_ff;
extern cpuop_func op_e1d0_11_nf;
extern cpuop_func op_e1d0_11_ff;
extern cpuop_func op_e1d8_11_nf;
extern cpuop_func op_e1d8_11_ff;
extern cpuop_func op_e1e0_11_nf;
extern cpuop_func op_e1e0_11_ff;
extern cpuop_func op_e1e8_11_nf;
extern cpuop_func op_e1e8_11_ff;
extern cpuop_func op_e1f0_11_nf;
extern cpuop_func op_e1f0_11_ff;
extern cpuop_func op_e1f8_11_nf;
extern cpuop_func op_e1f8_11_ff;
extern cpuop_func op_e1f9_11_nf;
extern cpuop_func op_e1f9_11_ff;
extern cpuop_func op_e2d0_11_nf;
extern cpuop_func op_e2d0_11_ff;
extern cpuop_func op_e2d8_11_nf;
extern cpuop_func op_e2d8_11_ff;
extern cpuop_func op_e2e0_11_nf;
extern cpuop_func op_e2e0_11_ff;
extern cpuop_func op_e2e8_11_nf;
extern cpuop_func op_e2e8_11_ff;
extern cpuop_func op_e2f0_11_nf;
extern cpuop_func op_e2f0_11_ff;
extern cpuop_func op_e2f8_11_nf;
extern cpuop_func op_e2f8_11_ff;
extern cpuop_func op_e2f9_11_nf;
extern cpuop_func op_e2f9_11_ff;
extern cpuop_func op_e3d0_11_nf;
extern cpuop_func op_e3d0_11_ff;
extern cpuop_func op_e3d8_11_nf;
extern cpuop_func op_e3d8_11_ff;
extern cpuop_func op_e3e0_11_nf;
extern cpuop_func op_e3e0_11_ff;
extern cpuop_func op_e3e8_11_nf;
extern cpuop_func op_e3e8_11_ff;
extern cpuop_func op_e3f0_11_nf;
extern cpuop_func op_e3f0_11_ff;
extern cpuop_func op_e3f8_11_nf;
extern cpuop_func op_e3f8_11_ff;
extern cpuop_func op_e3f9_11_nf;
extern cpuop_func op_e3f9_11_ff;
extern cpuop_func op_e4d0_11_nf;
extern cpuop_func op_e4d0_11_ff;
extern cpuop_func op_e4d8_11_nf;
extern cpuop_func op_e4d8_11_ff;
extern cpuop_func op_e4e0_11_nf;
extern cpuop_func op_e4e0_11_ff;
extern cpuop_func op_e4e8_11_nf;
extern cpuop_func op_e4e8_11_ff;
extern cpuop_func op_e4f0_11_nf;
extern cpuop_func op_e4f0_11_ff;
extern cpuop_func op_e4f8_11_nf;
extern cpuop_func op_e4f8_11_ff;
extern cpuop_func op_e4f9_11_nf;
extern cpuop_func op_e4f9_11_ff;
extern cpuop_func op_e5d0_11_nf;
extern cpuop_func op_e5d0_11_ff;
extern cpuop_func op_e5d8_11_nf;
extern cpuop_func op_e5d8_11_ff;
extern cpuop_func op_e5e0_11_nf;
extern cpuop_func op_e5e0_11_ff;
extern cpuop_func op_e5e8_11_nf;
extern cpuop_func op_e5e8_11_ff;
extern cpuop_func op_e5f0_11_nf;
extern cpuop_func op_e5f0_11_ff;
extern cpuop_func op_e5f8_11_nf;
extern cpuop_func op_e5f8_11_ff;
extern cpuop_func op_e5f9_11_nf;
extern cpuop_func op_e5f9_11_ff;
extern cpuop_func op_e6d0_11_nf;
extern cpuop_func op_e6d0_11_ff;
extern cpuop_func op_e6d8_11_nf;
extern cpuop_func op_e6d8_11_ff;
extern cpuop_func op_e6e0_11_nf;
extern cpuop_func op_e6e0_11_ff;
extern cpuop_func op_e6e8_11_nf;
extern cpuop_func op_e6e8_11_ff;
extern cpuop_func op_e6f0_11_nf;
extern cpuop_func op_e6f0_11_ff;
extern cpuop_func op_e6f8_11_nf;
extern cpuop_func op_e6f8_11_ff;
extern cpuop_func op_e6f9_11_nf;
extern cpuop_func op_e6f9_11_ff;
extern cpuop_func op_e7d0_11_nf;
extern cpuop_func op_e7d0_11_ff;
extern cpuop_func op_e7d8_11_nf;
extern cpuop_func op_e7d8_11_ff;
extern cpuop_func op_e7e0_11_nf;
extern cpuop_func op_e7e0_11_ff;
extern cpuop_func op_e7e8_11_nf;
extern cpuop_func op_e7e8_11_ff;
extern cpuop_func op_e7f0_11_nf;
extern cpuop_func op_e7f0_11_ff;
extern cpuop_func op_e7f8_11_nf;
extern cpuop_func op_e7f8_11_ff;
extern cpuop_func op_e7f9_11_nf;
extern cpuop_func op_e7f9_11_ff;
extern cpuop_func op_40c0_12_nf;
extern cpuop_func op_40c0_12_ff;
extern cpuop_func op_40d0_12_nf;
extern cpuop_func op_40d0_12_ff;
extern cpuop_func op_40d8_12_nf;
extern cpuop_func op_40d8_12_ff;
extern cpuop_func op_40e0_12_nf;
extern cpuop_func op_40e0_12_ff;
extern cpuop_func op_40e8_12_nf;
extern cpuop_func op_40e8_12_ff;
extern cpuop_func op_40f0_12_nf;
extern cpuop_func op_40f0_12_ff;
extern cpuop_func op_40f8_12_nf;
extern cpuop_func op_40f8_12_ff;
extern cpuop_func op_40f9_12_nf;
extern cpuop_func op_40f9_12_ff;
extern cpuop_func op_4200_12_nf;
extern cpuop_func op_4200_12_ff;
extern cpuop_func op_4210_12_nf;
extern cpuop_func op_4210_12_ff;
extern cpuop_func op_4218_12_nf;
extern cpuop_func op_4218_12_ff;
extern cpuop_func op_4220_12_nf;
extern cpuop_func op_4220_12_ff;
extern cpuop_func op_4228_12_nf;
extern cpuop_func op_4228_12_ff;
extern cpuop_func op_4230_12_nf;
extern cpuop_func op_4230_12_ff;
extern cpuop_func op_4238_12_nf;
extern cpuop_func op_4238_12_ff;
extern cpuop_func op_4239_12_nf;
extern cpuop_func op_4239_12_ff;
extern cpuop_func op_4240_12_nf;
extern cpuop_func op_4240_12_ff;
extern cpuop_func op_4250_12_nf;
extern cpuop_func op_4250_12_ff;
extern cpuop_func op_4258_12_nf;
extern cpuop_func op_4258_12_ff;
extern cpuop_func op_4260_12_nf;
extern cpuop_func op_4260_12_ff;
extern cpuop_func op_4268_12_nf;
extern cpuop_func op_4268_12_ff;
extern cpuop_func op_4270_12_nf;
extern cpuop_func op_4270_12_ff;
extern cpuop_func op_4278_12_nf;
extern cpuop_func op_4278_12_ff;
extern cpuop_func op_4279_12_nf;
extern cpuop_func op_4279_12_ff;
extern cpuop_func op_4280_12_nf;
extern cpuop_func op_4280_12_ff;
extern cpuop_func op_4290_12_nf;
extern cpuop_func op_4290_12_ff;
extern cpuop_func op_4298_12_nf;
extern cpuop_func op_4298_12_ff;
extern cpuop_func op_42a0_12_nf;
extern cpuop_func op_42a0_12_ff;
extern cpuop_func op_42a8_12_nf;
extern cpuop_func op_42a8_12_ff;
extern cpuop_func op_42b0_12_nf;
extern cpuop_func op_42b0_12_ff;
extern cpuop_func op_42b8_12_nf;
extern cpuop_func op_42b8_12_ff;
extern cpuop_func op_42b9_12_nf;
extern cpuop_func op_42b9_12_ff;
extern cpuop_func op_4e73_12_nf;
extern cpuop_func op_4e73_12_ff;
extern cpuop_func op_50c0_12_nf;
extern cpuop_func op_50c0_12_ff;
extern cpuop_func op_50d0_12_nf;
extern cpuop_func op_50d0_12_ff;
extern cpuop_func op_50d8_12_nf;
extern cpuop_func op_50d8_12_ff;
extern cpuop_func op_50e0_12_nf;
extern cpuop_func op_50e0_12_ff;
extern cpuop_func op_50e8_12_nf;
extern cpuop_func op_50e8_12_ff;
extern cpuop_func op_50f0_12_nf;
extern cpuop_func op_50f0_12_ff;
extern cpuop_func op_50f8_12_nf;
extern cpuop_func op_50f8_12_ff;
extern cpuop_func op_50f9_12_nf;
extern cpuop_func op_50f9_12_ff;
extern cpuop_func op_51c0_12_nf;
extern cpuop_func op_51c0_12_ff;
extern cpuop_func op_51d0_12_nf;
extern cpuop_func op_51d0_12_ff;
extern cpuop_func op_51d8_12_nf;
extern cpuop_func op_51d8_12_ff;
extern cpuop_func op_51e0_12_nf;
extern cpuop_func op_51e0_12_ff;
extern cpuop_func op_51e8_12_nf;
extern cpuop_func op_51e8_12_ff;
extern cpuop_func op_51f0_12_nf;
extern cpuop_func op_51f0_12_ff;
extern cpuop_func op_51f8_12_nf;
extern cpuop_func op_51f8_12_ff;
extern cpuop_func op_51f9_12_nf;
extern cpuop_func op_51f9_12_ff;
extern cpuop_func op_52c0_12_nf;
extern cpuop_func op_52c0_12_ff;
extern cpuop_func op_52d0_12_nf;
extern cpuop_func op_52d0_12_ff;
extern cpuop_func op_52d8_12_nf;
extern cpuop_func op_52d8_12_ff;
extern cpuop_func op_52e0_12_nf;
extern cpuop_func op_52e0_12_ff;
extern cpuop_func op_52e8_12_nf;
extern cpuop_func op_52e8_12_ff;
extern cpuop_func op_52f0_12_nf;
extern cpuop_func op_52f0_12_ff;
extern cpuop_func op_52f8_12_nf;
extern cpuop_func op_52f8_12_ff;
extern cpuop_func op_52f9_12_nf;
extern cpuop_func op_52f9_12_ff;
extern cpuop_func op_53c0_12_nf;
extern cpuop_func op_53c0_12_ff;
extern cpuop_func op_53d0_12_nf;
extern cpuop_func op_53d0_12_ff;
extern cpuop_func op_53d8_12_nf;
extern cpuop_func op_53d8_12_ff;
extern cpuop_func op_53e0_12_nf;
extern cpuop_func op_53e0_12_ff;
extern cpuop_func op_53e8_12_nf;
extern cpuop_func op_53e8_12_ff;
extern cpuop_func op_53f0_12_nf;
extern cpuop_func op_53f0_12_ff;
extern cpuop_func op_53f8_12_nf;
extern cpuop_func op_53f8_12_ff;
extern cpuop_func op_53f9_12_nf;
extern cpuop_func op_53f9_12_ff;
extern cpuop_func op_54c0_12_nf;
extern cpuop_func op_54c0_12_ff;
extern cpuop_func op_54d0_12_nf;
extern cpuop_func op_54d0_12_ff;
extern cpuop_func op_54d8_12_nf;
extern cpuop_func op_54d8_12_ff;
extern cpuop_func op_54e0_12_nf;
extern cpuop_func op_54e0_12_ff;
extern cpuop_func op_54e8_12_nf;
extern cpuop_func op_54e8_12_ff;
extern cpuop_func op_54f0_12_nf;
extern cpuop_func op_54f0_12_ff;
extern cpuop_func op_54f8_12_nf;
extern cpuop_func op_54f8_12_ff;
extern cpuop_func op_54f9_12_nf;
extern cpuop_func op_54f9_12_ff;
extern cpuop_func op_55c0_12_nf;
extern cpuop_func op_55c0_12_ff;
extern cpuop_func op_55d0_12_nf;
extern cpuop_func op_55d0_12_ff;
extern cpuop_func op_55d8_12_nf;
extern cpuop_func op_55d8_12_ff;
extern cpuop_func op_55e0_12_nf;
extern cpuop_func op_55e0_12_ff;
extern cpuop_func op_55e8_12_nf;
extern cpuop_func op_55e8_12_ff;
extern cpuop_func op_55f0_12_nf;
extern cpuop_func op_55f0_12_ff;
extern cpuop_func op_55f8_12_nf;
extern cpuop_func op_55f8_12_ff;
extern cpuop_func op_55f9_12_nf;
extern cpuop_func op_55f9_12_ff;
extern cpuop_func op_56c0_12_nf;
extern cpuop_func op_56c0_12_ff;
extern cpuop_func op_56d0_12_nf;
extern cpuop_func op_56d0_12_ff;
extern cpuop_func op_56d8_12_nf;
extern cpuop_func op_56d8_12_ff;
extern cpuop_func op_56e0_12_nf;
extern cpuop_func op_56e0_12_ff;
extern cpuop_func op_56e8_12_nf;
extern cpuop_func op_56e8_12_ff;
extern cpuop_func op_56f0_12_nf;
extern cpuop_func op_56f0_12_ff;
extern cpuop_func op_56f8_12_nf;
extern cpuop_func op_56f8_12_ff;
extern cpuop_func op_56f9_12_nf;
extern cpuop_func op_56f9_12_ff;
extern cpuop_func op_57c0_12_nf;
extern cpuop_func op_57c0_12_ff;
extern cpuop_func op_57d0_12_nf;
extern cpuop_func op_57d0_12_ff;
extern cpuop_func op_57d8_12_nf;
extern cpuop_func op_57d8_12_ff;
extern cpuop_func op_57e0_12_nf;
extern cpuop_func op_57e0_12_ff;
extern cpuop_func op_57e8_12_nf;
extern cpuop_func op_57e8_12_ff;
extern cpuop_func op_57f0_12_nf;
extern cpuop_func op_57f0_12_ff;
extern cpuop_func op_57f8_12_nf;
extern cpuop_func op_57f8_12_ff;
extern cpuop_func op_57f9_12_nf;
extern cpuop_func op_57f9_12_ff;
extern cpuop_func op_58c0_12_nf;
extern cpuop_func op_58c0_12_ff;
extern cpuop_func op_58d0_12_nf;
extern cpuop_func op_58d0_12_ff;
extern cpuop_func op_58d8_12_nf;
extern cpuop_func op_58d8_12_ff;
extern cpuop_func op_58e0_12_nf;
extern cpuop_func op_58e0_12_ff;
extern cpuop_func op_58e8_12_nf;
extern cpuop_func op_58e8_12_ff;
extern cpuop_func op_58f0_12_nf;
extern cpuop_func op_58f0_12_ff;
extern cpuop_func op_58f8_12_nf;
extern cpuop_func op_58f8_12_ff;
extern cpuop_func op_58f9_12_nf;
extern cpuop_func op_58f9_12_ff;
extern cpuop_func op_59c0_12_nf;
extern cpuop_func op_59c0_12_ff;
extern cpuop_func op_59d0_12_nf;
extern cpuop_func op_59d0_12_ff;
extern cpuop_func op_59d8_12_nf;
extern cpuop_func op_59d8_12_ff;
extern cpuop_func op_59e0_12_nf;
extern cpuop_func op_59e0_12_ff;
extern cpuop_func op_59e8_12_nf;
extern cpuop_func op_59e8_12_ff;
extern cpuop_func op_59f0_12_nf;
extern cpuop_func op_59f0_12_ff;
extern cpuop_func op_59f8_12_nf;
extern cpuop_func op_59f8_12_ff;
extern cpuop_func op_59f9_12_nf;
extern cpuop_func op_59f9_12_ff;
extern cpuop_func op_5ac0_12_nf;
extern cpuop_func op_5ac0_12_ff;
extern cpuop_func op_5ad0_12_nf;
extern cpuop_func op_5ad0_12_ff;
extern cpuop_func op_5ad8_12_nf;
extern cpuop_func op_5ad8_12_ff;
extern cpuop_func op_5ae0_12_nf;
extern cpuop_func op_5ae0_12_ff;
extern cpuop_func op_5ae8_12_nf;
extern cpuop_func op_5ae8_12_ff;
extern cpuop_func op_5af0_12_nf;
extern cpuop_func op_5af0_12_ff;
extern cpuop_func op_5af8_12_nf;
extern cpuop_func op_5af8_12_ff;
extern cpuop_func op_5af9_12_nf;
extern cpuop_func op_5af9_12_ff;
extern cpuop_func op_5bc0_12_nf;
extern cpuop_func op_5bc0_12_ff;
extern cpuop_func op_5bd0_12_nf;
extern cpuop_func op_5bd0_12_ff;
extern cpuop_func op_5bd8_12_nf;
extern cpuop_func op_5bd8_12_ff;
extern cpuop_func op_5be0_12_nf;
extern cpuop_func op_5be0_12_ff;
extern cpuop_func op_5be8_12_nf;
extern cpuop_func op_5be8_12_ff;
extern cpuop_func op_5bf0_12_nf;
extern cpuop_func op_5bf0_12_ff;
extern cpuop_func op_5bf8_12_nf;
extern cpuop_func op_5bf8_12_ff;
extern cpuop_func op_5bf9_12_nf;
extern cpuop_func op_5bf9_12_ff;
extern cpuop_func op_5cc0_12_nf;
extern cpuop_func op_5cc0_12_ff;
extern cpuop_func op_5cd0_12_nf;
extern cpuop_func op_5cd0_12_ff;
extern cpuop_func op_5cd8_12_nf;
extern cpuop_func op_5cd8_12_ff;
extern cpuop_func op_5ce0_12_nf;
extern cpuop_func op_5ce0_12_ff;
extern cpuop_func op_5ce8_12_nf;
extern cpuop_func op_5ce8_12_ff;
extern cpuop_func op_5cf0_12_nf;
extern cpuop_func op_5cf0_12_ff;
extern cpuop_func op_5cf8_12_nf;
extern cpuop_func op_5cf8_12_ff;
extern cpuop_func op_5cf9_12_nf;
extern cpuop_func op_5cf9_12_ff;
extern cpuop_func op_5dc0_12_nf;
extern cpuop_func op_5dc0_12_ff;
extern cpuop_func op_5dd0_12_nf;
extern cpuop_func op_5dd0_12_ff;
extern cpuop_func op_5dd8_12_nf;
extern cpuop_func op_5dd8_12_ff;
extern cpuop_func op_5de0_12_nf;
extern cpuop_func op_5de0_12_ff;
extern cpuop_func op_5de8_12_nf;
extern cpuop_func op_5de8_12_ff;
extern cpuop_func op_5df0_12_nf;
extern cpuop_func op_5df0_12_ff;
extern cpuop_func op_5df8_12_nf;
extern cpuop_func op_5df8_12_ff;
extern cpuop_func op_5df9_12_nf;
extern cpuop_func op_5df9_12_ff;
extern cpuop_func op_5ec0_12_nf;
extern cpuop_func op_5ec0_12_ff;
extern cpuop_func op_5ed0_12_nf;
extern cpuop_func op_5ed0_12_ff;
extern cpuop_func op_5ed8_12_nf;
extern cpuop_func op_5ed8_12_ff;
extern cpuop_func op_5ee0_12_nf;
extern cpuop_func op_5ee0_12_ff;
extern cpuop_func op_5ee8_12_nf;
extern cpuop_func op_5ee8_12_ff;
extern cpuop_func op_5ef0_12_nf;
extern cpuop_func op_5ef0_12_ff;
extern cpuop_func op_5ef8_12_nf;
extern cpuop_func op_5ef8_12_ff;
extern cpuop_func op_5ef9_12_nf;
extern cpuop_func op_5ef9_12_ff;
extern cpuop_func op_5fc0_12_nf;
extern cpuop_func op_5fc0_12_ff;
extern cpuop_func op_5fd0_12_nf;
extern cpuop_func op_5fd0_12_ff;
extern cpuop_func op_5fd8_12_nf;
extern cpuop_func op_5fd8_12_ff;
extern cpuop_func op_5fe0_12_nf;
extern cpuop_func op_5fe0_12_ff;
extern cpuop_func op_5fe8_12_nf;
extern cpuop_func op_5fe8_12_ff;
extern cpuop_func op_5ff0_12_nf;
extern cpuop_func op_5ff0_12_ff;
extern cpuop_func op_5ff8_12_nf;
extern cpuop_func op_5ff8_12_ff;
extern cpuop_func op_5ff9_12_nf;
extern cpuop_func op_5ff9_12_ff;
